package charite.christo;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**
   A graphical representation of a graph
   adjusting the node distances dynamically.
   @author Christoph Gille
   @author Sun Microsystems
*/
public class GraphSun extends JPanel implements Disposable,Runnable,ActionListener {
    // ========================================
    private transient IsEnabled isEnabled;

    public void setEnabled(IsEnabled e) {isEnabled=e;}
    private final UniqueList<GraphNode> vNodes=new UniqueList(GraphNode.class);
    private final UniqueList<GraphEdge> vEdges=new UniqueList(GraphEdge.class);
    // ----------------------------------------
    private void addNode(GraphNode n) {
        final int w2=200/2,h2=200/2;
        n.x = w2*Math.random();
        n.y = h2*Math.random();
        n.x=200;n.y=200;
        vNodes.add(n);
    }
    // ----------------------------------------
    public void addEdge(GraphEdge e) {
        vEdges.add(e);
        if (!vNodes.contains(e.getNodeFrom())) addNode(e.getNodeFrom());
        if (!vNodes.contains(e.getNodeTo())) addNode(e.getNodeTo());
    }
    // ----------------------------------------
    public void run() {
        if (isEnabled!=null && !isEnabled.isEnabled(this)) return;
        preventFixedNodesFromfloatingOutTheScreen();
        relax();
        repaint();
    }
    // ----------------------------------------
    private void preventFixedNodesFromfloatingOutTheScreen() {
        final int w=getWidth(),h=getHeight();
        for(GraphNode n :vNodes.asArray()) {
            if (n.isFixed()) {
                if (n.x<0) n.x=0;
                if (n.y<40) n.y=40;
                if (n.x+20>w) n.x=w-20;
                if (n.y+20>h) n.y=h-20;
            }
        }
    }
    // ----------------------------------------
    private boolean isReciprocal() { return false; }
    private double maxDistance,minDistance;
    // ----------------------------------------
    private double scale() {
        double mv=maxDistance-minDistance; if (mv==0) mv=1;
        return sliderScale.getValue()/300f/mv*getWidth();
    }
    private void relax() {
        final double thisScale=scale();
        maxDistance=Double.MIN_VALUE;minDistance=Double.MAX_VALUE;
        final boolean isReciprocal=isReciprocal();
        for(GraphEdge e:vEdges.asArray()) {
            final GraphNode to=e.getNodeFrom(),from=e.getNodeTo();
            final double
                origDistance=e.getLen(),
                distance= isReciprocal ? (origDistance==0 ? Double.MAX_VALUE: 1/origDistance) : origDistance;
            if (Double.isNaN(distance)) continue;
            if (maxDistance<distance) maxDistance=distance;
            if (minDistance>distance) minDistance=distance;
            final double
                vx=to.x-from.x, vy=to.y-from.y,
                len=Math.sqrt(vx * vx + vy * vy),
                f=len<2 ? 0 : (distance-minDistance)*thisScale/len-1 ;
            if (Double.isNaN(f)) continue;
            final double dx=f*vx, dy=f*vy;
            to.dx+=dx; to.dy+=dy;
            from.dx+=-dx; from.dy+=-dy;
        }
        for(GraphNode n1 : vNodes.asArray()) {
            double dx=0,dy=0;
            for(GraphNode n2 : vNodes.asArray()) {
                if (n1==n2)continue;
                final double
                    vx=n1.x-n2.x,
                    vy=n1.y-n2.y,
                    len=vx * vx + vy * vy;
                if (len==0) {dx+=Math.random();dy+=Math.random();
                } else if (len < 100*100) {dx+=vx / len;dy+=vy / len;}
            }
            double dlen=dx*dx + dy*dy;
            if (dlen>0) {
                dlen=Math.sqrt(dlen) / 2;
                n1.dx+=dx/dlen;
                n1.dy+=dy/dlen;
            }
        }
        Dimension d=getSize();
        for(GraphNode n : vNodes.asArray()) {
            if (!n.isFixed() && n!=pick) {
                n.x+=Math.max(-5, Math.min(5, n.dx))- n.x/(d.width+1)*10;
                n.y+=Math.max(-5, Math.min(5, n.dy))- n.y/(d.height+1)*10;
                //if (n.x < 0) n.x=0; else if (n.x > d.width) n.x=d.width;
                //if (n.y < 0) n.y=0; else if (n.y > d.height) n.y=d.height;
            }
            n.dx /= 2;
            n.dy /= 2;
        }
    }
    // ----------------------------------------
    private transient GraphNode pick;
    private final static Color nodeColor=new Color(250, 220, 100);
    // ----------------------------------------
    private void paintNode(Graphics g, GraphNode n, FontMetrics fm) {
        final int x=getWidth()/2 + (int)n.x, y=getHeight()/2 + (int)n.y;
        Image image=null;
        if (butImage.isSelected()) { image= n.getImage(); }
        g.setColor((n==pick) ? Color.PINK : (n.isFixed() ? Color.RED : nodeColor));
        if (image==null) {
            String s=n.getText();
            int w=fm.stringWidth(s) + 10;
            int h=fm.getHeight() + 4;
            g.fill3DRect(x - w/2, y - h / 2, w, h,n!=closestNode);
            g.setColor(Color.black);
            g.drawString(s, x - (w-10)/2, (y - (h-4)/2) + fm.getAscent());
        } else {
            final int w=image.getWidth(this);
            final int h=image.getHeight(this);
            g.drawImage(image,x-ICON_HEIGHT/2, y-ICON_HEIGHT/2,ICON_HEIGHT,ICON_HEIGHT, this);
        }
    }
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        final long time=System.currentTimeMillis();
        final int w=getWidth(),h=getHeight();
        final boolean isLines=butLines.isSelected();
        if (h<10 || w<10) return;
        final double scale=scale();
        final boolean isReciprocal=isReciprocal();
        final int xm=w/2,ym=h/2;
        for(GraphEdge e:vEdges.asArray()) {
            final GraphNode from=e.getNodeFrom(),to=e.getNodeTo();
            final int x1=(int)from.x, y1=(int)from.y;
            final int x2=(int)to.x, y2=(int)to.y;
            final double origDistance=e.getLen();
            final double distance = isReciprocal ? (origDistance==0 ? Double.MAX_VALUE: 1/origDistance) : origDistance;
            if (Double.isNaN(distance)) continue;
            final int len=(int)Math.abs(((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2)) - (distance-minDistance)*scale);
            g.setColor(blueYellowRed(len/100f));
            if (isLines) g.drawLine(xm+x1, ym+y1, xm+x2, ym+y2);
        }
        final FontMetrics fm=g.getFontMetrics();
        for(GraphNode n:vNodes.asArray()) {
            paintNode(g, n, fm);
        }
        ChThread.callEvery(50+(int)(System.currentTimeMillis()-time),this,"GraphSun",0);
    }
    @Override public void paintChildren(Graphics g) {
        if (gcp(KEY_IS_PRINTING,this)==null)
            try {
                super.paintChildren(g);
            } catch(Throwable e){}
    }
    // ----------------------------------------
    @Override public void processEvent(AWTEvent e) {
        if (e instanceof MouseEvent) {
            final MouseEvent ev=(MouseEvent)e;
            final int x=ev.getX()-getWidth()/2, y=ev.getY()-getHeight()/2;
            switch(ev.getID()) {
            case MOUSE_MOVED:
                closestNode=closestNode(x,y);
                setCursor(getCursr(closestNode !=null ? 'H':'D'));
                break;
            case MOUSE_DRAGGED:
                setCursor(getCursr('M'));
                if (pick==null) return;
                pick.x=x; pick.y=y;
                break;
            case MOUSE_PRESSED:
                pick=closestNode(x,y);
                if (pick!=null) { pick.x=x; pick.y=y;}
                break;
            case MOUSE_RELEASED:
                if (pick!=null) {pick.x=x; pick.y=y;}
                pick=null;
                break;
            }
        }
        super.processEvent(e);
    }
    { this.enableEvents(MOUSE_EVENT_MASK | MOUSE_MOTION_EVENT_MASK);}

    // ----------------------------------------
    private transient GraphNode closestNode;
    private GraphNode closestNode(int x,int y) {
        double bestdist=9999;
        GraphNode closest=null;
        for(GraphNode n : vNodes.asArray()) {
            double dx=n.x-x, dy=n.y-y, dist=dx*dx+dy*dy;
            if (dist < bestdist) {
                closest=n;
                bestdist=dist;
            }
        }
        return closest;
    }
    public void start() { ChThread.callEvery(999,this,"GraphSun",0);}
    public void dispose() { ChThread.callEvery(-1,this,"",0);}
    public ChSlider sliderScale=new ChSlider("sliderScale",0,400,100).endLabels("scale",null);
    private Container buttonBar;
    private final Container buttonBar() {
        if (buttonBar==null) {
            buttonBar=hB("#",new ChButton("SCRAMBLE",ChButton.NO_MARGIN).t("scramble").li(this), butLines,butImage);
        }
        return buttonBar;
    }

    public void addButton(JComponent c,int pos) { buttonBar.add(c,pos);}
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final int w=getWidth(),h=getHeight();
        final double w2=w/2.0,h2=h/2.0;
        for(GraphNode n : vNodes.asArray()) {
            if (!n.isFixed()) {
                if (cmd=="SHAKE") {
                    n.x+=w*(Math.random()-0.5);
                    n.y+=h*(Math.random()-0.5);
                }
                if (cmd=="SCRAMBLE") {
                    n.x= w2*Math.random();
                    n.y= h2*Math.random();
                }
            }
        }
    }
    private final AbstractButton
        butLines=toggl("connect").s(true).cb(),
        butImage=toggl().t("icons").tt("show images rather protein names<br>when defined").cb();
    // ----------------------------------------
    public GraphSun() {
        Container pSouth=vB(sliderScale,buttonBar());
        cnsew(this,null,null,pSouth,null,null);
        makeTranslucent(this);
    }
    /*
      public static void main(String argv[]) {
      GraphSun g=new GraphSun();
      new ChFrame("").ad(g).shw(0);
      GraphNode
      n1=new GraphNode("alpha"),
      n2=new GraphNode("beta"),
      n3=new GraphNode("gamma"),
      n4=new GraphNode("delta");
      GraphEdge
      e1=new GraphEdge(n1,n2) { public double getLen() { return 2;} },
      e2=new GraphEdge(n1,n3) { public double getLen() { return 1;} },
      e3=new GraphEdge(n1,n4) { public double getLen() { return 4;} },
      e4=new GraphEdge(n2,n3) { public double getLen() { return 2;} },
      e5=new GraphEdge(n1,n4) { public double getLen() { return 8;} },
      e6=new GraphEdge(n3,n4) { public double getLen() { return 9;} };
      g.addEdge(e1);
      g.addEdge(e2);
      g.addEdge(e3);
      g.addEdge(e4);
      g.addEdge(e5);
      g.addEdge(e6);
      g.repaint();
      g.start();
      }
    */
}
