package charite.christo.blast;
import charite.christo.*;
import java.util.*;
import java.io.*;

import static charite.christo.ChUtils.*;
/**
   This class provids an object oriented structure of BLAST results.
   It can be initialized with an XML BLAST output.
   A human readable text can be created with the method toText().
   There is a significant difference between the numbering in BLAST output files and this API:
   In contrast to the
   output written by the BLAST program counting in this model starts
   at zero and not at one.  For example  queryStart 1 queryEnd 100 in the BLAST output would
   means 100  residues with indices 0 to index 99 here in this class.
   There is a difference between the XML format at NCBI and at ebi.ac.uk. This parser can cope with both formats.
*/
final public class BlastResult {
/**
   The instance is created with the XML output of a blast run.
   This can either be in NCBI or in Wu-blast format
*/
    public BlastResult(ByteArray xml) {
        if (xml!=null)
        new BlastParser().parseXML(xml,60,this);

    }
    private final UniqueList<BlastHit> vHits=new UniqueList(BlastHit.class);
    public BlastHit[] getHits(){return vHits.asArray();}
    private UniqueList<BlastDatabase> vDatabases=new UniqueList(BlastDatabase.class);

    public void addDatabase(BlastDatabase database) {vDatabases.add(database);}
    public BlastDatabase[] getDatabases() {return vDatabases.asArray();}
    private String querySeq;
    public void setQuerySeq(String s) {querySeq=s.intern();}
    public String getQuerySeq(String s) {return querySeq;}
    public void addHit(BlastHit bh) { vHits.add(bh); bh.setBlastResult(this);}
    private String programName;
    public void setProgramName(String s){programName=s;}
    public String getProgramName(){return programName;}
    private String programVersion;
    public void setProgramVersion(String s){programVersion=s;}
    public String getProgramVersion(){return programVersion;}
    public void setMatrix(String s){}
    public BlastResult cln() {
        BlastResult br=new BlastResult(null);
        br.vHits.addAll(vHits);
        br.querySeq=querySeq;
        return br;
    }
    //private String created;
    //public void setCreated(String s){created=s;}
    //public String getCreated(){return created;}
    public void writeFile(File f, int outputWidth) throws IOException {
        OutputStream fOut=null;
        fOut=fileOutStream(f);
        writeFile(fOut,outputWidth);
        closeStrm(fOut);
    }
    public void writeFile(OutputStream fo,int outputWidth) throws IOException {
        if (fo==null)  return;
        int count=0;
        final ByteArray sb=new ByteArray("Program="+getProgramName()+" Version="+getProgramVersion()+" Databases="+vDatabases+"\n\n");
        sb.write(fo);
        for(BlastHit bh:getHits())
            for(BlastAlignment bAli: bh.getAlignments()) {
                sb.setEnd(0).append(++count);
                bAli.toText(sb,outputWidth);
                sb.write(fo);
            }

    }

    //----------------------------------------
    public int countAlignments() {
        int count=0;
        for(BlastHit hit:getHits()) count+=hit.getAlignments().length;
        return count;
    }
    public void dispose() { vHits.clear(); }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */

    public boolean[] getMatchingAminoacids(byte[] sequence) {
        boolean sel[]=null;
        for(BlastHit h : getHits()) {
            for (BlastAlignment a : h.getAlignments()) {
                final byte mSeq[]=a.getMatchSeq();
                final byte qSeq[]=a.getQuerySeq();
                final int pos=strstr(STRSTR_IC, onlyLetters(mSeq), sequence);
                if (pos>=0) {
                    if (sel==null) sel=new boolean[sequence.length];
                    int iA=pos;
                    final byte[][] midline=charite.christo.protein.Blosum.BLASTMIDLINE;
                    for(int iCol=0; iA<sel.length && iCol<mSeq.length && iCol<qSeq.length; iCol++) {
                        if (midline[mSeq[iCol]&127][qSeq[iCol]&127]!=' ') sel[iA]=true;
                        iA++;
                    }
                }
            }
        }
        return sel;
    }

}
