package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.io.File;
import java.util.*;
import javax.swing.*;
import static charite.christo.ChConstants.*;
import static charite.christo.ChUtils.*;
import static java.awt.Color.*;
import static java.awt.event.KeyEvent.*;
import static java.awt.event.MouseEvent.*;

/**HELP

   <i>PACKAGE:charite.christo.strap.</i>

   The simple 3D-viewer visualizes the C-alpha trace of one or several proteins as a polygon.
   Helices may be  drawn in red and sheets in yellow.
   <br>
   <b>Mouse Actions:</b>
   <ul>
   <li>CTRL+drag or mouse-wheel: rotate in screen pane</li>
   <li>CTRL+mouse wheel or CTRL-plus: zoom. (This is consistent with office programs). </li>
   <li>clicking an C-alpha atom: moves alignment cursor to the clicked residue</li>
   </ul>

   <b>Picking amino acids:</b>

   Amino acids can be picked by clicking the C-alpha atom.
   A tool-panel will appear which acts on the picked protein.

   <br><br>

   <b>List of proteins:</b>

   The toggle "List of proteins" can be activated to show a list of all proteins in this view.
   Each protein in the list comes with a color-chooser.

    <br><br>

   <b>Color-chooser:</b>
   The color chooser is shown if <i>CB:"List proteins"</i> is activated.
   Four color-schemes are available. <i>JCOMPONENT:Protein3d#docuColorChooser()</i>
   <ul>
   <li>Each chain in a different color (striped box)</li>
   <li>Secondary structure: helices red, sheets yellow (yellow-red box)</li>
   <li>Entire protein in one color (monochrome boxes)</li>
   <li>Protein entirely hidden (black box)</li>
   </ul>
   With the CTRL key the color of all proteins can be changed at once.

   <br><br><b>Adding proteins to the 3D-view:</b>
   Proteins are added by WIKI:Drag_and_drop (See MOVIE_Drag_protein_to_3D ).

   <br><br><b>Full featured 3D-views (Astex, Jv, Pymol...)</b><br>

   In the View-Menu, full featured 3D-Views with powerful rendering capabilities are offered.

   <br><br><b>Identifying similar proteins</b><br><br>

   The fold of a protein is much more conserved in evolution than the
   amino acid sequence.

   Therefore it is possible to identify remote homologs for a given protein by looking for  structures
   with the same fold  regarding only the c-Alpha atom positions (Menubar <i>DIALOG:charite.christo.strap.DialogNeighbors</i>).

   @author Christoph Gille
   <li>CTRL + clicking: toggles selection</li>
   <li>SHIFT + clicking: selects range</li>
*/
public class Protein3d extends ChPanel implements Disposable,ChRunnable,TooltipProvider, ChProcessEvent, PaintHook {
    private final static byte  LINE=1,DISC=12,PLUS=13,CROSS=14,LETTER=15, SEC_STRU=-1, CHAINS=-2, INVISIBLE=-3;
    public final static String
        KEY_INITIALLY_SHOW="P3$$IS", KEY_BIO_MOLECULES="P3$$M", KEY_INTERPRET_AFTER_LOADING="P3$$IAL",
        ID_DISPOSED="P3D$$3", ID_PROTEIN_VIEWER="P3D$$5", SPIN="P3D$$spin";
    protected final static String  RUN_LAYOUT="U";
    private final ChEventAdapter LI=new ChEventAdapter(this);
    private final UniqueList<PView> vDisplayed=new UniqueList(PView.class);
    private final UniqueList<JComponent> vNW=new UniqueList(JComponent.class), vNE=new UniqueList(JComponent.class);
    private static volatile java.lang.ref.Reference<Line[]> refLines;
    private static float[] xyzT;
    private int durationPaint=33, meanDuration=33, lastSize, countLines;
    private long whenKeyEvent;
    private final Runnable runLayout=ChDelay.newThread(this,RUN_LAYOUT,null);
    private final Matrix3D tmat=new Matrix3D();
    private final Matrix3D matrix3D=new Matrix3D().scale(5f);
    public Matrix3D getRotationAndTranslationOfView() {return matrix3D;}
    private JMenuBar menubar;
    public JMenuBar getMenuBar() { return menubar;}
    public void setMenuBar(JMenuBar mb) { this.menubar=mb; }
    /* --- Near to mouse arrow --- */
    private int  prevx, prevy;
    private final Line MOUSE=new Line(),MOUS_HET=new Line();
    /* --- Bounding Box --- */
    private double mean_dx,mean_dy,mean_dz,mzT;
    private int ox,oy;
    /* --- End Bounding Box --- */
    private final ChCombo comboBio;
    private final ChButton
        togBioColor=toggl(null,ChButton.ICON_SIZE).li(LI).i(IC_COLOR).tt("Each chain in a different color"),
        CB_GUI=toggl("List proteins",ChButton.NO_MARGIN).r(runLayout),
        cbDetailBio=toggl("Chain-specific settings").li(LI)
        .tt("Toggle detailed/simple view.<br>The check-box is enabled if more than one molecules exists.<br>For most PDB entries, however, there is only one molecule with number 1");

    private Object[][] dataBio[]=new Object[2][][];
    private ChJTable tableBio;
    private final JComponent  PAN_SE=pnl(OPAQUE_FALSE, C(DEFAULT_BACKGROUND));
    private final JScrollPane SP_SOUTH=scrllpn(0L,PAN_SE);
    private final JComponent transp[]=new JComponent[99];
    {
        final String tt="The four options are <ol>"+
            "<li>The  Asymmetric_unit* as recorded in the ATOM-lines of the PDB-file.</li>"+
            "<li>The WIKI:Biological_unit is  reconstructed from the BIOMT Matrices.</li>"+
            "<li>Same as above. Each chain gets a different color to contrast adjacent chains.</li>"+
            "<li>The user can specify what bio-molecules he wants to see.</li>"+
            "</ol>",
            oo[]={"Asymmetrical unit","Biological unit 1","Customized"};
        comboBio=new ChCombo(oo,ChJTable.DEFAULT_RENDERER|ChJTable.ICON_ROW_HEIGHT).li(LI).tt(tt).s(1);
        comboBio.addMouseListener(LI);
        pcp(KEY_PREF_SIZE_IF_DISABLED,dim(0,0),comboBio);

    }
    private boolean isAutoRotation, isDragged, isWheel, notYetScaled=true, mouseIn;
    private float scale=1, xyzCursor[];
    /* ---------------------------------------- */
    /* >>> Instance >>> */
    public Protein3d() {
        super(null);
        pcp(KEY_PRINTABLE_COMPONENTS, weakRef(this),this);
        rtt(this);
        SP_SOUTH.getViewport().setOpaque(false);
        SP_SOUTH.setOpaque(false);
        vNW.add(comboBio);
        addMouseListener(LI);
        addMouseMotionListener(LI);
        addMouseWheelListener(LI);
        addFocusListener(LI);
        addKeyListener(LI);
        setMinimumSize(dim(1,1));
        shiftTooltip(this);

    }
    public boolean isPreview() {
        for(PView dp : vDisplayed.asArray()) if (dp._m3dPrev!=null) return true;
        return false;
    }
    private Matrix3D mx3d(PView pv) {
        if (pv==null) return null;
        if (isPreview()) return pv._m3dPrev;
        final Protein p=pv.getProtein();
        return pv._m3dPrev!=null ? pv._m3dPrev : p!=null ? p.getRotationAndTranslation() : null;
    }

    @Override protected void finalize() throws Throwable {
        logFinalize(ANSI_RED+ANSI_FG_WHITE+"Protein3d","");
        super.finalize();
    }
    public void dispose() {
        isAutoRotation=false;
        assertEDT();
        new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
        vDisplayed.clear();
        run(ID_DISPOSED,null);
        for(PView pv : vDisplayed.asArray()) {
            final Protein p=pv.getProtein();
            if (p!=null) p.removeProteinViewer(pv);
        }
        lastG=null;
    }
    /* <<<  Instance <<< */
    /* ---------------------------------------- */
    /* >>> getProteins >>> */
    private  java.lang.ref.Reference<Protein[]> ref_pp;
    private int getP_mc;
    public Protein[] getProteins() {
        final PView dd[]=vDisplayed.asArray();
        final int mc=vDisplayed.getModificationCount();
        Protein pp[]=ref_pp!=null ? ref_pp.get() : null;
        if (pp==null || getP_mc!=mc)  {
            getP_mc=mc;
            ref_pp=new java.lang.ref.SoftReference(pp=new Protein[dd.length]);
            for(int i=dd.length; --i>=0;) pp[i]=dd[i].getProtein();
        }
        return pp;
    }
    private PView initDP(PView dp, Protein p, boolean doInit) {
        dp._protein=p;
        dp.p3d=this;
        vDisplayed.add(dp);
        final Object weakRef=weakRef(dp);
        dp.cbActiveSite=toggl("AS").t("Active","site",dim(48,32)).paintIfEnabled().tt("show active site residues").s(true);
        for(ResidueSelection s : p.getAllResidueSelections()) {
            if (gcp(KEY_INITIALLY_SHOW,s)!=null) {
                dp.cbActiveSite.s(true);
                break;
            }
        }
        int count=0;
        final JComponent bb[]=new JComponent[99];
        bb[count++]=dp.cbHetero=toggl().t("Hetero","groups",dim(48,32)).s(true).paintIfEnabled()
            .tt("Show hetero groups such as flavins<br>With SHIFT key: List all hetero compounds<br>SHIFT-key and CTRL-key: Toggle atom color");
        bb[count++]=dp.cbActiveSite;
        bb[count++]=addPaintHook(this,dp.colorChooser=new ChButton().ps(dim(( (COLORS.length-1)/3+2) * ICON_HEIGHT/3,ICON_HEIGHT)).tt(this));
        bb[count++]=dp.butNucl=new ChButton("N",ChButton.ICON_SIZE).i(IC_DNA1).paintIfEnabled().tt("Show nucleotide acid<br>With SHIFT key: List of DNAs/RNAs.");
        bb[count++]=new ChButton("CLOSE",ChButton.ICON_SIZE).t(null).magnifyUnderMouse(IC_CLOSE).tt("Remove the protein from this 3D-view");
        for(Component b : bb) {
            if (b==null) continue;
            pcp(PView.class, weakRef, b);
            if (b instanceof ChButton) ((ChButton)b).li(LI);
            b.addMouseListener(LI);
        }
        final ProteinAlignment align=p.getAlignment();
        dp.proteinLabel=align==null?null:align.newProteinLabel(p);
        dp.panToolButtons=newCNSEW(pnl(dp.proteinLabel),null,null, pnl(bb),"#LBB");
        dp.panToolButtons.addMouseListener(LI);
        addMoliRefreshTip(dp.colorChooser);
        dp.iColor= sze(p.getChainsAsString())<2  ? SEC_STRU : CHAINS;
        if (doInit) afterAddProteins(p);
        return dp;
    }

    public void addProteins(Protein...pp0) {
        if (countNotNull(pp0)==0) return;
        if (!isEDT()) {  ChDelay.invokeAndWait(this,"ADD_PROTEIN",pp0); return; }
        final Protein[] ppLoaded=getProteins(), pp=pp0.clone();
        for(int iP=sze(pp); --iP>=0; ) {
            if (cntains(pp[iP],ppLoaded)) pp[iP]=null;
        }
        final Protein[] ppTrim=rmNull(pp,Protein.class);
        if (ppTrim.length>0) {
            for(Protein p : ppTrim) {
                final PView pv=new PView();
                p.addProteinViewer(pv);
                initDP(pv,p,false);
            }
            afterAddProteins(ppTrim);
        }
    }

    public void afterAddProteins(Protein... pp) {
        if (pp==null) return;
        if (!isEDT()) {  ChDelay.invokeAndWait(this,"ADD_PROTEIN",pp); return; }
        boolean hasMolecule1=false;
        int propMol=MIN_INT;
        for(int i=0; i<pp.length; i++) {
            final Protein p=pp[i];
            if (p.hasCalpha() || p.getHeteroCompounds().length+p.getDnaAndRnaStructures().length>0) {
                final String pdb=p.getPdbID();
                for(Protein prot : getProteins()) {
                    final String pdbId=prot.getPdbID();
                    if (pdb!=null && pdbId==pdb || pdbId==null) {
                        for(Matrix3D m3d : prot.getBioMatrices()) {
                            if (m3d.getBiomolecule()==1)  hasMolecule1=true;
                        }
                    }
                }

                //final Object Molecules=gcp(KEY_BIO_MOLECULES,p); if (Molecules!=null && dp0!=null) dp0.molecules=propMol=atoi(Molecules);
            }
        }
        if (propMol!=MIN_INT)  comboBio.s( propMol==-1?0:2);
        else if (hasMolecule1)  {
            for(PView dp : vDisplayed.asArray()) dp.molecules=1<<1;
            comboBio.s(1);
        }
        drawProteins(null);
        if (this.countLines==0) comboBio.s(0);
        notYetScaled=true;
        String lastPdb=null;
        for(Protein p : getProteins()) {
            final String pdb=p.getPdbID();
            if (lastPdb!=null && !lastPdb.equals(pdb)) { comboBio.s(0); break; }
            lastPdb=pdb;
            final PView v=getView(p);
            for(Object cmd : oo(gcp(KEY_INTERPRET_AFTER_LOADING,p))) {
                if (v!=null) v.interpret(toStrg(cmd),null,0L);
            }

        }
    }

    public void removeProtein(Protein p) {
        assertEDT();
        if (p==null) return;
        PView dp=getView(p);
        vDisplayed.remove(dp);
        ChDelay.runAfterMS(EDT,runLayout,111);
        addProteins((Protein[])null);
    }
    /* <<< getProteins <<< */
    /* ---------------------------------------- */
    /* >>> Protein information  >>> */
    public boolean isProteinShown(Protein p) {
        for(PView dp : vDisplayed.asArray())  if (dp.getProtein()==p) return dp.iColor!=INVISIBLE;
        return false;
    }
    public boolean eachProteinInDifferentColor(boolean asChain, boolean testOnly) {
        boolean changed=false;
        final PView dd[]=vDisplayed.asArray();
        for(int i=0; i<dd.length; i++) {
            final PView dp=dd[i];
            final Protein p=dp!=null ? dp.getProtein() : null;
            if (p==null) continue;
            final int iColor;
            if (asChain) {
                String chains=p.getChainsAsString();
                iColor= sze(chains)==1 ? chains.charAt(0)+(26-'A'%26) : i;
            } else  iColor=i;
            final int newCol=iColor % COLORS.length;
            changed|=newCol!=dp.iColor;
            if (!testOnly) dp.iColor=newCol;
        }
        return changed;
    }
    /* <<< Protein information <<< */
    /* ---------------------------------------- */
    /* >>> Thread >>> */
    public Object run(String id,Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id=="mouseMsg") {
            if (mouseMsgRect.width>0) repaint(mouseMsgRect);
            mouseMsgPrepare();
            if (mouseMsgRect.width>0) repaint(mouseMsgRect);
        }
        if (id==ChTableLayout.RUN_COLUMN_TITLE) return Protein.titleFor3dViewers(getProteins());
        if (id==KEY_BIO_MOLECULES) {
            final PView dp=getView(get(0,arg,Protein.class));
            if (dp!=null) {
                final int i=atoi(get(1,arg));
                dp.molecules=i==-1 ? 0 : i;
                comboBio.s(i==-1 ? 0 : 2);
                repaint();
            }
            notYetScaled=true;
        }
        if (id=="ADD_PROTEIN") addProteins((Protein[])arg);
        if (id==SPIN) {
            isAutoRotation=true;
            final Matrix3D mSpin=new Matrix3D();
            while(isAutoRotation) {
                if (getWidth()<ICON_HEIGHT) sleep(100);
                else {
                    matrix3D.mult(mSpin.unit().rotate('y',.02).rotate('x',.02*Math.sin(System.currentTimeMillis()/(1000*10))));
                    repaint();
                    sleep(maxi(99,5*meanDuration));
                    MOUSE.molecule=null;
                    MOUS_HET.molecule=null;
                }
            }
        }
        if (id==RUN_LAYOUT) {
            removeAll();
            final int w=getWidth(), h=getHeight();
            final boolean showTools=CB_GUI.s() && w>16*EX;
            final float lastScale=scale;
            MOUSE.molecule=null;
            MOUS_HET.molecule=null;
            if (w>16*EX) {
                final Dimension d=CB_GUI.getPreferredSize();
                CB_GUI.setBounds(0,h-d.height,d.width,d.height);
                add(CB_GUI);
            }
            if (showTools) {
                final PView dd[]=vDisplayed.asArray();
                add(SP_SOUTH);
                PAN_SE.removeAll();
                int maxWidth=1;
                for(PView dp: dd)  maxWidth=maxi(maxWidth,dp.panToolButtons.getPreferredSize().width);
                final int gridH=maxi(1,(w-CB_GUI.getWidth())/maxWidth), gridV=(dd.length-1)/ gridH+1;
                PAN_SE.setLayout(new GridLayout(gridV,gridH));
                //SP_SOUTH.setBorder(null);
                for(int i=mini(transp.length, gridH*gridV-dd.length); --i>=0;) {
                    if (transp[i]==null) transp[i]=pnl(OPAQUE_FALSE);
                    PAN_SE.add(transp[i]);
                }
                for(PView dp: dd)  PAN_SE.add(dp.panToolButtons);
                PAN_SE.repaint();
                {
                    final Dimension ps=PAN_SE.getPreferredSize();
                    int
                        vsb=JScrollPane.VERTICAL_SCROLLBAR_NEVER,
                        hsb=JScrollPane.HORIZONTAL_SCROLLBAR_NEVER,
                        ht=ps.height,
                        wt=ps.width;
                    if (wt>w-32) { wt=w-32; ht+=EX; hsb=JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS; }
                    if (ht>h/2) { ht=h/2; wt+=EX; vsb=JScrollPane.VERTICAL_SCROLLBAR_ALWAYS; }
                    SP_SOUTH.setVerticalScrollBarPolicy(vsb);
                    SP_SOUTH.setHorizontalScrollBarPolicy(hsb);
                    SP_SOUTH.setBounds(w-wt-3, h-ht-3, wt, ht);
                    if (w>0) scale=(w-(float)ht)/w;
                }
            } else scale=1;
            if (!Float.isNaN(scale) && scale!=lastScale) matrix3D.scale(scale/lastScale);
            enableDisable();
            final JComponent ccNE[]=vNE.asArray();
            int x=w;
            for(int i=ccNE.length; --i>=0;) {
                final JComponent c=ccNE[i];
                final Dimension d=c.getPreferredSize();
                if ((x-=d.width-1)>0 || i==ccNE.length-1) {
                    c.setBounds(x,0,d.width,d.height);
                    add(c);
                    if (i==ccNE.length-1 && c instanceof ChButton)  ((ChButton)c).setContentAreaFilled(w<4*EX);
                } else remove(ccNE[i]);
            }
            int y=x<16*EM ? ICON_HEIGHT:0;
            for(JComponent c : vNW.asArray()) {
                if (w>16*EM) {
                    final Dimension d=c.getPreferredSize();
                    c.setBounds(0,y,d.width,d.height);
                    y+=d.height+1;
                    add(c);
                } else remove(c);
            }
            repaint();
        }
        return null;
    }
    /* <<< Thread <<< */
    /* ---------------------------------------- */
    /* >>>  Single Shapes >>> */
    private static class Line implements Comparable  {
        Object molecule;
        float x0,y0,z0, x1,y1;
        Color color[];
        short idx;
        Matrix3D m3d;
        byte style;
        public int compareTo(Object o) {
            if (o==null) return -1;
            float z=((Line)o).z0;
            return z<z0 ? -1 : z>z0 ? 1 : 0;
        }
    }
    private void boundingBox(final Line lines[],final int nLines) {
        double  mxT,myT;
        float xminT,xmaxT, yminT,ymaxT, zminT,zmaxT;
        xminT=yminT=zminT=MAX_INT;
        xmaxT=ymaxT=zmaxT=MIN_INT;
        mxT=myT=mzT=mean_dx=mean_dy=mean_dz=0;
        int styleLine=0;
        for(int i=0;i<nLines;i++) {
            Line l=lines[i];
            assertNotNull(l);
            if (l!=null && l.style==LINE) styleLine++;
        }
        int iLines=0;
        for(int i=0;i<nLines;i++) {
            Line l=lines[i];
            if (l==null || (styleLine>2 && l.style!=LINE)) continue;
            if (l.x0<xminT) xminT=l.x0;
            if (l.x0>xmaxT) xmaxT=l.x0;
            if (l.y0<yminT) yminT=l.y0;
            if (l.y0>ymaxT) ymaxT=l.y0;
            if (l.z0<zminT) zminT=l.z0;
            if (l.z0>zmaxT) zmaxT=l.z0;
            mxT+=l.x0;
            myT+=l.y0;
            mzT+=l.z0;
            iLines++;
        }
        if (iLines==0) return;
        mxT/=iLines; myT/=iLines; mzT/=iLines;
        // --- Mean derivation ---
        for(int i=0;i<nLines;i++) {
            final Line l=lines[i];
            if (l==null || (styleLine>2 && l.style!=LINE)) continue;
            final float
                dx=xminT-l.x0,
                dy=yminT-l.y0,
                dz=zminT-l.z0;
            mean_dx+= (dx>0 ?  dx : -dx);
            mean_dy+= (dy>0 ?  dy : -dy);
            mean_dz+= (dz>0 ?  dz : -dz);
        }
        mean_dx/=iLines;
        mean_dy/=iLines;
        mean_dz/=iLines;
        if (mean_dz==0) mean_dz=.1f;
        ox=(int)(getWidth()/2-mxT);
        final int visH=getHeight()- (CB_GUI.s() ? SP_SOUTH.getHeight() : 0);
        oy=(int)(visH/2-myT);
    }
    private void drawProteins(Graphics g) {
        final PView[] dpp=vDisplayed.asArray();
        Line[] lines=refLines!=null ? refLines.get() : null;
        if (lines==null) refLines=new java.lang.ref.SoftReference(lines=new Line[100*1000]);
        int iLine;
        while(true) {
            iLine=0;
            for(PView dp:dpp) {
                if (dp.cbHetero.s()) {
                    for(HeteroCompound h:dp.getProtein().getHeteroCompounds())  iLine=makeHeteros(lines,h,dp,iLine,true);
                }
                if (dp.showNucl>0) {
                    for(HeteroCompound dna:dp.getProtein().getDnaAndRnaStructures()) {
                        iLine=makeHeteros(lines, dna, dp,iLine, dp.showNucl>1);
                    }
                }
                iLine=makeHeteros(lines, null, dp,iLine, false);
            }
            if (iLine+99>lines.length) refLines=new java.lang.ref.SoftReference(lines=new Line[iLine+9999]); else break;
        }
        this.countLines=iLine;
        if (g!=null) {
            boundingBox(lines,iLine);
            Arrays.sort(lines,0,iLine);
            drawLines(lines, iLine,g);
        }
    }
    private int makeHeteros(Line[] lines, HeteroCompound h, PView dp,int iLine0, boolean allAtoms) {
        final int iColor=dp!=null ? dp.iColor : 0;
        if (iColor==INVISIBLE) return iLine0;
        final Protein p=dp!=null ? dp.getProtein() : null;
        int iL=iLine0;
        final int bioUnit=comboBio.i();
        assertNotNull(dp);
        final Matrix3D rotTrans=mx3d(dp);
        final Matrix3D mm3d[]= bioUnit==0 ? null : p.getBioMatrices();
        final int nMat=sze(mm3d);
        final float[] xyz=h!=null ? h.getCoordinates(null) : p.getResidueCalpha(null);
        if (xyz==null) return iL;
        final float[] xyzT= this.xyzT=redim(this.xyzT,xyz.length,99);
        final int distDna2=this.hideDistantDNS && sliderDistDna!=null ? sliderDistDna.getValue()*sliderDistDna.getValue() : -1;
        if (h!=null) {
            final boolean underMouse=h==MOUS_HET.molecule;
            final boolean nuclBackbone=!allAtoms && h.isNucleotideChain(),  colorHeteros= dp==null|| (dp.colorHeteros&1)!=0;
            final int nA=h.countAtoms();
            final byte[] base=h.getNucleotideBase();
            final byte[] aType=h.getAtomType();
            final char chain= h.getCompoundChain();
            final int name=h.getCompoundName32(),number=h.getCompoundNumber();
            for(int iM=maxi(1,nMat); --iM>=0;) {
                final Matrix3D m3d=nMat>iM && (nMat>1 || !mm3d[0].isUnit()) ? mm3d[iM]:null; // see pdb1r9o
                final String m3dChains=m3d!=null ? m3d.getChains() : null;
                if (m3dChains!=null) {
                    if (m3dChains.indexOf(chain)<0 && chain!=' ') continue;
                    final int m=m3d.getBiomolecule();
                    if (m>=0 && dp.molecules!=-1 &&  (dp.molecules&(1L<<m))==0) continue;
                }
                tmat.unit().mult(m3d).mult(rotTrans).mult(matrix3D).transformPoints(xyz, xyzT, xyz.length/3);
                int lastNum=0, lastNam=0;
                float lastX=Float.NaN ,lastY=0;
                float[] cAlphaXYZ=null;
                //float[] dnaXYZ=null;
                for(int iA=0,i3=0; i3<xyz.length && iA<nA;  iA++,i3+=3) {
                    if (nuclBackbone  &&  aType[iA]!='P') continue;
                    if (distDna2>=0 && p!=null) {
                        /*
                        if (dnaXYZ==null) {
                            dnaXYZ=h.getCoordinates(m3d);
                            cAlphaXYZ=p.getResidueCalpha(m3d);
                        }
                        if (cAlphaXYZ!=null) {
                            final float dist2=squareDistance(dnaXYZ[i3],dnaXYZ[i3+1],dnaXYZ[i3+2],cAlphaXYZ,0,p.countResidues());
                            if (dist2>distDna2) { lastX=Float.NaN; continue;}
                        }
                        */
                    }
                    final Line l=iL>=lines.length ? null : lines[iL]==null ? lines[iL]=new Line() : lines[iL];
                    iL++;
                    if (l==null) continue;
                    l.idx=(short)iA;
                    l.molecule=h;
                    final float x0=l.x0=xyzT[i3+0], y0=l.y0=xyzT[i3+1], z0=l.z0=xyzT[i3+2];
                    l.color=underMouse?YELLOWs : colorHeteros ?  COL_ATOMS[aType[iA]] : (iColor>=0 ? COLORS[iColor%COLORS.length]:WHITEs);
                    l.style=DISC;
                    {
                        final int n=name;
                        if (n==HETERO_NA||n==HETERO_MG||n==HETERO_MN||n==HETERO_ZN||n==HETERO_SE||n==HETERO_CO||n==HETERO_NI||n==HETERO_CA) l.style=PLUS;
                        else if (n==HETERO_FE) l.style=CROSS;
                    }
                    if (nuclBackbone) {
                        if (name!=lastNam || number!=lastNum) { lastNam=name; lastNum=number; }
                        if (!Float.isNaN(lastX)) {
                            final Line c=iL>=lines.length ? null : lines[iL]==null ? lines[iL]=new Line() : lines[iL];
                            iL++;
                            if (c!=null) {
                                c.molecule=null;
                                c.m3d=m3d;
                                c.x0=x0; c.x1=lastX;
                                c.y0=y0; c.y1=lastY;
                                c.z0=z0;
                                c.color=l.color;
                                c.style=LINE;
                            }
                        }
                        lastX=x0; lastY=y0;
                        if (underMouse) {
                            float dx=MOUS_HET.x0-x0, dy=MOUS_HET.y0-y0;
                            if (dx*dx+dy*dy<64*64) {
                                final Line c=iL>=lines.length ? null : lines[iL]==null ? lines[iL]=new Line() : lines[iL];
                                iL++;
                                if (c!=null) {
                                    c.molecule=null;
                                    c.style=LETTER;
                                    c.x0=x0;
                                    c.y0=y0;
                                    c.z0=z0;
                                    c.x1= base!=null ? base[iA] : 0;
                                    c.color=l.color;
                                }
                            }
                        }
                    }
                }
            }
        }
        if (h==null) {
            final byte[] ss=p.getResidueSecStrType(), cc=p.getResidueChain();
            final int rn[]=p.getResidueNumber();
            final int nR=p.countResidues(), nR3=mini(xyz.length,nR*3);
            final boolean cbActiveSite=isSelctd(dp.cbActiveSite);
            boolean hasActiveSite=false;
            for(char chain=' '; chain<='z'; chain++) {
                if (!is(LETTR_DIGT_US,chain) && chain!=' ') continue;
                final int first=p.getChainFirstResidueIdx(chain), last=mini(nR-1,p.getChainLastResidueIdx(chain));
                if (first<0 || first>=last) continue;
                for(int iM=maxi(1,nMat); --iM>=0;) {
                    final Matrix3D m3d=nMat>iM && (nMat>1 || !mm3d[0].isUnit()) ? mm3d[iM]:null;
                    final String m3dChains=m3d!=null ? m3d.getChains() : null;
                    if (m3dChains!=null) {
                        if (m3dChains.indexOf(chain)<0) continue;
                        final int m=m3d.getBiomolecule();
                        if (m>=0 && dp.molecules!=-1 &&  (dp.molecules&(1L<<m))==0) continue;
                    }

                    tmat.unit().mult(m3d).mult(rotTrans).mult(matrix3D).transformPoints(xyz,xyzT,nR3);
                    for(int iA=first,i3=first*3; i3<nR3-3 && iA<=last;  iA++,i3+=3) {
                        if (Float.isNaN(xyzT[i3]) || Float.isNaN(xyzT[i3+4])) continue;
                        if (sze(cc)>iA+1 && cc[iA]!=cc[iA+1]) continue;
                        if (sze(rn)>iA+1 && rn[iA+1]-rn[iA]!=1) {
                            final float dx=xyz[i3]-xyz[3+i3], dy=xyz[1+i3]-xyz[1+3+i3], dz=xyz[2+i3]-xyz[2+3+i3];
                            if (dx*dx + dy*dy + dz*dz>27)  continue;
                        }
                        final Line l=iL>=lines.length ? null : lines[iL]==null ? lines[iL]=new Line() : lines[iL];
                        iL++;
                        if (l!=null) {
                            l.idx=(short)iA;
                            l.m3d=m3d;
                            l.molecule=dp;
                            l.x0=xyzT[i3+0]; l.y0=xyzT[i3+1];  l.z0=xyzT[i3+2];
                            l.x1=xyzT[i3+3]; l.y1=xyzT[i3+4];
                            l.style=LINE;
                            if (iColor==INVISIBLE) l.color=null;
                            else {
                                final boolean bioCol=bioUnit==2 && isSelctd(togBioColor);
                                if ( (!bioCol)  && iColor==SEC_STRU && iA<sze(ss)) l.color=COL_SECSTRU[ss[iA]];
                                else if (!bioCol && iColor==CHAINS && iA<sze(cc)) l.color=COLORS[ (cc[iA]+ (26-'A'%26)) % COLORS.length];
                                else if (bioCol) l.color=COLORS[((iColor>=0?iColor:0)+iM)%COLORS.length];
                                else l.color=iColor>=0  ? COLORS[iColor%COLORS.length] : WHITEs;
                            }
                        }
                    }
                    for(ResidueSelection s:p.getAllResidueSelections()){
                        if (s==null) continue;
                        if (gcp(FEATURE_ACTIVE_SITE,s)!=null) {
                            hasActiveSite=true;
                            if (!cbActiveSite) continue;
                        }
                        final Color colors[]=farbReihe(s instanceof Colored ? ((Colored)s).getColor() : WHITE);
                        if (s instanceof VisibleIn123 && (((VisibleIn123)s).getVisibleWhere()&VisibleIn123.STRUCTURE)==0) continue;
                        final boolean bb[]=s.getSelectedAminoacids();
                        if (bb==null) continue;
                        final int offset=s.getSelectedAminoacidsOffset();
                        final int to=mini(last, bb.length+offset, xyzT.length/3);
                        for(int iA=maxi(0,offset,first); iA<to; iA++) {
                            if (!bb[iA-offset]) continue;
                            final Line l=iL>=lines.length ? null : lines[iL]==null ? lines[iL]=new Line() : lines[iL];
                            if (l!=null) {
                                l.molecule=null;
                                iL++;
                                l.style=LETTER;
                                final int iA3=3*iA;
                                l.x0=xyzT[iA3];
                                l.y0=xyzT[iA3+1];
                                l.z0=xyzT[iA3+2];
                                l.x1=p.getResidueType(iA);
                                l.color=colors;
                            }
                        }
                    }
                }
            }
            if (hasActiveSite!=dp.cbActiveSite.isEnabled()) dp.cbActiveSite.setEnabled(hasActiveSite);
        }
        return iL;
    }
    private void drawLines(final Line lines[],final int n,final Graphics g) {
        g.setFont(getFnt(12,true,Font.BOLD));
        Color lastColor=null;
        final int ox=this.ox, oy=this.oy;
        final String strgLen1[]=getStringsLenght1();
        final Rectangle clip=g.getClipBounds();
        final int clipX1,clipY1,clipX2,clipY2;
        if (clip!=null) {
            clipX1=clip.x;
            clipY1=clip.y;
            clipX2=clipX1+clip.width;
            clipY2=clipY1+clip.height;
        } else {
            clipX1=clipY1=0;
            clipX2=clipY2=MAX_INT;
        }
        for(int i=0;i<n;i++) {
            final Line l=lines[i];
            final Color[] cc=l!=null?l.color : null;
            if (sze(cc)==0) continue;
            final int z=(int) ((l.z0-mzT)/mean_dz*DEPTH + DEPTH/4);
            final Color color= cc[maxi(0,mini(cc.length-1,z))];
            if (color==null) continue;
            if (lastColor!=color) g.setColor(lastColor=color);
            if (( isDragged||isAutoRotation) && meanDuration>33 &&  z>(meanDuration>55?DEPTH/3 : DEPTH/2)) continue;
            final int x=ox+(int)l.x0, y=oy+(int)l.y0, x1=ox+(int)l.x1, y1=oy+(int)l.y1;
            final int style=l.style;
            final int ch=style==l.style?0:EM;
            if (
                (x<clipX1-ch && (style!=LINE || x1<clipX1)) ||
                (x>clipX2+ch && (style!=LINE || x1>clipX2)) ||
                (y<clipY1-ch && (style!=LINE || y1<clipY1)) ||
                (y>clipY2+ch && (style!=LINE || y1>clipY2))
                ) {
                continue;
            }

           switch(l.style) {
            case LINE:  g.drawLine(x,y, x1,y1);  break;
            case PLUS:  g.drawLine(x-3,y,x+3,y); g.drawLine(x,y-3,x,y+3); break;
            case CROSS: g.drawLine(x-3,y-3,x+3,y+3); g.drawLine(x+3,y-3,x-3,y+3); break;
            case LETTER:g.drawString(strgLen1[(byte)( ((int)l.x1) & 127)],x,y); break;
            default:
                g.fillOval(x-2,y-2,4,4);
            }
        }
    }
    /* <<< Shapes <<< */
    /* ---------------------------------------- */
    /* >>> HeteroCompound >>> */
    public boolean isNucleotideChainVisible(HeteroCompound ha) {
        for(PView dp: vDisplayed.asArray())
            if (dp.showNucl>0 && cntains(ha,dp.getProtein().getDnaAndRnaStructures())) return true;
        return false;
    }
    public void showListOfHeterosAndNucleotides(Protein p) {}
    /* <<< HeteroCompound <<< */
    /* ---------------------------------------- */
    /* >>> Paint >>> */
    private Rectangle mouseMsgRect=new Rectangle();

    private String[] mouseMsg=new String[3];
    private int[] mouseMsgX=new int[3], mouseMsgW=new int[3];
    private Runnable mouseMsgRun=ChDelay.newThread(this,"mouseMsg",null);
    private void mouseMsgPrepare() {
        final int iA=MOUSE.idx;
        final int W=getWidth();
        final PView dp=(PView)MOUSE.molecule;
        final Graphics g=this.lastG;
        if (g==null || dp==null) {
            mouseMsg[0]=mouseMsg[1]=mouseMsg[2]=null;
            mouseMsgRect.setBounds(0,0,0,0);
            return;
        }
        int wMax=0;
        final int x0=ox+(int)MOUSE.x0, y0=oy+(int)MOUSE.y0;
        int bx1=mini(x0-3, ox+ (int)MOUSE.x1-3);
        int bx2=maxi(x0+3, ox+ (int)MOUSE.x1+3);
        final int by1=mini(y0-3, oy+ (int)MOUSE.y1-3, y0-3*charH(g));
        final int by2=maxi(y0+3, oy+ (int)MOUSE.y1+3);
        final Protein p=dp.getProtein();
        Protein.mouseOver(p, iA);

        for(int iLine=3; --iLine>=0;) {
            final ByteArray sb=byteArr(SB_TIP);
            switch(iLine) {
            case 0:
                if (sze(vDisplayed)<2) continue;
                sb.append(p);
                break;
            case 1:
                if (iA>=0) {
                    sb.append(' ').append(iA).append(' ').appendSomeBytes(p.getResidueName32(iA),4);
                    sb.append(' ').append(p.getResidueNumber(iA)).append(':').append((char)p.getResidueChain(iA));
                }
                break;
            case 2:
                if (comboBio.i()>0 && MOUSE.m3d!=null) sb.append("Bio unit ").append(MOUSE.m3d.getBiomolecule()).append(" BIOMT:").append(MOUSE.m3d.getName());
                break;
            }
            final int textWidth=strgWidth(g, mouseMsg[iLine]=sb.newString());
            wMax=maxi(wMax,textWidth);
            final int x=mini(x0,W-textWidth);
            bx1=mini(bx1,x);
            bx2=maxi(bx2,x+textWidth);
            mouseMsgX[iLine]=x;
            mouseMsgW[iLine]=textWidth;

        }
        mouseMsgRect.setBounds(bx1,by1, bx2-bx1, by2-by1);
    }

    private Graphics lastG;
    @Override public void paintComponent(Graphics g){
        final int W=getWidth(), H=getHeight();
        if (mouseIn && !isDragged && lastSize!=W+H) { mouseIn=false;  ChDelay.runAfterMS(EDT,runLayout,111); lastSize=W+H;}
        final long time=System.currentTimeMillis();
        final Rectangle clip=g.getClipBounds();
        g.setColor(isWhiteBG() ? WHITE : BLACK);
        if (clip!=null) g.fillRect(clip.x, clip.y, clip.width, clip.height); else g.fillRect(0,0, W,H);
        g.setColor(getForeground());
        drawProteins(g);
        if (gcp(KEY_IS_PRINTING,this)==null) {
            final PView dp=(PView)MOUSE.molecule;
            if (dp!=null) {
                final int charA=charA(g), charH=charH(g);
                g.setColor(PINK);
                final Protein p=dp.getProtein();
                final int x0=ox+(int)MOUSE.x0, y0=oy+(int)MOUSE.y0;
                g.fillOval(x0-3,y0-3,6,6);
                g.drawOval(ox+ (int)MOUSE.x1-3, oy+(int)MOUSE.y1-3,6,6);
                if (W>ICON_HEIGHT && p!=null && !isWheel && !isDragged && System.currentTimeMillis()-whenKeyEvent>999) {
                    for(int iLine=3; --iLine>=0;) {
                        final String msg=mouseMsg[iLine];
                        if (msg==null) continue;
                        g.setColor(C(0x0,99));
                        final int y=y0+ (iLine-3)*charH;
                        g.fillRect(mouseMsgX[iLine],y,mouseMsgW[iLine],charH);
                        g.setColor(WHITE);
                        g.drawString(msg,mouseMsgX[iLine],y+charA);
                    }
                }
            }
        }
        durationPaint=(int)(System.currentTimeMillis()-time);
        meanDuration= (meanDuration*9+durationPaint)/10;
        highlightAminoAcid(xyzCursor);
        //if (hasFocus()) { g.setColor(RED); g.drawLine(0,0,9999,0);}
        if (notYetScaled) {
            notYetScaled=false;
            final float d=(float)(mean_dx+mean_dy+mean_dz);
            if (d!=0) matrix3D.mult(tmat.unit().scale(Math.min(1,(H+W)/d/3)));
            repaint();
        }

        if (W<=ICON_HEIGHT) {
            final Protein[] pp=getProteins();
            final int tYY[]=this.thumbY=chSze(this.thumbY, 2*pp.length);
            int y=H;
            nextProt:
            for(int i=0; i<pp.length; i++) {
                final Image im=pp[i].getIconImage();
                if (im==null) continue nextProt;
                for(int j=0; j<i; j++) if (im==pp[j].getIconImage()) continue nextProt;
                tYY[2*i]=y-=W+1;
                tYY[2*i+1]=y+W;
                g.drawImage(im,0,y,W,W,this);
                lastG=g;
            }
            pcp(ChTableLayout.KEY_HEIGHT_SOUTH_PANEL, intObjct(y),this);
        } else pcp(ChTableLayout.KEY_HEIGHT_SOUTH_PANEL,null,this);
        paintHooks(this, g, true);
    }
    private int[] thumbY;
    public void paintChildren(Graphics g){
        try {
            if (mouseIn && !isDragged && gcp(KEY_IS_PRINTING,this)==null) super.paintChildren(g);
        } catch(Exception e){}
    }
    public boolean paintHook(JComponent component,Graphics g, boolean after) {
        final PView dp=(PView)gcp(PView.class,component);
        if (dp!=null && component==dp.colorChooser) {
            if (!after) return false;
            final boolean isDoc=gcp(HtmlDoc.KEY_DO_NOT_CLONE,component)!=null;
            final int ny=3, u=ICON_HEIGHT/ny, iColor=dp.iColor;
            g.setColor(iColor<0 ? BLACK :  COLORS[iColor%COLORS.length][0]);
            g.fillRect(0,0,999,999);
            final Protein p=dp.getProtein();
            for(int i=COLORS.length+3; --i>=0; ) {
                final int ix=i/ny, iy=i%ny, x=ix*u+4, y=iy*u+1;
                switch(i) {
                case 0:
                    if (p.getResidueSecStrType()!=null || isDoc){
                        g.setColor(RED);
                        g.fill3DRect(x,y,u-1,u-1,iColor!=SEC_STRU);
                        g.setColor(YELLOW);
                        g.fillRect(x,y+u/2,u-2,u/2-1);
                        break;
                    }
                case 1:
                    if (sze(p.getChainsAsString())>1 || isDoc) {
                        g.setColor(BLACK);
                        g.fill3DRect(x,y,u-1,u-1,iColor!=CHAINS);
                        for(int j=y+2;j<y+u-1;j++) { g.setColor(COLORS[j%COLORS.length][0]); g.drawLine(x+2,j,x+u-2,j);}
                    }
                    break;
                case 2:
                    g.setColor(BLACK);
                    g.fill3DRect(x,y,u-2,u-2,iColor!=INVISIBLE);
                    g.setColor(WHITE);
                    g.drawRect(x,y,u-2,u-2);
                    break;
                default:
                    g.setColor(COLORS[i-3][0]);
                    g.fill3DRect(x,y,u-1,u-1,i!=iColor+3);
                }
            }
        }
        return true;
    }
    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Tooltip >>> */
    private String tipLast;
    @Override public String getToolTipText(MouseEvent ev) {

        final PView pv=ev2thumb(ev);
        final Protein prot=pv!=null ? pv.getProtein() : null;
        if (prot!=null) return addHtmlTagsAsStrg(prot.getTooltipWithoutHtmlTags(isShift(ev)));
        final HeteroCompound ha=(HeteroCompound)MOUS_HET.molecule;
        if (ha!=null) {
            if (tipLast==null) {
                final String rt=ha.getRendererText(0);
                final Matrix3D m3d=MOUS_HET.m3d;
                final ByteArray sb=byteArr(SB_TIP).append("<html><body><pre>").append(rt).append(" Matrix:" ).append(m3d!=null ? m3d.getName() : null);
                int isFirst=0;
                for(Protein p : getProteins()) {
                    for(HeteroCompound h:p.getHeteroCompounds()) {
                        if (ha!=h) continue;
                        if (isFirst++==0) sb.append('\n');
                        sb.append(p);
                        break;
                    }
                }

                tipLast=sb.append("</pre></html></body>").newString();
            }
            return tipLast;
        }
        final PView dp=(PView)MOUSE.molecule;
        if (dp!=null && dp.getProtein()!=null) {
            final boolean cbActiveSite=isSelctd(dp.cbActiveSite);
            for(ResidueSelection sel : dp.getProtein().getResidueSelectionsAt(0L, MOUSE.idx, MOUSE.idx+1,VisibleIn123.STRUCTURE)) {
                if (gcp(FEATURE_ACTIVE_SITE,sel)!=null && !cbActiveSite) continue;
                if (sel==Protein.getMouseOverSelection()) continue;
                final String tip=TabItemTipIcon.tip(sel);
                return  addHtmlTagsAsStrg(  (tip!=null ? tip : toStrg(sel)));
            }
        }
        return null;
    }
    public String getToolTipText(JComponent component, AWTEvent ev) {
        final PView dp=(PView)gcp(PView.class,component);
        if (dp!=null && component==dp.colorChooser) {
            final int
                ny=3,
                u=ICON_HEIGHT/ny,
                mx=x(ev),
                my=y(ev);
            for(int i=0;i<COLORS.length;i++) {
                final int ix=i/ny,iy=i%ny,x=ix*u,y=iy*u;
                if (x<=mx && mx<=x+u  &&  y<=my && my<=y+u) {
                    return
                        i==0 ? "<html><body>Helices <font color=\"#FF0000\"> red</font> and sheets <font color=\"#AAAA00\"> yellow</font></body></html>" :
                        i==1 ? "Each chain in a different color" :
                        i==2 ? "Hide protein" : "The entire protein in the respective color";
                }
            }
        }
        return "hallo";
    }
    /* <<<  Tooltip <<< */
    /* ---------------------------------------- */
    /* >>>  GUI >>> */
    public void enableDisable() {
        assertEDT();
        boolean isBioUnits=false;
        for(PView d:vDisplayed.asArray()) {
            final Protein p=d.getProtein();
            d.cbHetero.setEnabled(p.getHeteroCompounds().length>0);
            d.butNucl.enabled(p.getDnaAndRnaStructures().length>0);
            if (sze(p.getBioMatrices())>1) isBioUnits=true;
        }
        eachProteinInDifferentColor(true,true);
        boolean update=false;
        if (isBioUnits!=comboBio.isEnabled()) {
            comboBio.setEnabled(isBioUnits);
            update=true;
        }
        if (update) ChDelay.runAfterMS(EDT,runLayout,333);
    }
    public void addOrRemoveButton(int pos,JComponent b, int where ) {
        final UniqueList v= where==SwingConstants.WEST ? vNW : vNE;
        if (b!=null) {
            b.removeMouseListener(LI);
            if (pos>=0) {
                v.add(mini(sze(v),pos),b);
                b.addMouseListener(LI);
            } else v.remove(b);
        }
        ChDelay.runAfterMS(EDT,runLayout,222);
    }

    private JSlider sliderDistDna;
    private boolean hideDistantDNS;
    public void setHideDistantDNS(boolean b) {
        hideDistantDNS=b;
        if (sliderDistDna==null) addActLi(LI,sliderDistDna=new ChSlider("DIST_DNA",0,100,30).endLabels("Close", "Distant"));
        ChFrame f=ChFrame.frame("Distant DNA",sliderDistDna, ChFrame.CENTER).size(100,64);
        if (b) f.shw(ChFrame.ALWAYS_ON_TOP);
        else f.setVisible(false);
    }
    /* <<< GUI <<< */
    /* ---------------------------------------- */
    /* >>>  Event >>> */
    private PView ev2thumb(AWTEvent ev) {
        if (getWidth()<=ICON_HEIGHT) {
            final int tYY[]=this.thumbY, y=y(ev);
            for(int i=sze(tYY)/2; --i>=0;) {
                if (tYY[2*i]<y && y<tYY[2*i+1]) {
                    return vDisplayed.get(i);
                }
            }
        }
        return null;
    }
    public AWTEvent processEvent(JComponent c,AWTEvent ev) {
        final boolean mouseInLast=this.mouseIn;
        final Object q=ev.getSource();
        final int id=ev.getID();
        final MouseEvent mev= ev instanceof MouseEvent ? (MouseEvent)ev : null;
        final int modi=modifiers(ev);
        final boolean shift=isShift(ev), ctrl=isCtrl(ev), but1=(modi&BUTTON1_MASK)!=0, but2=(modi&BUTTON2_MASK)!=0&&!isMacintosh(), but3=(modi&BUTTON3_MASK)!=0&&!isMacintosh();
        final int x=mev!=null ? mev.getX():0, y=mev!=null?mev.getY():0;
        final int H=maxi(1,getHeight());
        final int W=maxi(1,getWidth());
        final String cmd=actionCommand(ev);
        if (q==togBioColor||q instanceof ChTableModel || cmd=="RST_BIO" || q==cbDetailBio) comboBio.s(2);
        final boolean showTableBio=(cmd!=null || id==MOUSE_PRESSED) && ((q==cbDetailBio || q==comboBio || cmd=="RST_BIO") && comboBio.i()==2);
        final boolean rstDataBio=showTableBio && dataBio[0]==null || cmd=="RST_BIO";
        final PView dd[]=vDisplayed.asArray();
        final Protein[] pp=getProteins();
        if (q==this) {
            final PView pv=ev2thumb(ev);
            final Protein p=pv!=null ? pv.getProtein() : null;
            if (p!=null) {
                if ((id==MOUSE_CLICKED) && processEvt(ev,pv.proteinLabel)) return null;
                if (id==MOUSE_DRAGGED && !isDragged) {
                    getDndDateien().clear();
                    adUniq(p.getFile(), getDndDateien());
                    exportDrg(this , ev, TransferHandler.COPY);
                }
            }
            if (gcp(KEY_IS_RESIZING,this)!=null) return ev;
            final int scaleUpDown=scaleUpDown(ev);
            if (scaleUpDown!=0) {
                tmat.unit();
                matrix3D.mult(tmat.unit().scale(1f+(scaleUpDown<0  ? -20f : 20f)/H));
                repaint();
            }
        }
        if (q==sliderDistDna) repaint();
        if (rstDataBio) {
            final UniqueList<Object[]> v=new UniqueList(Object[].class);
            final Map<String,long[]> mapMol=new HashMap();
            boolean moreMolecules=false, hasPdbId=false;
            for(Protein p : pp)  {
                for(Matrix3D m3d:p.getBioMatrices()) moreMolecules|=m3d.getBiomolecule()>1;
                hasPdbId|=p.getPdbID()!=null;
            }
            cbDetailBio.setEnabled(moreMolecules);
            if (!moreMolecules) cbDetailBio.s(true);
            for(int detailed=2; --detailed>=0;) {
                mapMol.clear();
                v.clear();
                for(Protein p : pp) {
                    final String pdb=p.getPdbID();
                    final Object propMol=gcp(KEY_BIO_MOLECULES,p);
                    final long iPropMol=atol(propMol);
                    long mols=0;
                    for(Matrix3D m3d: p.getBioMatrices()) mols|= 1L<<m3d.getBiomolecule();
                    if (mols==0) continue;
                    if (hasPdbId && detailed==0) {
                        if (pdb!=null) {
                            long[] m=mapMol.get(pdb);
                            if (m==null) mapMol.put(pdb,m=new long[2]);
                            m[0]|=mols;
                            if (propMol!=null && iPropMol>0) m[1]|=iPropMol;
                        }
                    } else {
                        v.add(new Object[]{p,boolToTxt(mols,0,"-",byteArr(SB_TIP)).newString(), boolToTxt(propMol!=null ? iPropMol : mols, 0, "-", byteArr(SB_TIP)).newString(),null});
                    }
                }
                if (hasPdbId && detailed==0) {
                    for(Object pdbId : keyArry(mapMol)) {
                        final long mols=mapMol.get(pdbId)[0], iPropMol=mapMol.get(pdbId)[1];
                        final String s=boolToTxt(mols, 0, "-", byteArr(SB_TIP)).newString();
                        v.add(new Object[]{pdbId, s ,  iPropMol>0 ? boolToTxt(iPropMol,0,"-",byteArr(SB_TIP)).newString() :  (mols&2)!=0 ? "1" : s});
                    }
                }
                dataBio[detailed]=v.asArray();
            }
        }
        final int detailedBio01=cbDetailBio.s()?1:0;
        if (showTableBio) {
            if (tableBio==null) {
                final ChTableModel tm=new ChTableModel("Protein or PDB-ID","Available biomolecules","Displayed bio unit").editable(false,false,true,true);
                addActLi(LI,tm);
                tableBio=new ChJTable(tm,ChJTable.DEFAULT_RENDERER|ChJTable.ICON_ROW_HEIGHT)
                    .headerTip(null, "Molecules recorded in the PDB file", "List of molecule numbers entered by the user");
                tm.setJTable(tableBio,ChJTable.SET_VALUE_SELECTED_ROWS);
                installPopupListener(tableBio);
                final String msg="Biological units are reconstructed from the Asymmetric_unit* with transformation matrices. "+
                    "Matrices are recorded in PDB files like<pre class=\"data\">REMARK 350 BIOMOLECULE ...</pre>"+
                    "<u>Third table column:</u> Specify the molecules to be shown.<br>"+
                    "Leave table cell empty to see the Asymmetric_unit*.<br>"+
                    "Enter a non-existing biological unit number to hide.<br>"+
                    "To selectively deactivate a matrix, add a leading space into the  respective PDB file lines.<br>";
                pcp(KEY_NORTH_PANEL,msg,tableBio);
                pcp(KEY_SOUTH_PANEL,pnl(togBioColor, " "," ",cbDetailBio.cb(), new ChButton("RST_BIO",ChButton.NO_MARGIN).t("Reset table").li(LI)),tableBio);
            } else {
                tableBio.setRowHeight(32+detailedBio01);
                tableBio.repaint();
            }
            final ChFrame f=ChFrame.frame("Customized", tableBio, ChFrame.SCROLLPANE|ChFrame.AT_CLICK).shw(ChFrame.TO_FRONT);
            f.setLocation( comboBio.getLocationOnScreen().x-2*ICON_HEIGHT, comboBio.getLocationOnScreen().y+3*ICON_HEIGHT);
            for(int i=3;--i>=0;) tableBio.setColumnWidth(i, i==0?200:i==1?64:200, true);
            repaint();
        }
        if ( (showTableBio||rstDataBio) && tableBio!=null) {
            ((ChTableModel) tableBio.getModel()).setData(dataBio[detailedBio01]);
        }
        if (q instanceof ChTableModel || q==comboBio || showTableBio || rstDataBio || q==cbDetailBio ) {
            if (comboBio.i()==2) {
                final Object[][] data=dataBio[detailedBio01];
                for(int iRow=0; iRow<data.length; iRow++) {
                    final Object row[]=data[iRow],  o0=row[0];
                    for(Protein p:pp) {
                        if (p==o0 || o0==p.getPdbID()) {
                            final PView dp=getView(p);
                            final String s=toStrgTrim(row[2]);
                            if (dp!=null && s!=null)  dp.molecules=sze(s)==0 ? -1 : boolToLong(parseSet(s,999,0));
                        }
                    }
                }
                if (tableBio!=null) tableBio.repaint();
            } else {
                for(PView dp : dd) dp.molecules=2;
            }
            repaint();
        }
        if (q==togBioColor) repaint();
        final PView dp=(PView)gcp(PView.class,q);
        if (mev!=null && dp!=null) {
            if (id==MOUSE_CLICKED && q==dp.colorChooser) {
                final int ny=3, u=ICON_HEIGHT/3;
                final int f=((x-4)/u)*ny+y/u;
                dp.iColor= f==0 ? SEC_STRU : f==1 ? CHAINS : f==2 ? INVISIBLE : f-3;
                enableDisable();            
            }
            if (ctrl) for(PView d : dd) d.iColor=dp.iColor;
            repaint();
        }
        if (id==ActionEvent.ACTION_PERFORMED) {
            final boolean isSelected=isSelctd(q);
            if (dp!=null) {
                if (cmd=="N" && ++dp.showNucl>2) dp.showNucl=0;
                for(PView d : dd) {
                    if (ctrl || dp==d) {
                        if (cmd=="CLOSE") removeProtein(d.getProtein());
                        if (cmd=="N" && !shift) d.showNucl=dp.showNucl;
                        if (cmd=="AS") d.cbActiveSite.s(isSelected);
                        if (q==dp.cbHetero) {
                            if (shift&&ctrl) { dp.colorHeteros++; dp.cbHetero.s(!isSelected); }
                            if (!shift) { d.cbHetero.s(dp.cbHetero.s()); d.colorHeteros=dp.colorHeteros;}
                        }
                    }
                }
                final ChButton cbH=dp.cbHetero;
                if (shift && !ctrl && (cmd=="N" || q==cbH)) { if (q==cbH) cbH.s(!isSelected); showListOfHeterosAndNucleotides(dp.getProtein());}
                if (q==cbH || cmd=="AS" || cmd=="N") repaint();
            }

        }
        if (q==this) {
            final int kcode=keyCode(ev);
            if (kcode!=0 && id==KEY_PRESSED) {
                tmat.unit();
                if (ctrl && kcode==VK_W) {  dispose();  return null;}
                final boolean L=kcode==VK_LEFT || kcode==VK_UP;
                if(kcode==VK_LEFT || kcode==VK_RIGHT) { matrix3D.mult(ctrl ? tmat.rotate('z',L?-.1:.1):tmat.rotate('y',L?.1:-.1)); repaint(); }
                if(kcode== VK_UP || kcode==VK_DOWN) {   matrix3D.mult(ctrl ? tmat.scale(L?1.1f:.9f) :  tmat.rotate('x',L?.1:-.1)); repaint(); }
                if (kcode==VK_PAGE_UP||kcode==VK_PAGE_DOWN) { matrix3D.mult(tmat.rotate('x',kcode==VK_PAGE_UP ? 3.1415 : -3.1415)); repaint(); }
                whenKeyEvent=System.currentTimeMillis();
            }
            if (id==FocusEvent.FOCUS_LOST || id==FocusEvent.FOCUS_GAINED) repaint();
            if (mev!=null) {
                final PView mdp=(PView)MOUSE.molecule;
                setTipTiming(id==MOUSE_MOVED && !shift && !ctrl ? 500 : -1,0);
                final Protein p=mdp!=null ? mdp.getProtein() : null;
                final ProteinAlignment align=p==null?null:p.getAlignment();
                final int idx=MOUSE.idx;
                if (id==MOUSE_CLICKED) {
                    isAutoRotation=false;
                    requestFocus();
                    if (p!=null && idx>=0 && align!=null) {
                        new StrapEvent(mdp,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(idx,p).run();
                        final String pdbId=p.getPdbID();
                        if (pdbId==null) return null;
                        final int resNum=p.getResidueNumber(idx);
                        final byte resChain=p.getResidueChain(idx);
                        final byte resIns=p.getResidueInsertionCode(idx);
                        final File file=p.getFile();
                        boolean success=false;
                        for(int iP=align.countRows(); --iP>=0;) {
                            if (p==align.getProtein(iP)) {
                                new StrapEvent(mdp,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(idx,p).run();
                                success=true;
                            }
                        }
                        if (!success) {
                            for(int iP=align.countRows(); --iP>=0;) {
                                final Protein pVis=align.getProtein(iP);
                                if (pVis!=null && file!=null && file.equals(pVis.getFile())) {
                                    final int rnn[]=pVis.getResidueNumber();
                                    final byte rcc[]=pVis.getResidueChain();
                                    final byte rii[]=pVis.getResidueInsertionCode();
                                    if (rnn==null || rcc==null) continue;
                                    for(int iA=rnn.length; --iA>=0;) {
                                        if (iA>=rcc.length || rii!=null && iA>=rii.length) continue;
                                        if (rnn[iA]==resNum && rcc[iA]==resChain && (rii==null || rii.length<=iA || rii[iA]==resIns)) {
                                            new StrapEvent(mdp,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(iA,pVis).run();
                                            break;
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                final int wheel=wheelRotation(ev);
                if (wheel!=0 && !ctrl) {
                    isWheel=true;
                    matrix3D.mult(tmat.unit().rotate('z',wheel*.1f));
                    MOUSE.molecule=null;
                    MOUS_HET.molecule=null;
                    repaint();
                }
                if (id==MOUSE_RELEASED || (id==MOUSE_MOVED&&!shift)) {
                    isWheel=false;
                    if (isDragged) { isDragged=false; repaint(); }
                }
                if (id==MOUSE_MOVED && !shift) {
                    if (findMouse(x-ox,y-oy)) {
                        ChDelay.runAfterMS(EDT, mouseMsgRun,11);
                        mouseMsg[0]=mouseMsg[1]=mouseMsg[2]=null;
                    }
                }
                if (id==MOUSE_MOVED) {
                    Cursor cursor=null;
                    if (p!=null) {
                        final boolean cbActiveSite=isSelctd(mdp.cbActiveSite);
                        for(ResidueSelection sel : p.getResidueSelectionsAt(0L,idx ,idx+1,VisibleIn123.STRUCTURE)) {
                            if (gcp(FEATURE_ACTIVE_SITE,sel)!=null && !cbActiveSite) continue;
                            if (sel==Protein.getMouseOverSelection()) continue;
                            if (sel instanceof PView || (_CCS+"ResidueAnnotation").equals(sel.getClass().getName())) {
                                cursor=ChIcon.getCursor(IC_ANNO,this);
                                continue;
                            }
                        }
                    }
                    setCursor(cursor!=null ? cursor : getCursr('D'));
                }
                if (!but2 && (id==MOUSE_DRAGGED || id==MOUSE_MOVED&&shift)) {
                    if (!isDragged) {
                        prevx=x;
                        prevy=y;
                        isDragged=true;
                        isAutoRotation=false;
                        return null;
                    }
                    if (!isAutoRotation) {
                        tmat.unit();
                        final float
                            dx=(prevx-x)*2f/W,dy=(prevy-y)*2f/H,
                            xtheta=4f*dy, ytheta=4f*dx, ztheta=-4f*dx;
                        if (!ctrl && !shift && !but3 && !but2 || id==MOUSE_MOVED)  {
                            tmat.rotate('x',xtheta);
                            tmat.rotate('y',ytheta);
                        }
                        else {
                            if (ctrl && !shift) tmat.rotate('z',ztheta);
                            if (shift && !ctrl) tmat.scale(1f-dy*2).rotate('z',ztheta);
                        }
                        matrix3D.mult(tmat);
                    }
                    prevx=x;prevy=y;
                    MOUSE.molecule=null;
                    MOUS_HET.molecule=null;
                    repaint();
                }
                if (id==MOUSE_ENTERED || id==MOUSE_EXITED) xyzT=null;
            }
        }
        if (!isDragged) {
            if (id==MOUSE_MOVED || id==MOUSE_ENTERED) mouseIn=true;
            if (id==MOUSE_EXITED && !but1 && (q==this || c!=null && c.getParent()==this)) {
                //final int x1=q==this ? x : c.getX()+x;
                //final int y1=q==this ? y : c.getY()+y;
                if (( x<2||y<2 || x>W-3 || y>H-3)) mouseIn=false;
            }
            if (mouseInLast!=mouseIn) repaint();
        }
        return ev;
    }
    /* <<< Event <<< */
    /* ---------------------------------------- */
    /* >>> Mouse >>> */
    private boolean findMouse(int mx,int my) {
        int distCA=MAX_INT, distHA=MAX_INT;
        short  dp_idx=0;
        Object ha=null, dp=null;
        final Line[] lines=refLines!=null ? refLines.get() : null;
        if (lines==null) return false;
        for(int i=this.countLines; --i>=0;) {
            final Line l=lines[i];
            assertNotNull(l);
            if (l==null || l.style==LETTER) continue;
            final float x=l.x0, y=l.y0;
            final Object mol=l.molecule;
            if (mol!=null) {
                final float dx=x-mx, dy=y-my;
                final int dd=(int)(dx*dx+dy*dy);
                if (dd<8*8 && mol instanceof HeteroCompound) {
                    if (distHA>dd) {
                        distHA=dd;
                        if (MOUS_HET.molecule!=mol) tipLast=null;
                        MOUS_HET.x0=x;
                        MOUS_HET.y0=y;
                        MOUS_HET.m3d=l.m3d;
                        ha=mol;
                    }
                }
                else if (dd<8*8 && mol instanceof PView && distCA>dd) {
                    distCA=dd;
                    MOUSE.x0=x;
                    MOUSE.y0=y;
                    MOUSE.x1=l.x1;
                    MOUSE.y1=l.y1;
                    MOUSE.m3d=l.m3d;
                    dp_idx=l.idx;
                    dp=mol;
                }
            }
        }
        final boolean changed= dp_idx!=MOUSE.idx || dp!=MOUSE.molecule || ha!=MOUS_HET.molecule;
        MOUS_HET.molecule=ha;
        MOUSE.molecule=dp;
        MOUSE.idx=dp_idx;
        return changed;
    }
    public Protein getProteinAtMouse() { final PView dp=(PView)MOUSE.molecule;  return dp!=null ? dp.getProtein() : null; }
    public int getResidueIndexAtMouse() {return MOUSE.molecule!=null ? MOUSE.idx : -1; }
    /* <<< Mouse  <<< */
    /* ---------------------------------------- */
    /* >>> Highlight Residue >>> */
    public void highlightAminoAcid(Protein p,int iA) {

        for(PView dp : vDisplayed.asArray()) {
            if (p==dp.getProtein()) {
                final float[] xyz=p.getResidueCalpha(mx3d(dp));
                if (xyz!=null && iA>=0 && iA*3+2<xyz.length) {
                    highlightAminoAcid(xyzCursor=new float[]{xyz[iA*3],xyz[iA*3+1],xyz[iA*3+2]});
                } else xyzCursor=null;
                break;
            }
        }
    }
    public void highlightAminoAcid(float xyz[]) {
        final Graphics g=getGraphics();
        if (g==null || xyz==null || !isShowing()) return;
        final float xyzT[]=new float[3];
        matrix3D.transformPoints(xyz,xyzT,3);
        if (!Float.isNaN(xyzT[0]+xyzT[1])) {
            // g.setColor(BLACK);
            ChIcon.drawCursor(g, (int)(ox+xyzT[0]-16/2), (int)(oy+xyzT[1]-16/2),16,16,this);
        }
    }

    /* <<<  Highlight Residue <<< */
    /* ---------------------------------------- */
    /* >>> Color >>> */
    private final static Map<Color,Color[]> mapFarbReihe=new WeakHashMap();
    private static Color[] farbReihe(Color f0) {
        final Color f= f0!=null ? f0 : WHITE;
        final int n=DEPTH,n1=n+1;
        Color[] ret=mapFarbReihe.get(f);
        if (ret==null) {
            mapFarbReihe.put(f,ret=new Color[n1]);
            double fR,fG,fB;
            fR=f.getRed();
            fG=f.getGreen();
            fB=f.getBlue();
            for(int i=0;i<n+1;i++) {
                double k=n1-i; k=k/n1;
                ret[i]=new Color((int)(k*fR),(int)(k*fG),(int)(k*fB));
            }
        }
        return ret;
    }
    private final static int DEPTH=64;
    public static Color chain2color(int i) {
        return COLORS[(i + (26-'A'%26)) % COLORS.length][0];
    }
    private final static Color[]
        REDs=farbReihe(RED),
        YELLOWs=farbReihe(YELLOW.darker()),
        WHITEs=farbReihe(WHITE),
        COL_SECSTRU[],
        COL_ATOMS[],
        COLORS[]=new Color['Z'-'A'+1][];
    static {
        final int[] chainRGBS={0xC0D0FF, 0xB0FFB0, 0xFFC0C8, 0xFFFF80, 0xFFC0FF, 0xB0F0F0, 0xFFD070, 0xF08080, 0xF5DEB3, 0x00BFFF, 0xCD5C5C, 0x66CDAA, 0x9ACD32, 0xEE82EE, 0x00CED1, 0x00FF7F, 0x3CB371,
                               0x33448B/*brighter*/,0xBDB76B, 0x336444/*brighter*/, 0x803344/*brighter*/, 0x808000, 0x800080, 0x008080, 0xB8860B, 0xB22222};
        for(int i=chainRGBS.length; --i>=0;) COLORS[i]=farbReihe(new Color(chainRGBS[i]));
    }
    static {
        final int TRANSPARENCY=128;
        Color[][] cc=COL_SECSTRU=new Color[256][];
        Arrays.fill(cc,WHITEs);
        cc['h']=cc['H']=REDs;
        cc['e']=cc['E']=YELLOWs;
        cc=COL_ATOMS=new Color[256][];
        Arrays.fill(cc,farbReihe(new Color(255,255,255,TRANSPARENCY)));
        cc['n']=cc['N']=farbReihe(new Color(0,0,255,TRANSPARENCY));
        cc['o']=cc['O']=farbReihe(new Color(255,0,0,TRANSPARENCY));
        cc['s']=cc['S']=farbReihe(new Color(255,255,0,TRANSPARENCY));
        cc['p']=cc['P']=farbReihe(new Color(0,255,0,TRANSPARENCY));
    }
    public static JComponent docuColorChooser() {
        try {
            ChButton b=new Protein3d().initDP(new PView(), new Protein(), false).colorChooser;
            pcp(HtmlDoc.KEY_DO_NOT_CLONE,"",b);
            pcp(TOP_TEXT,"",b);
            pcp(BOTTOM_TEXT,"",b);
            return b;
        }catch(Exception e) {return null;}
    }
    /* <<< Paint <<< */
    /* ---------------------------------------- */
    /* >>> Class Single Peptide chain  >>> */
    public static class PView implements ProteinViewer {
        private Matrix3D /* _m3d,*/ _m3dPrev;
        private int showNucl=1, iColor,colorHeteros=1;
        private ChButton  colorChooser, butNucl, cbHetero, cbActiveSite;
        private JComponent panToolButtons, proteinLabel;
        private long molecules=-1;
        private Protein _protein;
        private Protein3d p3d;
        public Color getColor() { return iColor>=0 ? COLORS[iColor%COLORS.length][0] : null; }
        /* --- Interface ProteinViewer --- */
        public void setSelectedResidues(long options,boolean[] bb,int atomType, String representation_id) {
            final Protein p=getProtein();
            if (representation_id==ProteinViewer.SELECTION_CURSOR) {
                p3d.highlightAminoAcid(getProtein(),firstTrue(bb));
            } else if (p!=null) {
                final String KEY_SEL="P3D$$S";
                AbstractResidueSelection sel=(AbstractResidueSelection) gcp(KEY_SEL,p);
                if (sel==null) {
                    pcp(KEY_SEL, sel=new AbstractResidueSelection(), p);
                    sel.setName(ResidueSelection.NAME_BACKBONE);
                    sel.setProtein(p);
                    sel.setColor(C(0xFFff00));
                }
                sel.setSelectedAminoacids(bb,0);
                p.addResidueSelection(sel);
                new StrapEvent(this,StrapEvent.RESIDUE_SELECTION_ADDED).setParameter(sel).run();
            }
        }
        public Protein getProtein() { return _protein;}
        public boolean setProtein(Protein p, ProteinViewer inSameView,  Runnable runAfterInstallation, long options) {
            if (p==null) return false;
            final PView pvSame= inSameView instanceof PView ? (PView) inSameView : null;
            Protein3d protein3D= pvSame!=null ? pvSame.p3d : null;
            if (protein3D==null) protein3D=new Protein3d();
            protein3D.initDP(this, p,true);
            return true;
        }
        public Matrix3D getTransformationPreview() { return _m3dPrev;}
        public void transformPreview(Matrix3D m) { _m3dPrev=m;}
        public void transform(Matrix3D m) {}
        public Protein3d getViewerPanel() {return p3d;}
        public ProteinViewer[] getViewersSharingView(){return p3d.vDisplayed.asArray(); }
        public JMenuBar getMenuBar() {return p3d.getMenuBar();}
        public long getPropertyFlags() { return PROPERTY_MULTI_MOLECULE|PROPERTY_MULTI_VIEW;}
        public boolean getSupportsCommand(String cmd) { return COMMANDbiomolecule.equals(cmd); }
        public void interpret(String command0,String id, long options) {
            final String command=command0.trim();
            final String word0=toStrgIntrn(frstTkn(command));
            final int L0=sze(word0);
            if (L0==0) return;
            if (getSupportsCommand(word0)) {
                if (word0==COMMANDbiomolecule) {
                    final long m=atolDecHex(command,L0);
                    this.molecules=m<0 ? 0 : m;
                    if (p3d!=null) p3d.comboBio.s(m<0 ? 0 : 2);
                }
                ChDelay.repaint(p3d,99);
            }
        }
        public Container getControlPanel() {return null;}
        public void dispose() {
            final Protein p=getProtein();
            p3d.removeProtein( p);
            if (p!=null) p.removeProteinViewer(this);
            if (sze(p3d.vDisplayed)==0) {
                p3d.dispose();
                rmFromParent(p3d);
            }
        }
    }
    public PView getView(Protein prot) {
        if (prot!=null)
            for(PView dp:vDisplayed.asArray())
                if (dp.getProtein()==prot) return dp;
        return null;
    }
    int count=0;
}
