package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import javax.swing.*;
import java.util.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
/**HELP

WIKI:PyMOL <i>HOME_PAGE:Pymol</i> is an excellent and widely applied protein
structure viewer. It has been developed by Warren Delano and
many tools and scripts are added by the scientific community.

<br><br><b>Residue selections:</b> Dragging the mouse over a part of
the amino acid sequence in the STRAP alignment creates a
residue selection in Pymol.

The STRAP cursor position refers to another Pymol selection.

Users can refer to these selections to show <b>[S]</b> or hide
<b>[H]</b> representations of residues like <I>sphere</I>,
<I>cartoon</I> or <I>stick</I>.

For example to show all atoms of the amino acid at the current cursor position as spheres
open the corresponding [S] menu of the selection "cursor" and select "spheres".

    <div class="figure">
      <table style="caption-side: bottom">
        <CAPTION><b>Figure: </b>Side pane of Pymol containing a list of loaded proteins and selections.
          [H]Hide style, [S]Show style, [L]Add Label, [C]Change Color </CAPTION>
        <tr><td><i>IMAGE:"pymolSelections"</i></td></tr>
      </table>
    </div>

<br><br><b>Removing a protein from Pymol:</b> If a protein
would be removed in Pymol this would be unnoticed by STRAP.
Therefore it is recommended to use the STRAP user interface and not the Pymol user interface to remove proteins.

There are two possibilities:
<ol>
<li>The STRAP menu item <i>ITEM:charite.christo.strap.ProteinPopup#ACTION_close3d</i></li>
<li>The STRAP object tree which is  visible at the left of the STRAP application frame in advanced user mode.</li>
</ol>

<div class="figure">
<table style="caption-side: bottom">
<CAPTION><b>Figure: </b>Automated installation on Windows, Linux and Mac</CAPTION>
<tr><td><i>JCOMPONENT:Pymol#canInstall(null)!</i></td></tr>
</table>
</div>

<br><br><b>Pymol commands:</b>

Pymol commands  (see http://pymol.sourceforge.net/newman/ref/S1000comref.html) are typed directly into the Pymol-window.

Most important commands:

<ul>

<li><b>reset</b><br>  Zooms the window and clipping planes to cover all objects.</li>

<li><b>ray</b><br> Creates a ray-traced image of the current frame.</li>

<li><b>png my_file.png</b><br> Writes an image file.</li>

<li><b>bg white</b><br>White background.</li>

<li><b>delete myProtein</b><br>Deletes an object. Understands wildcard</li>

</ul>

<br><br><b>Related links:</b> http://www.rubor.de/bioinf/ &nbsp;

http://www.rubor.de/anlagen/PyMOL_Tutorial.pdf

<i>PACKAGE:charite.christo.strap.</i>
<i>SEE_CLASS:ResidueAnnotation</i>
<i>SEE_CLASS:AddAnnotation</i>
@author Christoph Gille, Kristian Rother

 h+="If PyMOLWin.exe does not work replace it by PyMOL.exe.<br>\n"+
                    "If this does not work either comment out all lines with \"#\".<br>\n"+
                    "This will  force STRAP to install Pymol.<br>"+
                    "You can also try  PyMOL.exe instead of PyMOLWin.exe";
*/
public class Pymol implements ProteinViewer,ChRunnable,ActionListener {
    public final static String
        DEFINE_THIS_COLOR="DEFINE_THIS_COLOR",
        PYMOL_COLOR_NAME="$PYMOL_COLOR_NAME";

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Shared fields >>> */
    private static ChExec share_exec;
    private static UniqueList<Pymol> share_vView;
    /* <<< Shared fields <<< */
    /* ---------------------------------------- */
    /* >>> Constructor  >>> */

    public Pymol() {
        ProteinUtils.initCustomizableSettings(this);
        if (share_vView==null) share_vView=new UniqueList(Pymol.class);
        pcp(KEY_FRAME_TITLE, "PyMOL Viewer",this);
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Launch >>> */
    private static Runnable runAfterAllLoaded;
    private long timeExeStarted;
    private static ChFrame fCanInstall;
    private static boolean isWaitingForInput;
    private boolean showStdoutAndStderr, isRunning;
    public void showStdoutAndStderr(boolean yesNo) { showStdoutAndStderr=yesNo;}
    private boolean launch() {
        final boolean isWin=isWindows(), isMac=isMacintosh(), isLinux=isSystProprty(IS_LINUX386);
        ChExec ex=this.share_exec;
        if (ex!=null) ex.dispose();
        this.share_exec=ex=null;
        assertEDT();
        final boolean logPanel=getNextProcessHasLogPanel();
        final Customize cust=Customize.getCustomizableForClass(getClass(), Customize.CLASS_ExecPath);
        success:
        for(int iTry=isWin||isLinux||isMac?2:1; --iTry>=0;) {
            isWaitingForInput=false;
            Object para=cust.getSetting();
            if (iTry==0) (ex=new ChExec(0L).setCommandLines((String[])(cust.getSettings()))).log(logPanel?cust:null);
            else {
                final long options=ChExec.STDOUT|ChExec.STDERR|ChExec.IGNORE_ERROR|ChExec.LOG;
                final File fExe=file(downloadedBinary);
                if (sze(fExe)>0) {
                    makeExecutable(fExe);
                    makeShellScript();
                    para=new Object[]{fExe,"-p","XXXXXXXXXX"};
                    ex=new ChExec(options|(logPanel?ChExec.SHOW_STREAMS:0)).setCommandLineV((Object[])para);
                    final Map<String,String> m=new HashMap();
                    if (isWin) m.put("PATH", ChEnvironment.get("PATH")+";"+fExe.getParent());
                    for(String s : new String[]{"HOME","LD_LIBRARY_PATH" , "DISPLAY" }) m.put(s,ChEnvironment.get(s));
                    m.put("PYMOL_PATH",isMac ?  dirBin()+"/MacPyMOL.app/pymol" : fExe.getParent());
                    ex.addToEnvironement(m);
                }
            }
            if (ex!=null) {
                ex.printlnStdout((ChRunnable)this);
                if (showStdoutAndStderr) ex.showStdoutAndStderr(null);

                ChDelay.newThread(this,"EXEC_PYMOL",ex).start();
                timeExeStarted=System.currentTimeMillis();
                addActLi(this,ex);
                for(int i=0;i<6; i++) {
                    if (ex.couldNotLaunch()) { putln(RED_WARNING+"Pymol: could not launch "); break; }
                    if (ex.exitValue()!=Integer.MIN_VALUE) { putln(RED_WARNING+"Pymol: terminated "); break; }
                    if (ex.isRunning()) {
                        for(int wait=20;--wait>=0;) {
                            puts(ANSI_YELLOW+"Pymol: wait"+ANSI_RESET+" ");
                            sleep(555);
                            if (isWaitingForInput) {
                                putln(GREEN_SUCCESS+"Pymol: accepting commands"); sleep(99);
                                break;
                            }
                        }
                        break success;
                    } else putln(ANSI_MAGENTA+"Pymol:"+ANSI_RESET+" Not yet running");
                    sleep(999);
                }
                interpretNow("from pymol import cmd");
            }
        }
        final String error=ex==null || ex.couldNotLaunch() ? "Pymol not started" :
            ex.hasFinished() ? "could start Pymol but the process already terminated" : null;
        putln("error="+error+"<");

        if (error!=null) {
            final JComponent
                more=vBpnl(
                           getCtrlPanel(ex),
                           !isWin && !isLinux && !isMac ? pnl("Install Pymol: ",TabItemTipIcon.homePage(this)) :
                           " ",
                           pnl("If you already have Pymol, the<br>problem a wrong path: ",cust.newButton()),
                           showStdoutAndStderr ? null : "To see details of the pymol-process<br>  hold CTRL when pressing the pymol-button"
                           ),
                msg=vBpnl(error,canInstall(this), hB("Details ",toggl().actionCollapse(false,more).actionPack(),"#"),more);
            error(0L,error,msg);
            return false;
        }
        isRunning=true;
        ChDelay.newThread(this,"J",null).start();
        this.share_exec=ex;
        return true;
    }
    /* <<< Launch <<< */
    /* ---------------------------------------- */
    /* >>> ChRunnable >>> */
    private static Runnable afterInstHook;
    public Object run(final String id, Object arg) {
        if (id=="EXEC_PYMOL") {
            final ChExec ex=((ChExec)arg);

            ex.run();

            dispose();
        }
        if (id==ChRunnable.INTERPRET_LINE && arg!=null) {

            final String s=toStrg(arg).trim();
            if (s.startsWith("ALL_OBJECTS"))  allObjects=s.substring(11);
            if (s.indexOf("XXXXX")>=0) isWaitingForInput=true;
            if (s.startsWith("PyMOL: ") && s.indexOf("program termination.")>=0) { // brauch ich das?
                share_exec.dispose();
                dispose();
            }
            setTextStatic(s);
        }
        if (id=="J") {
            while(isRunning) {
                final String s=(String)get(0,jobs);
                if (s!=null) {
                    jobs.remove(0);
                    interpretNow(s);
                }
                sleep(33);
                if (share_exec==null || !share_exec.isRunning() ) jobs.clear();
            }
        }
        return null;
    }
    /* <<< ChRunnable <<< */
    /* ---------------------------------------- */
    /* >>> ActionEvent >>> */
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        final Protein protein=getProtein();
        final ProteinAlignment align=protein!=null ? protein.getAlignment() : null;

        if (cmd==ChExec.ACTION_TERMINATED  && isRunning) {
            for(int i=align==null?0:align.countRows(); --i>=0;) {
                Protein p=align.getProtein(i);
                if (p==null) continue;
                                for(ProteinViewer v : (ProteinViewer[])p.getProteinViewers(Pymol.class)) {
                    new StrapEvent(v,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
                }
            }
            isRunning=false;
            jobs.clear();
            if (System.currentTimeMillis()-timeExeStarted<4444) {
                if (fCanInstall==null) fCanInstall=new ChFrame("Pymol").ad(canInstall(this));
                fCanInstall.shw(0).pack();
            }
        }
        if (cmd=="DL") {
            final String zipFile=isWindows() ? "PyMOL099" : isMacintosh() ? "macpymol-0_99rc6" : "pymol-0_99rc6-bin-linux-x86-glibc23";
            InteractiveDownload.downloadFilesAndUnzip(Web.urls(urlOfBinariesForThisOS(),zipFile+".zip.gz"),dirBin());
            final File fZip=file(dirBin()+"/"+zipFile+".zip");
            if (ChZip.extractAll(fZip,dirBin())) {
                if (sze(file(downloadedBinary))>0) launch();
            } else {
                error("Error: could not install PyMOL "+fZip+" "+sze(fZip));
                renamFile(fZip, fZip+".error");
            }
            final Runnable hook=this.afterInstHook;
            if (hook!=null) hook.run();
        }
    }
    /* <<< ActionEvent  <<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private String proteinName;
    private Protein  protein;
    public Protein getProtein() {return protein;}
    public long getPropertyFlags() { return PROPERTY_NAMED_SELECTIONS|PROPERTY_MULTI_MOLECULE|PROPERTY_EXTERNAL_PROCESS|PROPERTY_NOTIFY_ALIGNMENT_CURSOR_DELAY; }
    public boolean setProtein(final Protein p, ProteinViewer inSameView, Runnable runAfterInstallation, long options) {
        interpretNow("print");/* otherwise waitFor does not terminate */
        afterInstHook=runAfterInstallation;
        final File pdb=p!=null ? p.getFileMayBeWithSideChains() : null;
        if (pdb==null) return false;
        for(ProteinViewer v:p.getProteinViewers(Pymol.class))  if (v!=this) return false;

        protein=p;
        if (share_exec==null || !share_exec.isRunning()) launch();
        if (share_exec==null || !share_exec.isRunning()) {
            putln(ANSI_RED+"Could not launch PyMOL"+ANSI_RESET+"\n");
            return false;
        }
        String fn=p.toString().replace('!','_');
        final int l=sze(fn);
        char chain=0;
        if (l>2 && fn.charAt(l-2)==':') { chain=fn.charAt(l-1); fn=fn.substring(0,l-2);}
        fn=delSfx(".ent",delSfx(".pdb",fn));
        if (chain!=0) fn+="_"+chain;
        this.proteinName=fn;
        loadObject(pdb, this.proteinName);
        for(HeteroCompound h: p.getDnaAndRnaAndHeteros()) {
            final String n=h.getNameForPymol();
            final File f=h.getFile();
            if (f==null) continue;
            loadObject(f, h.getNameForPymol());
            sendToPymol("hide everything, "+n+"; show "+ (h.isNucleotideChain()?"cartoon, ":"sticks, ")+n);
        }

        final Matrix3D m3d=p.getRotationAndTranslation();
        if (m3d!=null) transform(m3d);
        //for(String s : Customize.getCustomizableForClass(Pymol.class, Customize.CLASS_InitCommands).getSettings()) interpret(s,null);
        ChDelay.runAfterMS(EDT,runAfterAllLoaded,3333);
        share_vView.add(this);
        thread_setWndwStateT('I',0,0,"PyMOL Tcl/Tk GUI").start();
        return true;
    }

    /* <<< Protein  <<< */
    /* ---------------------------------------- */
    /* >>> CommandInterpreter >>> */
    private final static Vector<String> jobs=new Vector();
    private final static String supported[]={COMMANDcenter,COMMANDspheres,COMMANDlines,COMMANDsticks,COMMANDribbons,COMMANDcolor,COMMANDsurface};
    public boolean getSupportsCommand(String cmd) { return idxOfStrg(cmd,supported)>=0; }

    public void interpret(String command,String id, long options) {
        final String word0=toStrgIntrn(toStrgTrim(frstTkn(command)));
        final int L0=sze(word0);
        if (L0==0) return;
        String send=idxOfStrg(word0,COMMANDS)>=0 ? null : rplcToStrg(PROTEIN,this.proteinName,  rplcToStrg("COLOR","Color"+id,command));
        if (getSupportsCommand(word0)) {
            final boolean off=command.endsWith(" off");
            final String pfx=off ?"hide " : "hide everything, RESIDUES; show ";
            if (word0==COMMANDdots)         send=pfx+"dots,RESIDUES\n";
            else if (word0==COMMANDspheres) send=pfx+"spheres,RESIDUES\n";
            else if (word0==COMMANDmesh)    send=pfx+" mesh,RESIDUES\n";
            else if (word0==COMMANDsurface) send=pfx+"surface,RESIDUES\n";
            else if (word0==COMMANDlines)   send=pfx+"lines,RESIDUES\n";
            else if (word0==COMMANDsticks)  send=pfx+"sticks,RESIDUES\n";
            else if (word0==COMMANDribbons) send=pfx+"ribbon,RESIDUES\n";
            else if (word0==COMMANDcartoon) send=pfx+"cartoon,RESIDUES\n";
            else if (word0==COMMANDcolor)   send="color "+delPfx(word0,command)+" RESIDUES";
        }
        if (sze(send)>0) {
            String s=rplcToStrg(RESIDUES,id, send);
            putln(ANSI_RED+"Pymol>"+ANSI_RESET+s+ANSI_RED+"<"+ANSI_RESET);
            sendToPymol(s);
        }
    }
    private void sendToPymol(final String cmd0) {
        if (cmd0==null) return;
        final String cmd=rplcToStrg(PROTEIN,this.proteinName,cmd0).trim();
        String startsWith=null;
        for(int i=0; i<3; i++) {
            final String s=i==0?"select cursor " : i==1 ? "select selected " : "cmd.enable(";
            if (cmd.startsWith(s)) startsWith=s;
        }
        if (startsWith!=null) {
            for(Object job:jobs.toArray()) {
                final String s=(String)job;
                if (s.startsWith(startsWith) || "reset".equals(cmd) && "reset".equals(s))  jobs.remove(s);
            }
        }
        jobs.add(cmd);
    }
    private static String interpretNow(String s) {
        final ChExec ex=share_exec;
        if (ex!=null) {
            try {
                // debug("interpretNow "); debug(s);
                share_exec.write(s);
                share_exec.write("\r\n");
            } catch(IOException e){}
        }
        return "";
    }
    /* <<< CommandInterpreter <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    public void dispose() {
        interpretNow("delete "+this.proteinName);
        interpretNow("delete "+this.proteinName+"_*");
        share_vView.remove(this);
        new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> Object Management >>> */
    private static String allObjects;
    public String[] getAllObjects() {
        if (share_exec==null || !share_exec.isRunning()) return NO_STRING;
        allObjects=null;
        /* ALL_OBJECTS ['hs_EscherichiaColi'] */
        interpretNow("print 'ALL_OBJECTS', cmd.get_names('objects')");
        for(int i=0; i<100 && allObjects==null; i++) sleep(20);
        return allObjects!=null ? splitTokns(0L, delPfx("[",delSfx("]",allObjects.trim())), ",'") : NO_STRING;
    }
    public void loadObject(File f, String n0) {
        final String path=toStrg(f);
        if (path!=null && n0!=null) {
            final String n=rplcToStrg(PROTEIN,this.proteinName, n0);
            interpretNow("if not '"+n+"' in cmd.get_names('objects'): cmd.load( '"+path.replace('\\','/')+"' , '"+n+"')");
        }
    }
    private void setTextStatic(final String s) {
        final Protein protein=getProtein();
        if (protein==null) return;
        final ProteinAlignment align=protein.getAlignment();
        if (s!=null && s.startsWith("You clicked") && align!=null) {
            final int
                i1=s.indexOf('/'),
                i2=s.indexOf('/',i1+1),
                i3=s.indexOf('/',i2+1),
                i4=s.indexOf('/',i3+1),
                L=sze(s);
            if (i1>0 && L>i1+1 && i3+1<L && i4+1<L) {
                final String s1=s.substring(i1+1);
                char chain=s.charAt(i3+1);
                if (chain=='/') chain=0;
                final int num=atoi(s,nxt(DIGT,s,i4,MAX_INT));
                for(int i=align.countRows(); --i>=0;) {
                    final Protein p=align.getProtein(i);
                    if (p==null) continue;
                    for(ProteinViewer pv : p.getProteinViewers(Pymol.class)) {
                        final Pymol pm=(Pymol)pv;

                        if (s1.startsWith(pm.proteinName+"/")) {
                            final int iA=p.residueNumberAndChain2residueIdx(num,chain,' ');

                            if (iA>=0) new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(iA,p).run();
                        }
                    }
                }
            }
        }
    }
    /* <<< Object Management <<< */
    /* ---------------------------------------- */
    /* >>> Customize >>> */
    public final static String FIRST_Aaa="$FIRST_Aaa";
    public static String getCustomizableExamples() {
        return
            "hide everything, RESIDUES; show dots,RESIDUES\n"+
            "hide everything, RESIDUES; show spheres,RESIDUES\n"+
            "hide everything, RESIDUES; show mesh,RESIDUES\n"+
            "hide everything, RESIDUES; show surface,RESIDUES\n"+
            "hide everything, RESIDUES; show lines,RESIDUES\n"+
            "hide everything, RESIDUES; show sticks,RESIDUES\n"+
            "hide everything, RESIDUES; show ribbon,RESIDUES\n"+
            "hide everything, RESIDUES; show cartoon,RESIDUES\n"+
            "color magenta, RESIDUES\n"+
            DEFINE_THIS_COLOR+", color "+PYMOL_COLOR_NAME+",+ RESIDUES\n"+
            sATOM_TYPE_CA+" label (RESIDUES), \"%s\" % (resn)\n"+
            sATOM_TYPE_CA+" label (RESIDUES), \""+FIRST_Aaa+"\"";
    }
    public static String getCustomizableInitCommands() {
        return
            "hide everything,PROTEIN\n"+
            "show cartoon, PROTEIN\n"+
            "show spheres , PROTEIN and Het and not resn hoh and not resn hho and not  resn ohh and not resn h2o\n"+
            "show lines, (PROTEIN and name ca,c,n)\n"+
            "hide everything,PROTEIN and (resn hoh or resn hho or resn ohh or resn h2o) ";
    }
    public static String getCustomizableLateCommands() {
        return
            "select allWater , resn hoh or resn hho or resn ohh or resn h2o\n"+
            "select allHeteros , Het and not allWater\n"+
            "select allDNA , resn DA or resn DC or resn DG or resn DT  \n"+
            "select allRNA , resn A or resn C or resn G or resn T or resn U  \n"+
            "set selection_width=10\n"+
            "reset";
    }

    public static String getCustomizableExecPath() {
        return
            isWindows() ? "%ProgramFiles%/Delano%20Scientific/PyMOL/PyMOLWin.exe -p  XXXXXXXXXX \n"+"pymol -p XXXXXXXXXX \n" :
            isMacintosh() ?
            "/Applications/PyMOL.app/Contents/MacOS/PyMOL -p\n"+
            "/Applications/MacPyMOL.app/Contents/MacOS/MacPyMOL -p\n"+
            "/Applications/PyMOLX11Hybrid.app/Contents/MacOS/PyMOL -p" :
            "pymol -p XXXXXXXXXX \n";
    }
    /* <<< Customize <<< */
    /* ---------------------------------------- */
    /* >>> TransformsCoordinates  >>> */
    private final Map<String,Matrix3D>mapMatrix0=new HashMap();
    public void setMatrix0(String object, Matrix3D m) { mapMatrix0.put(rplcToStrg(PROTEIN,this.proteinName,object),m);}
    public void transform(Matrix3D m3d) {
        final Protein p=getProtein();
        //final Matrix3D m3d=p.getRotationAndTranslation();
        final String loaded[]=getAllObjects(), namePfx=this.proteinName+"_";
        final Matrix3D matrix=new Matrix3D();
        final HeteroCompound[] dna=p.getDnaAndRnaStructures(), het=p.getHeteroCompounds();
        for(String l:loaded) {
            boolean isHet=false;
            for(HeteroCompound h : dna) isHet=isHet||l.equals(h.getNameForPymol());
            for(HeteroCompound h : het) isHet=isHet||l.equals(h.getNameForPymol());
            if (isHet || l.equals(this.proteinName) || l.startsWith(namePfx)) {
                sendToPymol("matrix_reset "+l);
                final Matrix3D m0=mapMatrix0.get(l);
                if (m3d!=null || m0!=null) sendToPymol("cmd.transform_object('"+l+"',( "+matrix.unit().mult(m0).mult(m3d).toPymol()+" ) )");
            }
        }
        sendToPymol("reset");
    }
    /* <<< TransformsCoordinates <<< */
    /* ---------------------------------------- */
    /* >>> Installation >>> */
    private void makeShellScript() {
        if (sze(file(downloadedBinary))==0) return;
        final String
            downlDir= dirBin()+(isWindows() ? "\\PyMOL099\\" : "/pymol/"),
            txt=
            isWindows() ? "set PYMOL_PATH="+downlDir+"\r\nset PATH=%PATH%;%PYMOL_PATH%\r\n"+downloadedBinary +" %* \r\n":
            isMacintosh() ? "#!/bin/bash\nexport PYMOL_PATH="+dirBin()+"/MacPyMOL.app/pymol\n"+downloadedBinary +" $* & \n":
            " export PYMOL_PATH="+downlDir+"\n"+downloadedBinary +" $* \n";
        for(String dir: new String[]{ downlDir , dirHome()+"/Desktop/strap_utils"}) {
            if (!file(dir).isDirectory()) continue;
            final File f=file(dir+"/Pymol."+ (isWindows() ? "bat" : isMacintosh() ? "command" : "sh"));
            wrte(f,txt);
        }
    }
    private final String downloadedBinary=dirBin()+(
                                                    isWindows() ? "\\PyMOL099\\PyMOL.exe" :
                                                    isMacintosh() ? "/MacPyMOL.app/Contents/MacOS/MacPyMOL" :
                                                    "/pymol/pymol.exe");
    public static JComponent canInstall(Pymol p) {
        return pnl("STRAP can install Pymol automatically (13MB)", new ChButton("DL").t("Download & install").li(p));
    }
    /* <<< Installation <<< */
    /* ---------------------------------------- */
    /* >>> ProteinViewer methods >>> */

    public Container getControlPanel() {  return getCtrlPanel(share_exec); }
    public Component getViewerPanel() {return null;}
    public ProteinViewer[] getViewersSharingView() { return share_vView.asArray(); }
    public JMenuBar getMenuBar() { return null; }
    /* <<< ProteinViewer methods <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */

    public void setSelectedResidues(long options,final boolean[] bb, final int atomType, final String representation_id) {
        final Protein p=this.protein;
        if (sze(bb)==0) {
            if (sze(representation_id)>0) sendToPymol("delete "+representation_id);
        } else {
            final ByteArray sb=new ByteArray(333);
            for(char iChain=' '; iChain<128; iChain++) {
                int countAA=0;
                for(int iA=bb.length; --iA>=0;) {
                    if (bb[iA] && p.getResidueChain(iA)==iChain) {
                        final int resNum=p.getResidueNumber(iA);
                        if (resNum==MIN_INT) continue;
                        if (countAA==0) {
                            if (sb.getEnd()>0) sb.append(" or ");
                            sb.append('/',3);
                            if (is(LETTR_DIGT,iChain)) sb.append(iChain);
                            sb.append('/');
                        } else sb.append(',');
                        sb.append(resNum);
                        final char insert=(char)p.getResidueInsertionCode(iA);
                        if (insert>0) sb.append(insert);
                        countAA++;
                    }
                }
                if (countAA>0)  sb.append('/').append(atomType==ATOM_CA ? "ca" :  atomType==ATOM_CB ? "cb" :"");
            }
            if (sb.getEnd()>0) {
                final String cmd="select "+representation_id+" = (PROTEIN and ( "+ sb.newString() +") )\r\n";
                sendToPymol(cmd);
                sendToPymol("cmd.enable(\""+representation_id+"\")");
            }
        }
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

}

// http://biodesign.asu.edu/labs/assets/lindsay/molvis/MOL.html
/*
if 'pdb1a34_mtx42' in cmd.get_names('objects'): matrix_reset pdb1a34_mtx42
*/
