package charite.christo.protein;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.awt.*;
import static charite.christo.ChUtils.*;
/**
   Instances of StrapEvent are used to notify changes within the alignment.
   Therefore StrapEvent is an important class for developers of plugins.
   StrapEvent's are broadcasted using
   <i>JAVADOC:StrapEvent#run()</i>.
   When a class is interested in receiving StrapEvent's it must implement <i>JAVADOC:StrapListener</i>.
   Plugins are usually automatically registered and developers do not need to take care that the
   method <i>JAVADOC:StrapAlign#addListener(StrapListener)</i> is called.
   Using the debug menu all listeners of
   StrapEvent's can be viewed and all StrapEvent's watched.
   These are important debugging tools.
   Each StrapEvent has a type such as <i>JAVADOC:StrapEvent#CURSOR_CHANGED_PROTEIN</i> obtainable with <i>JAVADOC:StrapEvent#getType()</i>
   and may have associated objects <i>JAVADOC:StrapEvent#oo()</i>
   Events are received via the method <i>JAVADOC:StrapListener#handleEvent(StrapEvent)</i>
   The instances of StrapEvent also provides a reference to the <i>JAVADOC:StrapAlign</i> object <i>JAVADOC:StrapEvent#getStrapAlign()</i>

   <b>Performance:</b> The same StrapEvent instance can be reused several times.
   See <i>JAVADOC:StrapListener#handleEvent(StrapEvent)</i>
   @author Christoph Gille
*/
public class StrapEvent implements Runnable {
    private int typeOfEvent;

    /**
        Dispatch the event to all Listeners. All registered
        <i>JAVADOC:StrapListener</i> -objects will be sent the event
        via the method <I>StrapListener#handleEvent(StrapEvent)</I>.
        The run method is thread-safe.
    */
/**
       @param source Source object, important because the event is never send back to its source
       @param typeOfEvent of event e.g. StrapEvent.RESIDUES_CHANGED_COLUMN
    */

    public StrapEvent(Object source,int typeOfEvent) {
        setSource(source);
        this.typeOfEvent=typeOfEvent;
    }
    private static ProteinAlignment proteinAlignment;
    public static void setProteinAlignmentForAll(ProteinAlignment pa) { proteinAlignment=pa;}
    public void run() {
        if (proteinAlignment!=null) proteinAlignment.dispatch(this);
    }
    /**
       Get the type of the event such as StrapEvent.PROTEINS_KILLED
    */

    public final int getType() {return typeOfEvent;}
    public final static int
        FLAG_ALIGNMENT_CHANGED=1<<16,
        FLAG_RESIDUE_TYPES_CHANGED=1<<17,
        FLAG_ROW_HEADER_CHANGED=1<<18,
        FLAG_ATOM_COORDINATES_CHANGED=1<<19,
        FLAG_CHILDS_OF_PROTEINS_CHANGED=1<<20,

        CURSOR_MOVED_WITHIN_PROTEIN=1,
        CURSOR_MOVED_WITHIN_PROTEIN_DELAYED=2,
        CURSOR_CHANGED_PROTEIN=3,
        CURSOR_CLICKED=4,

        STRAP_LISTENER_ADDED=9,

        PROTEIN_RENAMED=11|FLAG_ROW_HEADER_CHANGED,
        PROTEIN_ICON_CHANGED=12|FLAG_ROW_HEADER_CHANGED,

        OBJECTS_SELECTED=13,

        RESIDUE_TYPES_CHANGED=20 | FLAG_ALIGNMENT_CHANGED|FLAG_RESIDUE_TYPES_CHANGED,
        CHARACTER_SEQUENCE_CHANGED=21 | FLAG_ALIGNMENT_CHANGED|FLAG_RESIDUE_TYPES_CHANGED,
        NUCL_TRANSLATION_CHANGED=22 | FLAG_ALIGNMENT_CHANGED | FLAG_RESIDUE_TYPES_CHANGED,
        ATOM_COORDINATES_CHANGED=23 | FLAG_ATOM_COORDINATES_CHANGED,
        PROTEIN_3D_MOVED=24 | FLAG_ATOM_COORDINATES_CHANGED,
        ALIGNMENT_CHANGED=25|FLAG_ALIGNMENT_CHANGED,

        ORDER_OF_PROTEINS_CHANGED=30 |FLAG_ALIGNMENT_CHANGED|FLAG_ROW_HEADER_CHANGED,
        PROTEINS_HIDDEN=31 |FLAG_ALIGNMENT_CHANGED|FLAG_ROW_HEADER_CHANGED,
        PROTEINS_SHOWN=32 |FLAG_ALIGNMENT_CHANGED|FLAG_ROW_HEADER_CHANGED,
        PROTEINS_KILLED=33 |FLAG_ALIGNMENT_CHANGED|FLAG_ROW_HEADER_CHANGED,
        PROTEINS_ADDED=34 |FLAG_ALIGNMENT_CHANGED|FLAG_ROW_HEADER_CHANGED,

        VALUE_OF_RESIDUE_CHANGED=50,
        VALUE_OF_ALIGN_POSITION_CHANGED=51,
        VALUE_OF_RESIDUE_CHANGED_COLOR=52,
        VALUE_OF_ALIGN_POSITION_CHANGED_COLOR=53,
        VALUE_OF_PROTEIN_CHANGED=54,
        PROTEIN_PROTEIN_DISTANCE_CHANGED=55,

        RESIDUE_SELECTION_CHANGED_COLOR=60|FLAG_CHILDS_OF_PROTEINS_CHANGED,
        RESIDUE_SELECTION_CHANGED=61,
        RESIDUE_SELECTION_ADDED=62|FLAG_CHILDS_OF_PROTEINS_CHANGED,
        RESIDUE_SELECTION_DELETED=63|FLAG_CHILDS_OF_PROTEINS_CHANGED,
        RESIDUE_SELECTION_RENAMED=64|FLAG_CHILDS_OF_PROTEINS_CHANGED,

        HETERO_OR_NUCLEOTIDE_STRUCTURE_ADDED_OR_REMOVED=101|FLAG_CHILDS_OF_PROTEINS_CHANGED,
        PLUGIN_CHANGED=102,

        BACKUP_WRITTEN=103,

        AA_SHADING_CHANGED=110,
        BACKGROUND_CHANGED=111,
        FONT_CHANGED=112,

        ALIGNMENT_SCROLLED=120,
        COMPUTATION_STARTED=121,

        USER_PROFILE_CHANGED=122,

        IMAGE_DOWNLOADED=123,

        PROGRESS=124,
        PRINT_COMPUTATION_STARTED=130,

        PROTEIN_VIEWER_LAUNCHED=132|FLAG_CHILDS_OF_PROTEINS_CHANGED,
        PROTEIN_VIEWER_CLOSED=133|FLAG_CHILDS_OF_PROTEINS_CHANGED,
        PROTEIN_VIEWER_PICKED=134,
        PROTEIN_XREF_CHANGED=136,

        NEED_SCROLL_DOWN_ALIGNMENT=137,
    //BIOMOLECULES_SPECIFIED=140,
        SCRIPT_LINE=150;

    private Object source;
    public StrapEvent setSource(Object o) { source=o; return this;}
    /**
       get the object where the event was formed
    */
    public final Object getSource() {return source;}
    // ----------------------------------------
    private Object objects[];
    /** get associated objects */
    public Object[] getParameters() { return objects!=null ? objects : NO_OBJECT; }
    public boolean hasParameters() { return objects!=null;}
    /**  set associated objects*/
    public final StrapEvent setParameters(Object[] oo) { objects=oo; return this;}
    public final StrapEvent setParameter(Object o) { objects=new Object[]{o}; return this;}
    public final StrapEvent setAminoAcid(int res, Protein p) { setParameters(new Object[]{new Integer(res),p}); return this;}
    // ----------------------------------------
    /**
       A text-representation of the event
    */
    @Override public String toString() {
        final StringBuffer sb=new StringBuffer("StrapEvent.")
            .append(getFinalStaticFields(StrapEvent.class).get(longObjct(getType())))
            .append(": source=").append(shrtClasNam(getSource())).append(" oo={");
        final Object oo[]=getParameters();
        for(Object o:oo) sb.append(o).append('\n');
        sb.append('}');
        return sb.toString();
    }
    // ----------------------------------------
    public ProteinAlignment getProteinAlignment() {return proteinAlignment;}

}
