package charite.christo.protein;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import charite.christo.*;
import charite.christo.interfaces.*;
import java.io.File;
import javax.swing.*;
import java.util.*;
import java.awt.*;
import java.awt.event.*;

/**HELP

VMD is an excellent  molecular visualization program for displaying, animating,
and analyzing large biomolecular systems  (see WIKI:Visual_molecular_dynamics &nbsp;
http://www.ks.uiuc.edu/Research/vmd/).

<br><br><b>Compatibility:</b>
This wrapper class  works on X-Windows but not on Windows.
<br><br><b>Requirements:</b>
Vmd  and TCL-TK.

<br><br><b>Principle of inter-process communication:</b>
Commands are passed to VMD via the TK/TCL-interface wish.
VMD is identified by a TK-window identifier, usually "tk". If a second
Vmd is started it gets the identifier "tk #2".
<br>
STRAP starts a TK-shell, the program <b><I>wish</I></b>, which appears in a
little window. This window is essential for the interprocess
communication but is not needed by the user and can therefore be
iconified.
<br><br><b>Mouse selection:</b>
When the mouse pointer is dragged in STRAP over a series of residues these will be selected in Vmd.
To highlight the selection in Vmd a command is send.
This command can be edited/deleted in the customize dialog <i>ICON:IC_CUSTOM</i>.

   @author Christoph Gille
*/
public class Vmd implements ProteinViewer,ActionListener, ChRunnable, HasControlPanel  {
    public Vmd() {
        ProteinUtils.initCustomizableSettings(this);
    }
    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection  >>> */
    public void setSelectedResidues(long options,boolean[] bb,int atomType,String representation_id) {
        final StringBuffer sb=new StringBuffer("resid ");
        char chain=0;
        for(int i=0;i<bb.length;i++) {
            if (bb[i]) {
                chain=(char)protein.getResidueChain(i);
                int iPos=protein.getResidueNumber(i);
                sb.append(" ").append(iPos);
            }
        }
        if (is(LETTR_DIGT,chain)) sb.append(" and chain ").append(chain);
        final String d=display();
        Wish.wish(d,tkWindow,"catch { mol delrep $"+representation_id+" $"+protVariable()+" } ");
        Wish.wish(d,tkWindow,"set "+representation_id+" [molinfo $"+protVariable()+" get numreps];");
        Wish.wish(d,tkWindow,"mol selection "+sb);
        Wish.wish(d,tkWindow,"mol addrep $"+protVariable()+";");

    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Disposable >>> */
    private final static UniqueList<Vmd> vVmds=new UniqueList(Vmd.class);
    public void dispose() {
        Wish.wish(display(),tkWindow," mol delete $"+protVariable());
        vVmds.remove(this);
        new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_CLOSED).run();
        if(myExec!=null) myExec.dispose();
    }
    /* <<< Disposable <<< */
    /* ---------------------------------------- */
    /* >>> Display >>> */
    private static String  tkWindow;
    private String display() { return tfDisplay!=null?toStrg(tfDisplay) : "";}
    private static ChTextField tfDisplay;
    public Container getSharedControlPanel() {
        if (shared==null) {
            tfDisplay=new ChTextField().cols(60,true,true);
            shared=vB("Optionally enter an X-display (Example 141.42.56.114:0.0)", tfDisplay);
        }
        return shared;
    }
    /* <<< Display <<< */
    /* ---------------------------------------- */
    /* >>> Control Panel >>> */
    private JComponent  panel, shared;
    public Container getControlPanel() {
        if (panel!=null) return panel;
        ChButton  butConnect=new ChButton("CLOSE_ALL").t("close all and dis-connect").li(this);
        Container pButtons=pnl(butConnect,panHS(this));
        panel=newCNSEW(Wish.getControlPanel(display()),null,pButtons,null,null);
        panel.setPreferredSize(dim(500,500));
        return panel;
    }
    /* <<< Control Panel <<< */
    /* ---------------------------------------- */
    /* >>> Execute >>> */
    private static ChExec myExec;
    public void actionPerformed(ActionEvent ev) {
        final String cmd=ev.getActionCommand();
        if (cmd==ChExec.ACTION_TERMINATED) {
            new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_CLOSED).setParameter(this).run();
        }
        if (cmd=="CLOSE_ALL") {for(Vmd v:vVmds.asArray())v.dispose();}
    }
    private boolean launch() {
        final int iOption=ChMsg.option("You can either connect to a running VMD <br>or start a new VMD process",
                                       new String[] {"Launch new Vmd process","Connect to running vmd","Cancel"});
        if (iOption<0||iOption==2) return false;
        final String display=display();
        if (iOption==0){
            if (myExec!=null) myExec.dispose();
            String cmd= sze(display)==0 ? "vmd" : "xterm -display "+display+" -e vmd";
            myExec=new ChExec(ChExec.STDOUT | ChExec.STDERR | ChExec.IGNORE_ERROR).printlnStdout((ChRunnable)this).setCommandLine(cmd);
            addActLi(this,myExec);
            new Thread(myExec,"exec Vmd").start();
            myExec.waitFor();
            final String error= myExec.hasFinished() ? "could launch Vmd but the prosess  already terminated" :
                !myExec.isRunning() ? "could not launch Vmd" : null;
            if (error!=null) {
                error(0L,"", vBpnl("Error in "+getClass()+": "+error, getCtrlPanel(myExec)));
                return false;
            }
            sleep(2222);
        }
        String oldWish=rplcToStrg(" #","#",Wish.getWinfo(display)+"");
        StringTokenizer t=new StringTokenizer(oldWish," \n{}");
        int n=t.countTokens();
        if (n<2) { error("Cannot connect to VMD.\nNo TK-windows detected"); return false;}
        String tkWindows[]=new String[n];
        for(int i=0;i<n;i++) tkWindows[i]=rplcToStrg("#"," #",t.nextToken());
        ChCombo comboWindow=new ChCombo(tkWindows);
        for(int i=0;i<n;i++) if (tkWindows[i].startsWith("tk") || tkWindows[i].startsWith("vmd")) comboWindow.s(i);
        Container  msg=vB(
                          "Going to connect to a running VMD.",
                          " ",
                          "If you have not started VMD yet you can ",
                          "start it now and  and open this dialog again.",
                          " ",
                          "Important: when absolute protein files are used (default)",
                          " vmd does not need to be  started in the project directory !",
                          " ",
                          "Select the TK-window, VMD is running in.\n(usually 'TK'",
                          comboWindow
                          );
        final int iOp=ChMsg.option(msg,new String[]{"Connect","Cancel"});
        if (iOp<0 || iOp==1) return false;
        tkWindow=comboWindow.getSelectedItem().toString();
        return true;
    }
    /* <<<  Execute<<< */
    /* ---------------------------------------- */
    /* >>> Protein >>> */
    private Protein protein=null;
    public Protein getProtein() {return protein;}
    public boolean setProtein(Protein p, ProteinViewer inSameView, Runnable runAfterInstallation, long options) {
        final File pdb=p!=null ? p.getFileMayBeWithSideChains() : null;
        if (pdb==null) return false;
        protein=p;
        for(ProteinViewer v:p.getProteinViewers(Vmd.class))  if (v!=this) return false;
        if ((myExec==null || !myExec.isRunning()) &&  !launch()) return false;
        Wish.wish(display(),tkWindow,"set "+protVariable()+" [ mol load pdb "+pdb+"]");
        vVmds.add(this);
        //for(String s : getCustomSettings(getClass(), Customize.CLASS_InitCommands)) interpret(s,null);
        return true;
    }
    /* <<< Protein <<< */
    /* ---------------------------------------- */
    /* >>> interpret >>> */
    private String protVariable() { return "strapProtein_"+variableName(protein.toString());}
    private static String variableName(String s) {
        byte bb[]=s.getBytes();
        for(int i=0;i<bb.length;i++) {
            char c=(char)bb[i];
            if (!is(LETTR,c) && !is(DIGT,c))
                bb[i]=(byte)'_';
        }
        return new String(bb);
    }
    public boolean getSupportsCommand(String cmd) { return false; }
    public void interpret(String str,String representation_id, long options) {
        String s=rplcToStrg(PROTEIN,"$"+protVariable(),str);
        if (representation_id==null)  {
            Wish.wish(display(),tkWindow,s);
        }
        else {
            //        String s1="set "+representation_id+" [molinfo $"+protVariable()+" get numreps];\n  mol addrep $"+protVariable()+";";
            Wish.wish(display(),tkWindow,"mol "+s);
            Wish.wish(display(),tkWindow,"mol modrep $"+representation_id+" $"+protVariable());
        }
    }
    /* <<< Interpret <<< */
    /* ---------------------------------------- */
    /* >>> Customize >>> */
    public long getPropertyFlags() { return PROPERTY_NAMED_SELECTIONS|PROPERTY_MULTI_MOLECULE|PROPERTY_EXTERNAL_PROCESS|PROPERTY_TK|PROPERTY_NOTIFY_ALIGNMENT_CURSOR_DELAY; }
    public static String getCustomizableExamples() {
        return
            "representation VDW 1 8\n"+
            "representation CPK 1 0.3 8 6\n"+
            "representation Lines 1\n"+
            "representation Bonds 0.3 6\n"+
            "representation DynamicBonds 1.6 0.3 6\n"+
            "representation Licorice 0.3 8 6\n"+
            "representation Trace 0.3 6\n"+
            "representation Tube 0.3 6\n"+
            "representation Cartoon 2.1 12 5\n"+
            "representation Surf 1.4 0\n"+
            "representation Dotted 1 8\n"+
            "representation Solvent 0 7 1\n"+
            "material Transparent\n"+
            "material Opaque";
    }
    /*
       if (v instanceof Vmd) commands[0]=get(0,Vmd.selectionCommand().getSettings());
    */
    /*
    private static Customize selectionCommand;
    public static Customize selectionCommand() {
        if (selectionCommand==null) {
            String ss[]={"representation VDW 1 8"};
            selectionCommand=new Customize("VMD_mouseSelectionCommand",ss,"Rendering command that is applied for the mouse selection\nwhen the mouse is dragged in the alignment pane.\nThe goal is to highlight the mouse selection in the alignment pane also in Vmd.",Customize.TEXT_LINES);
        }
        return selectionCommand;
    }
    */
    public static String getCustomizableInitCommands() {
        return
            "mol selection  not (protein or nucleic or water)\n\n"+
            "mol material Opaque\n\n"+
            "mol addrep PROTEIN\n\n"+
            "mol modstyle 1 PROTEIN VDW 1 8\n\n"+
            "mol color Name\n\n"+
            "mol selection  all\n\n"+
            "mol material Opaque\n\n"+
            "mol addrep PROTEIN\n\n"+
            "mol modstyle 2 PROTEIN cartoon 2.1 12 5\n\n"+
            "mol color Chain\n\n"+
            "mol delrep 0 PROTEIN";
    }
    /* <<< Customize <<< */
    /* ---------------------------------------- */
    /* >>> Cursor  >>> */
    /*
      Info) index: 559
      Info) resname: ALA
      Info) resid: 76
    */
    private int residueIndexCursor=-1;
    public Object run(String id, Object arg) {
        if (id==ChRunnable.INTERPRET_LINE && arg!=null){
            int num=0;
            char chain=0;
            residueIndexCursor=protein.residueNumberAndChain2residueIdx(num,chain,' ');
            new StrapEvent(this,StrapEvent.PROTEIN_VIEWER_PICKED).setAminoAcid(residueIndexCursor,getProtein()).run();
        }
        return null;
    }
    public void transform(Matrix3D m){}
    public Component getViewerPanel() {return null;}
    public ProteinViewer[] getViewersSharingView() { return null; }
    public JMenuBar getMenuBar() { return null; }
}
/*
  Christoph,
  If you want to get residues selected by VMD, a much better way than
  reading the VMD text output is to use a Tcl callback for this purpose:
  http://www.ks.uiuc.edu/Research/vmd/current/ug/node140.html#7338
  You'll want to use the vmd_pick_atom callback.
  See the "mol_weight" example script in that section to see how it works.
  John
*/
