package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import java.awt.*;

import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import java.lang.ref.Reference;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
/**HELP

A dot-plot is a correlogram of two sequence WIKI:Dot_plot_(bioinformatics).

Each dot in the pane represents a pair of residues from either
protein.  The brightness of the dot is mainly determined by the
similarity of both amino acids according to the blosum62 matrix.  A
number of positions up and downstream are taken into account
with declining weight.

<br><br><b>GUI-controls:</b>
<ul>
<li><b>Slider:</b> The width of the low pass filter with Co-sinus weights.</li>
<li><b>Slider:</b> The brightness of the entire plot.</li>
<li><b>Combo box:</b> The proteins of the X axis and Y axis.</li>
<li><b>Toggle <i>CB:DialogDotPlot#CB_LAB_Trace</i>:</b> View the trace of the sequence alignment.</li>
</ul>

<br><br><b>Anchor points:</b><br>

Anchor points can be set by clicking into the pane and are marked
in the multiple sequence alignment and the scrollbar.  The idea is
to chose an anchor point on a bright oblique line with high
sequence similarity.  Subsequently, one can introduce gaps into the
proteins until the corresponding marks coincide in the multiple
sequence alignment pane.

<br><br><b>Periodicity</b> Sequence periodicity becomes evident by more than one white oblique trace. See PDB:1g61 for an example.

<br><br> Save ink while printing: <i>JCOMPONENT:ChUtils#togWhiteBG()!</i>

<br><br><b>Related links:</b> PUBMED:15724278  MOL_TOOLKIT:dnadot
@author Christoph Gille
*/
public class DialogDotPlot extends AbstractDialogJPanel implements Disposable, HasControlPanel, StrapListener, PaintHook, ChProcessEvent  {
    private static int WEIGHTS_AVERAGE_VALUE=1024;

    private Backtranslate.Codonusage usage;
    private Image imageBall;
    private final DialogDotPlot_Sequence dotplotSeqX=new DialogDotPlot_Sequence(this,'X'), dotplotSeqY=new DialogDotPlot_Sequence(this,'Y');

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>> Weights >>> */
    private final static int[][] COSINUS=new int[50][];
    private final static byte[][] DISTANCE_NT=new byte[128][128], DISTANCE_RC=new byte[128][128];
    static {
        for(int w=COSINUS.length; --w>0;) {
            final int ww[]=COSINUS[w]=new int[w+1];

            for(int i=0;i<=w;i++) {
                double v=100000 *(1+Math.cos(i*Math.PI/w));
                ww[i]=(int)v;
            }
            normalizeWeights(ww);
        }
        for(int i=128; --i>=0;) {
            for(int j=128; --j>=0;) {
                int ci=i|32, cj=j|32;
                if (ci=='u') ci='t';
                if (cj=='u') cj='t';
                DISTANCE_NT[i][j]= (byte)(ci==cj && ( ci=='a' || ci=='c' || ci=='t' || ci=='g') ? 3 :-3);
                final boolean compl= ci=='a'&&cj=='t' || ci=='a'&&cj=='u' || ci=='c'&&cj=='g' ||  ci=='g'&&cj=='c' ||  ci=='t'&&cj=='a' ||  ci=='u'&&cj=='a';
                DISTANCE_RC[i][j]= (byte) (compl ? 3 :-3);
            }
        }
    }

    private static void normalizeWeights(int ww[]) {
        double mean=0, min=MAX_INT;
        for(int w : ww) {
            if (w<min) min=w;
            mean+=w;
        }
        mean/=ww.length;
        for(int i=ww.length; mean>0 && --i>=0;)  {
            ww[i]=(int) (WEIGHTS_AVERAGE_VALUE*(ww[i]-min)/mean);
            //if (ww.length==20) putln("i="+i+"  "+ww[i]);
        }
    }
    /* <<< Weights <<< */
    /* ---------------------------------------- */
    /* >>> Constructor >>> */
    public DialogDotPlot() { this("",null,null);  }
    private final String mode;
    public DialogDotPlot(String mode,StrapProtein pX, StrapProtein pY) {
        this.mode=mode;
        if (pX!=null && pY!=null) {
            comboProtX.s(pX);
            comboProtY.s(pY);
            comboProtX.setEnabled(false);
            comboProtY.setEnabled(false);
        }
								final boolean filter=mode.indexOf('f')<0;
								if (filter) {
												sliderWidth.setPaintLabels(true);
												sliderWidth.setPaintTicks(true);
												sliderWidth.setLabelTable(sliderWidth.createStandardLabels(10));
												sliderWidth.setMajorTickSpacing(8);
												sliderWidth.setMinorTickSpacing(2);
								}
        if (mode.indexOf('2')>=0) bl2seq=new int[0][];
								if (mode.indexOf('t')>0) sliderThreshold.setValue(0);
        final boolean removeTopPanel=mode.indexOf('I')>=0, noNT=mode.indexOf('n')>=0;

        final Color bg=sliderWidth.getBackground();
        final JComponent pSlider=vB(
                                    filter ? sliderWidth : null,
                                    filter ? pnl(bg,"Width of low pass filter") : null,
                                    filter ? comboFilter : null,
                                    " ",
                                    sliderBrightness,
                                    pnl(bg,"Brightness"),
                                    " ",
                                    sliderThreshold,
                                    pnl(bg,"Threshhold")
                                    );
        pButtons=vB(pnl("Marks: ",butMark,butClear,butColor)," ",togTrace," ");
        JComponent pWest=newCNSEW(null,pSlider,pButtons,null,SCROLLABLE_TRACKS_VIEWPORT_WIDTH);
        dotplotSeqX.addMouseWheelListener(LI);
        dotplotSeqY.addMouseWheelListener(LI);
        dotplotSeqX.addMouseListener(LI);
        dotplotSeqY.addMouseListener(LI);
        ChDelay.runAfterMS(EDT,dotplotSeqX,2500);
        ChDelay.runAfterMS(EDT,dotplotSeqY,2500);
        final String ttNT="Nucleotide sequence.<br>If no nt-sequence is defined in the file than backtranslate using the selected coding table.";
        JComponent pNorth=vB(
                             removeTopPanel?null:pTop(),
                             removeTopPanel?null:pnl(new GridLayout(1,2),comboProtX,comboProtY),
                             vB(
                                newCNSEW(dotplotSeqX.getPanel(),null,null,null,noNT?null:togNT_X.tt(ttNT)),
                                newCNSEW(dotplotSeqY.getPanel(),null,null,null,noNT?null:togNT_Y.tt(ttNT))
                                )
                             );
        dotplot.setPreferredSize(dim(999999,99999));
        cnsew(this,dotplot,pNorth,null,null,new JScrollPane(pWest));
        dotplot.addMouseListener(LI);
        dotplot.addMouseWheelListener(LI);
        dotplot.addMouseMotionListener(LI);
        align.addListener(this);
        enableDisable();
        pcp(StrapAlign.KEY_CHOICE_PROTEIN,new Object[]{comboProtX, comboProtY},this);
    }
        @Override public void dispose() {
        remove(selectorX);
        remove(selectorY);
    }
    /* <<< Constructor <<< */
    /* ---------------------------------------- */
    /* >>> Alignment >>> */
    protected int alignmentPos(char X_or_Y) { return X_or_Y=='X'? alignmentPosX : alignmentPosY;}
    public byte[] getSeq(char X_or_Y) {
        final StrapProtein p=protein(X_or_Y);
        if (p==null) return null;
        final byte aa[]=p.getResidueTypeExactLength();
        final byte nt[]=p.getNucleotidesCurrentStrandExactLength();
        if (usage==null) usage=Backtranslate.codonUsage(comboUsage.toString());
        final boolean isNuc=(X_or_Y=='X' ? togNT_X:togNT_Y).s();
        final boolean isBT=isNuc && nt==null;
        final byte seq[]= isBT ?  Backtranslate.backtranslate(aa,aa.length,usage,false) : isNuc ? nt : aa;
        sequenceType[X_or_Y=='X'?0:1]=isBT?'B':isNuc?'N':'A';
        return seq!=null ? seq:NO_BYTE;
    }
    private char[] sequenceType={'A','A'};
    protected char sequenceType(char X_or_Y) { getSeq(X_or_Y); return sequenceType[X_or_Y=='X' ? 0:1];}
    public boolean isNucleotide(char X_or_Y) { return sequenceType(X_or_Y)!='A'; }
    protected void setAlignmentPos(char X_or_Y,int pos) {
        if (nx*ny==0) return;
        final int xOld=plotWidth*alignmentPosX/nx;
        final int yOld=plotHeight*alignmentPosY/ny;
        if (X_or_Y=='X') {
            final int xNew=plotWidth*(alignmentPosX=pos)/nx;
            dotplot.paintImmediately(xOld,0,1, this.plotHeight);
            dotplot.paintImmediately(xNew,0,1, this.plotHeight);
        } else {
            final int yNew=plotHeight*(alignmentPosY=pos)/ny;
            dotplot.paintImmediately(0,yNew,this.plotWidth,1);
            dotplot.paintImmediately(0,yOld,this.plotWidth,1);
        }
        dotplot.repaint(xOld,yOld,textWidthPxl+EM,2*EX);
    }
    protected StrapProtein protein(char X_or_Y) {return (X_or_Y=='X' ? comboProtX : comboProtY).getStrapProtein();}

    private void resetImage(){
        refValues=null;
        dotplot.repaint();
        dotplotSeqX.repaint();
        dotplotSeqY.repaint();
        if (this instanceof Runnable) new Thread((Runnable)this).start();
        pcp(KEY_IMAGE_FILE_NAME, "dotplot-"+protein('X')+"-"+protein('Y'),this);
    }

    /* <<< Alignment <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent  >>> */
    private final ChEventAdapter LI=new ChEventAdapter(this);
    public AWTEvent processEvent(JComponent component,AWTEvent ev) {
        final Object q=ev.getSource();
        final int id=ev.getID();
        final StrapProtein px=protein('X'), py=protein('Y');
        {
            final int wheel=wheelRotation(ev), x=x(ev), y=y(ev);
            if (q==dotplot &&  (id==MOUSE_CLICKED || id==MOUSE_DRAGGED) && !(plotWidth*plotHeight==0||px==null||py==null)) {
                setAlignmentPos('X',x*nx/plotWidth);
                setAlignmentPos('Y',y*ny/plotHeight);
                dotplotSeqX.run();
                dotplotSeqY.run();
                butMark.enabled(true);
            }
            if (wheel!=0) {
                if (q==dotplotSeqY || q==dotplotSeqY) {
                    if (isCtrl(ev)) {
                        dotplotSeqX.setFontSize(wheel,false);
                        dotplotSeqY.setFontSize(wheel,false);
                    } else if (q instanceof DialogDotPlot_Sequence) {
                        final JScrollBar sb=((DialogDotPlot_Sequence)q).sb();
                        sb.setValue(sb.getValue()+wheel*EM);
                    }
                }
                if (q==dotplot) {
                    (isShift(ev)?dotplotSeqY:dotplotSeqX).move(5*wheel);
                    ChDelay.repaint(dotplot,333);
                }
            }
            if (id==MOUSE_RELEASED) {
                dotplotSeqX.repaint();
                dotplotSeqY.repaint();
            }
        }
        final String cmd=actionCommand(ev);
        if (cmd!=null) {
            if (q==butColor) {
                final Color c=this.color=((ButColor)q).getColor();
                selectorX.setColor(c);
                selectorY.setColor(c);
                imageBall=null;
                highlight(dotplot.getGraphics());
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,111);
            }
            if (q==comboUsage) {
                resetImage();
                if (q==comboUsage) usage=null;
            }
            if (q==togNT_X || q==togNT_Y) { dotplotSeqX.revalidate(); dotplotSeqY.revalidate(); resetImage(); }
            if (q instanceof ProteinCombo) {
                clearSelection();
                resetImage();
            }
            if (cmd==CB_LAB_Trace) dotplot.repaint();
            if (q==butClear) {clearSelection();dotplotSeqX.repaint(); dotplotSeqY.repaint(); butClear.enabled(false);butColor.enabled(false);}
            if (q==butMark)  {
                if (nSelected+2<selected.length) {
                    selected[nSelected++]=alignmentPosX;
                    selected[nSelected++]=alignmentPosY;
                }
                boolean[] selX=selectorX.getSelectedAminoacids();
                if (selX==null) selX=new boolean[alignmentPosX+1];
                else if (alignmentPosX>=selX.length) selX=chSze(selX,alignmentPosX+1);
                selX[alignmentPosX]=true;
                selectorX.setSelectedAminoacids(selX,0);
                px.addResidueSelection(selectorX);
                selectorX.setName("dotplot");
                selectorX.setStyle(VisibleIn123.CIRCLE);
                selectorX.setColor(color);
                boolean[] selY=selectorY.getSelectedAminoacids();
                if (selY==null) selY=new boolean[alignmentPosY+1];
                else if (alignmentPosY>=selY.length) selY=chSze(selY,alignmentPosY+1);
                selY[alignmentPosY]=true;
                selectorY.setSelectedAminoacids(selY,0);
                py.addResidueSelection(selectorY);
                selectorY.setName("dotplot");
                selectorY.setColor(color);
                selectorY.setStyle(VisibleIn123.CIRCLE);
                dotplotSeqX.run();
                dotplotSeqY.run();
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
                butColor.enabled(true);
                butClear.enabled(true);
            }
            if (q==sliderWidth || q==comboFilter) resetImage();
            if (q instanceof JSlider) ChDelay.repaint(dotplot,99);
            enableDisable();
        }
        return ev;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> GUI Elements >>> */
    public final static String CB_LAB_Trace="trace of the alignment";
    private final AbstractButton togTrace=toggl(CB_LAB_Trace).li(LI).s(true).cb();
    private final ChButton
        togNT_X=toggl("nt",ChButton.ICON_SIZE).li(LI),
        togNT_Y=toggl("nt",ChButton.ICON_SIZE).li(LI);
    {
        String tt="When comparing an amino acid sequence to nucleotide sequence the amino acid sequence must be back-translated. <br>The codon table is specified in the control panel.";
        togNT_X.tt(tt);
        togNT_Y.tt(tt);
    }
    private final ChCombo comboUsage=new ChCombo(Backtranslate.ORGANISMS).li(LI);
    private final ChCombo comboFilter=new ChCombo("Cosinus weights","Triangular weights","Constant weights").li(LI);
    private final JComponent pButtons, dotplot=new ChPanel(); {addPaintHook(this,dotplot); }
    private int nx,ny,plotWidth,plotHeight, valuesW,valuesH;
    private int alignmentPosX=10,alignmentPosY=10, textWidthPxl;
    private final ProteinCombo
        comboProtX=(ProteinCombo)new ProteinCombo(align,0).li(LI),
        comboProtY=(ProteinCombo)new ProteinCombo(align,0).li(LI);
     private final JSlider
        sliderBrightness=new ChSlider("",0,1024,256)
        .tt("set the intensity of the plot<br>Values are first multiplied <br>and then turned into gray color"),
        sliderThreshold=new ChSlider("",0,MAX_WERT_SQRT, MAX_WERT_SQRT/2),
        sliderWidth=new ChSlider("",1,COSINUS.length-1,COSINUS.length/2)
        .tt("we apply a <br>moving average low pass filter<br>set here the width of the cosinus curve <br> #<sub>weights</sub>=2*n+1");
    {
        addActLi(LI,sliderBrightness);
        addActLi(LI,sliderThreshold);
        addActLi(LI,sliderWidth);
        pcp(KEY_PRINTABLE_COMPONENTS, new Object[]{dotplotSeqX,dotplotSeqY,dotplot},this);
    }
    public void addButton(JComponent button) { pButtons.add(button);}
    private void enableDisable() {
        final StrapProtein px=protein('X'), py=protein('Y');
        final boolean onlyNX=px!=null && px.containsOnlyACTGNX();
        final boolean onlyNY=py!=null && py.containsOnlyACTGNX();
        togNT_X.setEnabled(!onlyNX);
        togNT_Y.setEnabled(!onlyNY);
        if (onlyNX) togNT_X.s(false);
        if (onlyNY) togNT_Y.s(false);
    }
    /* <<< GUI Elements <<< */
    /* ---------------------------------------- */
    /* >>> Highlight  >>> */
    private void highlight(int ix,int iy,Graphics g) {
        if (nx*ny==0) return;
        if (imageBall==null)  imageBall=ChIcon.coloredBall(color,this);
        if (imageBall==null || g==null) return;
        final int radius=imageBall.getWidth(this), x=plotWidth*ix/nx-radius/2, y=plotHeight*iy/ny-radius/2;
        ChIcon.drawImage(g,imageBall,x,y,radius,radius, ChIcon.getAlphaComposite(70),this);
    }
    private Rectangle rectCursor;
    private Rectangle rectCursor() {
        final StrapProtein px=protein('X'), py=protein('Y');
        final char tX=sequenceType('X'), tY=sequenceType('Y');
        final int cursorCol=align.getColumnAtCursor();
        if (nx*ny==0 || px==null || py==null || cursorCol<0) return null;
        int iAx=px.column2index(cursorCol), iAy=py.column2index(cursorCol);
        if (iAx<0) iAx=-99;
        if (iAy<0) iAy=-99;
        final int ix=tX=='N' ? px.translatedNucleotideIndex2nucleotideIndex(iAx*3) : tX=='B' ?  iAx*3 : iAx;
        final int iy=tY=='N' ? py.translatedNucleotideIndex2nucleotideIndex(iAy*3) : tY=='B' ?  iAy*3 : iAy;
        final int RADIUS=16;
        return new Rectangle(plotWidth*ix/nx-RADIUS/2, plotHeight*iy/ny-RADIUS/2 ,RADIUS,RADIUS);
    }
    private void highlight(Graphics g) { for(int i=0; i<nSelected;i+=2) highlight(selected[i],selected[i+1],g);}
    private Color color=Color.PINK;
    private final ChButton
        butColor=new ButColor(0L,Color.PINK,LI).enabled(false).t("").magnifyUnderMouse(IC_COLOR),
        butClear=new ChButton("Clear", ChButton.NO_MARGIN|ChButton.DISABLED).li(LI).tt("clear selection"),
        butMark=new ChButton("Add",ChButton.NO_MARGIN|ChButton.DISABLED).li(LI);
    /* <<< Highlight <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private final int selected[]=new int[66];
    private int nSelected=0;
    private final AbstractResidueSelection selectorX=new AbstractResidueSelection(), selectorY=new AbstractResidueSelection();
    { selectorX.imageBall(true);selectorY.imageBall(true);}
    private void clearSelection(){
        selectorX.setSelectedAminoacids(null,0);
        selectorY.setSelectedAminoacids(null,0);
        remove(selectorX);
        remove(selectorY);
        nSelected=0;
        dotplot.repaint();
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
    }
    private static void remove(ResidueSelection s) {
        final Protein p=s!=null ? s.getProtein() : null;
        if (p!=null) p.removeResidueSelection(s);
    }
    public boolean setProtein(StrapProtein p){comboProtX.s(p); return true; }
    public void setOtherProtein(StrapProtein p){comboProtY.s(p);}
    private static void addPoint( int x, int y, double score, int[][] vHits) {
        if (vHits.length<=y) {
            return;
        }
        if (vHits[y]==null) vHits[y]=new int[3];
        final int newEntry=x| ( (int)(10*score)<<16 ) | (1<<31);
        final int[] hy=vHits[y];
        {   /* --- Point already set? Replace if score is higher. --- */
            for(int i=hy.length; --i>=0;) {
                if (hy[i]!=0 && (hy[i]&0xffFF)==x) {
                    if (hy[i]<newEntry) hy[i]=newEntry;
                    return;
                }
            }
        }
        int idx=0;
        while(idx<vHits[y].length && vHits[y][idx]==0) idx++;
        (hy.length<=idx ?  vHits[y]=chSze(hy,idx+3) : hy)[idx]=newEntry;
    }
    public void addPoints(byte[] tX,byte[] tY, int startX,   int startY, boolean revX, boolean revY,   int endX, int endY, int fromX, int fromY,double score, int[][] vHits) {
        int seqX=startX, seqY=startY;
        final char typX=sequenceType('X');
        final char typY=sequenceType('Y');
        /* --- In case bl2seq detected nucleotides --- */
        final boolean ntAutoX= typX=='A' && nxt(NO_ACTGUN,getSeq('X'))<0;
        final boolean ntAutoY= typY=='A' && nxt(NO_ACTGUN,getSeq('Y'))<0;
        final int stepX=(revX?-1:1) * (typX=='B'|| (ntAutoX&&!ntAutoY) ? 3:1);
        final int stepY=(revY?-1:1) * (typY=='B' ? 3:1);
        for(int x=fromX, y= (ntAutoY&&!ntAutoX)?fromY/3:fromY;  seqX<endX && seqX<tX.length && seqY<endX && seqY<tY.length;  seqX++,seqY++) {
            final byte cx=tX[seqX];
            final byte cy=tY[seqY];
            if (cx=='\r'||cx=='\n' || cy=='\r'||cy=='\n') break;
            final boolean lx='a'<=cx && cx<='z' || 'A'<=cx && cx<='Z';
            final boolean ly='a'<=cy && cy<='z' || 'A'<=cy && cy<='Z';
            if (lx&&ly) addPoint(x,y,score, vHits);
            if (lx) x+=stepX;
            if (ly) y+=stepY;
        }
    }
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> getValues >>> */
    private int[][] bl2seq;
    private ChExec bl2seq_ex;
    public void setBl2seq(int[][] data,ChExec ex) {
        controlPanel=null;
        bl2seq_ex=ex; refValues=null; bl2seq=data; dotplot.repaint();
    }
    private Reference<short[][]>refValues;
    public short[][] getValues(byte[] seqX,char tX, byte[] seqY, char tY,  int w, int h, int weights[]) {
        short min=Short.MAX_VALUE, max=Short.MIN_VALUE;
        short[][] values=deref(refValues,short[][].class);
        if (sze(values)!=w || w<1 ||values[0].length!=h) values=new short[w][h];
        final int b2s[][]=this.bl2seq;
        if (b2s!=null) {
            if (b2s.length==0) return null;
            for(int iy=b2s.length; --iy>0;) {
                final int[] line=b2s[iy];
                if (line==null) continue;
                for(int d:line) {
                    if (d==0) continue;
                    final int ix=d&0xffff;
                    final int score=(d>>16) &0xfff;
                    final int x=ix*w/nx;
                    final int y=iy*h/ny;
                    final int v=score;
                    if (x>=0 && y>=0 && x<w && y<h && values[x][y]<v)  values[x][y]= (short)(v>Short.MAX_VALUE ? Short.MAX_VALUE:v);
                }
            }
            for(int x=w;--x>=0;) {
                for(int y=h; --y>=0;) {
                    int v=values[x][y];
                    if (v<min) v=min;
                    if (v>max) v=max;
                }
            }
        } else {
            final byte[][] DISTANCE_BLOSUM=Blosum.BLOSUM62;
            final int w2=weights.length/2;
            final boolean isNT= tX!='A' || nxt(NO_ACTGUN,seqX)<0;
            final int filter=comboFilter.i();
            for(int x=w;--x>=0;) {
                final int ix=nx*x/w;
                final short[] vv=values[x];
                Arrays.fill(vv, Short.MIN_VALUE);
                for(int y=h; --y>=0;) {
                    final int iy=ny*y/h;
                    int sum=0;
                    for(int isRC= (isNT?2:1); --isRC>=0;) {
                        final byte distance[][]= isRC!=0 ? DISTANCE_RC : isNT ? DISTANCE_NT : DISTANCE_BLOSUM;
                        for(int i=weights.length, rx0=ix+i,ry=iy+i,  lx0=ix-i, ly=iy-i; --i>=0;  rx0--, ry--, lx0++, ly++) {
                            final int rx,lx;
                            if (isRC!=0) { rx=lx0; lx=rx0;} else { rx=rx0; lx=lx0;}
                            final int dr= (rx>=0 && ry>=0 && rx<seqX.length && ry<seqY.length) ?  distance[seqX[rx]][seqY[ry]] : 0;
                            final int dl=(i>0 && lx>=0 && ly>=0 && lx<seqX.length && ly<seqY.length) ?  distance[seqX[lx]][seqY[ly]] : 0;
                            switch(filter) {
                            case 0: sum+=weights[i]* (dr+dl); break;
                            case 1: sum+=(weights.length-i)*(dr+dl); break;
                            case 2: sum+=dr+dl; break;
                            }
                            sum+=(dr+dl);
                        }
                    }
                    final short v=sum>Short.MAX_VALUE ? Short.MAX_VALUE : sum<Short.MIN_VALUE ? Short.MIN_VALUE  : (short)sum;
                    vv[y]=v;
                    if (x>w2 && y>w2 && x<w-w2 && y<h-w2) {
                        if (v<min) min=v;
                        if (v>max) max=v;
                    }
                }
            }
        }
        if (max>min) {
            for(int x=w;--x>=0;) {
                final short[] vvX=values[x];
                for(int y=h; --y>=0;) {
                    int v=vvX[y];
                    v=MAX_WERT*(v-min)/(max-min);
                    vvX[y]=(short)v;
                    if ( v<0 || v>MAX_WERT)
                        if (isMyComputer()) putln("Warning DialogDotPlot v="+v+" MAX_WERT="+MAX_WERT);
                }
            }
        }
        return values;
    }
    private final static int MAX_WERT_SQRT=100, MAX_WERT=MAX_WERT_SQRT*MAX_WERT_SQRT;
    /* <<< getValues <<< */
    /* ---------------------------------------- */
    /* >>> StrapEvent >>> */
    public final void handleEvent(StrapEvent ev) {
        final int t=ev.getType();

        if (t==StrapEvent.AA_SHADING_CHANGED) { dotplotSeqX.repaint(); dotplotSeqY.repaint(); }

        if ((t==StrapEvent.BACKGROUND_CHANGED || t==StrapEvent.ALIGNMENT_CHANGED) && togTrace.isSelected()) dotplot.repaint();
        if (t==StrapEvent.BACKGROUND_CHANGED || t== StrapEvent.ALIGNMENT_CHANGED || t==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN || t==StrapEvent.CURSOR_CHANGED_PROTEIN) {
            if (rectCursor!=null) dotplot.repaint(rectCursor);
            rectCursor=rectCursor();
            if (rectCursor!=null) dotplot.repaint(rectCursor);
												dotplotSeqX.repaint();
												dotplotSeqY.repaint();
        }
        if (t==StrapEvent.RESIDUE_SELECTION_CHANGED || t==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR) {
            final Object protein=get(0,ev.getParameters());
            if (protein==protein('X') || protein==protein('Y') || protein==null)
                dotplot.repaint();
												dotplotSeqX.repaint();
												dotplotSeqY.repaint();
        }
        if ( (t&StrapEvent.FLAG_RESIDUE_TYPES_CHANGED)!=0) resetImage();
    }
    /* <<< StrapEvent <<< */
    /* ---------------------------------------- */
    /* >>> PaintHook >>> */
    public boolean paintHook(JComponent component, Graphics g, boolean after) {
        final boolean isWhiteBG=isWhiteBG();
        if (component==dotplot && after){
            final int plotWidth=dotplot.getWidth();
            final int plotHeight=this.plotHeight=dotplot.getHeight();
            final Rectangle clip;
            if (this.plotWidth!=plotWidth || this.plotHeight!=plotHeight) {
                this.plotWidth=plotWidth;
                this.plotHeight=plotHeight;
                clip=null;
            } else clip=g.getClipBounds();

            final int minX=clip!=null ? maxi(0,clip.x):0;
            final int maxX=clip!=null ? mini(plotWidth,clip.x+clip.width+1):plotWidth;
            final int minY=clip!=null ? maxi(0,clip.y):0;
            final int maxY=clip!=null ? mini(plotHeight,clip.y+clip.height+1):plotHeight;

            if (plotWidth<3 || plotHeight<3) return true;
            g.setFont(getFnt(16,true,Font.BOLD));
            final StrapProtein px=comboProtX.getStrapProtein();
            final StrapProtein py=comboProtY.getStrapProtein();
            if (px==null || py==null) return true;
            final char tX=sequenceType('X'), tY=sequenceType('Y');
            this.nx= tX=='N' ? px.countNucleotides() : (tX=='B'?3:1)*px.countResidues();
            this.ny= tY=='N' ? py.countNucleotides() : (tY=='B'?3:1)*py.countResidues();
            String msg=null;
            if (px==null || py==null)  msg="Two proteins required !";
            else if (nx*ny==0) msg="Erro: Sequence is empty !";
            else {
                final int brightness=sliderBrightness.getValue();
                final int threshhold=sliderThreshold.getValue()*sliderThreshold.getValue();
                final int[] weights=COSINUS[mini(COSINUS.length-1,maxi(1, sliderWidth.getParent()==null ? 1 : sliderWidth.getValue()))];
                short values[][]=(short[][]) deref(refValues);
                {
                    final int vW=mini(nx,plotWidth), vH=mini(ny,plotHeight);
                    if (values==null || valuesH!=vH || valuesW!=vW) {
                        refValues=new java.lang.ref.SoftReference(values=getValues(getSeq('X'),tX, getSeq('Y'),tY,valuesW=vW,valuesH=vH,weights));
                    }
                }
                final int vWidth=sze(values),  vHeight=vWidth>0 ? values[0].length : 0;
                if (vHeight>0 ) {
                    final int vMinX=maxi(0,minX*vWidth/plotWidth-1);
                    final int vMaxX=mini(vWidth,maxX*vWidth/plotWidth+1);
                    final int vMinY=maxi(0,minY*vHeight/plotHeight-1);
                    final int vMaxY=mini(vHeight,maxY*vHeight/plotHeight+1);
                    final java.awt.image.BufferedImage image=sharedImage('v',vHeight);
                    final int[] rgb=sharedRGB(vHeight);
                    for(int x=vMinX; x<vMaxX; x++) {
                        final short[] valuesX=values[x];
                        Arrays.fill(rgb,vMinY,vMaxY,0);
                        for(int y=vMinY; y<vMaxY; y++) {
                            final int v0=(int)valuesX[y];
                            if (v0>threshhold) {
                                int v=(v0-threshhold)*brightness/(MAX_WERT-threshhold);
                                if (v>255) v=255;
                                if (v<0) v=0;
                                if (isWhiteBG) v=255-v;
                                rgb[y]=(v << 16)  | (v<<8) |  v;
                                //if (isWhiteBG) rgb[y]= (~rgb[y])&0xffFFff;
                            }
                        }
                        image.setRGB(0, 0, 1,vMaxY-vMinY, rgb,  vMinY,1);
                        final int dx1=x*plotWidth/vWidth;
                        final int dy1=vMinY*plotHeight/vHeight;
                        final int dx2=(1+x)*plotWidth/vWidth;
                        final int dy2=vMaxY*plotHeight/vHeight;
                        g.drawImage(image,dx1,dy1,dx2,dy2,  0,0,1,vMaxY-vMinY,this);
                    }
                } else {
                 msg=mode.indexOf('I')>=0? " Activate check-boxes of hits " : " calculating...";
                }
            }
            if (msg!=null) {
                g.fillRect(0,0,component.getWidth(),component.getHeight());
                g.setColor(Color.PINK);
                g.drawString(msg,20,20);
            }

            if (togTrace.isSelected()) {
                g.setColor(C(0xFF64ff,99));
                int lastX=-1,lastY=-1;
                int lineWidth=plotWidth/50+1;
                for(int x=minX;x<maxX;x++) {
                    final int ix=(int) (x* (long)nx/plotWidth);
                    final int iAx= tX=='N' ? px.nucleotideIdx2translatedNucleotideIdx(ix)/3 : tX=='B' ? ix/3 : ix;
                    final int col=px.getResidueColumn(iAx);
                    final int iAy=py.column2nextIndex(col);
                    if (py.getResidueColumn(iAy)!=col) continue;
                    final int iy=tY=='N' ? py.translatedNucleotideIndex2nucleotideIndex(iAy*3) : tY=='B' ?  iAy*3 : iAy;
                    final int y=plotHeight*iy/ny;
                    g.fillRect(lastX,lastY-lineWidth/2, x-lastX,lineWidth);
                    lastX=x;lastY=y;
                }
            }
            highlight(g);
            final Rectangle r=rectCursor();
            if (r!=null) ChIcon.drawCursor(g,r.x,r.y,r.width,r.height,this);
            rectCursor=r;
            if (alignmentPosX>=0 && alignmentPosY>0 && nx>0 && ny>0) {
                g.setColor(C(0xFF0000));
                final int x=plotWidth*alignmentPosX/nx, y=plotHeight*alignmentPosY/ny;
                g.drawLine(0,y,plotWidth,y);
                g.drawLine(x,0,x,plotHeight);
                this.textWidthPxl=0;
                for(int j=0;j<2;j++) {
                    final int alPos=j==0 ? alignmentPosX : alignmentPosY;
                    if (alPos< (j==0 ? nx:ny)) {
                        final ByteArray s=(j==0?px:py).cursorPosAsString(alPos,false);
                        textWidthPxl=maxi(textWidthPxl, charBounds(g.getFont()).width*s.getEnd());
                        g.drawBytes(s.getBytes(),0,s.getEnd(),x,y+(j+1)*EX);
                    }
                }
            }
            final int secStr2color[]=colors2rgb(ShadingAA.SECSTRU);
            for(boolean isVertical : FALSE_TRUE) {
                final int THICKNES=4;
                if ((isVertical ? minX:minY)>THICKNES) continue;
                final StrapProtein p=isVertical?py:px;
                final byte ss[]=p.getResidueSecStrType();
                if (ss==null) continue;
                final int nR=p.countResidues();
                final java.awt.image.BufferedImage im=sharedImage(isVertical ? 'V':'H',nR);
                final int rgb[]=sharedRGB(nR);
                for(int i=nR;--i>=0;) rgb[i]= i<ss.length ? secStr2color[ss[i]] : 0;
                if (isVertical) {
                    im.setRGB(0, 0, 1,nR, rgb,  0,1);
                    g.drawImage(im,0,0,THICKNES-1,plotHeight,  0,0,1,nR,this);
                } else {
                    im.setRGB(0,0,nR,1, rgb, 0, nR);
                    g.drawImage(im,  0,0, plotWidth,THICKNES-1,    0,0,nR,1, this);
                }
            }
         }
         return true;
    }
    /* <<< PaintHook <<< */
    /* ---------------------------------------- */
    /* >>> Logging >>> */
    private ByteArray log_=new ByteArray(0);
    public void log(Object... oo) {  if (oo!=null) for(Object o:oo) log_.append(o); }
    /* <<< Logging <<< */
    /* ---------------------------------------- */
    /* >>> ControlPanel >>> */
    private Container controlPanel;
    public Container getControlPanel() {
        if (controlPanel==null) {
            JComponent pNorth= mode.indexOf('n')>=0 ? null : vB( pnl("In case of translation of aminos to triplets use the codon table: ",comboUsage));
            controlPanel=newCNSEW( new AsciiView(log_).tools().scrollPane(),pNorth, getCtrlPanel(bl2seq_ex),null,null);
        }
        return controlPanel;
    }

}

/*
tblastn aa , n
blastx n , aa
*/
