package charite.christo.strap;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.*;
import java.io.File;
import java.awt.*;
import java.util.*;
import java.util.List;
import java.lang.ref.*;
import static charite.christo.protein.ProteinModiCount.*;

/**HELP

Residue annotations, allow the assignment of information to specific
amino acids or nucleotides of proteins.

They are special types of residue selections.

A residue annotation has a list of entries each having a name, some
user typed text and a toggle button.

The user can change these entries and add new entries.

The entries can contain several types of information:

<ul>
<li>Rendering commands for 3D-viewers,</li>
<li>LaTeX-code for the PDF-output,</li>
<li>Hyperlinks, PUBMED-references, database references,</li>
<li>Notes, for example literature citations, experimental conditions.</li>
</ul>

A ResidueAnnotation can be created in different ways:
<ul>

<li>A residue selection can be created by dragging the mouse over the sequence. Right-click opens the context menu of the residue
selection.  With the menu item
<i>ITEM:ResidueSelectionPopup#ACTION_convertToResidueAnnotation</i> a
residue annotation can be created that has the same amino acid
positions.</li>

Alternatively, use the context menu of the alignment panel to create an annotation for the protein under the mouse pointer.

</ul>

Residue annotations which are highlighted in the alignment can be selected in two different ways: By CTRL left-click or by dragging  a rectangular box.
Selected residue annotations have WIKI:Marching_ants.

<br><br><b>Sequence Features: </b>

Features like phosporylation sites and active sites are special
annotations which are retrieved from a computational service and are
therefore not saved to hard disk.  Their color cannot be changed.

<h2>Background Images</h2>

<i>INCLUDE_DOC:ChIcon</i>

<i>SEE_DIALOG:DialogResidueAnnotationChanges</i>
<i>SEE_DIALOG:DialogResidueAnnotationList</i>
<i>SEE_CLASS:ResidueAnnotationView</i>
<i>SEE_CLASS:ActivateDeactivate</i>
<i>SEE_CLASS:AddAnnotation</i>
*/
public final class ResidueAnnotation
    implements HasWeakRef, HasName, ResidueSelection,charite.christo.strap.interfaces.SelectorOfNucleotides,VisibleIn123,Colored,Disposable,
                                                TabItemTipIcon.Interface, HasRendererText, HasRendererOptions, HasRendererModiCount, HasImage, HasPropertyMap,
                                                ChRunnable, HasModiCount {

    public final static String
        NAME="Name",POS="Positions",GROUP="Group",
        COLOR="Color", BG_IMAGE="BG_IMAGE", VIEW3D="3D_view", SPECIFIC_VIEW3D="Specific_3D_view",
        mainKeys[]={NAME,GROUP,POS};
    public final static ResidueAnnotation NONE[]={};
    private static int maxLengthOfUnderlining=7;
    protected static void setMaxLengthOfUnderlining(int l) { maxLengthOfUnderlining=l;}

    final Reference WEAK_REF=new WeakReference(this);
    public Reference getWeakRef() { return WEAK_REF;}

    private int modiCount;
    public int getModificationCount() { return this.modiCount;}
    /* ---------------------------------------- */
    /* >>> Instance  >>> */
    public ResidueAnnotation(StrapProtein p) {
        refProtein=weakRef(p);
        for(int i=0; i<2; i++) {
            final String v=i==0?"new annotation" : i==1?"new selection group" : "";
            final String k=i==0?NAME : i==1?GROUP : POS;
            if (getEntryWithKey(k)==null) add(k,v,true);
        }
        vAll.add(WEAK_REF);
    }
    private final long created=System.currentTimeMillis();
    public long getWhenCreated() { return created;}
    public void dispose() {
        final Protein protein=getProtein();
        if (protein!=null) protein.removeResidueSelection(this);
        rendererText=" Discarded: "+getName();
        vAll.remove(WEAK_REF);
    }
    /**  ResidueAnnotation-Objects belong to a certain protein */
    public StrapProtein getProtein() { return (StrapProtein)deref(refProtein); }
    private Reference<StrapProtein> refProtein;
    public boolean setProtein(Protein p) { refProtein=weakRef(p);  return true; }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Name and ID >>> */
    /**  A ResidueAnnotation-object has a name. */
    public String getName() {return getValue(NAME);}
    private ByteArray baID;
    public String getID() {
        if (baID==null) baID=new ByteArray(99);
        return baID.setEnd(0).filter(LETTR_DIGT, getProtein(), 0 , MAX_INT, '_').append('_').filter(LETTR_DIGT, getName(), 0 , MAX_INT, '_').newString();
    }
    /* <<< Name and ID  <<< */
    /* ---------------------------------------- */
    /* >>> View >>> */
    private Reference view;
    protected Reference getView() { return view;}
    public void setView(Reference v) { view=v;}
    /* <<< View  <<< */
    /* ---------------------------------------- */
    /* >>> Selected Nucleotides and Aminoacids >>> */
    private final static int IS_SEL_NT=0, IS_SEL=1, IS_SEL_V=2, REND_TEXT=3, REND_TEXT_JLIST=4;
    private final long cacheMC[]=new long[6];
    private transient Reference<Object[]> rCached;
    private Object[] cached() {
        Object[] r=rCached!=null ? rCached.get() : null;
        if (r==null) rCached=new java.lang.ref.SoftReference(r=new Object[6]);
        return r;
    }
    public void setSelectedAminoacids(boolean bb[],int offset) {
        final StrapProtein p= getProtein();
        setValue(POS,Protein.selectedPositionsToText(bb,0,p));
        setEnabled(true,ResidueAnnotation.POS);
    }
    /** e.g. s.setSelectedAminoacids("3-5"); */
    public void setSelectedAminoacids(String selection) { setValue(POS,selection); setEnabled(true,POS);}
    public void setSelectedNucleotides(String selection) { setValue(POS,selection);setEnabled(false,POS);}
    private int selectedAAOffset;
    public int getSelectedAminoacidsOffset() {
        getSelectedAminoacids();
        int offset=this.selectedAAOffset;
//         final Protein p=getProtein();
//         if (p!=null) {
//             final String pos=getValue(POS);
//             if (pos!=null) offset-=p.getFirstResidueIndex()*0;
//         }
        return offset;
    }
    public static boolean[] getSelectedAminoacidsDisplayed(ResidueSelection s) {
        if (s instanceof ResidueAnnotation) {
            final ResidueAnnotation a=(ResidueAnnotation)s;
            final boolean sa[]=a.getSelectedAminoacids();
            final boolean saViewed[]=(boolean[])a.cached()[IS_SEL_V];
            return saViewed!=null ? saViewed : sa;
        }
        return s!=null ? s.getSelectedAminoacids() : NO_BOOLEAN;
    }
    public boolean[] getSelectedAminoacids() {
        final Object[] cached=cached();
        parseResidueAnnotationIfNecessary(cached);
        boolean bb[]= (boolean[])cached[IS_SEL];
        return bb!=null ? bb : NO_BOOLEAN;
    }
    private int isSelectedNtOffset;
    public int getSelectedNucleotidesOffset() {
        final Object[] cached=cached();
        parseResidueAnnotationIfNecessary(cached);
        return isSelectedNtOffset;
    }
    public boolean[] getSelectedNucleotides() {
        final Object[] cached=cached();
        parseResidueAnnotationIfNecessary(cached);
        return (boolean[])cached[IS_SEL_NT];
    }

    /* <<< Selected Nucleotides and Aminoacids <<< */
    /* ---------------------------------------- */
    /* >>> Rendering >>> */
    /**  Rendering style  */
    public int getStyle() {  return style>=0 ? style : iconImage==null ? VisibleIn123.LOWER_HALF_BACKGROUND : VisibleIn123.IMAGE; }
    /**  Rendering style  */
    public void setStyle(int i) {style=i;}
    private int style=-1;
    /** sets the color */
    public void setColor(Color c) {
        if (c!=this.color) {
            this.color=c;
            modiCount++;
            Protein.incrementModificationCount(MC_RESIDUE_SELECTION_LABELS, getProtein());
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,111);
        }
    }
    public Color getColor() { return color; }
    protected void setColorForWhiteBG(Color c) { colorPaper=c; }
    protected Color getColorForWhiteBG() { return colorPaper!=null ? colorPaper : color; }
    private Color color=Color.PINK, colorPaper;
    private int shownWhere=0xffffffff;
    public int getVisibleWhere() { return style!=VisibleIn123.HIDDEN && isEnabled() ? shownWhere : 0;}
    public void setVisibleWhere(int where) { shownWhere=where;}
    /* <<< Rendering  <<< */
    /* ---------------------------------------- */
    /* >>> Image Icon >>> */
    private Image iconImage;
    public Image getImage(Component observer) {  return iconImage; }

    private final static Object[] provideStrings={null,null}, provideWC={null,null};
    public Object run(String id, Object arg) {
        if (id==DialogStringMatch.PROVIDE_STRINGS) {
            provideStrings[0]=getName();
            provideStrings[1]=vEntries;
            return provideStrings;
        }
        if (id==ChRunnable.SET_ICON_IMAGE) {
            Image im=(Image)arg;
            if (iconImage!=im) {
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,444);
                iconImage=im;
                icon=null;
                this.modiCount++;
            }
        }
        if (id==PROVIDE_WORD_COMPLETION_LIST){
            provideWC[0]=getRendererText(HasRendererText.JLIST);
            provideWC[1]=getName();
            return provideWC;
        }
        return null;
    }
    private String imageId, lastImage;
    protected String getLastImage() { return lastImage;}
    protected void findIconImage() {
        lastImage=getLastValue(BG_IMAGE);
        final String newImageId=lastImage!=null ? lastImage : null;
        if (imageId!=newImageId) {
            imageId=newImageId;
            icon=null;
            iconImage=null;
            requestImage(imageId,this);
        }
    }
    private javax.swing.ImageIcon icon;
    public javax.swing.ImageIcon getIcon() {
        final Image im=iconImage;
        if (im==null) return null;
        if (icon==null) {
            Image imSmall=Toolkit.getDefaultToolkit().createImage(new java.awt.image.FilteredImageSource(im.getSource(),   new java.awt.image.ReplicateScaleFilter(ICON_HEIGHT, ICON_HEIGHT)));
            if (imSmall!=null) icon=new javax.swing.ImageIcon(imSmall);
        }
        return icon;
    }
    /* <<< Image Icon <<< */
    /* ---------------------------------------- */
    /* >>> Tooltip  >>> */
    private TabItemTipIcon titi;
    private boolean titiShift;
    private int titi_mc;
    public TabItemTipIcon getTabItemTipIcon(AWTEvent ev) {
        final boolean shift=isShift(ev);
        final int mc=getModificationCount();
        if (titi==null || titiShift!=shift || titi_mc!=mc) {
            titi_mc=mc;
            titiShift=shift;
            final Color color=getColorForWhiteBG();
            final boolean[] sel=getSelectedAminoacids();
            final int offset=getSelectedAminoacidsOffset();
            final ByteArray sb=new ByteArray(99).append("<span style=\"background-color: ");
            rgbHex(color!=null ? color.getRGB()&0xFFffFF : 0,sb).append(";\"> &nbsp; </span>&nbsp;");
            final Protein p=getProtein();
            final String pN=p!=null ? p.toString() : "", n=getName();
            final String f=getFeatureName();
            if (f!=null) sb.append(" feature ").append(f).append(" ");
            else sb.append("annotation ").append(getName());
            sb.append("<small>");
            for(Entry e : getEntries()) {
                final String k=e.getKey();
                if (k=="Balloon" || shift && (k=="Remark" || k=="Note")) {
                    final String v=e.getValue();
                    if (v!=null) sb.append(f==null ? "<br>" :"").append(htmlEncod(v));
                }
            }
            if (simRegion!=null && shift) {
                sb.append("<pre>");
                simRegion.getLocalAlignment(offset+firstTrue(sel)-5, offset+lastTrue(sel)+5, sb);
                sb.append("\n</pre>");
            } else sb.append("<br>");
            if (cached()[IS_SEL_V]!=null) sb.append("Positions ").append(getValue(POS)).append("<br>");
            sb.append("</small>");
            final String tab="<sub>"+
                (pN!=null ? pN.substring(0,mini(10,sze(pN))) : null)+
                "<br>"+
                (n!=null ? n.substring(0,mini(10,sze(n))) : null)+
                "</sub>";
            titi= new TabItemTipIcon(tab,null, sb.newString(), getIcon());
        }
        return titi;
    }
    private String rendererText;
    public void setRendererText(String s) { rendererText=s; }
    public long getRendererOptions() {
        final Protein p=getProtein();
        return p==null || !cntains(this, p.getAllResidueSelections()) ? HasRendererOptions.STRIKE_THROUGH : 0L;
    }

    public CharSequence getRendererText(int options) {
        final Object[] cached=cached();
        final boolean jList=(options&HasRendererText.JLIST)!=0;
        final int cacheID=jList ? REND_TEXT_JLIST : REND_TEXT;
        final Protein p=getProtein();
        long hc=hashCd(rendererText)+getModificationCount();
        if (p!=null) hc+=p.getModificationCount(MC_PROTEIN_LABELS)+p.getModificationCount(MC_RESIDUE_SELECTIONS);
        String s=(String)cached[cacheID];
        if (s==null || cacheMC[cacheID]!=hc) {
            cacheMC[cacheID]=hc;
            final String rt=rendererText!=null ?  rendererText : getName();
            cached[cacheID]=s= jList ? getProtein()+"/"+rt : rt;
        }
        return s;
    }
    private static boolean _rendererInit;
    {
    if (!_rendererInit) {
        _rendererInit=true;
        TabItemTipIcon.set("",VIEW3D,"Script lines, suitable for different protein viewers",iicon(IC_JMOL),VIEW3D);
        TabItemTipIcon.set("",SPECIFIC_VIEW3D,"Script lines for one particular protein viewers",iicon(IC_JMOL),SPECIFIC_VIEW3D);
    }
    }
    final long hashCodeShift=hashCode()<<32;
    public long getRendererModiCount() {
      final StrapProtein p=(StrapProtein)getProtein();
      long mc=hashCodeShift+modiCount;
      if (p!=null) { p.getResidueAnnotations(); mc+=p.getModificationCount(MC_RESIDUE_ANNOTATIONS_V)<<16; }
      return mc;
    }

    /* <<< Tooltip <<< */
    /* ---------------------------------------- */
    /* >>>  Get Entry >>> */
    protected String getLastValue(String key) {
        final Entry ee[]=getEntries();
        for(int i=ee.length; --i>=0;) {
            if (ee[i].isEnabled() && key==ee[i].getKey()) {
                final String v=ee[i].getValue();
                if (v.length()>0) return v;
            }
        }
        return null;
    }
   /**  Return the 1st value with the key. */
    public String getValue(String key) {
        final Entry e=key!=null ? getEntryWithKey(key) : null;
        return e!=null ? e.getValue(): null;
    }

    public boolean isAmino() { return isEnabled(ResidueAnnotation.POS);}

     /** Returns the status of the i-th key-value pair or null. */
    protected boolean isEnabled(String key) {
        for(Entry e:vEntries.asArray()) {
            if (key==e.getKey() && e.isEnabled()) return true;
        }
        return false;
    }
    /* <<< Get Value <<< */
    /* ---------------------------------------- */
    /* >>> Set Value >>> */
    /** Adds a new key-value-pair */
    public void setValue(String key,CharSequence textValue) {
        if (key==null) return;
        final String value=toStrg(textValue);
        boolean found=false;
        for(Entry e:getEntries()){
            if (key.equals(e.getKey())) {
                setValue(value,e,true);
                found=true;
            }
        }
        if (!found && value!=null) add(key,value,true);
        if (BG_IMAGE==key) findIconImage();
        Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,getProtein());
    }

    boolean setValue(final String value, Entry e, boolean alsoTextField) {
        if (value==null || e==null) removeEntry(e);
        else if (!value.equals(e.getValue())) {
            final String k=e.getKey();
            String v=value.trim();
            if (k==NAME || k==GROUP) {
                v=rplcNoFilenameChars(v);
            }
            e._value=v;
            if (alsoTextField) {
            final ChTextField tf=deref(e.TEXTFIELD[0],ChTextField.class);
            if (tf!=null) tf.t(v);
            }
            this.modiCount++;
            final Protein p=getProtein();
            Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,p);
            if (k==NAME) Protein.incrementModificationCount(MC_RESIDUE_SELECTION_LABELS, p);
            if (k==GROUP) Protein.incrementModificationCount(MC_RESIDUE_ANNOTATION_GROUPS, p);
            return true;
        }
        return false;
    }
    public void setEnabled(boolean b) {
        setEnabled(b,NAME);
    }
    public void setEnabled(boolean b, String key) {
        final Protein p=getProtein();
        for(Entry e:getEntries()) {
            if (key==e.getKey() && e.isEnabled()!=b) {
                e.setEnabled(b);
                Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS,p);
                if (key==NAME) Protein.incrementModificationCount(MC_RESIDUE_SELECTION_LABELS, p);
                if (key==GROUP) Protein.incrementModificationCount(MC_RESIDUE_ANNOTATION_GROUPS,p);
            }
        }
    }
    /* <<< Set Value <<< */
    /* ---------------------------------------- */
    /* >>> Get Entry Objects >>> */
    private Entry onlyEntry;
    void setOnlyEntry(Entry e) {onlyEntry=e;}
    private final UniqueList<Entry> vEntries=new UniqueList(Entry.class);
    public Entry[] getEntries() {
        Entry oe=this.onlyEntry;
        return oe!=null ? new Entry[]{oe}:vEntries.asArray();
    }
    public void removeEntry(Entry entry) { vEntries.remove(entry); modiCount++; }
    public void add(String key0, String value, boolean isEnabled) {
        if (key0==null || value==null) return;
        final String key=key0.intern();
        if (cntains(key,mainKeys)) {
            final Entry e=getEntryWithKey(key);
            if (e!=null) {
                e.setEnabled(isEnabled);
                setValue(value, e,true);
                return;
            }
        }
        vEntries.add(newEntry(key,value,isEnabled));
        modiCount++;
        if (BG_IMAGE==key) findIconImage();
    }
    Entry getEntryWithKey(String key) {
        for(Entry e:getEntries()) if (key==e.getKey()) return e;
        return null;
    }
    Entry getEntryAt(int i) { return vEntries.get(i);}

    /* <<< Entry List  <<< */
    /* ---------------------------------------- */
    /* >>> Parse >>> */
    /**
       Example
       texshade=  X\feature...
       ^
       i1234
       value is substring(i+4);
       X/x  is substring(i+3);
    */
    private void parse(String[] lines) {
        for(String s:lines){
            if (s.startsWith("charite.christo.strap.BG_Image")) s=BG_IMAGE+s.substring(30);
            final int i=s.indexOf('=');
            if (i<2 || s.length()<i+4) continue;
            final String k=s.substring(0,i).intern(),v=s.substring(i+4);
            if (k==COLOR) {
                color=str2color(v);
                if (color==null) color=Color.PINK;
            } else {
                add(k,v,(s.charAt(i+3)=='X'));
            }
        }
    }

    private static Boolean allEnabled;
    static void setEnabledAll(Boolean b) {
        allEnabled=b;
        Protein.incrementModificationCount(MC_RESIDUE_SELECTIONS, null);
    }
    public boolean isEnabled() {
        if (getFeatureName()==null) return isEnabled(NAME);
        final boolean ret=allEnabled==Boolean.TRUE ? true : allEnabled==Boolean.FALSE ? false : isEnabled(NAME);
        return ret;

    }
    private Range _range;
    protected Range getRange() {
        getSelectedAminoacids();
        return _range;
    }
    private void parseResidueAnnotationNT(String sSet, Object[] cached) {
        final StrapProtein protein=getProtein();
        if (protein==null) {
            cached[IS_SEL_NT]=NO_BOOLEAN;
            isSelectedNtOffset=0;
        }
        else {
            final boolean bb[]=parseSet(sSet,protein.countNucleotides(),-1, null);
            final int first=firstTrue(bb);
            cached[IS_SEL_NT]=trimArray(bb,first);
            isSelectedNtOffset=first<0 ? 0 : first;
        }
    }
    /* <<< Parse <<< */
    /* ---------------------------------------- */
    /* >>> Parse >>> */
    private int  aa_modificationCount;
    private void parseResidueAnnotationIfNecessary(Object[] cached) {
        final String txt=getValue(ResidueAnnotation.POS);
        final StrapProtein p=(StrapProtein)getProtein();
        if (txt!=null && p!=null) {
            final int modi=p.getModificationCount(MC_RESIDUE_TYPE)+this.modiCount+(maxLengthOfUnderlining<<10);
            final boolean isA=isAmino();
            if (cached[IS_SEL]==null || !isA && cached[IS_SEL_NT]==null || aa_modificationCount!=modi) {
                aa_modificationCount=modi;
                if (isA) {
                    cached[IS_SEL]=parseResidueAnnotationAA(txt);
                    isSelectedNtOffset=0;
                } else {
                    parseResidueAnnotationNT(txt,cached);
                    cached[IS_SEL]=p.nucleotideSelection2aminoAcidSelection( (boolean[])cached[IS_SEL_NT],isSelectedNtOffset);
                }
            }

        }
    }

    private boolean[] parseResidueAnnotationAA(String residueRange) {
        if (_range==null)_range=new Range();
        final StrapProtein p=getProtein();
        final int nR=p.countResidues();
        final ReferenceSequence refS;
        final String rr;
        {
            final String s=residueRange.trim();
            final int ref=s.indexOf("REFERENCE=");
            if (ref>=0) {
                final String uniprot=wordAt(ref+10,s);
                refS=p.getReferenceSequence(uniprot,(ByteArray)null, true, (Runnable)null);
                rr=s.substring(0,ref)+substrg(s,  ref+10+uniprot.length(),MAX_INT);
            } else {
                refS=null;
                rr=s;
            }
        }
        final StrapProtein refP=refS!=null ? refS.getReferenceProtein() : null;
        final boolean bb[]=
            refS==null ? p.residueSubsetAsBooleanArray(rr,1, _range) :
            refP!=null ? refS.uniprotIdx2Idx(refP.residueSubsetAsBooleanArray(rr,1)) :
            null;

        if (refP!=null) { _range.set(firstTrue(bb), lastTrue(bb)+1); }
        final int firstTrue=firstTrue(bb);
        if (firstTrue<0) return NO_BOOLEAN;
        this.selectedAAOffset=firstTrue;
        final boolean selAA[]=trimArray(bb,firstTrue);
        final String fn=getFeatureName();
        final Object[] cached=cached();
        if (fn!=null && selAA.length>maxLengthOfUnderlining && countTrue(selAA)==selAA.length) {
            final boolean lastResIsSelected=selAA.length+selectedAAOffset>=nR;
            final boolean[] selAA_d=new boolean[lastResIsSelected ? 2 : selAA.length];
            selAA_d[0]=selAA_d[1]=selAA_d[0]=selAA_d[selAA_d.length-1]=selAA_d[selAA_d.length-2]=true;
            cached[IS_SEL_V]=selAA_d;
        } else cached[IS_SEL_V]=null;
        return selAA;
    }
    public static void loadResidueAnnotation(StrapProtein p,File  file) {
        if (sze(file)==0) return;
        final String txt[]=readLines(file);
        if (txt==null) return;
        int n=0, lineNumber[]=new int[txt.length];
        for(int i=0;i<txt.length;i++) {
            txt[i]=rplcToStrg(CLASS_Hyperrefs,"Note",txt[i]);
            if (txt[i].startsWith(NAME+"=")) lineNumber[n++]=i;
        }
        for(int i=0;i<n;i++) {
            final ResidueAnnotation s=new ResidueAnnotation(p);
            final int
                from=lineNumber[i],
                to= (i==n-1) ? txt.length : lineNumber[i+1];
            final String lines[]=new String[to-from];
            System.arraycopy(txt,from,lines,0,to-from);
            s.parse(lines);
            p.addResidueSelection(s);
        }
    }
    /* <<< Parse <<< */
    /* ---------------------------------------- */
    /* >>> Save >>> */
    private static String line(Entry e) {
        return
            e==null ? null :
            e.getKey()+"="+
            ( e.isEnabled() ? "  X" : "  x")+
            e.getValue()+"\n";
    }
    public static File save(StrapProtein p,File dir) {
        final ResidueAnnotation[] aa=p.getResidueAnnotations();
        final File f=Protein.strapFile(dir,p,Protein.mANNO);
        final StringBuffer sb=new StringBuffer();
        int count=0;
        for(ResidueAnnotation a:aa) {
            if (!isAnno(a)) continue;
            sb.append(line(a.getEntryWithKey(NAME)));
            for(Entry e: a.getEntries())
                if (NAME!=e.getKey())
                    sb.append(line(e));
            final Color c=a.getColor();
            if (c!=null) {
                final int rgb=c.getRGB();
                sb.append(line(a.newEntry(COLOR, ((rgb>>16)&255)+","+ ((rgb>>8)&255)+","+(rgb&255)+","+c.getAlpha(),true)));
            }
            count++;
        }
        if (count==0) delFile(f);
        return wrte(f,sb) ? null : f;
    }
    /* <<<  Save <<< */
    /* ---------------------------------------- */
    /* >>> Entry Instance >>> */
    private Entry newEntry(String key,String value, boolean status) {
        final Entry e=new Entry();
        e.resAnno=WEAK_REF;
        e._status=status;
        String k=key;
        final String sl[][]=short_long();
        for(int i=sl[0].length; --i>=0;) if (k==sl[0][i]) k=sl[1][i];
        e._key=k;
        e._value=value.trim();
        return e;
    }
    public final static class Entry implements ChRunnable {
        private boolean _status;
        private String _key, _value;
        private Object resAnno;
        final Object TEXTFIELD[]={null};
        public void setEnabled(boolean b) {
            if (_status!=b) {
                _status=b;
                final ResidueAnnotation a=getResidueAnnotation();
                if (a!=null) a.modiCount++;
            }
        }
        public ResidueAnnotation getResidueAnnotation() { return deref(resAnno,ResidueAnnotation.class);}
        public String getValue() { final String v=this._value; return v!=null ? v : "";}
        public boolean setValue(final String value) {
            final ResidueAnnotation a=getResidueAnnotation();
            if (a!=null) return a.setValue(value,this,true);
            return false;
        }
        public boolean isEnabled() { return _status;}
        public String getKey() { return _key;}
        public Object run(String id, Object arg) {
            if (id==DialogStringMatch.PROVIDE_STRINGS || id==PROVIDE_WORD_COMPLETION_LIST) return getValue();
            return null;
        }

    }
    /* <<<  Entry Class <<< */
    /* ---------------------------------------- */
    /* >>> Client Property >>> */
    private Map _clientObjects;
    public Map getPropertyMap(boolean create) {return _clientObjects==null&&create?_clientObjects=new HashMap() : _clientObjects;}
    /* <<< Client Property <<< */
    /* ---------------------------------------- */
    /* >>> Sequence Features >>> */
    private String featureName, featureSrc="";
    private int whereLoadedFrom;
    public void setFeatureName(String name, int whereLoadedFrom) {
        featureName=name!=null ? name.intern() : null;  this.whereLoadedFrom=whereLoadedFrom;
    }
    public String getFeatureName() { return featureName;}
    public int getWhereFeatureLoadedFrom() { return whereLoadedFrom;}
    private String cssClass;
    protected static String toCssClass(String s) {
        String css=s!=null ? s.replaceAll("[^0-9A-Za-z_]","_") : "";
        if (css.length()==0) return "annotation";
        return  (is(LETTR,css,0)) ? css : "x"+css;
    }
    public String getCssClass() {
        if (cssClass==null)  cssClass=toCssClass(getFeatureName());
        return cssClass;
    }
    protected void setFeatureSrc(String s) { featureSrc=s;}
    protected String getFeatureSrc() { return featureSrc;}
    private ReferenceSequence simRegion;
    private float simRegionScore=Float.NaN;
    protected float getFeatureAlignmentScore() {  return simRegionScore;}
    protected void setReferenceSequence(ReferenceSequence s, int start, int end) {
        if (s!=null) simRegionScore=s.getAlignmentScore();
        simRegion=s;
    }
   private CharSequence srcText;
    public CharSequence getSrcText() { return srcText;}
    public void setSrcText(CharSequence t) {  srcText=t;}

    /* <<< Sequence Features <<< */
    /* ---------------------------------------- */
    /* >>> Variables >>> */
    public final static String
        FIRST_INDEX="$FIRST_INDEX",
        LAST_INDEX="$LAST_INDEX",
        FIRST_AAA="$FIRST_AAA",
        LAST_AAA="$LAST_AAA",
        FIRST_aaa="$FIRST_aaa",
        LAST_aaa="$LAST_aaa",
        FIRST_Aaa="$FIRST_Aaa",
        LAST_Aaa="$LAST_Aaa",
        FIRST_A="$FIRST_A",
        LAST_A="$LAST_A",
        CHAIN="$CHAIN",
        FIRST_NUMBER="$FIRST_NUMBER",
        LAST_NUMBER="$LAST_NUMBER",
        SORE_NAME="$SORE_NAME",
        TEXSHADE_RESIDUES="$TEX_RESIDUES",
        PROTEIN_NO="$TEX_PROTEIN",
        TEXSHADE_RESIDUES_OLD="{RESIDUES}",
        PROTEIN_NO_OLD="{PROTEIN}";
    public final static String VARIABLES[]={FIRST_INDEX,LAST_INDEX,FIRST_AAA,LAST_AAA,FIRST_aaa,LAST_aaa,FIRST_Aaa,LAST_Aaa,FIRST_A,LAST_A,CHAIN,FIRST_NUMBER,LAST_NUMBER,SORE_NAME,TEXSHADE_RESIDUES,PROTEIN_NO};

    public static  String replaceVariables(ResidueAnnotation a, String s0) {
        return a==null ? null : replaceVariables(a,a.getProtein(), s0,null);

    }
    public static  String replaceVariables(Protein p, String s0,Protein allProteins[]) {
        return replaceVariables(null, p, s0,allProteins);

    }

    private static  String replaceVariables(ResidueAnnotation a,Protein p, String s0,Protein allProteins[]) {
        if (s0==null) return null;
        String s= replaceVariables(a!=null ? a.getSelectedAminoacids():null, a!=null ? a.getSelectedAminoacidsOffset():0, p,s0, a!=null ? a.getName():null,allProteins);
        final Color color= a!=null ? a.getColor() : null;
        if (color!=null && s.indexOf(Pymol.DEFINE_THIS_COLOR)>=0 ) {
            final int rgb=color.getRGB();
            s=rplcToStrg(Pymol.DEFINE_THIS_COLOR, "set_color "+pymolColorName(color)+" , [ "+  ( (rgb>>16)&255)/256f+","+(((rgb>>8)&255)/256f)+","+((rgb&255)/256f)+" ] ", s);
        }
        if (s.indexOf(Pymol.PYMOL_COLOR_NAME)>=0) {
            s=rplcToStrg(Pymol.PYMOL_COLOR_NAME, pymolColorName(color), s);
        }
        return s;
    }
    private static String pymolColorName(Color c) { return rgbHex(c.getRGB(),new StringBuffer(7).append('c')).toString();  }
    public static String replaceVariables(boolean selectedAa[],int offset,Protein p0,String s0,String selectionName,Protein allProteins[]) {
        final StrapProtein p=(StrapProtein)p0;
        String s=s0;
        if (s==null) return null;
        final boolean
            isFIRST_INDEX=s.indexOf(FIRST_INDEX)>=0,
            isLAST_INDEX=s.indexOf(LAST_INDEX)>=0,
            isFIRST_AAA=s.indexOf(FIRST_AAA)>=0,
            isLAST_AAA=s.indexOf(LAST_AAA)>=0,
            isFIRST_aaa=s.indexOf(FIRST_aaa)>=0,
            isLAST_aaa=s.indexOf(LAST_aaa)>=0,
            isFIRST_Aaa=s.indexOf(FIRST_Aaa)>=0,
            isLAST_Aaa=s.indexOf(LAST_Aaa)>=0,
            isFIRST_A=s.indexOf(FIRST_A)>=0,
            isLAST_A=s.indexOf(LAST_A)>=0,
            isCHAIN=s.indexOf(CHAIN)>=0,
            isFIRST_NUMBER=s.indexOf(FIRST_NUMBER)>=0,
            isLAST_NUMBER=s.indexOf(LAST_NUMBER)>=0,
            isSORE_NAME=s.indexOf(SORE_NAME)>=0,
            isTEXSHADE_RESIDUES=s.indexOf(TEXSHADE_RESIDUES)>=0,
            isPROTEIN_NO=s.indexOf(PROTEIN_NO)>=0,
            isTEXSHADE_RESIDUES_OLD=s.indexOf(TEXSHADE_RESIDUES_OLD)>=0,
            isPROTEIN_NO_OLD=s.indexOf(PROTEIN_NO_OLD)>=0;
        if (!(isFIRST_INDEX||isLAST_INDEX||isFIRST_AAA||isLAST_AAA||isFIRST_aaa||isLAST_aaa||isFIRST_Aaa||isLAST_Aaa||isFIRST_A||isLAST_A||isCHAIN||isFIRST_NUMBER||isLAST_NUMBER||isSORE_NAME||isTEXSHADE_RESIDUES||isPROTEIN_NO||isTEXSHADE_RESIDUES_OLD||isPROTEIN_NO_OLD)) return s;
        final StrapAlign align=StrapAlign.getInstance();
        final int
            idxFirst=firstTrue(selectedAa)+offset,
            idxLast=lastTrue(selectedAa)+offset,
            nF=idxFirst<0 || p==null ? -1  : p.getResidueIndexForExport(idxFirst),
            nL=idxFirst<0 || p==null ? -1  : p.getResidueIndexForExport(idxLast);
        final char
            aF=idxFirst<0 || p==null ? '?' : (char) p.getResidueType(idxFirst),
            aL=idxFirst<0 || p==null ? '?' : (char) p.getResidueType(idxLast),
            c=idxFirst<0 || p==null  ? '?' : (char) p.getResidueChain(idxFirst);
        final String aaaF=Protein.toThreeLetterCode(aF),aaaL=Protein.toThreeLetterCode(aL);
        if (isFIRST_INDEX)          s=rplcToStrg(FIRST_INDEX,String.valueOf(idxFirst+1),s);
        if (isLAST_INDEX)           s=rplcToStrg(LAST_INDEX,String.valueOf(idxLast+1),s);
        if (isFIRST_AAA)            s=rplcToStrg(FIRST_AAA,String.valueOf(aaaF),s);
        if (isLAST_AAA)             s=rplcToStrg(LAST_AAA,String.valueOf(aaaL),s);
        if (isFIRST_aaa)            s=rplcToStrg(FIRST_aaa,aaaF.toLowerCase(),s);
        if (isLAST_aaa)             s=rplcToStrg(LAST_aaa,aaaL.toLowerCase(),s);
        if (isFIRST_Aaa)            s=rplcToStrg(FIRST_Aaa,aaaF.substring(0,1)+aaaF.substring(1).toLowerCase(),s);
        if (isLAST_Aaa)             s=rplcToStrg(LAST_Aaa,aaaL.substring(0,1)+aaaL.substring(1).toLowerCase(),s);
        if (isFIRST_A)              s=rplcToStrg(FIRST_A,String.valueOf(aF),s);
        if (isLAST_A)               s=rplcToStrg(LAST_A,String.valueOf(aL),s);
        if (isCHAIN)                s=rplcToStrg(CHAIN,String.valueOf(c),s);
        if (isFIRST_NUMBER)         s=rplcToStrg(FIRST_NUMBER,String.valueOf(nF),s);
        if (isLAST_NUMBER)          s=rplcToStrg(LAST_NUMBER,String.valueOf(nL),s);
        if (isSORE_NAME)            s=rplcToStrg(SORE_NAME,selectionName,s);
        if (isTEXSHADE_RESIDUES)    s=rplcToStrg(TEXSHADE_RESIDUES,ResidueSelectionUtils.texshadeRange(selectedAa,offset,p),s);
        if (isPROTEIN_NO)           s=rplcToStrg(PROTEIN_NO,String.valueOf(allProteins==null ? 1 : idxOf(p,allProteins)+1),s);
    if (isTEXSHADE_RESIDUES_OLD)s=rplcToStrg(TEXSHADE_RESIDUES_OLD,"{"+ResidueSelectionUtils.texshadeRange(selectedAa,offset,p)+"}",s);
        if (isPROTEIN_NO_OLD)       s=rplcToStrg(PROTEIN_NO_OLD,"{"+(idxOf(p,align.getSelectedProteins())+1)+"}",s);
        return s;
    }
     /* <<< Variables <<< */
    /* ---------------------------------------- */
    /* >>> Utils >>> */
    private static String short_long[][];
    protected static String[][] short_long() {
        String sl[][]=short_long;
        if (sl==null) {
            final String[] l={CLASS_ChJmolPROXY, CLASS_Texshade, CLASS_Pymol, CLASS_Vmd, CLASS_Rasmol},s=new String[l.length];
            for(int i=l.length; --i>=0;) {
                s[i]=lastComponent(delSfx("PROXY",l[i]), '.').intern();
            }
            short_long=sl=new String[][]{s,l};
        }
        return sl;
    }
    protected static boolean isAnno(Object o) {return o instanceof ResidueAnnotation && ((ResidueAnnotation)o).getFeatureName()==null;}
    protected static boolean isEnabledAnno(Object o) {return isAnno(o) && ((ResidueAnnotation)o).isEnabled();}

    /* <<< Utils <<< */
    static List vAll=new ArrayList(999);

}

