package charite.christo.strap;
import charite.christo.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import charite.christo.protein.*;
import charite.christo.strap.extensions.Blaster_ebi_ac_uk;
import java.io.File;
import java.awt.*;
//import java.awt.event.*;
import java.net.URL;
import java.util.*;
import java.util.List;
import javax.swing.*;
import javax.swing.table.*;
import static charite.christo.ChUtils.*;
import static charite.christo.ChConstants.*;
import static java.awt.Color.*;
import static java.awt.event.MouseEvent.*;
// (setq fill-column 100 )
/**HELP
   <h2>Sequence Features</h2>

   Sequence features are annotations of signal peptides, disulfide bonds,
   glycosylated or phosphorylated amino acids at sequence level.  In
   Strap they are loaded from the protein file, from BioDAS* servers, from Expasy or from the
   Catalytic Site Atlas at the EBI.  They are underlined in the alignment
   panel and in the alignment output.  They have a context menu
   (right-click), a balloon message and they can be dragged to other proteins or 3D-views.

   <br><br>

   <h3>By sequence similarity</h3>

   Sequence features my be indirectly inferred from a homologous protein ("By similarity").  In this
   case the sequence positions are assigned using a sequence alignment method

   If the alignment is poor, the sequence features is deaktivated and not shown.

   Sequence features are special types of residue annotations with the difference that they are not
   saved when the project is saved.

   <h3>Colors</h3>

   Each feature type has a color for black background (screen), and one for white background (paper).
   This table can be customized <i>CUSTOMIZE:SequenceFeatures#customizeColors()!</i>.

   <h3>Deactivated Features</h3>

   Deactivated features are hidden unless the respective choice-menu item is activated. Reasons why
   features are deaktivated:

   <ul>
   <li>Poor alignment with the homologous sequence, where the feature is indirectly inferred from.</li>

   <li>The Sequence positions do not fall into the shown sequence fragment. This is often the case for
   N-terminal signal sequences in Pfam alignments showing only the conserved part.</li>

   <li>Secondary structure when the the structure information is already obtained from the protein
   file. Files with secondary structure information: PDB, Pfam (Stockholm-Format), DSSP.</li>

   <li>Phospho-tyrosine falling on a non-'Y' residue. This is controlled by the third column of the
   customizable table which contains a set of allowed letters or a dash if all letters are allowed
   <i>CUSTOMIZE:SequenceFeatures#customizeColors()!</i> </li>

   </ul>

   <h3>Mouse-control</h3>

   The following mouse actions  apply to the sequence features in the sequence feature table and to the alignment panel.
   <ul>

   <li>Right click opens a context menu for the feature.</li>

   <li>CTRL left-click  selects and deselects the sequence features.</li>

   <li>Dragging the mouse copies sequence features to other proteins or to 3D-views (See  WIKI:Drag_and_drop).</li>

   <li>Double clicking a feature opens a table with detailed data.</li>

   <li>Moving the mouse provides a balloon text with information on the sequence feature under the mouse.</li>

   </ul>

   <h3>Examples</h3>
   The following Swissprot entries have many sequence features and can be used for testing/demonstration:
   SWISS:Q9Y4K3 &nbsp; SWISS:Q9H9D4 &nbsp;  SWISS:Q05516 &nbsp;  SWISS:P28825 &nbsp;  SWISS:P08045 &nbsp;  SWISS:Q13105
   <br>
   However, for many other protein files no sequence features is recorded yet.
   @author Christoph Gille
   <h3>Problems</h3>
   There are rare cases where mapping goes wrong because the sequence of Uniprot entries changed slightly.
   Example: In earlier versions PRC3_HUMAN and UNIPROT:P25787 was lacking the initial Methionine.
   <br><br><u>Related documents</u>
   <ul>
   <li>http://www.uniprot.org/docs/humsavar</li>
   <li>BIOWIKI:GffFormat</li>
   </ul>
*/
public class SequenceFeatures extends DefaultTableModel  implements Comparator,ChRunnable,ChProcessEvent, TooltipProvider {
    final static int  CSA_EBI=1<<1, GFF_BLAST_IF_NO_ID=1<<5;
    private final static int
        GFF_PROTEIN_FILE=1<<0, GFF_EXPASY=1<<2, GFF_DAS=1<<3,  GFF_BLAST=1<<4,
        MAX_DAYS=33, ROW_HEIGHT=3*EX/2,
        SEQ=0, WIRE=1, PDF=2, HTML=3, PYMOL=4, JALVIEW=5,  BOOLEANS=6, BOOL_START=3, NAVI_START=BOOL_START+1+BOOLEANS;
    final static String ALI_SCORE=" Alignment-Score=";
    private static String[] featureNames=NO_STRING;
    private final static String TT_default=
        "<u>Note:</u> The sequence features such as \"Phosphorylated Tyrosine\"<br>"+
        "will be underlined in the alignment panel.";
    final static ByteArray log=new ByteArray(99);

    /* ---------------------------------------- */
    /* >>> Instance  >>> */
    private ChEventAdapter LI=new ChEventAdapter(this);
    private static SequenceFeatures _inst;
    public static SequenceFeatures instance() {if (_inst==null) _inst=new SequenceFeatures(); return _inst;}
    static ChButton getButton() { return _inst!=null ? _inst.butDefault : null; }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Table >>> */
    public SequenceFeatures() {
        super(rplcToStrg(">>","</body></html>",
                         rplcToStrg("<<","<html><body>",
                                    "\n"+
                                    "Feature\n"+
                                    "#\n"+
                                    "<<Sequ-<br>ence>>\n"+
                                    "<<3D-<br>wire>>\n"+
                                    "PDF\n"+
                                    "HTML\n"+
                                    "<<Py-<br>mol>>\n"+
                                    "<<Jal<br>view>>\n"+
                                    "<<Web-<br>links>>\n"+
                                    "<<Go<br>first>>\n"+
                                    "<<Go<br>previous>>\n"+
                                    "<<Go<br>next>>\n"+
                                    "<<Go<br>last>>")).split("\n"),0);

    }
    @Override public boolean isCellEditable(int row, int col){ return col>=BOOL_START; }
    private ChButton BUT_WEB,BUT_WEB_EDITOR, butExpasy, butDefault;
    private StrapProtein[] proteins() {
        final StrapAlign a=StrapAlign.getInstance();
        return onlySelected() ? a.getSelectedProteins() : a.getProteins();
    }
    @Override public Object getValueAt(int row, int col){
        final String ft[]=featureNames;
        final String f=ft.length>row ? ft[row] : "error";
        if (col==0) return new Object[]{ChRenderer.KEY_TOGGLE_SELECTED,screenColor(featureName2color(f))};
        if (col==1) return f;
        if (col==BOOL_START+BOOLEANS) {
            for(StrapProtein p : proteins())
                for(ResidueAnnotation a :p.getResidueAnnotations())
                    if (a.getLastValue(CLASS_Hyperrefs)!=null)  return BUT_WEB;
            return null;
        }

        if (NAVI_START<=col && col<NAVI_START+4) {
            final Color color=paperColor(featureName2color(f));

            return bbNext[col-NAVI_START].fg(color!=null && color!=WHITE ? color : BLACK);
        }
        final int iB=col-BOOL_START;
        int count=0;
        for(StrapProtein p : proteins()) {
            for(ResidueAnnotation a :p.getResidueAnnotations()) {
                if (a.getFeatureName()!=f) continue;
                count++;
                if (col>=BOOL_START) {
                    if (
                        (ib2vw(iB) &  a.getVisibleWhere())!=0 ||
                        iB==PYMOL &&  a.isEnabled(CLASS_Pymol) ||
                        iB==PDF   && a.isEnabled(CLASS_Texshade)
                        ) return Boolean.TRUE;
                }
            }
        }
        if (col==2) return intObjct(count);
        return Boolean.FALSE;
    }
    private final static int ib2vw(int F) {
        return
            F==WIRE ? VisibleIn123.STRUCTURE :
            F==HTML ? VisibleIn123.HTML :
            F==SEQ ? VisibleIn123.SEQUENCE :
            F==JALVIEW ? VisibleIn123.JALVIEW: 0;
    }
    private void getFeaturesAtRow(int rows[], List v) {
        final String ft[]=this.featureNames;
        for(StrapProtein p : proteins()) {
            for(ResidueAnnotation a :p.getResidueAnnotations()) {
                for(int r : rows) {
                    if (ft.length>r && r>=0 && a.getFeatureName()==ft[r]) v.add(a);
                }
            }
        }
    }

    private final static List<ResidueAnnotation.Entry> v_setValueAt=new ArrayList();
    @Override public void setValueAt(Object o, int row, int col){
        final String ft[]=featureNames;
        final int iB=col-BOOL_START;
        if (iB<0 || iB>=BOOLEANS ||  row>=ft.length) return;
        final String fName=ft[row];
        final StrapProtein pp[]=proteins();
        if (pp.length==0) StrapAlign.errorMsg("SequenceFeatures Error: no protein");
        final boolean on=o==Boolean.TRUE;
        for(StrapProtein p : pp) {
            for(ResidueAnnotation a :p.getResidueAnnotations()) {
                if (fName!=a.getFeatureName()) continue;
                if (on) a.setEnabled(true);
                final int vw0=ib2vw(iB);
                for(int i=0;i<2;i++) {
                    final int vw=i==0 ?  vw0  : vw0==VisibleIn123.SEQUENCE ? VisibleIn123.SB : 0;
                    if (vw!=0) a.setVisibleWhere( on ? (a.getVisibleWhere()|vw) : (a.getVisibleWhere() &~ vw));
                }
                final String key=
                    iB==PYMOL ? CLASS_Pymol :
                    iB==PDF ? CLASS_Texshade :
                    null;
                if (key!=null) {
                    for(ResidueAnnotation.Entry e : a.getEntries()) {
                        if (key==e.getKey() && e.isEnabled()) v_setValueAt.add(e);
                    }
                    for(int iV= sze(v_setValueAt); --iV>=0;) v_setValueAt.get(iV).setEnabled(on);
                    v_setValueAt.clear();
                }
            }
        }
        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,11);
    }
    static void updateTable(int delay) {
        final ChJTable t=_inst!=null ? _inst.jTable : null;
        if (t!=null) { ChDelay.revalidate(t,delay); ChDelay.repaint(t,delay); }
    }
    private int tt_last_row;
    private String tt_last;
    public String getToolTipText(JComponent component, AWTEvent ev) {
        final Point point=point(ev);
        if (point==null) return null;
        if (jTable.columnAtPoint(point)>1) return null;
        final int row=jTable.rowAtPoint(point); // !!!
        if (row==tt_last_row && tt_last!=null) return tt_last;
        final StringBuffer sb=new StringBuffer();
        final Set setRemarks=new HashSet();
        final String[] nn=this.featureNames;
        if (row>=0 && row<nn.length) {
            for(StrapProtein p:proteins()) {
                int isFirstA=-1;
                setRemarks.clear();
                for (ResidueAnnotation a:p.getResidueAnnotations()) {
                    if (a.getFeatureName()!=nn[row] || !a.isEnabled() || a.getVisibleWhere()==0) continue;
                    if (isFirstA<0) { isFirstA=sze(sb); sb.append("<br>").append(p).append("  "); }
                    final int from=a.getSelectedAminoacidsOffset();
                    final int to=from+a.getSelectedAminoacids().length;
                    final String attribute=a.getValue("Remark");
                    sb.append(from);
                    if (to-from>1) sb.append('-').append(to-1);
                    if (sze(sb)-isFirstA>110) { sb.append(" ...."); break;}
                    if (sze(attribute)>3 && !setRemarks.contains(attribute)) {
                        setRemarks.add(attribute);
                        sb.append("<small>(").append(htmlEncod(attribute)).append(")</small>");
                    }
                    sb.append("   ");
                }
            }
        }
        return tt_last=addHtmlTagsAsStrg(sb);
    }
    /* <<< Table <<< */
    /* ---------------------------------------- */
    /* >>> GUI >>> */
    private JComponent pnl, labMsg=new JLabel(" ");
    private ChJTable jTable;
    private ChFrame frame, frameWeb;
    private ChTextArea taWeb;
    private ChButton cbNotAll, cbArrows, cbBlast, bbNext[]=new ChButton[4];
    private ChCombo comboAll;
    private JProgressBar progress=new JProgressBar(); { progress.setPreferredSize(dim(0,0)); }
    private ChRenderer renderer=new ChRenderer();
    protected static boolean onlySelected() { return _inst!=null && isSelctd(_inst.cbNotAll);}
    public void showPanel() {
        DAS.readRegistry((ChRunnable)this);
        if (pnl==null) {
            for(int i=bbNext.length; --i>=0;) {
                final String
                    t=i==0 ? "|<" : i==1 ? "<" : i==2 ? ">" : ">|",
                    tt="Go to "+ (i==0 ? "first": i==1 ? "previous" : i==2 ? "next" : "last")+" selected amino acid";
                final ChButton b=bbNext[i]=new ChButton(t,ChButton.NO_MARGIN|ChButton.PAINT_IN_TABLE).li(LI).tt(tt);
                pcp(ChRenderer.KEY_KEEP_COLOR,"",b);
                pcp(ChRenderer.KEY_EDITOR_COMPONENT, b.cln().options(ChButton.NO_MARGIN), b);
            }
            BUT_WEB_EDITOR=new ChButton(null,ChButton.NO_FILL|ChButton.NO_BORDER).i(IC_FIREFOX16).li(LI);
            BUT_WEB=new ChButton(null,ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE).i(IC_FIREFOX16);
            pcp(ChRenderer.KEY_EDITOR_COMPONENT,BUT_WEB_EDITOR,BUT_WEB);
            jTable=new ChJTable(this, ChJTable.NO_REORDER);
            jTable.setRowHeight(ROW_HEIGHT);
            setProcessEvt(this,jTable);
            addActLi(LI,jTable);
            final Comparator[] comparators=new Comparator[BOOLEANS+BOOL_START+1];
            for(int i=comparators.length; --i>=0;) comparators[i]=jTable;
            ChJTable.setRowSorter(comparators,jTable);
            rtt(jTable);
            jTable.setDefaultRenderer(Object.class,renderer);
            jTable.setDefaultEditor(Object.class,renderer);
            cbBlast=toggl("Blast search to find missing Uniprot IDs");
            cbArrows=toggl("Show only arrow heads for  long underlinings").s(true).li(LI);
            cbNotAll=toggl("Act not on all but only on selected proteins")
                .tt("Select proteins by left-click the protein-label in the alignment").li(LI);
            comboAll=new ChCombo("Show those and only those features that are activated","Show all features, including those with poor alignment","Hide all features").li(LI)
                .tt("Features with poor alignment to the reference sequence are deactivated but can be activated by the user");
            butExpasy=new ChButton("Expasy",ChButton.NO_MARGIN).li(LI)
                .tt("GFF files are loaded from Expasy. These files contain lines of the form"+
                    "<pre class=\"data\">P25787	UniProtKB	Modified residue	24	24	.	.	.	Note=Phosphotyrosine</pre>"+
                    "As long as one job is runnting the button remains deactivated.<br>"+
                    "The process can be interrupt by closing this frame."
                    );
            butDefault=new ChButton("Load sequence features of "+StrapAlign.LABEL_ALL_OR_SELECTED_P,ChButton.NO_MARGIN).li(LI).tt(TT_default);
            StrapAlign.setNumOfProteins(butDefault);
            final JComponent
                hb=pnl(Customize.newButton(customizeColors()).options(ChButton.NO_FILL|ChButton.NO_BORDER|ChButton.ICON_SIZE).magnifyUnderMouse(IC_CUSTOM), panHS(SequenceFeatures.class)),
                panAboveT=vB(hB("#",WHITE,cbNotAll.cb()),hB("#",WHITE, comboAll)),

                pTop=newCNSEW(pnl("<h2>Sequence Features</h2>"),null,panAboveT, hb,pnl(new ChButton("Log",ChButton.NO_MARGIN).li(LI))),
                b1=new ChButton("FROM_FILE",ChButton.NO_MARGIN).li(LI).t("Protein files")
                .tt("Some protein files contain sequence features lines like <pre class=\"data\">FT   METAL       141    141       Copper.\n\nSITE     1 CAT  3 ASP A  32  HIS A  64  SER A 221\n\nMODRES 1IVO ASN A   32  ASN  GLYCOSYLATION</pre>"),
                b3=new ChButton("FROM_DAS",ChButton.NO_MARGIN).li(LI).t("BioDAS...")
                .tt("The user can chose from a large number of BioDAS Servers"),
                pMoreAdvanced=vB(
                                 pnl(monospc("Spice-Browser* on selected protein"),new ChButton("SPICE",ChButton.NO_MARGIN).li(LI).t(ChButton.GO)),
                                 pnl(monospc("Dasty on selected protein"),new ChButton("DASTY",ChButton.NO_MARGIN).li(LI).t(ChButton.GO))
                                 ),
                pLoadFrom=vBpnl(pnl("Load from ",b1,butExpasy,b3), cbArrows.cb(),  cbBlast.cb()),
                pLoadDefault=vBpnl(butDefault,TT_default,WATCH_MOVIE+MOVIE_Sequence_Features_in_3D),
                pAdvanced=vB(
                             hB("Other options ",toggl().actionCollapse(false,pMoreAdvanced).actionCollapse(true,pLoadFrom),"#",pnl(pLoadFrom),"#"),
                             pMoreAdvanced
                             ),
                togA=toggl().actionCollapse(false,new Object[]{ pAdvanced, panAboveT}).actionCollapse(true, new Object[]{butDefault,pLoadDefault}),
                pSouth0=vB( hB(pLoadDefault,"#","Options ", togA),pAdvanced ),
                pSouth=newCNSEW(pSouth0, progress,labMsg,null,null);

            repaintOnMouseEnterAndExit(labMsg);
            pnl=newCNSEW(scrllpn(SCRLLPN_SEARCH_BUTTON,jTable),pTop,pSouth, null,null);
            setProcessEvt(instance(),pnl);
            frame=new ChFrame("feature").ad(pnl).size(EM*70,EX*30).shw(CLOSE_ON_CtrlW);
            for(int i=NAVI_START+4;--i>=0;) {
                if (i!=1) jTable.getColumnModel().getColumn(i).setMaxWidth(i==2 ? EM*5 : EM*3);
            }
            jTable.getTableHeader().setPreferredSize(dim(32,32));
            enlargeFont(jTable.getTableHeader(),.8);
        }
        setSettings(instance(),customizeCSS_IE(), customizeColors());
        frame.shw(ChFrame.CENTER);
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Sort >>> */
    public int compare(Object o1, Object o2) {
        if (o1 instanceof ResidueAnnotation && o2 instanceof ResidueAnnotation) {
            final ResidueAnnotation a1=(ResidueAnnotation)o1, a2=(ResidueAnnotation)o2;
            final String f1=a1.getFeatureName(), f2=a2.getFeatureName();
            if (f1!=null && f2!=null) {
                final int diffLen=a1.getSelectedAminoacids().length-a2.getSelectedAminoacids().length;
                if (diffLen!=0) return -diffLen;
                final int diffOff=a1.getSelectedAminoacidsOffset()-a2.getSelectedAminoacidsOffset();
                if (diffOff!=0) return -diffOff;
                final int diffName=f1.compareTo(f2);
                if (diffName!=0) return diffName;
            }
        }
        return 0;
    }
    /* <<< Sort <<< */
    /* ---------------------------------------- */
    /* >>> Ontology >>> */
    private static Set<String> setSeqOnt;
    private static Set<String> setSequenceOntologyTermsLC() {
        if (setSeqOnt==null) {
            setSeqOnt=new HashSet();
            final ByteArray ba=readBytes(rscAsStream(_CC,"sequence_ontology_terms.rsc"));
            if (ba!=null) {
                final byte T[]=ba.getBytes();
                for(int i=ba.getEnd(); --i>=0;)  if ('A'<=T[i] && T[i]<'Z') T[i]=(byte)(T[i]|32);
                final ChTokenizer TOK=new ChTokenizer().setText(ba);
                while(TOK.nextToken()) setSeqOnt.add(TOK.asString());
            } else stckTrc();
        }
        return setSeqOnt;
    }
    /* <<< Ontology <<< */
    /* ---------------------------------------- */
    /* >>> AWTEvent >>> */
    private int lastRow;
    private final Map<String,NextAndPrevSelection> mapNextPrev=new HashMap();
    private final static List<ResidueAnnotation> vTMP=new ArrayList();
    public AWTEvent processEvent(JComponent component, AWTEvent ev) {
        final int id=ev.getID(), modi=modifiers(ev);
        final String cmd=actionCommand(ev);
        final Object q=ev.getSource();
        final StrapAlign align=StrapAlign.getInstance();
        final StrapView view=align.getAlignmentPanelWithFocus();
        vTMP.clear();
        if (cmd=="<" || cmd==">" || cmd=="|<" || cmd==">|") {
            final int row=ChJTable.rowIndexToModel(renderer.editedRow(),jTable);
            final String name=get(row,featureNames, String.class);
            if (name!=null) {
                NextAndPrevSelection nextPrev=mapNextPrev.get(name);
                if (nextPrev==null) mapNextPrev.put(name,nextPrev=new NextAndPrevSelection());
                vTMP.clear();
                getFeaturesAtRow(new int[]{row}, vTMP);
                final ResidueAnnotation ss[]=toArry(vTMP,ResidueAnnotation.class);
                nextPrev.goNextMatch(cmd,ss);
            }
        }
        if (component!=null && component==jTable && cmd==null) {
            final int row=y(ev)/ROW_HEIGHT;
            if (isPopupTrigger(ev)) {
                if (0<=row && row<getRowCount()) {
                    final int selected[]=ChJTable.selectedRowsConverted(jTable);
                    getFeaturesAtRow(idxOf(row, selected,0,MAX_INT)>=0 ? selected : new int[]{row} ,(List)vTMP);
                    StrapAlign.showContextMenu('F', toArry(vTMP,ResidueAnnotation.class),ev);
                }
            }
            if (id==MOUSE_MOVED && view!=null && lastRow!=row) {
                getFeaturesAtRow(new int[]{row},vTMP);
                final ResidueAnnotation ss[]=toArry(vTMP,ResidueAnnotation.class);
                final StrapProtein pp[]=align.getVisibleProteins();
                final int iiP[]=new int[ss.length];
                Color color=null;
                for(int i=ss.length; --i>=0;) {
                    iiP[i]=idxOf(ss[i].getProtein(),pp);
                    if (color==null) color=ss[i].getColor();
                }
                view.highlightProteins(iiP,new Color[]{color});
                lastRow=row;
            }
            if (id==MOUSE_DRAGGED && !isCellEditable(row, jTable.columnAtPoint(point(ev)))) {
                final List v=getDndFilesV(jTable);
                v.clear();
                getFeaturesAtRow(new int[]{row},v);
                exportDrg(jTable, ev, TransferHandler.COPY);
            }
        }
        if (id==MOUSE_EXITED) lastRow=-1;
        if (id==MOUSE_ENTERED && view!=null) { lastRow=-1;  view.highlightProteins(null,null);}
        if (cmd!=null) {
            StrapAlign.errorMsg("");

            final StrapProtein ppSel[]=align.getSelectedProteins();
            if (cmd=="SPICE" || cmd=="DASTY") {
                if (ppSel.length!=1) error("For Spice-Browser* or Dusty please select exactly one protein");
                else {
                    if (cmd=="SPICE") toSpiceViewer(ppSel[0]);
                    if (cmd=="DASTY") {
                        final String uniprot=ppSel[0].getUniprotID();
                        if (uniprot==null) error("No uniprot ID associated with "+ppSel[0]);
                        else visitURL(Hyperrefs.toUrlString("DASTY:"+uniprot.substring(uniprot.indexOf(':')+1),Hyperrefs.WEB_LINK), modi);
                    }
                }
            }
            if (q==jTable && cmd==ACTION_SELECTION_CHANGED) {
                StrapAlign.getSelectedObjectsV().clear();
                getFeaturesAtRow(ChJTable.selectedRowsConverted(jTable), StrapAlign.getSelectedObjectsV());
                StrapAlign.dispatchLater(StrapEvent.OBJECTS_SELECTED,11);
            }
            if (q==frame && cmd==ACTION_WINDOW_CLOSING && butExpasy!=null) butExpasy.enabled(true);
            if (q==cbNotAll || q==comboAll || q==cbArrows) {
                if (q==cbArrows) ResidueAnnotation.setMaxLengthOfUnderlining(cbArrows.s() ?6:99999);
                Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED,111);
            }
            final boolean fromFile=cmd=="FROM_FILE" || q==butDefault;
            if (fromFile || q==butExpasy || q==butDefault) {
                final StrapProtein pp[]=ppSel.length>0 ? ppSel: align.getVisibleProteins();
                if (fromFile) ChDelay.newThread(this,"VAR", readFeaturesInProteinFile(null,MAX_INT, pp) ).start();
                final Thread run=
                    q==butExpasy  ? thread_loadGFF(pp, GFF_EXPASY|CSA_EBI| (cbBlast.s() ? GFF_BLAST_IF_NO_ID:0)) :
                    q==butDefault ? StrapDAS.threadSequenceFeatures(0L, DAS.IMPORTANT_FEATURES, pp, progress, (ChRunnable)this) :
                    null;
                if (run!=null) {
                    ((JComponent)q).setEnabled(false);
                    progress.setPreferredSize(dim(TF_HEIGHT,EX));
                    newThrd(run,
                            new ChDelay(EDT,ChDelay.ENABLED,q,Boolean.TRUE),
                            new ChDelay(EDT,ChDelay.SET_PREF_SIZE,progress,dim(0,0))
                            ).start();
                }
            }
            if (cmd=="FROM_DAS") StrapDAS.show(true);
            if (cmd=="Log") new AsciiView(log).showInFrame("Log",0); /* TODO update problem */
            if (q instanceof Customize && q==customizeColors()) {
                mapFeatureColors.clear();
                customizeCSS_IE=null;
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR,999);
                ChDelay.repaint(this,999);
            }
            if (q==BUT_WEB_EDITOR) {
                final int row=ChJTable.rowIndexToModel(renderer.editedRow(),jTable);
                final String[] nn=featureNames;
                if (row<0 || row>=nn.length) return null;
                if (frameWeb==null) frameWeb=new ChFrame("feature URLs").ad(new JScrollPane(taWeb=new ChTextArea(""))).shw(CLOSE_ON_CtrlW);
                final Set vURLs=new HashSet();
                int max=0;
                final StringBuffer sb=new StringBuffer().append(nn[row]).append('\n');
                for(int pass=2; --pass>=0;) {
                    for(StrapProtein p:proteins()) {
                        nextFeature:
                        for (ResidueAnnotation a:p.getResidueAnnotations()) {
                            if (a.getFeatureName()!=nn[row] || !a.isEnabled() || a.getVisibleWhere()==0) continue;
                            final int L=sze(p.getName());
                            if (pass>0) { max=maxi(max,L); continue;}
                            boolean firstUrl=true;
                            for(ResidueAnnotation.Entry e : a.getEntries()) {
                                if (CLASS_Hyperrefs!=e.getKey()) continue;
                                final String url=e.getValue();
                                if (firstUrl) {
                                    if (vURLs.contains(url)) continue nextFeature;
                                    vURLs.add(url);
                                    for(int j=max-L+1; --j>=0;) sb.append(' ');
                                    sb.append(a.getProtein()).append(' ');
                                    firstUrl=false;
                                }
                                sb.append(' ').append(url);
                            }
                            sb.append('\n');
                        }
                    }
                }
                taWeb.t(sb);
                taWeb.tools().underlineRefs(0);
                frameWeb.shw(0);
            }
        }
        return ev;
    }
    /* <<< AWTEvent <<< */
    /* ---------------------------------------- */
    /* >>> Color >>> */
    private final static Map<String,FeatureColor> mapFeatureColors=new HashMap();
    public static void cssColors(StrapProtein pp[], StringBuffer sb) {
        final Map<String,Color> map=new HashMap();
        for(StrapProtein p:pp) {
            for(ResidueAnnotation a:p.getResidueAnnotations()) {
                final String n=a.getFeatureName();
                if (n!=null) {
                    map.put(n,a.getColorForWhiteBG());
                }
            }
        }
        for(Map.Entry<String,Color> e : entryArry(map)) {
            final int color=e.getValue().getRGB();
            sb.append('.').append(e.getKey()).append(" { background-color: #");
            rgbHex(color,sb).append("; color: #");
            rgbHex(color,sb).append("; } \n");
        }
    }
    private static Customize customizeColors, customizeCSS_IE;
    public static Customize customizeCSS_IE() {
        if (customizeCSS_IE==null) {
            parseColors();
            final Object oo[]=toArry(new HashSet(mapFeatureColors.values()),Object.class);
            StringBuffer sbIE=null;
            for(int i=oo.length;--i>=0;) {
                final FeatureColor fc=(FeatureColor)oo[i];
                final String cssClass=ResidueAnnotation.toCssClass(fc.name);
                if (sbIE==null) sbIE=new StringBuffer(999).append("PRE { margin: 0mm;}\n PRE A, ");
                else sbIE.append(", ");
                sbIE.append(" .").append(cssClass);
            }
            sbIE.append(" { background-color: transparent;  } ");
            final String hlpIE="Workaround for IE bugs.";
            customizeCSS_IE=new Customize("cssInternetExplorer",new String[]{"/* "+hlpIE+" */",sbIE.toString()},hlpIE,Customize.LIST);
        }
        return customizeCSS_IE;
    }
    public static Customize customizeColors() {
        if (customizeColors==null) {
            final String
                help="# WIKI:Web_colors of features for black and white background.",
                rsc="charite/christo/strap/SequenceFeatures.rsc",
                txt0[]=readLines(CLASSLOADER.getResourceAsStream(rsc)),
                txt[]=txt0!=null ? txt0 : new String[]{"#Error not found: "+rsc};
            addActLi(instance().LI,customizeColors=new Customize("featureColors",txt,help,Customize.LIST|Customize.FLAG_NOTIFY_IMMEDIATELY));
        }
        return customizeColors;
    }
    private static class FeatureColor {
        String name,nameUS,  screen, paper, rule="-";
        Color cS, cP;
        FeatureColor(String n) { name=n;}
    }
    private static Color screenColor(FeatureColor fc) {
        if (fc.cS==null) {
            final long rgb=HtmlUtil.hexToInt(fc.screen,0);
            fc.cS=rgb>=0 ? new Color((int)rgb) : WHITE;
        }
        return fc.cS;
    }
    private static Color paperColor(FeatureColor fc) {
        if (fc.cP==null) {
            final long rgb=HtmlUtil.hexToInt(fc.paper,0);
            fc.cP=rgb>=0 ? new Color((int)rgb) : GRAY;
        }
        return fc.cP;
    }
    private static String nameUS(FeatureColor fc) {
        if (fc.nameUS==null && fc.name!=null) fc.nameUS=fc.name.replace(' ','_').replace('-','_');
        return fc.nameUS;
    }
    public static FeatureColor featureName2color(String f) {
        parseColors();
        FeatureColor fc=mapFeatureColors.get(f);
        if (fc==null) {
            mapFeatureColors.put(f,fc=new FeatureColor(f));
            mapFeatureColors.put(nameUS(fc),fc);
        }
        return fc;
    }
    private static void parseColors() {
        if (sze(mapFeatureColors)==0) {
            StringBuffer err=null;
            for(String s :customizeColors().getSettings()) {
                final StringTokenizer st=new StringTokenizer(s);
                final String fn=st.hasMoreTokens() ? st.nextToken().replace('_',' ') : null;
                if (fn==null || fn.startsWith("#")) continue;

                final String screen=st.hasMoreTokens() ? st.nextToken() : null;
                if (chrAt(0,screen)!='#') continue;
                final FeatureColor fc=new FeatureColor(fn);
                fc.screen=screen;
                fc.paper=st.hasMoreTokens() ? st.nextToken() : null;
                if (chrAt(0,fc.paper)!='#') continue;
                fc.rule=st.hasMoreTokens() ? st.nextToken().intern() : "-";
                mapFeatureColors.put(fn,fc);
                mapFeatureColors.put(nameUS(fc),fc);

            }
            error(err);
        }
    }
    private static void setFN(String n, int src, ResidueAnnotation a) {
        if (a!=null && n!=null) {
            a.setFeatureName(n,src);
            final FeatureColor featureC=featureName2color(n);
            a.setColor(screenColor(featureC));
            a.setColorForWhiteBG(paperColor(featureC));
        }
    }
    /* <<< Color <<< */
    /* ---------------------------------------- */
    /* >>> Layout Underlining >>> */
    private final static int  F_NO_SEC_STRU=1, F_SEC_STRU=2;
    private final static Set<String> setFeatureNames=new HashSet();
    protected static int layoutFeatures(StrapProtein pp[]) {
        assertEDT();
        int maxLine=0, max_nR=0;
        for(StrapProtein p : pp) max_nR=maxi(max_nR,p.countResidues());
        setFeatureNames.clear();
        final int features[]=new int[max_nR];
        final int SELECTED=0, ALL=1, NONE=2;
        final int activate=_inst!=null && _inst.comboAll!=null ? _inst.comboAll.i() : SELECTED;
        ResidueAnnotation.setEnabledAll(activate==ALL ? Boolean.TRUE : activate==NONE ? Boolean.FALSE : null);
        for(StrapProtein p : pp) {
            Arrays.fill(features,0);
            final ResidueAnnotation[] aa=p.getResidueAnnotations();
            if (cntains(null,aa)) putln(RED_ERROR,"SequenceFeatures aa contains null");
            ResidueAnnotation[] ff=null;
            while(true) {
                int countF=0;
                for(ResidueAnnotation a : aa) {
                    if (a==null) {
                        putln(RED_ERROR,"SequenceFeatures a==null");
                        continue;
                    }
                    final String f=a.getFeatureName();
                    if (f==null || activate==NONE ||activate!=ALL && !a.isEnabled()) continue;
                    if (ff!=null) ff[countF]=a;
                    countF++;
                }
                if (ff==null) ff=new ResidueAnnotation[countF];
                else break;
            }
            sortArry(ff,instance());
            String lastFn=null;
            int lastFrom=0,lastTo=0;
            for(ResidueAnnotation a : ff) {
                if (a==null) continue;
                final int from=a.getSelectedAminoacidsOffset();
                final boolean sel[]=ResidueAnnotation.getSelectedAminoacidsDisplayed(a);
                final int to=mini(features.length,from+sel.length );
                final String fn=a.getFeatureName();
                if (fn==null) continue;
                if (fn==lastFn && lastFrom==from && lastTo==to) {
                    a.setStyle(VisibleIn123.HIDDEN);
                    continue;
                }
                lastFn=fn;
                lastFrom=from;
                lastTo=to;
                int foundLine=-1;
                nextLine:
                for(int iLine=0; iLine<32; iLine++) {
                    final int bitMask=1<<iLine;
                    for(int iA=from; iA<to; iA++) {
                        if (iA<0 || iA<from || iA<features.length && sel[iA-from] && (features[iA]&bitMask)!=0)
                            continue nextLine;
                    }
                    foundLine=iLine;
                    for(int iA=from; iA<to; iA++) features[iA]|=bitMask;
                    break;
                }
                if (foundLine>=0) {
                    a.setStyle(VisibleIn123.LINES+ (foundLine<<VisibleIn123.BIT_SHIFT_LINE));
                    setFeatureNames.add(fn);
                    if (maxLine<foundLine) maxLine=foundLine;
                }
                setFeatureNames.add(fn);
            }
        }
        updateTable(33);
        featureNames=toArry(setFeatureNames,String.class);
        instance().setNumRows(featureNames.length);
        return maxLine+1;
    }
    /* <<< Layout Underlining <<< */
    /* ---------------------------------------- */
    /* >>> Threading  >>> */
    public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[]) arg : null;
        if (id==ChRunnable.APPEND) drawMsg(toStrg(arg), labMsg);
        if (id=="VAR") {
            ByteArray ba=null;
            final ResidueAnnotation aa[]=(ResidueAnnotation[])arg;
            for(ResidueAnnotation a : aa ) {
                for(ResidueAnnotation.Entry e : a.getEntries()) {
                    if ("Remark"!=e.getKey()) continue;
                    final String v=e.getValue();
                    if (v.startsWith("FTId=VAR")) {
                        final String u=Hyperrefs.toUrlString(v,Hyperrefs.WEB_LINK);
                        final File f=download_TS(url(u), CacheResult.isEnabled()?MAX_DAYS:0);
                        ba=readBytes(f,ba);
                        if (ba!=null) {
                            int i=strstr("<td>Disease",ba);
                            if (i<0) i=strstr("<td>Polymorphism",ba);
                            if (i<0) i=strstr("<td>Unclassified",ba);
                            if (i>0) {
                                final String type=ChUtils.wordAt(i+4, (CharSequence)ba, LETTR).toUpperCase();
                                final char c=chrAt(0,type);
                                setFN(c=='U' ? type+"_"+type : type, GFF_PROTEIN_FILE, a);
                                if (idxOfStrg(type,this.featureNames)<0) featureNames=ad(type,featureNames,0);
                                updateTable(999);
                                if (c!='U') {
                                    a.add(CLASS_Pymol,"hide everything, RESIDUES; show spheres,RESIDUES",true);
                                    a.add(CLASS_Pymol,"set_color COLOR,[ Rf,Gf,Bf ]; color COLOR, RESIDUES",true);
                                }
                            }
                            final int from=strstr(STRSTR_AFTER,"<td><b>From ",ba);
                            if (from>0) {
                                final int colon=strstr(",",ba,from, mini(ba.length(),from+30));
                                if (colon>0) a.setValue(ResidueAnnotation.NAME,wordAt(colon+2,ba,LETTR_DIGT));
                            }
                        }
                    }
                }
            }
        }
        return null;
    }
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> GFF  >>> */
    public static Thread thread_loadGFF(StrapProtein pp[], int sources) {
        final Object para[]={pp, intObjct(sources)};
        return threadM("loadGFF", SequenceFeatures.class, para);
    }
    public static void loadGFF(StrapProtein pp[],int sources) {
        StrapDAS.pdb2uniprot(pp, (ChRunnable)_inst);
        boolean hasFeatures=false;
        final JProgressBar progress=_inst!=null ? _inst.progress : null;
        List<StrapProtein> vNoGFF=null;
        StringBuffer sbNoGFF=null;
        final Set<String> vID_already=new HashSet();
        final StrapAlign align=StrapAlign.getInstance();
        ByteArray baGFF=null;
        try {
            final Set<String> vID=new HashSet();
            for(int step : new int[]{0, F_NO_SEC_STRU, F_SEC_STRU}) {
                nextProtein:
                for(int iP=0; iP<pp.length; iP++) {
                    if (_inst!=null && _inst.frame!=null && !_inst.frame.isVisible()) break;
                    final StrapProtein p=pp[iP];
                    if (p==null || p.containsOnlyACTGNX()) continue;
                    final int countAnnotationsOld=p.getResidueAnnotations().length;
                    int countFeatures=0;
                    final Object idsFromBlast0=(sources&GFF_BLAST)!=0 ? gcp("BL4FEATURE",p) : null;
                    final String[] idsFromBlast = idsFromBlast0 instanceof String[] ? (String[]) idsFromBlast0 : null;
                    if ((sources&GFF_BLAST)!=0 && sze(idsFromBlast0)==0) continue;
                    if ((sources&GFF_EXPASY)!=0 || idsFromBlast!=null) {
                        String[] ids=idsFromBlast!=null ? idsFromBlast : new String[]{p.getUniprotID(),p.getAccessionID()};
                        int count=0;
                        for(int j=sze(ids); --j>=0;) {
                            final String id=ids[j];
                            if (id==null || !id.startsWith("UNIPROT:") && !id.startsWith("SWISS:") && !id.startsWith("TREMBL:") && !id.startsWith("PDB:")) ids[j]=null;
                            if (ids[j]!=null) count++;
                        }
                        vID_already.clear();
                        boolean foundGff=false;
                        for(String id0 : idsFromBlast!=null ? idsFromBlast : count>0 ? ids: p.getSequenceRefs()) {
                            if (id0==null || !id0.startsWith("UNIPROT:") && !id0.startsWith("SWISS:") && !id0.startsWith("TREMBL:") && !id0.startsWith("PDB:")) continue;
                            String id=id0.startsWith("PDB:") && sze(id0)>=8 ? id0.substring(4,8) : id0.substring(id0.indexOf(':')+1);
                            if (!vID_already.contains(id))  vID_already.add(id); else continue;
                            final String gff_expasy_colon_Id="GFF_EXPASY:"+id;
                            final URL url=url(Hyperrefs.toUrlString(gff_expasy_colon_Id,Hyperrefs.PLAIN_TEXT));
                            final File f=download_TS(url(url),  CacheResult.isEnabled() ? MAX_DAYS:0);
                            if (f!=null && sze(f)==0 &&  sze(file(f+TMP_DOWNLOAD_SFX))==0) wrte(f,"0");
                            if (step==0) SequenceFeatures.log.append(p+" "+gff_expasy_colon_Id+" "+(f!=null ? sze(f) : -1)+"bytes\n");
                            baGFF=readBytes(f,baGFF);
                            if (baGFF!=null) {
                                hasFeatures=foundGff=true;
                                for(ResidueAnnotation aNew : parseFeatures(step, baGFF,  url, p, "", step==0 ? vID : null)) {
                                    aNew.add(CLASS_Hyperrefs,gff_expasy_colon_Id,true);
                                    if (aNew.isEnabled()) countFeatures++;
                                }
                                break;
                            }
                        }
                        if (step==0 && !foundGff) {
                            if (vNoGFF==null) vNoGFF=new ArrayList();
                            if (sbNoGFF==null) sbNoGFF=new StringBuffer(999).append("For the following proteins no Feature file was found:\n\n");
                            join(p.getSequenceRefs()," ",sbNoGFF.append(p+" protein-Ids: [")).append("]<br>\n");
                            if (gcp("BL4FEATURE",p)==null) { adUniq(p,vNoGFF); pcp("BL4FEATURE","",p); }
                        }
                        ChDelay.set(progress,iP+1,0,pp.length, step==0?"Fetching GFF files": step==F_SEC_STRU?"Secondary structures" : "Sequence features");
                    }
                    if (step==0) continue nextProtein;
                    if (step==F_NO_SEC_STRU && (sources&CSA_EBI)!=0 && p.getPdbID()!=null && p.hasCalpha()) hasFeatures|=catalyticSiteAtlas(p,null);
                    final ResidueAnnotation[] aa=p.getResidueAnnotations();
                    if (step>0 && countAnnotationsOld!=aa.length)  {
                        keepFeaturesWithBestScore(p);
                        disableSecondaryStructure(aa);
                    }
                    if (countFeatures>0 || iP+1==pp.length && hasFeatures) {
                        Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
                        StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED, 111);
                    }
                }
                if (step==0) {
                    final String msg="GFF SOAP uniprot "+sze(vID)+" files ... ";
                    StrapAlign.drawMessage("@2"+msg);
                    SOAP_WSDbfetch.download(SOAP_WSDbfetch.DB_UNIPROTKB,toArry(vID,String.class),"");
                    StrapAlign.drawMessage("@2"+ANSI_GREEN+msg);
                    if (sbNoGFF!=null && (sources&GFF_BLAST_IF_NO_ID)!=0) {
                        if (sze(vNoGFF)>0) {
                            sbNoGFF.append("<br><br>Blast against UNIPROT will be started to obtain [further] uniprot IDs. This will take some time.");
                            error(sbNoGFF);
                        }
                    }
                }
            }
        } finally {
            ChDelay.set(progress,0,0,1,pp!=null ? "Features of "+pp.length+" proteins loaded" : null);
        }
        if (sze(vNoGFF)>0 && (sources&GFF_BLAST_IF_NO_ID)!=0) {
            final StrapProtein ppNoGFF[]=toArry(vNoGFF,StrapProtein.class);
            for(StrapProtein p : ppNoGFF) {
                final String allIds[]=Blaster_ebi_ac_uk.uniprotIdForSequence(p.getResidueTypeAsStringUC(),50);
                final String ids[]=new String[mini(sze(allIds) , 3)];
                for(int i=ids.length;--i>=0;) ids[i]="UNIPROT:"+allIds[i];
                pcp("BL4FEATURE",ids,p);
            }
            loadGFF(ppNoGFF,GFF_BLAST|CSA_EBI);
        }
    }
    static ResidueAnnotation[] parseFeatures(int mode, ByteArray ba, URL url, StrapProtein p, String sequenceDatabase,  Set vOnlyCollectIDs) {
        if (0==sze(ba) || p==null) return ResidueAnnotation.NONE;
        final boolean isResnum="PDBresnum".equals(sequenceDatabase);
        final String sequenceDB= isResnum ? "PDB" : sequenceDatabase;
        final byte[] T=ba.getBytes();
        final int E=ba.getEnd();
        final int TABS[]=new int[9];
        final boolean isXML=strstr("</FEATURE>",T,0,E)>0;
        final String xmlSegment;
        if (isXML) {
            final int segment=strstr(STRSTR_w|STRSTR_AFTER,"<SEGMENT",T,0,E);
            xmlSegment=xmlAttribute("id",T,segment, E);
        } else xmlSegment=null;
        List<ResidueAnnotation> vRA=null;
        final ByteArray baTmp=new ByteArray(99);
        int b=0;
        nextFeature:
        for(int iL=0;    ;iL++) {
            final int e;
            if (isXML) {
                b=strstr(STRSTR_w|STRSTR_AFTER,"<FEATURE",T,b,E);
                e=strstr("</FEATURE>",T,b,E);
            } else {
                b=iL==0 ? 0 : ba.getLineEnds()[iL-1]+1;
                e=get(iL,ba.getLineEnds());
                if (get(b,T)=='#') continue;
            }
            if (e<0 || b<0) break;
            final CharSequence line=isXML ? new ByteArray(T,b,e) : ba.newString(b,e);
            if (p.containsInvalidFeatureLine(line)) continue;
            for(ResidueSelection a:p.getAllResidueSelections()) {
                final CharSequence srcTxt= a instanceof ResidueAnnotation ? ((ResidueAnnotation)a).getSrcText() : null;
                if (sze(srcTxt)==sze(line) && equls(srcTxt, line,0))  {  continue nextFeature; }
            }
            final String featureName, seqname, source, attribute, viewSrcText;
            final int origStart, origEnd;

            String disabledWhy=null;
            if (isXML) {
                final int featureGT=strchr('>',T,b,e);
                final String sLabel=xmlAttribute("label",T,b,featureGT);
                final int type=strstr(STRSTR_w|STRSTR_AFTER,"<TYPE",T,b,e), typeGT=strchr('>',T,type,e);
                final String sType=htmlDecodAsStrg( T[typeGT-1]!='/' ? ba.toString(typeGT+1,  strstr("</TYPE>",T,type,e), ByteArray.STRING_POOL) : xmlAttribute("id",T, type,typeGT));
                final String category=xmlAttribute("category",T, type,typeGT);
                if (category!=null && category.indexOf(" similarity")>0) disabledWhy="Occures 'similarity' in category";
                final boolean[] found={false};
                final String sTypeTrans=translateFeatureName(sType,found);
                final String
                  sTypeLC=lCase(sType),
                    sTypeLC1=delSfx("_one",delSfx("_two",delSfx("_three",delSfx("_four",delSfx("_five",delSfx("_six",delSfx("_seven",delSfx("_eight",sTypeLC))))))));
                if (found[0] || setSequenceOntologyTermsLC().contains(sTypeLC) || setSequenceOntologyTermsLC().contains(sTypeLC1)) {
                    featureName=sTypeTrans;
                    attribute=htmlDecodAsStrg(sLabel);
                } else {
                    featureName=translateFeatureName(htmlDecodAsStrg(sLabel), (boolean[])null );
                    attribute=htmlDecodAsStrg(capitalize(lCase(sType)));
                }
                origStart=atoi(T,strstr(STRSTR_AFTER,"<START>",T,b,e))-1;
                origEnd=atoi(T,strstr(STRSTR_AFTER,"<END>",T,b,e));
                source=uCase(sequenceDB);
                seqname=xmlSegment;
                if (isDeactivated(sType)) disabledWhy="isDeactivated(sType)";
                viewSrcText= url!=null ? file(url)+ChTextComponents.RANGE_SUFX_XML+(b-8)+"-"+(e+9) : null;
            } else { /* P25787	UniProtKB	Modified residue	24	24	.	.	.	Note=Phosphotyrosine	*/
                viewSrcText=null;
                if (tabulatrs('\t', T, b, e, TABS)<8) continue;
                seqname=ba.newString(b,TABS[0]);
                source=uCase(ba.newString(1+TABS[0],TABS[1]));
                if (vOnlyCollectIDs!=null) { if (source.startsWith("UNIPROT")) vOnlyCollectIDs.add(seqname); continue; }
                final String feature0=ba.newString(1+TABS[1], TABS[2]).replace(' ','_').intern();
                if (cntainsOnly(DIGT, T, 1+TABS[2],TABS[3])) origStart=atoi(T,1+TABS[2])-1; else continue;
                if (cntainsOnly(DIGT, T, 1+TABS[3],TABS[4])) origEnd=atoi(T,1+TABS[3]); else continue;
                attribute=ba.newString(1+TABS[7],TABS[8]>0 ? TABS[8]:e);
                if ( (feature0=="Modified_residue" || feature0=="Topological_domain" || feature0=="Region") && attribute.startsWith("Note=")) {
                    final int semicolon=attribute.indexOf(';');
                    featureName=attribute.substring(5,semicolon>0 ? semicolon : sze(attribute));
                } else featureName=feature0;
            }
            if (featureName==null) continue;
            final String fName=capitalize(lCase(featureName)).intern();
            if ("Chain"==fName) continue;
            final boolean isSecStru= fName=="Turn" || fName=="Helix" || fName=="Beta_strand";
            if ( (mode==F_SEC_STRU || mode==F_NO_SEC_STRU) &&  isSecStru != (F_SEC_STRU==mode)) continue;
            final ResidueAnnotation a=new ResidueAnnotation(p);
            if (isResnum) {
                final String sChain=is(LETTR_DIGT,p.getChainsAsString(),0) ? p.getChainsAsString().substring(0,1) : "";
                a.add(ResidueAnnotation.POS, (origStart+1)+":"+sChain+"-"+origEnd+":"+sChain,true);
                disabledWhy=null;
            } else {
                final ReferenceSequence uniprot=p.getReferenceSequence(source+":"+seqname,log,false,(Runnable)null);
                if (uniprot==null) continue;
                final float score=uniprot.getAlignmentScore();
                final int start=get(origStart,uniprot.getUniprotIdx2Idx());
                final int end=  get(origEnd-1, uniprot.getUniprotIdx2Idx())+1;
                if (start<0 || end<0) {  final String msg="Protein "+p+" "+start+".."+end+": failed to map positions   \n\n"; log.append(msg); continue; }
                final boolean poorAlignment;
                if (uniprot!=null && (score<50 ||  uniprot.countMismatch(start,end,null)/(end-start+4f)>.3f)) {
                    final String s="Protein "+p+" poor alignment "+uniprot.getLinkForAlignment()+ALI_SCORE+score+"\n\n";
                    if (log!=null && strstr(s,log.getBytes())<0) log.append(s);
                    poorAlignment=true;
                } else poorAlignment=false;
                final int idx0=p.getFirstResidueIndex();
                final String rule=featureName2color(fName).rule;
                final String interval=rule=="-" ? "-" : ",";
                baTmp.setEnd(0).append(idx0+start+1,4);
                if (start+1<end) baTmp.append(interval).append(idx0+end);
                final String positions=baTmp.newString();
                a.add(ResidueAnnotation.POS, positions,true);
                a.setRendererText(fName+" "+positions+" "+ALI_SCORE+score);
                a.add("Note",uniprot.getLinkForAlignment(),true);
                a.add("Remark",ALI_SCORE+score,true);
                a.setReferenceSequence(uniprot, start,end );
                final int selOffset=a.getSelectedAminoacidsOffset();
                if (selOffset>=p.countResidues()) disabledWhy="selOffset>=p.countResidues()";
                else if (poorAlignment) disabledWhy="Poor alignment ";
                else if (strstr("Status=Potential",T,b,e)>=0) disabledWhy="Occurs 'Status=Potential'";
                if (rule!=null && rule!="-") {
                    final char amino=(char) (p.getResidueType(selOffset)&~32);
                    if (rule.indexOf(amino)<0) disabledWhy="Only one amino and "+amino+" not in "+rule;
                }
                if (isSecStru && p.getResidueSecStrType()!=null) disabledWhy="Secondary structure is already infered";
            }
            if (viewSrcText!=null) a.add("SourceXML",viewSrcText,true);
            a.setSrcText(line);
            a.setStyle(VisibleIn123.LINES | (1<<VisibleIn123.BIT_SHIFT_LINE));
            a.add(ResidueAnnotation.NAME,fName,true);
            if (!isXML) a.add("SourceLine",line.toString().replace('\t',' '),true);
            a.add("Remark",attribute,true);
            setFN(fName, isXML ? GFF_DAS : GFF_EXPASY,a);
            a.setFeatureSrc(source);
            addFeature(a,p);
            a.setValue(ResidueAnnotation.GROUP, (disabledWhy==null ? "" : "deactivated ")+(isXML ? "DAS-":"GFF-")+"features");
            if (countTrue(a.getSelectedAminoacids())==0) disabledWhy="No aminos selected";
            if (disabledWhy!=null) {
                a.setEnabled(false);
                a.add("Disabled","Why disabled: "+disabledWhy,true);
            }
            (vRA==null ? vRA=new ArrayList() : vRA).add(a);
        }
        return toArry(vRA, ResidueAnnotation.class);
    }
    /* <<< GFF <<< */
    /* ---------------------------------------- */
    /* >>> CSA >>> */
    /*
      <td>ASP<td>A<td>  32 <td>  68<td>Sidechain</td>
      <td>HIS<td>A<td>  64 <td>  99<td>Sidechain</td>
      <td>SER<td>A<td> 221 <td> 256<td>Sidechain</td>
    */
    private final static WeakHashMap<StrapProtein,String> vAlreadyCSA=new WeakHashMap();
    public static boolean catalyticSiteAtlas(final StrapProtein p,String pdb) {
        if (vAlreadyCSA.get(p)!=null) return false;
        vAlreadyCSA.put(p,"");
        boolean hasFeatures=false;
        if (pdb==null && p!=null) pdb=p.getPdbID();
        if (p==null || pdb==null) return false;
        final URL url=url("http://www.ebi.ac.uk/thornton-srv/databases/cgi-bin/CSA/CSA_Site_Wrapper.pl?pdb="+ pdbID(pdb));
        final File f=download_TS(url,MAX_DAYS);
        final String txt[]=readLines(f);
        if (txt==null) return false;
        String alignmentOn=null;
        final List<ResidueAnnotation> vAnno=new ArrayList();
        boolean foundBy=false,literatureReference=false;
        for(String line : txt) {
            if (line==null) continue;
            line=line.toUpperCase();
            if (line.startsWith("FOUND BY:"))  foundBy=true;
            if (!foundBy) continue;
            literatureReference|=line.indexOf("LITERATURE REFERENCE")>0;
            if (line.indexOf("ALIGNMENT ON")>0) {
                final int gt=line.indexOf('>'),lt=line.indexOf('<',gt+1);
                if (gt>0 && lt>0) alignmentOn=line.substring(gt+1,lt);
            }
            if (line.startsWith("<TD>") && !line.startsWith("<TD><B>") ) {
                final String ss[]=line.split("<TD>");
                if (ss.length<4) continue;
                final String resType=ss[1].trim(),chain_=ss[2].trim(), resNum=ss[3].trim();
                final String chain="".equals(chain_) ? "?" : chain_;
                if (sze(resType)!=3 || StrapProtein.toOneLetterCode(chrAt(0,resType),chrAt(1,resType),chrAt(2,resType))=='X') continue;
                final ResidueAnnotation a=new ResidueAnnotation(p);
                a.setValue(ResidueAnnotation.NAME, (literatureReference ? "CATALYTIC_SITE" : "catalytic_site ")+resType);
                a.setValue(ResidueAnnotation.POS,resNum+":"+chain);
                vAnno.add(a);
                setFN(FEATURE_ACTIVE_SITE, CSA_EBI, a);
                continue;
            }
            if (line.indexOf("</TABLE>")>=0 && sze(vAnno)>0) {
                for(ResidueAnnotation a : toArry(vAnno,ResidueAnnotation.class)) {
                    a.add(ResidueAnnotation.GROUP,"CSA",true);
                    // a.setFeatureName(FEATURE_ACTIVE_SITE,CSA_EBI);
                    a.add(ResidueAnnotation.NAME, "Catalytic_Site_Atlas", true);
                    a.setFeatureSrc("CATALYTIC_SITE_ATLAS");
                    if (countTrue(a.getSelectedAminoacids())>0) {
                        addFeature(a,p);
                        hasFeatures=true;
                        //a.tableChanged=true;
                        a.add(CLASS_ChJmolPROXY,"spacefill on",true);
                        a.add(CLASS_ChJmolPROXY,"ATOM_TYPE_CA LABEL %n",true);
                        a.add(CLASS_Pymol,"hide everything, RESIDUES; show spheres,RESIDUES",true);
                        a.add(CLASS_Texshade,"\\shaderegion{$TEX_PROTEIN}{$TEX_RESIDUES}{Red}{Green}",true);
                        a.add(CLASS_Texshade,"\\feature{bbottom}{$TEX_PROTEIN}{$TEX_RESIDUES}{fill:$\\uparrow$}{active site}",true);
                        a.add(CLASS_Hyperrefs," PDB_CATALYTIC_SITE:"+pdbID(pdb),true);
                        a.add("Remark"," Cat Site Atlas",true);
                        if (literatureReference) {
                            a.add("Note","from literature",true);
                            pcp(Protein3d.KEY_INITIALLY_SHOW,"",a);
                        }
                        if (alignmentOn!=null) a.add("Note","indirectly, by alignment to "+alignmentOn,true);
                        if (!isMacintosh() && !isWindows()) {
                            a.add(CLASS_Rasmol,"spacefill on",true);
                            a.add(CLASS_Rasmol,"ATOM_TYPE_CA LABEL %n",true);
                        }
                        pcp(FEATURE_ACTIVE_SITE,"",a);
                    }
                }
                vAnno.clear();
                foundBy=literatureReference=false;
                alignmentOn=null;
            }
        }
        return hasFeatures;
    }
    /* <<< CSA <<< */
    /* ---------------------------------------- */
    /* >>> In Protein File >>> */
    /*
      pdb1vyp.ent:
      REMARK 800 SITE
      REMARK 800 SITE_IDENTIFIER: AC1
      REMARK 800 SITE_DESCRIPTION: TNF BINDING SITE FOR CHAIN X
      SITE     1 AC1 17 HOH C 699  HOH C 700  HOH C 701  HOH C 702
      SITE     2 AC1 17 ALA X  23  PRO X  24  LEU X  25  THR X  26
      SITE     3 AC1 17 ALA X  58  GLN X 100  HIS X 181  HIS X 184
      SITE     4 AC1 17 ARG X 233  ALA X 302  GLY X 323  ARG X 324
      SITE     5 AC1 17 TYR X 351
      SITE     1 AC2 10 HOH C 703  HOH C 704  THR X  26  ALA X  58
      SITE     2 AC2 10 TYR X  68  PHE X 102  HIS X 181  HIS X 184
      SITE     3 AC2 10 TYR X 186  TYR X 351
    */
    public static ResidueAnnotation[] readFeaturesInProteinFile(ByteArray text, int maxLen, StrapProtein... pp) {
        final List<ResidueAnnotation> vAnnotations=new ArrayList();
        ByteArray ba=null;
        boolean found=false;
        final StrapAlign align=StrapAlign.getInstance();
        for(StrapProtein p : pp) {
            boolean thisFound=false;
            ba=pp.length==1 && text!=null ? text : readBytes(p.getFile(),ba);
            if (ba==null) continue;
            final byte[] T=ba.getBytes();
            final int B=ba.getBegin(), E=ba.getEnd(),  ends[]=ba.getLineEnds();
            final ChTokenizer TOK=new ChTokenizer();
            Map<String,String> mapRemark=null;
            for(int iL=0; iL<ends.length; iL++) {
                final int b=iL==0 ? B :ends[iL-1]+1, e=ends[iL];
                if (e-b>28 && T[b+7]=='8' && T[b+8]=='0' && T[b+9]=='0' && T[b]=='R' && equls("REMARK 800 SITE_IDENTIFIER: ",T,b)) {
                    final String id=ba.toString(b+28,e, ByteArray.STRING_TRIM);
                    for(int jL=iL+1; jL<ends.length; jL++) {
                        final int lineBegin=ends[jL-1]+1;
                        if (is(LETTR_DIGT, T, lineBegin)) {
                            if (equls("REMARK 800 SITE_DESCRIPTION: ",T,lineBegin)) {
                                (mapRemark==null ? mapRemark=new HashMap() : mapRemark).put(id,ba.toString(lineBegin+29, ends[jL], ByteArray.STRING_TRIM));
                            }
                            break;
                        }
                    }
                }
            }
            int b=0, e=0;
            ResidueAnnotation lastAnno=null;
            nextLine:
            for(int iL=0; iL<ends.length; iL++, b=e+1) {
                e=prev(NO_SPC,T,ends[iL]-1,b-1)+1;
                final ResidueAnnotation a;
                final String featureName;
                if (e-b>10 && T[b]=='S' && equls("SITE: ",T,b)) {
                    final String siteID=tokn(T,b+11,e);
                    featureName=capitalize(lCase(siteID));
                    final StringBuffer sb=strBuf(SB_TIP);
                    for(int i=0; i<3; i++) {
                        final int resnum=b+26+i*11;
                        if (resnum>=e || T[resnum]==' ') break;
                        final char iCode=(char)T[resnum+1];
                        sb.append(atoi(T,resnum-3));
                        if (iCode!=' ') sb.append(iCode);
                        sb.append(':').append((char)T[resnum-4]).append(' ');
                    }
                    final String remark=mapRemark!=null ? mapRemark.get(siteID) : null;
                    a=new ResidueAnnotation(p);
                    a.add(ResidueAnnotation.POS, sb.toString(), true);
                    if (remark!=null) a.add("Remark",remark, true);
                    thisFound=found=true;
                } else if (T[b]=='M' && equls("MODRES ",T,b) && e-b>25) {
                    /*
                      MODRES 1IVO ASN A   32  ASN  GLYCOSYLATION SITE
                    */
                    a=new ResidueAnnotation(p);
                    featureName=capitalize(lCase(translateFeatureName(ba.toString(b+29, e, ByteArray.STRING_TRIM|ByteArray.STRING_POOL), (boolean[])null)));
                    a.add("Remark",featureName,true);
                    a.add("Note",ba.newString(b,e),true);
                    final char iCode=(char)T[b+22];
                    //final String pos= atoi(T,b+18)+(iCode!=' '?iCode+":":":")+(char)T[b+16];
                    a.add(ResidueAnnotation.POS, atoi(T,b+18)+(iCode!=' '?iCode+":":":")+(char)T[b+16],true);
                    thisFound=found=true;
                } else if (T[b]=='F' && T[b+1]=='T' &&  T[b+2]==' ') {
                    TOK.setText(T,b+3, e);
                    if (!TOK.nextToken()) continue nextLine;
                    if (lastAnno!=null && equls(" /FTId=",T,b+33)) {
                        lastAnno.add("Remark", ba.newString(b+35, T[e-1]=='.' ? e-1 : e), true);
                        continue nextLine;
                    }
                    {
                        final int f=TOK.getFrom();
                        final byte c0=T[f];
                        if ( (c0=='C'&&TOK.eq("CDS")||TOK.eq("CHAIN")) || c0=='e'&&TOK.eq("exon")  || !is(LETTR,c0) || !TOK.containsOnly(LETTR_DIGT_US)) continue nextLine;
                    }
                    final String feature0=TOK.asStringIP();
                    if (!TOK.nextToken() || !TOK.containsOnly(DIGT)) continue;
                    final int start=TOK.asInt()-1;
                    if (!TOK.nextToken() || !TOK.containsOnly(DIGT)) continue;
                    final int end=TOK.asInt();
                    if (end-start<1 || end-start>maxLen) continue;
                    featureName=capitalize(lCase(translateFeatureName(feature0, (boolean[])null)));
                    final String interval=featureName2color(featureName).rule=="-" ? "-" : ",";
                    a=new ResidueAnnotation(p);
                    a.add("Remark",ba.newString(b,e),true);
                    a.setStyle(VisibleIn123.LINES | (1<<VisibleIn123.BIT_SHIFT_LINE));
                    a.add(ResidueAnnotation.POS, start+1==end ? String.valueOf(end) : (1+start)+interval+(end),true);
                    thisFound=found=true;
                } else { a=null; featureName=null;}
                if (a!=null && featureName!=null) {
                    addFeature(a,p);
                    setFN(featureName, GFF_PROTEIN_FILE,a);
                    a.add(ResidueAnnotation.NAME,featureName,true);
                    if (isDeactivated(featureName)) {
                        a.setEnabled(false);
                        a.add("Disabled","Why disabled: isDeactivated(featureName)",true);
                    }
                    lastAnno=a;
                    vAnnotations.add(a);
                }
            }
            if (thisFound) keepFeaturesWithBestScore(p);
        }
        final ResidueAnnotation aa[]=toArry(vAnnotations,ResidueAnnotation.class);
        if (found) {
            disableSecondaryStructure(aa);
            Protein.incrementModificationCount(ProteinModiCount.MCA_SEQUENCE_FEATURES_V,null);
            StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED, 111);
        }
        return aa;
    }
    /* <<<  In Protein File <<< */
    /* ---------------------------------------- */
    /* >>>  Score >>> */
    private static void keepFeaturesWithBestScore(StrapProtein p) {
        final ResidueAnnotation[] aa=p.getResidueAnnotations();
        log.append("\n --- "+p+": Keep only features whith max alignment score --- \n");
        for(int i0=aa.length; --i0>=0;) {
            final ResidueAnnotation a0=aa[i0];
            if (a0==null) continue;
            final float score0=a0.getFeatureAlignmentScore();
            final String name=a0.getFeatureName();
            //final boolean isSecStru=name=="Turn" || name=="Helix" || name=="Beta_strand";
            final boolean[] sel0=a0.getSelectedAminoacids();
            final int from0=a0.getSelectedAminoacidsOffset(), to0=from0+lastTrue(sel0);
            final int where0=a0.getWhereFeatureLoadedFrom();
            for(int i=aa.length; --i>=0;) {
                final ResidueAnnotation a=aa[i];
                if (a==null || a==a0 || a.getFeatureName()!=name) continue;
                final int where=a.getWhereFeatureLoadedFrom();
                final float score=a.getFeatureAlignmentScore();
                final boolean[] sel=a.getSelectedAminoacids();
                if (where0==where && score0<=score) continue;
                final int from=a.getSelectedAminoacidsOffset(), to=from+lastTrue(sel);
                if (/* !isSecStru && */ from==from0 && to==to0 ||
                    from0<=from && from<to0 || from0<=to && to<=to0 ||  from<=from0 && from0<=to   || from<=to0 && to0<=to) {
                    log.append("   DELETE: "+name+" "+score+"  ["+from+"-"+to+"]   BECAUSE  OF "+score0+"  ["+from0+"-"+to0+"]\n");
                    p.removeResidueSelection(a);
                    p.addInvalidFeatureLine(a.getSrcText());
                    aa[i]=null;
                }
            }
        }
    }
    /* <<< Score <<< */
    /* ---------------------------------------- */
    /* >>> Enable / Disable >>> */
    private static boolean isDeactivated(String s) { return s=="Sequence_conflict" || s=="mature_protein_region"; }
    private static void  disableSecondaryStructure(ResidueAnnotation[] aa) {
        for(ResidueAnnotation a : aa) {
            final String n=a!=null ? a.getFeatureName() : null;
            final StrapProtein p=a!=null ? a.getProtein() : null;
            boolean changed=false;
            if (n!=null && a.isEnabled() && p!=null && p.getResidueSecStrType()!=null) {
                if ("helix".equalsIgnoreCase(n) || "beta_strand".equalsIgnoreCase(n) || "strand".equalsIgnoreCase(n) || "turn".equalsIgnoreCase(n)) {
                    a.setEnabled(false);
                    a.add("Disabled","Secondary structure already knwown",true);
                    changed=true;
                }
            }
            if  (changed) {
                StrapAlign.dispatchLater(StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR, 999);
            }
        }
    }
    /* <<< Enable / Disable <<< */
    /* ---------------------------------------- */
    /* >>> Translations >>> */
    private final static String[][] RPLC_FEATURE={
        {"GLYCOSYLATION SITE","Glycosylation"},
        {"CARBOHYD","Glycosylation"},
        {"ACT_SITE","Active_site"},
        {"Catalytic residue","Active_site"},
        {"BINDING","Binding_site"},
        {"COMPBIAS","Compositional_bias"},
        {"DISULFID","Disulfide_bond"},
        {"alpha-helix","Helix"},
        {"INIT_MET","Initiator_methionine"},
        {"METAL","Metal_binding"},
        {"Metal contact","Metal_binding"},
        {"MOD_RES","Modified_residue"},
        {"NP_BIND","Nucleotide_binding"},
        {"PROPEP","Propeptide"},
        {"REGION","Region"},
        {"SIGNAL","Signal_peptide"},
        {"STRAND","Beta_strand"},
        {"TOPO_DOM","Domain"},
        {"DOMAIN","Domain"},
        {"Polypeptide_domain","Domain"},
        {"TRANSMEM","Transmembrane"},
        {"Non cytoplasmic region","Non-cytoplasmic"},
        {"Natural_variant_site","Natural_variant"},
        {"protein protein contact","Protein_contact"},
        {"CONFLICT","Sequence_conflict"},

        {"?","Phosphoserine"},
        {"?","Phosphotyrosine"},
        {"?","Phosphothreonine"},
        {"?","Selenomethionine"},
        {"?","Alternative_sequence_site"},
        {"?","Mutated_variant_site"},
        {"?","Theoretical_tryptic_peptide"},
        {"?","Tryptic_peptide"},
        {"?","Solvent_accessibility"},
        // {"?","Cytoplasmic region"},
        // {"?","Low_complexity"},
        // {"?","Turn"},
        {"?","Peptide"},
        {"?","Swissmodel"},
        {"?","Polypeptide_conserved_region"},
        {"?","Merops_feature"},
        {"?","Merops_identifier"},
        {"?","Merops_clan"},
        {"?","Merops_family"},
        {"?","Variant"},
        {"","Repeat"},
        {"","Polypeptide_structural_domain"},
    };
    private static boolean equalsLetter(String s1, String s2) {
        if (s1==null || s2==null || s1.length()!=s2.length()) return false;
        for(int i=s1.length(); --i>=0;) {
            final char  c1=s1.charAt(i), c2=s2.charAt(i);
            if (is(LETTR,c1)!=is(LETTR,c2)) return false;
            if (is(LETTR,c1) && (c1|32) != (c2|32)) return false;
        }
        return true;
    }

    static void addFeature(ResidueAnnotation a, StrapProtein p) {
        if (a==null || p==null) return;
        final boolean bb[]=a.getSelectedAminoacids();
        if (countTrue(bb)>5) a.setVisibleWhere(a.getVisibleWhere()&~VisibleIn123.STRUCTURE);
        p.addResidueSelection(a);
    }
    private static String translateFeatureName(String f, boolean found[]) {
        //  NON_TER CROSSLNK DOMAIN MUTAGEN  SIMILAR SITE   VAR_SEQ
        if (f==null) return null;
        String ret=null;
        if (f.indexOf("phosphorylated")>=0) {
            ret=
                f.indexOf("tyrosine")>=0 ?  "Phosphotyrosine" :
                f.indexOf("threonine")>=0 ? "Phosphothreonine":
                f.indexOf("serine")>=0 ? "Phosphoserine" :
                // id="UNIPROTKB_P56817_DISULFID_330_380" label="UNIPROTKB_P56817_DISULFID_330_380">
                null;
        }
        if (ret==null && (f.indexOf("glycosylat")>=0 || f.indexOf("glycosolat")>=0) ) ret="Glycosylation";
        //        if (ret==null && f.indexOf("Swissmodel")>=0) ret="Swissmodel";
        for(String[] rplc : RPLC_FEATURE) {
            if (ret!=null) break;
            if (equalsLetter(rplc[0],f) || equalsLetter(rplc[1],f)) ret=rplc[1];
        }
        if (ret!=null && found!=null) found[0]=true;
        return ret!=null? ret : f;
    }
    public static String toShortFeatureName(String fn) {
        if (sze(fn)<4) return fn;
        final int space=fn.indexOf(' ');
        return space<0 || space==sze(fn)-1 ?  fn.substring(0,1) :  fn.substring(0,1)+fn.substring(space+1,space+2);
    }
    /* <<< Translations <<< */
    /* ---------------------------------------- */
    /* >>> Spice >>> */
    public static void toSpiceViewer(StrapProtein p) {
        if (p==null) return;
        final StringBuffer sb=new StringBuffer("http://www.dasregistry.org/runspice.jsp?");
        final String pdb=p.getPdbID(), uniprot=p.getUniprotID();

        if (pdb!=null) sb.append("pdb=").append(pdb.substring(pdb.indexOf(':')+1));
        else if (uniprot!=null) sb.append("uniprot=").append(uniprot.substring(uniprot.indexOf(':')+1));
        else {
            error("Error: cannot run Spice on "+p+"\nbecause it has neither a PDB ID nor a Uniprot ID");
            return;
        }
        visitURL(sb.append(VISIT_URL_JAVAWS),0);
    }
}
