package charite.christo.strap;
import charite.christo.*;
import charite.christo.protein.*;
import charite.christo.interfaces.*;
import charite.christo.strap.interfaces.*;
import java.awt.*;
import java.io.File;
import javax.swing.*;
import java.awt.event.*;
import java.util.List;
import java.util.*;
import static java.lang.System.currentTimeMillis;
import static charite.christo.ChUtils.*;
import static charite.christo.protein.ProteinModiCount.*;
import static charite.christo.ChConstants.*;
import static java.awt.event.MouseEvent.*;
import static java.awt.Color.*;
/**HELP

   The central view of the STRAP application is the alignment panel.  It
   shows the names and sequences of loaded proteins. The user can
   add and remove white space to
   align the sequences, so-called alignment gaps.

   Usually, only one sequence alignment panel is shown but additional
   alignment views can be opened <i>ITEM:StrapAlign#ACTION_newPan</i>.

   In any case only one alignment is present in one STRAP instance.
   Nevertheless, STRAP can be run several times in parallel and data
   exchange between STRAP sessions is conducted by Drag-n-Drop.

   <br><br><b>Shading</b>:
   Three  residue shadings are available in the tool-bar: "charge", "hydropathy" and
   "chemical".

   Alternatively, secondary structure can be highlighted:  Helices are
   painted red and sheets yellow.

   Further GUI controls are in the context-menu.

   <br><br><b>Editing</b>: Manipulation of the multiple sequence alignment is
   performed with the keyboard.
   STRAP has many  sophisticated keyboard commands.
   Please see <i>ITEM:StrapKeyStroke</i>.

   <br><br><b>Cursor</b>: The cursor position is highlighted in all views with the following symbol:
   <i>ICON:IC_CURSOR16</i>.

   Usually only the alignment row containing the cursor is changed when
   for example a gap is introduced by pressing the space bar.

   It is often necessary to add white space to
   a number of sequences and not only to one sequence.

   The number of adjacent rows edited simultaneously can be set by
   hitting "#" and subsequently typing a number.

   <br><br><b>Stacking proteins</b>: Usually, each row contains exactly
   one protein.  The proteins can be <b>dragged</b> up or down in the row
   header to change their order.  A number of proteins can be
   <b>stacked</b> into one single row to be manipulated simultaneously
   while only the sequence of one protein is shown:

   <i>DIALOG:DialogManyInOneRow</i>.

   <br><b>Residue selections</b> are highlighted in the multiple sequence
   alignment.  When the mouse is over a highlighted residue a WIKI:Tooltip
   with additional information appears.  Right-click opens the
   context meny and CTRL+left-click selects the selection. It can also be selected by dragging a rectangular region.

   <br><br><b>Scrolling:</b>
   The horizontal
   scroll-bar outlines the entire alignment and shows selections and
   plots. It can be enlarged with the mouse.
   If there are many proteins the vertical scroll-bar is visible.
   But also the horizontal
   scroll-bar can scroll vertically.
   Try  mouse wheel turning  with and without SHIFT and  CTRL.
   @author Christoph Gille
*/
public final class StrapView
    implements ChRunnable,StrapListener,Disposable,PreferredSize,  ActionListener, PaintHook, TooltipProvider, ChProcessEvent {
    final static String KEY_AP="SV$$AP", KEY_RH="SV$$RH";
    public final static int
        BUT_INFER_3D=11, BUT_INFER_3D_ALL=12, BUT_PDB=13, BUT_SIMIL_SEARCH=14,
        BUT_NUCL=15, BUT_SEQVISTA=16, BUT_WIRE=17, BUT_EXPORT=18,
        BUT_ALIGN=19, BUT_RESET_ALIGN=20, BUT_LAF=21,  BUT_FEATURE=23,
        TOG_MAXIMIZE=31,TOG_OUTLINE_HSB=32, TOG_BRIGHTER=33, TOG_SHADE_ONLY_SELECTED=34, TOG_PLOT_ENTIRE_PANEL=35,TOG_PROTECT_COL=36;
    private final static int
        PR_ONLY_CHANGES=1<<0, PR_IGNORE_CONSERVATION=1<<1, PR_ALL_SELECTED=1<<2,
        T_ENABLED=11, T_CLEAR_SB=12, T_WOBBLE=14, T_WOBBLE_SCROLL_DOWN=15,
        BUT_RESTORE_CURSOR=6, BUT_SAVE_CURSOR=7;
    private final static byte TRIPLET[]=new byte[3];
    private final static Rectangle RECT_EDT=new Rectangle();
    private static boolean  pCenterPaintedStatic;
    private static long mouseInToolbarWhen;
    private static int selIntervallRow=-1, selIntervallCol=-1;
    private Color[] shadingColors=ShadingAA.CHEMICAL;
    /* ---------------------------------------- */
    /* >>> Instance >>> */
    private static StrapView inst;
    private static StrapView instance() {if (inst==null) inst=new StrapView(false); return inst;}
    private final ChDelay threadPR[]=new ChDelay[20];
    private final Map<Object,Object[]> mapSavedCursor=new HashMap();
    private final HScrollBar hSB;
    private final ChScrollBar vSB;
    private final ChTableLayout tLayout;
    private final ChEventAdapter processEvent=new ChEventAdapter(this);
    public StrapView(boolean real) {
        super();
        if (!real) { rowHeader=null; tLayout=null; vSB=hSB=null; _panel=null; pCenter=null;  scrollPane=null; return;}
        addPaintHook(this,rowHeader=new ChPanel());
        //setTipDelay(rowHeader,0);
        pcp(KEY_TIP_LOCATION, new Point(100,20),rowHeader);
        setProcessEvt((ChProcessEvent)this,rowHeader);
        addProteins(getRowAtCursor()+1,a().getProteins(), NO_OPTION);
        vSB=new ChScrollBar(JScrollBar.VERTICAL);
        addPaintHook(this,vSB);
        addPaintHook(this,hSB=new HScrollBar(this));
        final String EXPLAIN_CURSOR=
            "<html><body>"+
            "the amino acid at the cursor position"+
            "<ul><li><b>i</b>: residue index starting with 1</li>"+
            "<li><b>c</b>: column (horizontal position) including gaps </li"+
            "<li><b>p</b> residue number in the structure (pdb)  file</li>"+
            "<li><b>n</b> nucleotide index and coding triplet</li></ul>"+
            "</body></html>";
        rowHeader.setMinimumSize(dim(33,33));
        rtt(rowHeader);
        addTooltip(this,rowHeader);
        shiftTooltip(rowHeader);
        setAutoscr(rowHeader);
        dragWithAltKey(rowHeader);
        setPrefSize(this, rowHeader);
        setPrefSize(this,pCenter=new ChPanel());
        addPaintHook(this, pCenter);
        shiftTooltip(pCenter);
        final java.lang.ref.Reference rThis=weakRef(this);
        pcp(KEY_AP,rThis,pCenter);
        pcp(KEY_RH,rThis,rowHeader);
        pcp(StrapView.class,rThis,pCenter);
        pcp(StrapView.class,rThis,rowHeader);
        pcp(StrapView.class,rThis,hSB);
        pcp(StrapView.class,rThis,vSB);
        addTooltip(this,pCenter);
        addMoliRefreshTip(pCenter);
        dragWithAltKey(pCenter);
        pCenter.setToolTipText("");
        tLayout=new ChTableLayout(ChJTable.NO_REORDER);
        setPrefSize(this,tLayout);
        final javax.swing.table.JTableHeader header=tLayout.getTableHeader();
        header.addMouseListener(processEvent);
        header.addMouseMotionListener(processEvent);
        scrollPane=scrllpn(0L,pCenter);
        scrollPane.setHorizontalScrollBar(hSB);
        (this._panel=newCNSEW(tLayout,null,hSB,null,null)).setMinimumSize(dim(32,32));
        setPrefSize(this,this._panel);
        addPaintHook(this, this._panel);
        setFont("",0,true);
        final JScrollPane spHeader=scrllpn(0L,rowHeader);
        rmFromParent(spHeader.getHorizontalScrollBar());
        spHeader.setMinimumSize(dim(1,1));
        scrollPane.setMinimumSize(dim(1,1));
        scrollPane.setPreferredSize(dim(3000,100));
        TabItemTipIcon.set(null,"protein names",null,null,spHeader);
        tLayout.addColumn(spHeader);
        tLayout.addColumn(scrollPane);
        pcp(KEY_DISPOSABLE, rThis,scrollPane);
        tLayout.setTitle(0,"Row headers");
        tLayout.setTitle(1,"Sequences");
        spHeader.setVerticalScrollBar(vSB);
        scrollPane.setVerticalScrollBar(vSB);
        header.setToolTipText(EXPLAIN_CURSOR);
        tLayout.setColWidth(true,0,97);
        pcp(KEY_LOOKS_LIKE_DIVIDER,"",header);
        pCenter.addKeyListener(new StrapKeyStroke(this));
        pCenter.addMouseWheelListener(processEvent);
        pCenter.addMouseListener(processEvent);
        pCenter.addMouseMotionListener(processEvent);
        pCenter.addFocusListener(processEvent);
        pCenter.setMinimumSize(dim(1,1));
        hSB.addMouseWheelListener(processEvent);
        hSB.addMouseMotionListener(processEvent);
        hSB.setCursor(getCursr('M'));
        vSB.setCursor(getCursr('M'));
        generateMiddleForLeftButton(hSB);
        generateMiddleForLeftButton(vSB);
        ChDelay.runAfterMS(EDT,thread(T_ENABLED),555);
    }
    public static StrapAlign a() { return StrapAlign.getInstance();}
    public void dispose() {
        if (isClosed) return;
        isClosed=true;
        ChThread.callEvery(-1,threadMarchingAnts,null,0);
        a().closeAlignmentPanel(this);
        recursiveRemoveAllListeners(getPanel());
        a().removeListener(this);
        UIupdateList.remove(getPanel());
        rmAllChilds(getPanel());
    }
    /* <<< Instance <<< */
    /* ---------------------------------------- */
    /* >>> Model >>> */
private final UniqueList<StrapProtein[]> vROWS=new UniqueList(StrapProtein[].class).rmOptions(UniqueList.UNIQUE).addOptions(UniqueList.ASSERT_EDT);
    { vROWS.ensureCapacity(5555); }
    public final int countRows() { return vROWS.size();}
    List<StrapProtein[]> getList() { return vROWS;}
    public final StrapProtein getProtein(int row) { return get(0,get(row,vROWS),StrapProtein.class);}
    private int modiCount;
    public final int getModificationCount() { return modiCount+vROWS.getModificationCount(); }
    public final int findRow(Protein p) {
        if (p!=null)
            for(int row=countRows();--row>=0;) {
                final StrapProtein[] ppRow=getProteinsInRow(row);
                if (ppRow.length==1 && ppRow[0]==p || cntains(p,ppRow)) return row;
            }
        return -1;
    }
    public final boolean addProteinToRow(StrapProtein p,int row) {
        if (p==null) return false;
        final List<StrapProtein[]> v=getList();
        StrapProtein[] ppRow=v.get(row);
        if (ppRow==null) return false;
        final StrapProtein[] ppNew=set1stElement(p,ppRow,0, StrapProtein.class);
        if(ppNew!=ppRow) v.set(row,ppNew);
        final boolean changed=ppRow!=ppNew || ppRow[0]!=ppNew[0];
        if (changed) modiCount++;
        return changed;
    }
    public final void addProtein(int afterRow,StrapProtein p) {
        if (p!=null)  getList().add(maxi(0,afterRow), new StrapProtein[]{p});
    }
    public final StrapProtein[] getProteinsInRow(int row) {
        final StrapProtein pp[]=getList().get(row);
        return pp!=null ? pp : StrapProtein.NONE;
    }
    public void removeProteins(StrapProtein... pp) {
        for(int row=countRows(); --row>=0; ) {
            final StrapProtein ppRow[]=getProteinsInRow(row);
            boolean changed=false;
            for(StrapProtein p : pp) changed|=setToNull(p,ppRow);
            if (changed) {
                final StrapProtein[] ppNew=rmNull(ppRow, StrapProtein.class);
                getList().set(row,ppNew);
            }
        }
    }
    int removeEmptyRows() {
        final List<StrapProtein[]> v=getList();
        int count=0;
        for(int i=countRows();--i>=0;) {
            final StrapProtein[] ppRow=v.get(i);
            if (ppRow==null || ppRow.length==0) v.remove(i);
            else count+=ppRow.length;
        }
        return count;
    }
    // ----------------------------------------
    private StrapProtein proteins[], visibleProteins[];
    private int proteins_mc, visibleProteins_mc;
    public StrapProtein[] getProteins() {
        if (!isEDT()) {
            final StrapProtein[] ret[]={null};
            inEDT(ChDelay.newThread(this, "getProteins",ret));
            return ret[0];
        }
        final int  mc=getModificationCount();
        final List<StrapProtein[]> v=getList();
        StrapProtein[] pp=this.proteins;
        if (pp==null||proteins_mc!=mc) {
            this.proteins_mc=mc;
            final int nR=v.size();
            final int nP=removeEmptyRows();
            if (pp==null || pp.length!=nP) pp=new StrapProtein[nP];
            String error=null;
            int iP=0;
            nextRow:
            for(int iR=0; iR<nR; iR++) {
                final StrapProtein[] ppRow=v.get(iR);
                if (ppRow==null || ppRow.length==0) error="sze(ppRow)==0";
                else {
                    for(StrapProtein p : ppRow) {
                        if (iP>=nP) {
                            error="iP>=nP "+iP;
                            break nextRow;
                        } else if (p==null) {
                            error="p==null "+iP;
                        } else pp[iP++]=p;
                    }
                }
            }
            if (error!=null) putln(RED_ERROR+"StrapView ",error);
            if (iP!=nP) error="iP!=nP";
            this.proteins= error!=null ? rmNull(pp, StrapProtein.class) : pp;
        }
        return pp;
    }
    public final StrapProtein[] getVisibleProteins() {
        final StrapProtein[] ppAll=this.getProteins();
        final List<StrapProtein[]> v=getList();
        final int nR=v.size();
        if (nR==ppAll.length) return ppAll;
        StrapProtein[] pp=this.visibleProteins;
        final int mc=getModificationCount();
        if (this.visibleProteins_mc!=mc || pp==null) {
            visibleProteins_mc=mc;
            if (pp==null || pp.length!=nR) pp=new StrapProtein[nR];
            boolean error=false;
            for(int iR=0; iR<nR; iR++) {
                final StrapProtein[] ppRow=v.get(iR);
                if (ppRow==null || ppRow.length==0) error=true;
                else pp[iR]=ppRow[0];
            }
            this.visibleProteins= error ? rmNull(pp, StrapProtein.class) : pp;
            assert !error;
        }
        return pp;
    }
    /* <<< Model, Proteins <<< */
    /* ---------------------------------------- */
    /* >>> GUI Elements >>> */
    private static ChSlider _sliderSim;
    public static ChSlider sliderSimilarity() {
        if (_sliderSim==null) {
            final ChSlider s=_sliderSim=new ChSlider("SLIDER_SIMIL",0,100,50)
                .tt("Conserved residues are drawn in bright color. The slider sets the threshold")
                .endLabels(" all "," only conserved ");
            pcp(KEY_PREF_SIZE_IF_DISABLED,dim(0,0),s);
            s.setPreferredSize(dim(6*ICON_HEIGHT,ICON_HEIGHT));
            s.setMaximumSize(dim(13*ICON_HEIGHT,ICON_HEIGHT));
            s.setPaintTicks(true);
            s.setMajorTickSpacing(10);
            s.setMinorTickSpacing(2);
            addActLi(instance(),s);
        }
        return _sliderSim;
    }
    private static String shadingColorsName;
    protected static String getShadingColorsName(){return shadingColorsName;}
    protected static JComponent pnlShading() { return choiceShading().cln();}
    private static ChCombo _cShading;
    public static ChCombo choiceShading() {
        if (_cShading==null)
            _cShading=new ChCombo(
                                  "white","charge","hydropathy",shadingColorsName="chemical","nucleotide",
                                  "<html><body><font color=\"cc0000\">second struct</font></body></html>",
                                  "<html><body><font color=\"cc0000\">solvent access</font></body></html>",
                                  "<html><body><font color=\"00AAAA\">organism</font></body></html>",
                                  "<html><body><font color=\"00AAAA\">compound</font></body></html>",
                                  "<html><body><font color=\"00AAAA\">ec_code</font></body></html>"
                                  ).li(instance())
                .tt("Determins how sequences of PDB-files are shaded<br>"+
                    "<ul><li>3D structure: red=helices   yellow=extended, beta sheets</li>"+
                    "<li>amino acid type. Set by other choice menu.<li>"+
                    "<li>solvent accessibility: Only if available. Currently only for DSSP-files</li></ul>")
                .li(instance()).s(3);
        return _cShading;
    }
    private static ChButton[] _button;
    public static ChButton button(int t) {
        if (_button==null) {
            _button=new ChButton[99];
            for(int i=99; --i>=0; ) {
                final Dimension small=dim(ICON_HEIGHT*3/2,ICON_HEIGHT);
                final ChButton b=_button[i]=
                    i==TOG_MAXIMIZE? toggl(ChButton.MAXIMIZE, ChButton.DISABLED).tt("Maximize alignment pane") :
                    i==TOG_OUTLINE_HSB? toggl("Color shade horizontal scroll-bar").tt("shade horizontal scroll bar").s(true) :
                    i==TOG_BRIGHTER? toggl("Amino acids appear brighter").tt("Unconservered amino acids are shown dark. Make them not that dark.") :
                    i==TOG_SHADE_ONLY_SELECTED? toggl("Colorize only the selected sequences").s(true) :
                    i==TOG_PLOT_ENTIRE_PANEL? toggl("Large").s(true) :
                    i==TOG_PROTECT_COL? toggl("Vertical line and protect alignment beyond this line")
                    .tt("The alignment is protected right from the black/white line.<br>Gaps can not be altered beyond this point.<br>Toggle on/off by typing curly brace \"}\"") :
                    i==BUT_INFER_3D? new ChButton("PDB structures listed in the protein file").tt("Infer coordinates from  PDB-file according to X-refs in protein file") :
                    i==BUT_INFER_3D_ALL? new ChButton("Infer PDB structure") :
                    i==BUT_PDB? new ChButton("ASS_3D").t("Associate","3D-structure").tt("Searching for homologous PDBs") :
                    i==BUT_SIMIL_SEARCH? new ChButton("Search homologous PDB structures").tt("Homologous PDBs are identified for the given Uniprot ID.") :
                    i==BUT_NUCL? new ChButton().t("DNA","sequence",small).tt("If the character sequence contains only the letters a c t g and n<br>a nucleotide view is opened.") :
                    i==BUT_SEQVISTA? new ChButton().t("Seq","VISTA",small).tt("SeqVISTA presents a holistic, graphical view of features annotated on nucleotide sequences.<br>http://zlab.bu.edu/SeqVISTA/") :
                    i==BUT_WIRE? new ChButton().t("3D","structure",small).tt("A backbone representation of the 3D-structure is opened.") :
                    i==BUT_EXPORT? new ChButton("W",ChButton.ICON_SIZE).bg(WHITE).tt("Exporting the alignment to your favorit Word-Processor (MS-Word, OpenOffice ... ) ") :
                    i==BUT_RESET_ALIGN? new ChButton().t("","",dim(0,0)).tt("Restore alignment gaps to what it was before pressing the align button") :
                    i==BUT_LAF? new ChButton().t("Op-","tions",small).tt("Appearance of the graphical user interface and user-mode ...") :
                    i==BUT_FEATURE? new ChButton().t("Sequence","feature",small).tt("Highlight Glycosylated and Phosphorylated residues, Signal peptides, Metal binding....") :
                    i==BUT_RESTORE_CURSOR? new ChButton(null,ChButton.DISABLED).t("Revert","Cursor",small) :
                    i==BUT_SAVE_CURSOR? new ChButton(null,ChButton.DISABLED).t("Save","Cursor",small) :
                    i==BUT_ALIGN? new ChButton(null,ChButton.DISABLED)
                    .tt("The selected or all protein sequences are aligned.<br>The alignment is computed from 3D-structures and AA sequences.<br>"+
                        "Hold shift for reordering.<br><br>If  more control is required then  use the Align-dialog instead.<br><br>Right-click for Result panel.") :
                    null;
                if (b!=null) {
                    if (i==BUT_RESTORE_CURSOR || i==BUT_SAVE_CURSOR) b.tt("[S]ave and [R]estore the cursor position<br>the modifier keys meta shift and control can be used in combination");
                    if (i==BUT_SEQVISTA||i==TOG_MAXIMIZE||i==BUT_PDB) pcp(KEY_PREF_SIZE_IF_DISABLED,dim(0,0),b);
                    b.li(instance());
                    final char up=
                        i==BUT_NUCL || i==BUT_SEQVISTA ? 'n' :
                        i==BUT_SAVE_CURSOR || i==BUT_RESTORE_CURSOR ? 'e' :
                        i==TOG_MAXIMIZE ? 'W' :
                        0;
                    if (up!=0) UserProfile.set(up,b);
                }
            }
        }
        return _button[t];
    }
    /* <<< GUI Elements <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */
    private final JComponent pCenter, rowHeader, _panel;

    private boolean isMouseButtonPressed, isNucleotides, mouseOverRectangle;
    private boolean isClosed,pCenterPainted, isResizing, looksLikeDivider,alreadySaidScroll;
    private boolean alreadyPainted, shadeSecStru=true, showSolvAcc,  isShadingScrollbar=true,isMouseDrag;
    private char isDragged;
    private int colDraggedFrom, rowDraggedFrom, colLastClicked;
    private int mouseDragFromX,mouseDragFromY;
    private int reduceHeight,reduceHeightSave;
    private static int prevMouseX, prevMouseY;
    private int lineSkip=1 ,lastLineSkip;
    private int fontSize=maxi(5,getPrpty(StrapView.class,"fontSize",28));
    private int tip_where;
    private int lastCountColumns, lastCountRows, lastScrollV;
    private int charWidths[];
    private int vSB_yCursor;
    private int frequencies_modificationCount=-1;
    private int iInfo=-1;
    private int mouseMovedCol=-1, mouseMovedRow=-1, mousePressedRow, mousePressedCol;
    private int countColumns, countColumns_modificationCount=-1;
    private final JScrollPane scrollPane;
    private Rectangle charBounds;
    private Point zoomZenter;
    private Runnable threadMarchingAnts;

    private Font paintRegion_lastFont;
    /* ---------------------------------------- */
    /* >>> Html >>>  */
    private static ChButton html_cbFileExt, html_cbPdbNumbering, html_cbUL, html_cbJalviewSS, html_cbFeatures, html_cbProteinIcons[], html_cb_imgAttributes;
    private static ChTextField html_tfFile, tfWordCharsPerLine, html_iconSize;
    final static List<java.lang.ref.Reference<ChLabel>> vLabelsRunningAnts=new ArrayList();
    private static ChCombo html_comboTarget;
    private static int html_count;
    private static ChButton html_butJalview;
    /* <<< Html <<<  */
    /* ---------------------------------------- */
    /* >>> Threading >>> */
    private final ChDelay threadPR(int option) {
        if (threadPR[option]==null) threadPR[option]=ChDelay.newThread(this,"paintRegion", intObjct(option), new long[ARRAY_DURATION]);
        return threadPR[option];
    }
    private final ChDelay thread(int option) {
        if (threadPR[option]==null) threadPR[option]=ChDelay.newThread(this,option+"thread",null);
        return threadPR[option];
    }
    @Override public Object run(String id, Object arg) {
        final Object argv[]=arg instanceof Object[] ? (Object[])arg : null;
        final StrapAlign align=a();
        if (id=="getProteins") ((StrapProtein[][])arg)[0]=getProteins();
        if (id=="scrRectToVis") { sleep(atoi(argv[3])); scrRectToVis((Rectangle)argv[0], argv[1]!=null, atoi(argv[2])); }
        if (id=="CURSOR_TXT") {
            final StrapProtein p=getProteinAtCursor();
            if (p!=null) {
                final String ch=p.getChainsAsString();
                ByteArray ba=p.cursorPosAsString(getIndexOfAminoAcidAtCursor(),true);
                if (sze(ch)>1) ba.append("  concatenated chains: ").append(ch);
                tLayout.setTitle(1,ba.newString());
                tLayout.setTitle(0,"Protein "+(1+getRowAtCursor()));
            }
        }
        if (id.endsWith("thread")) {
            final int atoi=atoi(id);
            switch(atoi) {
            case T_CLEAR_SB: hSB.clearImage().repaint(); break;
            case T_ENABLED:
                final StrapProtein
                    ppSel[]=align.getSelectedProteins(),
                    pCursor=getProteinAtCursor(),
                    pp[]=getVisibleProteins(),
                    ppSelOrAll[]=ppSel.length>0 ? ppSel : pp,
                    p1=pCursor!=null ? pCursor : ppSel.length>0 ? ppSel[0] : pp.length==1 ? pp[0] : null;
                if (!UserProfile.get('n') &&  ProteinList.isEnabled(pp, ProteinList.NT_or_ACTGN)) {
                    UserProfile.set('n',true);
                    getToolbars()[0].revalidate();
                }
                final boolean isNucEnabled=ProteinList.isEnabled(ppSel, ProteinList.NT_or_ACTGN) ||  ppSel.length==0 && ProteinList.isEnabled(p1, ProteinList.NT_or_ACTGN);
                setEnabld(button(BUT_NUCL), isNucEnabled);
                setEnabld(button(BUT_SEQVISTA), isNucEnabled);
                setEnabld(button(BUT_WIRE),ProteinList.isEnabled(ppSel, ProteinList.CALPHA) || ppSel.length==0 && ProteinList.isEnabled(p1, ProteinList.CALPHA));
                final boolean no_calpha=ProteinList.isEnabled(ppSelOrAll, ProteinList.NO_CALPHA);
                for(int i=4; --i>=0;)  setEnabld(button(i==0 ? BUT_INFER_3D : i==1 ? BUT_INFER_3D_ALL : i==2 ? BUT_PDB : BUT_SIMIL_SEARCH), no_calpha);
                for(int i=3; --i>=0;)  setEnabld(button(i==0 ? BUT_EXPORT : i==1 ? BUT_FEATURE : TOG_BRIGHTER),pp.length>0);
                for(int i=2; --i>=0;)  setEnabld(i==0 ? button(BUT_ALIGN) : sliderSimilarity(), pp.length>1);
                button(BUT_EXPORT).fg(pp.length>0 ? BLUE : GRAY).setBorder(new javax.swing.border.MatteBorder(new Insets(4,4,4,4),pp.length>0 ? BLUE : GRAY));
                button(BUT_ALIGN).t(ppSel.length>1 ? "Align "+ppSel.length : "Align all", "proteins",dim(64,32));
            }
            if (atoi==T_WOBBLE || atoi==T_WOBBLE_SCROLL_DOWN) {
                hSB.setPreferredSize(dim(1,hSB.getPreferredSize().height ^3));
                hSB.revalidate();
                if (atoi==T_WOBBLE_SCROLL_DOWN) {
                    ChDelay.scrollToEnd(pCenter,999, (long[])null);
                }
            }
        }
        if (id=="marchingAnt"/* && !needsRepaintAlignment()*/) {
            try {
                final Graphics2D g=(Graphics2D)pCenter.getGraphics(), gSB=(Graphics2D)hSB.getGraphics();
                marchingAnt(g,gSB);
            } catch(Throwable ex){ if (isMyComputer()) stckTrc(ex);}
        }
        if (id=="AFTER_ALIGN") {
            ChMsg.sound(ChMsg.SOUND_SUCCESS);
            setButAlign(false);
            button(BUT_RESET_ALIGN).t("Undo","align",dim(54,32)).revalidate();
            StrapAlign.drawMessage("@1"+ANSI_GREEN+"Alignment finished");
        }
        if (id=="paintRegion") {
            final Graphics g=pCenter.getGraphics();
            if (g!=null) {
                final Rectangle v=getVisibleRowsAndCols();
                paintRegion(g, v.x,v.x+v.width,  v.y,v.y+v.height, atoi(arg));
                plot(g);
            }
        }
        return null;
    }
    private static SequenceAligner lastAligner;
    /* <<< Threading <<< */
    /* ---------------------------------------- */
    /* >>> Rows and Columns >>> */
    private final Rectangle visibleRowsAndCols=new Rectangle(), visible=new Rectangle();
    public int countColumns() {
        if (rowHeader==null) return 0;
        final StrapProtein pp[]=getProteins();
        final int mc=getModificationCount()+Protein.MODI_COUNTS_ALL[MCA_ALIGNMENT];
        if (countColumns_modificationCount!=mc) {
            countColumns_modificationCount=mc;
            countColumns=StrapProtein.getMaxColumn(pp)+1;
        }
        return countColumns;
    }
    public Rectangle getVisibleRowsAndCols() {
        final Rectangle cb=charB(), v=getVisibleXYWH();
        setBnds(
                v.x/cb.width, v.y/(cb.height+lineSkip),
                (v.width-1)/cb.width+1,(v.height-1)/(cb.height+lineSkip)+1,
                visibleRowsAndCols);
        return visibleRowsAndCols;
    }
    public int row2y(int row) { return (charB().height+lineSkip)*row; }
    public int col2x(int col) { return charB().width*col; }
    public final int y2row(int y, boolean middle) {
        if (y<0) return -1;
        final Rectangle cb=charB();
        final int h=maxi(1,cb.height+lineSkip);
        return (y+ (middle ? h/2 : 0))/h;
    }
    private static void setProtectFromColumn(int column, boolean toggle) {
        boolean sel=button(TOG_PROTECT_COL).s();
        if (toggle) button(TOG_PROTECT_COL).s(sel=!sel);
        protectFromColumn=sel ? column : -1;
        for(StrapView v : a().getAlignmentPanels()) {
            v.pCenter.repaint();
            v.hSB.repaint();
        }
    }
    /* <<< Rows and Columns <<< */
    /* ---------------------------------------- */
    /* >>> Layout >>> */
    private boolean oneRowVisible;
    public Dimension getPreferredSize(JComponent component) {
        if (component==pCenter) return getPaneSize();
        if (component==rowHeader) return dim(tLayout.columnWidth(0)  ,row2y(countRows()));
        if (component==TOOLBARS[0]) return !UserProfile.isWebMode() && a().getAlignmentPanels().length==0 ? dim(0,0) : null;
        if (component==getPanel()) {
            if (a().TOG_MAXIMIZE_DIALOGS.s()) return dim(0,0);
            final int prefH=getPrefPanelHeight();
            if (prefH<(oneRowVisible?2*charB().height:0)) setReduceHeight(reduceHeight+prefH);
            final int h=maxi(hSB.getPreferredSize().height+EX, mini(a().getPanelHeight(), getPrefPanelHeight()));
            return dim(9999,h);
        }
        return dim(32,32);
    }
    private int seqFeat_mc;
    private void revalidateWhenSizeChanged() {
        final int mc=Protein.MODI_COUNTS_ALL[MCA_SEQUENCE_FEATURES_V];
        if (seqFeat_mc!=mc) {
            seqFeat_mc=mc;
            final int newSpc=SequenceFeatures.instance().layoutFeatures(getProteins())*3;
            if (this.lineSkip!=newSpc) { setSpaceBetweenRows(newSpc); pCenter.revalidate();}
        }
        if (lastCountColumns!=countColumns() || lastCountRows!=countRows()) {
            lastCountColumns=countColumns();
            lastCountRows=countRows();
            ChDelay.revalidate(getPanel(),222, (long[])null);
        }
    }
    private int getPrefPanelHeight() {
        final Dimension d=hSB.getPreferredSize();
        return (d!=null?d.height : EX)+row2y(countRows())+5-reduceHeight + EX;
    }
    public int getSpaceBetweenRows() { return lineSkip;}
    public void setSpaceBetweenRows(int space) {
        final int old=this.lineSkip;
        this.lineSkip=maxi(1,space);
        if (old!=this.lineSkip && alreadyPainted) {
            pCenter.repaint();
            vSB.setValue(vSB.getValue()*(charB().height+this.lineSkip)/(charB().height+old));
        }
    }
    protected Dimension getPaneSize() {
        final Rectangle cb=charB();
        final int w;
        if (!showsSequence()) {
            int max=1;
            for(StrapProtein p: getVisibleProteins()) {
                final String info=p.getInfo(iInfo);
                max=maxi(max,info!=null ? strgWidth(getPanel(),info)+32 : 1);
            }
            w=max;
        } else w= cb.width*(1+countColumns());
        return dim(w, row2y(countRows()));
    }
    public Rectangle getVisibleXYWH() {
        if (isEDT() && hSB!=null) {
            final JViewport vp=scrollPane.getViewport();
            setBnds(-pCenter.getX(),-pCenter.getY(),vp.getWidth(), vp.getHeight(), visible);
        }
        return visible;
    }
    protected Rectangle charB() {
        if (charBounds==null) { charBounds=new Rectangle(0,0,1,1); setFont("",0,true); }
        return charBounds;
    }
    private int zoom_reduceHeight[]=new int[100];
    protected void setFont(String fontName,int size,boolean isAbsolute) {
        final Rectangle rVis=this.visibleRowsAndCols;
        if (rVis!=null && alreadyPainted && zoomZenter==null) zoomZenter=new Point( rVis.x+rVis.width/2,rVis.y+rVis.height/2);
        final String fn=fontName!=null && fontName!="" ? fontName : getPrpty(StrapView.class,"fontName","Monospaced");
        if (size!=0) {
            final int fs=this.fontSize=maxi(3, isAbsolute ? size : fontSize+size);
            if (fs<zoom_reduceHeight.length && zoom_reduceHeight[fs]>=0) this.reduceHeight=mini(this.reduceHeight, zoom_reduceHeight[fs]);
        }
        final Font font=new Font(fn,Font.PLAIN,fontSize);
        if (fontSize<40) setPrpty(StrapView.class,"fontSize",fontSize);
        setPrpty(StrapView.class,"fontName",fn);
        if (getPanel()!=null)  getPanel().setFont(font);
        final Rectangle cbOld=this.charBounds, cb=this.charBounds=new Rectangle(charBounds(font));
        cb.height++;
        cb.width++;
        final Point zc=this.zoomZenter;
        if(zc!=null) scrRectToVis(new Rectangle( zc.x*cb.width, zc.y*(cb.height+lineSkip),  1,1), true);
        if (cbOld!=null) setReduceHeight(reduceHeight-(cbOld.height-cb.height)*countRows());
        hSB.setUnitIncrement(cb.width);
        vSB.setUnitIncrement(cb.height);
    }
    public JComponent getPanel() { return _panel; }
    public static void setReduceHeight() {
        if (!isEDT()) inEdtM("setReduceHeight",StrapView.class, NO_OBJECT);
        else {
            StrapView v=a().getAlignmentPanelWithFocus();
            if (v==null) v=get(0,a().getAlignmentPanels(),StrapView.class);
            if (v!=null) {
                final int old=v.reduceHeight;
                final int neu=v.row2y(v.countRows())-a().getFrame().getHeight()/3;
                if (neu>old) v.setReduceHeight(neu);
            }
        }
    }
 void setReduceHeight(final int h) {
        final int old=this.reduceHeight, neu=maxi(h,0);
        this.reduceHeight=neu;
        if (fontSize<zoom_reduceHeight.length) zoom_reduceHeight[fontSize]=this.reduceHeight;
        if (correctHeight() || old!=neu) getPanel().revalidate();
    }
    protected static boolean correctHeight() {
        long sumH=0;
        final StrapView[] vv=a().getAlignmentPanels();
        for(StrapView v: vv) sumH+=v.getPrefPanelHeight();
        final JComponent toolbar=TOOLBARS[0];
        final long tooHigh=  (sumH-a().getPanelHeight()+ (toolbar!=null ? toolbar.getHeight() : ICON_HEIGHT));
        if (tooHigh>0) {
            for(StrapView v: vv) {
                v.reduceHeight=(int)(v.reduceHeight+tooHigh*v.getPrefPanelHeight()/sumH);
                ChDelay.setEnabled(button(TOG_MAXIMIZE), v.reduceHeight>1 || v.reduceHeightSave>0 || a().getDividerLocation(true)>10 || a().getDividerLocation(false)>10, 99);
                ChDelay.revalidate(v.getPanel(),222,(long[])null);
            }
        }
        return tooHigh>0;
    }
    /* <<< Layout <<< */
    /* ---------------------------------------- */
    /* >>> Tip >>> */
    private String tipLast;
    private int tipLastRow;
    public String getToolTipText(JComponent c, AWTEvent ev) {
        final int modi=modifiers(ev);
        final boolean shift=isShift(ev);
        final Rectangle cb=charB();
        final int x=x(ev), y=y(ev), lineH=cb.height+getSpaceBetweenRows(), row=y/lineH, col=x/cb.width;
        final String RC4CM="<br><font color=#FF4455><b>Right-click to open context menu</b></font>";
        if (c==rowHeader) {
            final int where=row*1024+modi;
            if (x>2*EX) return null;
            else if ((y%lineH) > lineH-4 || x>EX) return "<html><body><font size=0> </font></body>/<html>";
            else if (tipLast==null || tipLastRow!=where) {
                tipLastRow=where;
                final StrapProtein pp[]=getProteinsInRow(row);
                final StringBuffer sb=
                    pp.length==1 && pp[0]!=null ?  pp[0].getTooltipWithoutHtmlTags(shift) :
                    join(pp,"<br>",strBuf(SB_TIP).append("This row contains ").append(pp.length).append(" proteins:<br><sup>")).append("</sup>");
                if (getPrpty(StrapView.class,"popupShownP",null)==null) sb.append(RC4CM);
                if (getPrpty(StrapView.class,"xrefShownP",null)==null) sb.append("<br><font color=#FF4455><b>Mouse Click at left edge of protein label to see cross-references</b></font>");
                tipLast=addHtmlTagsAsStrg(sb);
            }
            return tipLast;
        }
        if (c==pCenter) {
            final int where=col+row*1024+modi;
            if (tipLast==null || where!=tip_where) {
                final StringBuffer sb=strBuf(SB_TIP);
                tip_where=where;
                final ResidueSelection[] ssUM=this.resSelUnderM;
                int count=0;
                char type='S';
                sb.setLength(0);
                for(int iS=-1; iS<sze(ssUM); iS++) {
                    final ResidueSelection s=iS<0 ? featureUnderM : ssUM[iS];
                    if (s==null || s==resSel_cursor) continue;
                    final ResidueAnnotation a= s instanceof ResidueAnnotation ? (ResidueAnnotation)s : null;
                    if (a!=null && !a.isEnabled()) continue;
                    if (a!=null) {
                        if (ResidueAnnotation.isAnno(a)) type='A';
                        else if (type!='A') type='F';
                    }
                    if (ResidueSelectionUtils.isShownIn(s,VisibleIn123.SEQUENCE)==null) continue;
                    String tip=TabItemTipIcon.tip(s,ev);
                    if (tip==null) tip=TabItemTipIcon.item(s);
                    if (count++>0) sb.append("<hr>");
                    sb.append(sze(tip)>0 ? tip : s);
                    if (s==featureUnderM) break;
                }
                if (count>0) {
                    baloonHint(!shift?(char)0:type,sb);
                    this.tipLast=addHtmlTagsAsStrg(sb);
                } else if(this.mouseOverRectangle && getPrpty(StrapView.class,"popupShownR",null)==null) {
                    this.tipLast=addHtmlTagsAsStrg(RC4CM.substring(4)+" of rectangular region");
                } else if (Math.abs(x-cb.width* getProtectFromColumn((char)0) )<EM) {
                    this.tipLast=button(TOG_PROTECT_COL).getToolTipText();
                } else tipLast=null;
            }
            return this.tipLast;
        }
        return null;
    }
    private static void baloonHint(char type, StringBuffer sb) {
        if (type==0) sb.append("<br><sub><font color=\"#FF0000\">shift-key for details</font></sub>");
        else sb.append("Important Notes<ul>"+
                       "<li>Drag'n Drop: Residue selections can be dragged onto other proteins and 3D-structure views</li>"+
                       "<li>Right-click: The context menu provides many options.</li>"+
                       (type=='A' ? "<li>Annotations: The user can add annotations and 3D-rendering commands.</li>":"")+
                       "</ul><br>");
    }
    /* <<< Tip <<< */
    /* ---------------------------------------- */
    public boolean paintHook(JComponent c,Graphics g0, boolean after) {
        final Graphics2D g=(Graphics2D)g0;
        if (g==null || !after) return true;
        antiAliasing(g,false);
        final long TIME=currentTimeMillis();
        final boolean debugPaint=false;
        alreadyPainted=true;
        final Rectangle  clipG=g.getClipBounds(), cb=charB();
        final int w=c.getWidth(), h=c.getHeight(),  charW=cb.width, charA=cb.y, charH=cb.height, lineH=cb.height+lineSkip;
        final StrapProtein[] ppVis=getVisibleProteins();
        final StrapProtein[] ppSel=a().getSelectedProteins();
        if (c==rowHeader) {
            final boolean hlIdentical=a().CB_HIGHLIGHT_IDENTICAL_SEQUENCES.s();
            g.setColor(C(DEFAULT_BACKGROUND));
            final Color colSel=C(COLOR_SELECTED);
            fillBigRect(g, 0,0,w,h);
            if (lineH==0) return true;
            final int rowFrom=clipG!=null ? clipG.y/lineH : 0;
            final int rowTo=mini(ppVis.length,clipG==null ? ppVis.length : (clipG.height+clipG.y) /lineH+1);
            final Font font=getPanel().getFont();
            g.setFont(font);
            final int rowCursor=getRowAtCursor();
            for(int iP=rh_draggedProtein2row==null ? rowFrom : 0; iP< (rh_draggedProtein2row==null ? rowTo : ppVis.length); iP++) {
                final int row= sze(rh_draggedProtein2row)>iP ? rh_draggedProtein2row[iP]:iP;
                if (row<rowFrom || row>=rowTo) continue;
                final int y=row*lineH;
                final StrapProtein p=ppVis[iP];
                final String name=p.toString();
                final Rectangle labelSize=p.getLabelSize(font,cb);
                final int nProts=getProteinsInRow(iP).length;
                final int logN=nProts==1 ? 0: nProts <10 ? 1 : nProts<100 ?2 : 3 ;
                final int countH=p.getHeteroCompounds().length, countN=p.getDnaAndRnaStructures().length;
                final int offset=(int) ((rowHeader.getParent().getWidth()-labelSize.width)*rh_adjust)+(countN+countN>0 ? 3 : 0);
                {
                    final Image icIm=p.getIconImage();
                    if (icIm!=null) g.drawImage(icIm,offset+logN*charW, y, labelSize.x, charH,rowHeader);
                }
                final int x=offset+labelSize.x+logN*charW;
                if (hlIdentical && iP>0 && Protein.equalsResidueType(ppVis[iP-1],p)) {
                    g.setColor(C(0xFF0600,66));
                    g.fillRect(0,y-1,300,2);
                }
                p.paintChain(0L,g, p.getName(), x,y,charW,charH);
                g.setColor(cntains(p,ppSel) ? colSel : BLACK);
                g.drawString(name,x, y+charA);
                if (iP==rowCursor) { ChIcon.drawCursor(g,charH, y+ charH/4,charH/2,charH/2,rowHeader); }
                if (nProts>1) {
                    g.setFont(font);
                    g.setColor(BLACK);
                    final String s=String.valueOf(nProts);
                    for(int i=0;i<4;i++) g.drawString(s,offset+i%2*2-1,y+charA+2+(i&1)*2-1);
                    g.setColor(WHITE); g.drawString(s,offset,y+charA+2);
                }
                if (countN+countH>0) {
                    g.setColor(BLACK); g.fillRect(0,y+1,3,lineH-2);
                    if (countN>0) { g.setColor(GREEN); g.fillRect(0,y+2,  2, lineH/3); }
                    if (countH>0) { g.setColor(RED); g.fillRect(0,y+lineH/2+2,  2, lineH/3);}
                }
            }
            barChart(vopClass,vop_sharedInstance,g,ppVis,rowFrom,rowTo);
        }
        if (c==vSB && ppVis.length>0 && ppSel.length>0) {
            final Rectangle track=vSB.getTrack();
            final int v=-pCenter.getY();
            if (v>0 && !alreadySaidScroll && ((TIME-TIME_AT_START)>1000*20*60)) alreadySaidScroll=true;
            final int yCursor=vSB_yCursor();
            if (yCursor>=0) ChIcon.drawCursor(g,0,yCursor-8,16,16,getPanel());
            for(int iP=ppVis.length; --iP>=0;) ppVis[iP].setRow(iP);
            if (v!=lastScrollV) { lastScrollV=v;  hSB.repaint(hSB.getThumb());  }
            g.setColor(C(COLOR_SELECTED));
            final int thickness=mini(3,maxi(1,ppSel.length<5 ? 3 : track.height/ppVis.length-1));
            for(int iP=ppSel.length; --iP>=0;) {
                final StrapProtein p=ppSel[iP];
                if (p==null) continue;
                final int iRow=p.getRow();
                if (iRow<0 || iRow>=ppVis.length || ppVis[iRow]!=p) continue;
                final int y= track.y + track.height*iRow/ppVis.length;
                g.fillRect(3,y,99,thickness);
            }
        }
        if (c==pCenter) {
            if (reduceHeight<0) reduceHeight=0;
            if (!pCenterPainted) {
                debugTime("\nStrapView: ",TIME_AT_START);
                if (isBenchmark('S')) System.exit(0);
                pCenterPainted=pCenterPaintedStatic=true;
            }
            final Rectangle v=getVisibleXYWH(), clip=clipG!=null ? clipG : v;
            revalidateWhenSizeChanged();
            if (!showsSequence()) {
                g.setColor(BLACK);
                fillBigRect(g, 0,0,w,h);
                g.setColor(WHITE);
                g.setFont(getPanel().getFont());
                g.setClip(0,0,w,h);
                for(int row= ppVis.length; --row>=0;) {
                    final StrapProtein p=ppVis[row];
                    final String info=p.getInfo(iInfo);
                    if (info==null) continue;
                    final int y=row*lineH;
                    g.drawString(info,10,y+cb.y);
                }
                return true;
            }
            final int colLeft=clip.x/charW, colRight=(clip.x+clip.width-1)/charW+2;
            final int rowTop=clip.y/lineH, rowBot=(clip.y+clip.height-1)/lineH+1;
            paintRegion(g, colLeft,colRight, rowTop,rowBot, PR_ALL_SELECTED);
            plot(g);
            if (looksLikeDivider || isResizing) looksLikeDivider(null,g,-pCenter.getX(),0,EM*3/2,9999);
            if (threadMarchingAnts==null) ChThread.callEvery(200,threadMarchingAnts=ChDelay.newThread(this,"marchingAnt",null),"marchingAnt", EDT);
            marchingAnt(g,null);
        }
        if (c==hSB) marchingAnt(null, g);
        return true;
    }
    /* <<< Tip <<< */
    /* ---------------------------------------- */
    /* >>> Rectangle >>> */
    private void marchingAnt(Graphics2D gAli, Graphics2D gSB) {
        if (gAli==null && gSB==null) return;
        final StrapAlign A=a();
        final Stroke stroke0=gAli!=null ? gAli.getStroke() : gSB.getStroke();
        if (gAli!=null) gAli.translate(BUGFIX_OPENJDK_G2, BUGFIX_OPENJDK_G2);
        if (gSB!=null)  gSB.translate(BUGFIX_OPENJDK_G2, BUGFIX_OPENJDK_G2);
        final Rectangle rRowsCols=getVisibleRowsAndCols();
        final int cMin=rRowsCols.x, cMax=cMin+rRowsCols.width;
        final int rMin=rRowsCols.y, rMax=rMin+rRowsCols.height;
        final Stroke dashedStroke0=dashedStroke(0), dashedStroke1=dashedStroke(1);
        final int nRows=countRows();
        final int nCols=countColumns();
        if (nRows<=0||nCols<=0) return;
        final Rectangle cb=charB();
        final int charW=cb.width, charH=cb.height, lineH=charH+lineSkip;
        final StrapProtein[] ppVis=getVisibleProteins();
        final ResidueSelection ss[]=A.SELECTED_OBJECTS.residueSelections(' ');
        final Rectangle sbArea= gSB!=null ? hSB.getTrack() :null;
        for(int iS=-1; iS<ss.length; iS++) {
            final ResidueSelection s=iS>=0 ? ss[iS]: featureUnderM;
            if (s!=null) {
                if (s instanceof ResidueAnnotation && !((ResidueAnnotation)s).isEnabled()) continue;
                if (s instanceof VisibleIn123 && (((VisibleIn123)s).getVisibleWhere()&VisibleIn123.SEQUENCE)==0) continue;
                final int offset=s.getSelectedAminoacidsOffset();
                final boolean bb[]=s.getSelectedAminoacids();
                final StrapProtein p=(StrapProtein)s.getProtein();
                if (p==null || bb==null) continue;
                final int gaps[]=p.getResidueGap();
                final int cols[]=p.getResidueColumn();
                final int nR=p.countResidues();
                final int row=idxOf(p,ppVis);
                if (row<0) continue;
                final int style= isInstncOf(VisibleIn123.class,s) ? ((VisibleIn123)s).getStyle() : -1;
                final boolean lines=(style&255)==VisibleIn123.LINES;
                final int yT=lines || style>=0 ? ResidueSelectionUtils.yT(style,charH) : 0;
                final int yB=lines ? yT+2 : style>=0 ?   ResidueSelectionUtils.yB(style,charH) : lineH-1;
                final int iB_to=mini(bb.length, nR-offset);
                final Color color=s instanceof VisibleIn123 ? ((VisibleIn123)s).getColor() : WHITE;
                final Color complColor=color!=null ? C( 0xFFffFF&(~color.getRGB())) : BLACK;
                for(int iB=maxi(0,-offset); iB<iB_to; iB++) {
                    if (!bb[iB]) continue;
                    int j=iB+1;
                    while(j<iB_to && bb[j] && gaps[j+offset]==0) j++;
                    final int colFrom0=cols[iB+offset], colTo0=cols[j-1 +offset]+1;
                    iB=j-1;
                    for(boolean isSB : FALSE_TRUE) {
                        if (!isSB) {
                            if (cMax<colFrom0 || cMin>colTo0) continue;
                            if (rMax<row+1 || rMin>row) continue;
                        }
                        final int colFrom=isSB ? colFrom0 : maxi(cMin-1,colFrom0);
                        final int colTo=isSB ? colTo0 : mini(cMax+1,colTo0);
                        final Graphics2D g= isSB ? gSB : gAli;
                        if (g==null) continue;
                        final int x1, x2, y1, y2;
                        if (isSB) {
                            if (sbArea==null) continue;
                            x1=sbArea.x+colFrom*sbArea.width/nCols;
                            x2=sbArea.x+colTo  *sbArea.width/nCols-1;
                            y1=sbArea.y+row    *sbArea.height/nRows;
                            y2=sbArea.y+(row+1)*sbArea.height/nRows-1;
                        } else { x1=colFrom*charW; x2=colTo*charW-2; y1=row*lineH+yT; y2=y1-yT+yB; }
                        for(boolean isDashed:FALSE_TRUE) {
                            g.setColor(isDashed ? complColor : color);
                            g.setStroke(!isDashed ? stroke0 : isSB || y2-y1<3 ? dashedStroke1 : dashedStroke0);
                            if (y2-y1<3) {
                                if (isDashed) for(int y=y1; y<=y2; y++) g.drawLine(x1,y,x2,y);
                                else g.fillRect(x1,y1,x2-x1,y2-y1);
                            } else g.drawRect(x1,y1,x2-x1, y2-y1);
                        }
                    }
                }
            }
        }
        final int rectW,rectN,rectE, rectS;
        {
            final Rectangle r=getRectRubberBand();
            if (r!=null) {
                rectW=r.x;
                rectN=r.y;
                rectE=rectW+r.width;
                rectS=rectN+r.height;
            } else rectW=rectN=rectE=rectS=-1;
        }
        if (rectW<rectE && rectN<rectS) {
            for(boolean isSB : FALSE_TRUE) {
                final Graphics2D g= isSB ? gSB : gAli;
                if (g==null) continue;
                final int x1, x2, y1, y2;
                if (isSB) {
                    if (sbArea==null) continue;
                    x1=sbArea.x+rectW*sbArea.width/nCols;
                    x2=sbArea.x+rectE*sbArea.width/nCols-1;
                    y1=sbArea.y+rectN*sbArea.height/nRows+1;
                    y2=sbArea.y+rectS*sbArea.height/nRows;
                } else  { x1=rectW*charW; x2=rectE*charW-2; y1=rectN*lineH; y2=rectS*lineH-2; }
                g.setColor(RED);
                g.setStroke(stroke0);
                g.drawRect(x1-1,y1-1,x2-x1+2,y2-y1+2);
                for(boolean isDashed:FALSE_TRUE) {
                    g.setColor(isDashed ? RED : WHITE);
                    g.setStroke(!isDashed ?  stroke0 : isSB  ? dashedStroke1 : dashedStroke0);
                    g.drawRect(x1,y1,x2-x1, y2-y1);
                }
            }
        }
        drawProtectAlignment( stroke0, dashedStroke1, gAli,  gSB, 0,MAX_INT);
        if (gAli!=null) { gAli.setStroke(stroke0); gAli.translate(-BUGFIX_OPENJDK_G2,-BUGFIX_OPENJDK_G2); }
        if (gSB!=null)  { gSB.setStroke(stroke0); gSB.translate(-BUGFIX_OPENJDK_G2,-BUGFIX_OPENJDK_G2); }
    }
    private static int protectFromColumn;
    public static int getProtectFromColumn(char WorN) {
        final StrapAlign a=a();
        Gaps2Columns g2c=a!=null ? a.gap2col() : null;
        if (g2c==null) return -1;
        final char w_or_n=WorN!=0 ? WorN :  g2c.isWide() ? 'W' : 'N';
        if (w_or_n=='w') return protectFromColumn;
        else {
            return a!=null ? get(protectFromColumn, g2c.wide2narrowColumn()) : -1;
        }
    }
    private void drawProtectAlignment(Stroke stroke0, Stroke dashedStroke, Graphics2D gAli, Graphics2D gSB, int colFrom, int colTo) {
        final int colProtect=getProtectFromColumn((char)0),  nCols=countColumns();
        if (colProtect<=0 || colProtect<colFrom || colProtect>colTo || nCols==0) return;
        for(boolean isSB : FALSE_TRUE) {
             final Graphics2D g= isSB ? gSB : gAli;
            if (g==null) continue;
            final int x, y1,y2;
            if (isSB) {
                final Rectangle sbArea=hSB.getTrack();
                if (sbArea==null) continue;
                x=sbArea.x+colProtect*sbArea.width/nCols;
                y1=0;
                y2=hSB.getHeight();
            } else {
                final Rectangle rect= getVisibleXYWH();
                x=charB().width*colProtect;
                if ((x<rect.x-2 || x>2+rect.x+rect.width)) continue;
                y1=rect.y;
                y2=y1+rect.width;
            }
            if (stroke0!=null) g.setStroke(stroke0);
            g.setColor(WHITE);
            g.fillRect(x-1, 0, 3, Short.MAX_VALUE);
            g.setColor(BLACK);
            if (dashedStroke!=null) g.setStroke(dashedStroke);
            g.drawLine(x,y1, x,y2);
        }
    }
    static StrapProtein[] getProteinsInRectangle(StrapView view0) {
        final StrapView v=view0!=null ? view0 : a().getAlignmentPanelWithFocus();
        final Rectangle r=v==null?null:v.getRectRubberBand();
        final int w=r==null?0:r.width;
        final int h=r==null?0:r.height;
        final int y=r==null?0:r.y;
        final StrapProtein pp[]=w<=0 || h<=0 ? StrapProtein.NONE : new StrapProtein[h];
        for(int i=pp.length;  --i>=0;) pp[i]=v.getProtein(y+i);
        return rmNull(pp,StrapProtein.class);
    }
    protected static StrapProtein[] getProteinsInRectangleForAlignment(int minNum) {
        StrapView view=a().getAlignmentPanelWithFocus();
        if (view==null) return StrapProtein.NONE;
        StrapProtein pp[]=a().getSelectedProteinsInVisibleOrder();
        final StrapProtein ppRegion[]=getProteinsInRectangle(view);
        if (ppRegion.length>=0) {
            final UniqueList<StrapProtein> vP=new UniqueList(StrapProtein.class);
            for(StrapProtein p:pp) if (idxOf(p,ppRegion)>=0) vP.add(p);
            pp=sze(vP)>=minNum ? vP.asArray() : ppRegion.length>0 ? ppRegion : pp;
        }
        if (pp.length<minNum) pp=view.getVisibleProteins();
        return pp;
    }
    protected static ChLabel newLabelRectRegion() {
        final StrapView v=StrapAlign.getAlignmentPanelWithFocus();
        final ChLabel lab=new ChLabel("Warning: Rectangular alignment region is defined. ").fg(v!=null && v.getRectRubberBand()!=null ? BLACK:C(DEFAULT_BACKGROUND));
        vLabelsRunningAnts.add(weakRef(lab));
        return lab;
    }
    /* <<< Region <<< */
    /* ---------------------------------------- */
    /* >>> paint >>> */
    public static JComponent getAlignmentPane(StrapView v){  return v!=null ? v.pCenter : null; }
    public JComponent getAlignmentPane() { return pCenter;}
    public HScrollBar getHorizontalSB() { return hSB;}
    public static boolean alreadyPaintedStatic() { return pCenterPaintedStatic;}
    public boolean alreadyPainted() { return pCenterPainted;}
    protected boolean needsRepaintAlignment() {
        return
            isMouseButtonPressed || ChDelay.contains(this) || ChDelay.contains(hSB) ||
            ChDelay.contains(thread(T_CLEAR_SB)) || ChDelay.contains(threadPR(PR_ONLY_CHANGES));
    }
    protected Color[] getShadingColors() {return shadingColors;}
    protected boolean showsSequence() { return iInfo<0;}
    protected boolean isShadingSecStru() {return shadeSecStru;}
    private static void markDirtyRect(boolean fill, short ccc[][], int col,int row, int width, int height) {
        if (ccc==null) return;
        final int fromR=maxi(0,row);
        final int toR=mini(sze(ccc), row+height);
        for(int r=fromR; r<toR; r++) {
            final short cc[]=ccc[r];
            final int fromC=maxi(0,col);
            final int toC=maxi(sze(cc), col+width);
            for(int c=fromC; c<toC; c++) {
                if (!fill && !(c==fromC || r==toC-1 || r==fromR || r==toR-1)) continue;
                cc[c]=-1;
            }
        }
    }
    private void paintRegion(final Graphics g_, final int colLeft, final int colRight, final int rowTop, final int rowBot,int options) {
        final Object cached[]=cached();
        {
            final int w=pCenter.getWidth(), h=pCenter.getHeight();
            final Dimension panSize=getPaneSize();
            if (panSize.height!=vSB.getMaximum()) {
                ChDelay.runAfterMS(PAUSE,thread(T_WOBBLE),2222);
            }
            if (w!=panSize.width || h!=panSize.height || lastLineSkip!=lineSkip) {
                lastLineSkip=lineSkip;
                if (h!=panSize.height) vSB.invalidate();
                if (w!=panSize.width) hSB.invalidate();
                ChDelay.revalidate(pCenter,555, (long[])null);
            }
        }
        final String strgLen1[]=getStringsLenght1();
        revalidateWhenSizeChanged();
        final boolean ignoreConservation=(PR_IGNORE_CONSERVATION&options)!=0;
        final Graphics2D g=(Graphics2D)g_;
        int countPainted=0, countAll=0;
        final Rectangle cb=charB();
        final Color secStr2color[]=ShadingAA.SECSTRU,
            actg2color[]=ShadingAA.ACTG,
            secStr2color_LessConserved[]=button(TOG_BRIGHTER).s() ? darker(secStr2color) : darker(darker(secStr2color)),
            actg2color_LessConserved[]=button(TOG_BRIGHTER).s() ? darker(actg2color) : darker(darker(actg2color)),
            colorFg=WHITE,
            bgComponent=C(DEFAULT_BACKGROUND),
            colorFg_lessConserved=C(colorFg==WHITE ? 0xc0c0c0 : 0x030303);
        final boolean debugPaint=false;
        {
            final Font f=getPanel().getFont();
            g.setFont(f);
            if (charWidths==null || paintRegion_lastFont!=f) {
                paintRegion_lastFont=f;
                charWidths=getPanel().getFontMetrics(f).getWidths();
            }
        }
        antiAliasing(g,false);
        final StrapProtein[] ppVis=getVisibleProteinsReordered();
        final short oldCharacters[][];
        {
            short[][] old=(short[][])cached[MCSV_CHARACTERS];
            if (sze(old)<ppVis.length) cached[MCSV_CHARACTERS]=old=new short[ppVis.length][];
            oldCharacters=old;
        }
        final int charW4=cb.width/4, charW8=maxi(charW4/8,1), lineH=cb.height+lineSkip;
        final short[][] frequencies= ignoreConservation ? null : computeFrequenciesAtColumn(colLeft,colRight);
        final int shadingThreshold=sliderSimilarity().getValue()*sze(ppShading());
        for(int row=maxi(rowTop,0); row<=rowBot && row<ppVis.length; row++) {
            final int y=row*lineH;
            final StrapProtein p=ppVis[row];
            if (p==null) continue;
            final boolean isShading=frequencies!=null && ppShading==ppVis || cntains(p,ppShading);
            final ResidueSelection[] residueSelections=p.getAllResidueSelections();
            final boolean isACTG=isNucleotides && p.getNucleotides()!=null;
            final byte selectedAA[]=p.getSelectedAminoacids();
            final byte seq[]=p.getGappedSequence(),secStr[]=p.getGappedSecStru();
            if (sze(seq)==0) continue;
            final float access[]=p.getGappedSolventAccessibility();
            final int maxCol=p.getMaxColumn();
            final Color colorSolvAcc[]=ShadingAA.SOLVENT_ACCESSIBILITY2WHITE_BLUE;
            /* ---  gaps black --- */
            final int[] column2index=p.columns2indices();
            short[] ocr=oldCharacters[row];
            if (sze(ocr)<seq.length) ocr=oldCharacters[row]=chSze(ocr,seq.length+99);
            final short[] oldChRow=ocr;
            final int iColMax=mini(colRight,maxCol+1);
            for(int iCol=colLeft,x=colLeft*cb.width; iCol<iColMax; iCol++,x+=cb.width) {
                if (iCol<0 || iCol>=column2index.length) continue;
                final int iAa=column2index[iCol];
                final int hasSelection=iAa>=0 && iAa<selectedAA.length ?  selectedAA[iAa] : 0;
                final byte triplet[]=isACTG ? p.getResidueTriplet(iAa,TRIPLET) : null;
                final int triplet8;
                final short chr=seq[iCol];
                final byte chr7F=(byte)(chr&0x7F);
                final boolean isConserved;
                final short newCh;
                if (iAa>=0) {
                    int t8=0;
                    if (triplet!=null) {
                        final int
                            t0=triplet[0]|32,
                            t1=triplet[1]|32,
                            t2=triplet[2]|32;
                        t8=
                            t0=='a'  ? 0+0 : t0=='c' ? 1+0 : t0=='g' ? 2+0 : 3+0 +
                            t1=='a'  ? 0+4 : t1=='c' ? 1+4 : t1=='g' ? 2+4 : 3+4 +
                            t2=='a'  ? 0+8 : t2=='c' ? 1+8 : t2=='g' ? 2+8 : 3+8;
                    }
                    short nc=chr;
                    final boolean isCons;
                    if (isShading && sze(frequencies)>iCol) {
                        final int c=(chr7F|32)-'a';
                        final short ff[]=frequencies[iCol];
                        isCons=ff!=null ? shadingThreshold==0 || c>=0 && c<1+'z'-'a' && ff[c]*100>shadingThreshold : false;
                    } else isCons=false;
                    if (isCons) { nc|=0x80; t8|=0x80; }
                    nc|= (hasSelection<<8);
                    t8|= (hasSelection<<8);
                    newCh=nc;
                    triplet8=t8;
                    isConserved=isCons;
                }else {triplet8=0;newCh=0;isConserved=false;}
                final boolean isSame;
                if (ignoreConservation) isSame=(~0x80  & oldChRow[iCol]) == (~0x80  & (triplet==null ? newCh : triplet8));
                else isSame= oldChRow[iCol] == (triplet==null ? newCh : triplet8);
                if (!isSame) oldChRow[iCol]= triplet==null ? newCh : (short)triplet8;
                countAll++;
                if ( (options&PR_ONLY_CHANGES)==0  || !isSame  || hasSelection==Byte.MAX_VALUE || (hasSelection!=0 && (options&PR_ALL_SELECTED)!=0)) {
                    g.setColor(chr==' ' || chr==0 ? bgComponent : BLACK);
                    g.fillRect(x,y,cb.width,cb.height+lineSkip);
                    final Color newColor;
                    {
                        final Color nc;
                        if (shadeSecStru && iCol<sze(secStr)) nc=( isConserved ? secStr2color : secStr2color_LessConserved)[secStr[iCol]];
                        else
                            if (showSolvAcc && iCol<sze(access)) nc=colorSolvAcc[mini((int)access[iCol],colorSolvAcc.length-1)];
                            else nc=isConserved ? shadingColors[chr] : colorFg_lessConserved;
                        newColor=nc!=BLACK ? nc : colorFg;
                    }
                    if (chr!=0 && chr!='-')  {
                        cb.x=(cb.width-charWidths[chr7F])/2;
                        if (hasSelection!=0) {
                            setBnds(x,y,cb.width, lineH, RECT_EDT);
                            ResidueSelectionUtils.drawResidueSelection(StrapProtein.AMINO_ACIDS,triplet==null ? chr7F : 0,p,residueSelections,iAa,g, RECT_EDT,cb,getPanel());
                        }
                        if (triplet==null) {
                            g.setColor(newColor);
                            g.drawString(strgLen1[chr7F&127], cb.x+x,y+cb.y);
                            countPainted++;
                        }
                        if (triplet!=null) {
                            for(int i=0,xNt=x+charW8;i<3;i++) {
                                final int nt=triplet[i]|32;
                                final int h=nt=='c' || nt=='g' ? cb.height : cb.height/2;
                                final Color col=(isConserved ? actg2color : actg2color_LessConserved)[nt];
                                g.setColor(col!=BLACK ? col : colorFg);
                                g.fillRect(xNt,y+1,charW4-1,h-2);
                                xNt+=charW4;
                            }
                        }
                    }
                }
            }
            /* --- Clear after the sequence --- */
            if (colRight>maxCol) {
                g.setColor(bgComponent);
                g.fillRect((maxCol+1)*cb.width,y,(colRight-maxCol)*cb.width,cb.height+lineSkip);
                final int to=mini(colRight, ocr.length);
                if (maxCol+1<to) Arrays.fill(ocr,maxCol+1,to,(short)0);
            }
        }
    }
    /* <<< paint <<< */
    /* ---------------------------------------- */
    /* >>> Scroll >>> */
    protected void highlightProteins(int ii[],Color cc[]) { vSB.setSelectedY(ii,lineSkip+charB().height,cc); }
    protected void setHorizontalScrollPosition(int v) { try { hSB.setValue(v); } catch(Exception e){}} // ????
    protected boolean isShadingScrollbar() { return isShadingScrollbar;}
    private int vSB_yCursor() {
        final int row=getRowAtCursor(), cr=countRows();
        if (row<0 || cr==0) return -1;
        final Rectangle track=vSB.getTrack();
        return track.y+track.height*(row*2+1)/(2*cr);
    }
    protected void scrollRel(char v_or_h, int i,boolean charsOrPixels) {
        final Rectangle r=getVisibleXYWH(), cb=charB();
        if (v_or_h=='h') r.x+=  i* (charsOrPixels?cb.width:1);
        else r.y+=i* (charsOrPixels?(cb.height+lineSkip):1);
        scrRectToVis(r,false);
    }
    protected void scrollCursorToVisible() {
        if (pCenter==null) return;
        final int w=charB().width, h=charB().height+lineSkip;
        setBnds(getColumnAtCursor()*w, getRowAtCursor()*h,w,h, RECT_EDT);
        zoomZenter=null;
        scrRectToVis(RECT_EDT,false);
    }
    protected void scrollToVisible(int x0, int y0) {
        if (pCenter==null) return;
        final int x=x0>=0 ? x0 : 1-pCenter.getX();
        final int y=y0>=0 ? y0 : 1-pCenter.getY();
        setBnds(mini(x, pCenter.getWidth()-EX),mini(y, pCenter.getHeight()-EX), EX,EX, RECT_EDT);
        scrRectToVis(RECT_EDT,false);
    }
    protected void scrRectToVis(Rectangle rect, boolean center) { scrRectToVis(rect,center,1); }
    private void scrRectToVis(Rectangle rect, boolean center, int tryAgain) {
        if (!isEDT()) ChDelay.invokeAndWait(this,"scrRectToVis",new Object[]{rect, center ? "":null, intObjct(tryAgain), null});
        else {
            vSB.setLock(false);
            final int midx=rect.x+rect.width/2;
            final int midy=rect.y+rect.height/2;
            if (center) {
                final int h=tLayout.getHeight();
                final int w=tLayout.columnWidth(1);
                pCenter.scrollRectToVisible(new Rectangle( midx-w/2, midy-h/2,  w,h));
            } else pCenter.scrollRectToVisible(rect);
            if (tryAgain>0 && !scrollPane.getViewport().getViewRect().contains(midx,midy)) {
                ChDelay.newThread(this,"scrRectToVis",new Object[]{rect,"", intObjct(tryAgain-1), intObjct(555)}).start();
            }
        }
    }
    /* <<< Scroll <<< */
    /* ---------------------------------------- */
    /* >>> Order Of proteins >>> */
    protected void moveLines(int von, int nach, int n) {
        {
            int newCursorLine=getRowAtCursor();
            final int nRows=countRows(), row=newCursorLine;
            if (von+n>nRows || nach+n>nRows) return;
            final int[] newIndices=moveIndices(nRows,von,nach,n);
            if (newIndices==null) return;
            final StrapProtein[] newRows[]=new StrapProtein[nRows][];
            for(int i=0;i<nRows;i++) {
                final int j=newIndices[i];
                if (j<0) return;
                newRows[i]=getProteinsInRow(j);
                if (j==row) newCursorLine=i;
            }
            getList().clear();
            adAll(newRows,getList());
            setCursorToResidue(getProteinAtCursor(),cursorAA);
        }
        ChDelay.runAfterMS(EDT|PAUSE,threadPR(PR_ONLY_CHANGES),threadPR(PR_ONLY_CHANGES).LAST[DURATION]*4);
        final long durationRH[]=((ChPanel)rowHeader).getWhenPainted();
        ChDelay.repaint(rowHeader,4*durationRH[DURATION],durationRH);
        hSB.clearImage().repaint();
        StrapAlign.dispatchLater(StrapEvent.ORDER_OF_PROTEINS_CHANGED,111);
    }
    private static int[] moveIndices(int max, int von, int nach, int n) {
        int[] r=rh_indices;
        if (r==null || r.length!=max) r=rh_indices=new int[max];
        int i,j=0;
        if (max<2 || (von+n)>max ||  nach>=max || von<0 || nach<0) return null;
        for(i=0;i<max;i++) r[i]=i;
        if (von<nach) {
            if (nach+n>max) return r;
            for(i=0;i<von;i++) r[j++]=i;
            for(i=von+n;i<n+nach;i++) r[j++]=i;
            for(i=von;i<von+n;i++) r[j++]=i;
            for(i=nach+n;i<max;i++) r[j++]=i;
        }
        if (nach<von) {
            if (von+n>max) return r;
            for(i=0;i<nach;i++) r[j++]=i;
            for(i=von;i<von+n;i++) r[j++]=i;
            for(i=nach;i<von;i++) r[j++]=i;
            for(i=von+n;i<nach;i++) r[j++]=i;
        }
        return r;
    }
    /* <<< Order Of proteins <<< */
    /* ---------------------------------------- */
    /* >>> Event >>> */
    static long whenMouseInToolbar() { return mouseInToolbarWhen;}
    @Override public void handleEvent(StrapEvent ev) {
        final Object oo[]=ev.getParameters(), q=ev.getSource();
        if (isClosed) return;
        final int t=ev.getType();
        boolean doEnable=false, clrHSB=false;
        final boolean focused=StrapAlign.getAlignmentPanelWithFocus()==this;
        /*
          if (oo.length>2 && oo[0]==KEY_PROGRESS && oo[2] instanceof int[]) {
          final int[] progress=(int[])oo[2];
          final Object src=oo[1];
          final ChButton b=src==SequenceAligner.class ? button(BUT_ALIGN) : null;
          if (b!=null) b.setProgressBar(progress[0],progress[1]);
          }
        */
        if (t==StrapEvent.PROTEINS_HIDDEN || t==StrapEvent.PROTEINS_KILLED) {
            final int nRowsOld=countRows();
            if (oo.length>0) {
                removeProteins((StrapProtein[])oo[0]);
                removeEmptyRows();
            }
            if (idxOf(cursorP,getVisibleProteins())<0) {
                if (cursorP!=null) StrapAlign.dispatchLater(StrapEvent.CURSOR_CHANGED_PROTEIN,666);
                cursorP=null;
            }
            update_isInAlignment();
            clrHSB=doEnable=true;
            final int nRows=countRows();
            if (nRowsOld>nRows) setReduceHeight(maxi(0, this.reduceHeight- row2y(nRowsOld-nRows)));
            if (nRows<5) setReduceHeight(0);
        } else if (t==StrapEvent.PROTEINS_ADDED || t==StrapEvent.PROTEINS_SHOWN) {
            oneRowVisible=true;
            final int row=get(1,oo)!=null ? atoi(get(1,oo)) : -1;
            final StrapProtein pp[]=get(0,oo, StrapProtein[].class);
            if (sze(pp)>0) addProteins(row, pp, 0);
            update_isInAlignment();
            setCursorToResidue(cursorP,cursorAA);
            clrHSB=doEnable=true;
            //for(int i=sze(oo); --i>=0;) if (oo[i] instanceof Runnable) {  ((Runnable)oo[i]).run(); }
        } else if (t==StrapEvent.NEED_SCROLL_DOWN_ALIGNMENT) {
            scrRectToVis(new Rectangle(0,row2y(countRows())-2, 1, 1),false,4);
        }
        int paintRegion=0;
        final JComponent toolbar=TOOLBARS[0];
        if (getPanel()!=null && toolbar!=null && (alreadyPainted || (currentTimeMillis()-TIME_AT_START)>9999)) {
            if (t==StrapEvent.USER_PROFILE_CHANGED) {
                toolbar.revalidate();
                for(int i=3; --i>=0;) button(i==0 ? BUT_NUCL : i==1 ? BUT_SEQVISTA : TOG_MAXIMIZE).revalidate();
            }
            if ( (t&StrapEvent.FLAG_ALIGNMENT_CHANGED)!=0) {
                ChDelay.runAfterMS(EDT,threadPR(PR_ONLY_CHANGES|PR_IGNORE_CONSERVATION), 10);
                if (showsSequence()) {
                    clrHSB=true;
                    ChDelay.runAfterMS(EDT,threadPR(PR_ONLY_CHANGES), threadPR(PR_ONLY_CHANGES).LAST[DURATION]*3);
                } else pCenter.repaint();
            }
            if (t==StrapEvent.AA_SHADING_CHANGED) {
                threadPR(PR_ONLY_CHANGES).run();
                if (!button(TOG_OUTLINE_HSB).s()) clrHSB=true;
            }
            if (t==StrapEvent.ORDER_OF_PROTEINS_CHANGED || t==StrapEvent.PROTEINS_KILLED || t==StrapEvent.PROTEINS_HIDDEN || t==StrapEvent.PROTEINS_SHOWN || t==StrapEvent.PROTEINS_ADDED) {
                this.ppR=null;
                vSB.repaint();
            }
            if (focused && (t==StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN || t==StrapEvent.CURSOR_CHANGED_PROTEIN)) {
                threadPR(PR_ONLY_CHANGES).run();
                doEnable=true;
            }
            if (t==StrapEvent.RESIDUE_SELECTION_CHANGED_COLOR || t==StrapEvent.RESIDUE_SELECTION_CHANGED) threadPR(PR_ALL_SELECTED).run();
            if (t==StrapEvent.RESIDUE_SELECTION_CHANGED || t==StrapEvent.RESIDUE_SELECTION_DELETED || t==StrapEvent.RESIDUE_SELECTION_ADDED) {
                if (q!=Protein.class) {
                    final String name=getNam(get(0,oo));
                    if (name !=ResidueSelection.NAME_MOUSE_OVER && name!=ResidueSelection.NAME_BACKBONE) button(TOG_OUTLINE_HSB).s(isShadingScrollbar=false);
                }
                clrHSB=true;
                paintRegion=PR_ONLY_CHANGES;
            }
            if (t==StrapEvent.ATOM_COORDINATES_CHANGED || t==StrapEvent.RESIDUE_SELECTION_DELETED) {
                pCenter.repaint();
                ChDelay.runAfterMS(EDT,thread(T_ENABLED),555);
            }
            if (t==StrapEvent.VALUE_OF_ALIGN_POSITION_CHANGED || t==StrapEvent.VALUE_OF_ALIGN_POSITION_CHANGED_COLOR ||
                t==StrapEvent.VALUE_OF_RESIDUE_CHANGED || t==StrapEvent.VALUE_OF_RESIDUE_CHANGED_COLOR) {
                if (button(TOG_OUTLINE_HSB).s()) clrHSB=true;
                if (!UserProfile.isWebMode()) button(TOG_OUTLINE_HSB).s(isShadingScrollbar=false);
                if (t==StrapEvent.VALUE_OF_ALIGN_POSITION_CHANGED || t==StrapEvent.VALUE_OF_RESIDUE_CHANGED) pCenter.repaint();
                else plot(pCenter.getGraphics());
                hSB.repaint();
            }
            if (t==StrapEvent.OBJECTS_SELECTED) {
                doEnable=true;
                final Object clazz=get(0,oo);
                if (clazz==Protein.class || clazz==StrapProtein.class ) {clrHSB=true; paintRegion=PR_ONLY_CHANGES;}
                final Graphics2D g=(Graphics2D)pCenter.getGraphics();
                if (g!=null) {
                    final Rectangle v=getVisibleRowsAndCols();
                    paintRegion(g, v.x,v.x+v.width,  v.y,v.y+v.height,  PR_ALL_SELECTED);
                    marchingAnt(g, null);
                }
                vSB.repaint();
            }
            if ( (t& (StrapEvent.FLAG_ROW_HEADER_CHANGED|StrapEvent.HETERO_OR_NUCLEOTIDE_STRUCTURE_ADDED_OR_REMOVED))!=0) rowHeader.repaint();
            if ( (t==StrapEvent.VALUE_OF_PROTEIN_CHANGED || t==StrapEvent.PROTEIN_PROTEIN_DISTANCE_CHANGED) && sze(getPlottersV())>0) {
                final long durationRH[]=((ChPanel)rowHeader).getWhenPainted();
                ChDelay.repaint(rowHeader,4*durationRH[DURATION],durationRH);
            }
        }
        maybeUpdateCursor();
        if (doEnable) ChDelay.runAfterMS(EDT,thread(T_ENABLED),333);
        if (clrHSB) {
            final long duration=hSB.LAST[DURATION]*5+100;
            ChDelay.runAfterMS(duration<100 ? EDT|PAUSE : EDT, thread(T_CLEAR_SB),duration);
        }
        if (paintRegion>0) ChDelay.runAfterMS(EDT|PAUSE,threadPR(paintRegion),threadPR(paintRegion).LAST[DURATION]*4);
    }
    public AWTEvent processEvent(JComponent component, AWTEvent e) {
        final int id=e.getID(),kcode=keyCode(e), modi=modifiers(e);
        final Object q=e.getSource();
        if (isPopupTrigger(e) && BUT_ALIGN==idxOf(q,_button)) {
            final Container ctrlPan=getCtrlPanel(lastAligner);
            if (ctrlPan!=null) ChFrame.frame("Aligner",ctrlPan,0).shw(ChFrame.AT_CLICK);
        }
        if (tLayout==null) return null;
        final boolean ctrl= isCtrl(e), shift=(modi&SHIFT_MASK)!=0, alt=(modi&ALT_MASK)!=0;
        final Rectangle cb=charB();
        final MouseEvent mev=e instanceof MouseEvent ? (MouseEvent)e : null;
        final UniqueList vSEL=StrapAlign.getSelectedObjectsV();
        final int vSEL_mc=modiCount(vSEL);
        final int lineH=cb.height+lineSkip, charW=cb.width;
        final int nRows=countRows();
       final StrapAlign align=a();
        final javax.swing.table.JTableHeader header=tLayout.getTableHeader();
        if (id==FocusEvent.FOCUS_LOST || id==FocusEvent.FOCUS_GAINED) {header.setForeground(id==FocusEvent.FOCUS_GAINED?BLACK:GRAY);}
        if (q==rowHeader&& id==KeyEvent.KEY_PRESSED) {
            if (kcode==KeyEvent.VK_BACK_SPACE || kcode==KeyEvent.VK_DELETE) {
                final StrapProtein[] pSel=align.getSelectedProteins();
                if (ChMsg.yesNo("Remove "+pSel.length+" proteins?")) align.removeProteins(false,pSel);
                return null;
            }
            if (kcode==KeyEvent.VK_A && ctrl) {
                for(Object o: getProteins()) vSEL.add(o);
                new StrapEvent(this,StrapEvent.OBJECTS_SELECTED).setParameter(Protein.class).run();
                return null;
            }
        }
        if (mev!=null) {
            final int rectN,rectW,rectE,rectS;
            final Rectangle rect=getRectRubberBand();
            if (rect!=null) {
                rectW=rect.x;
                rectN=rect.y;
                rectE=rectW+rect.width;
                rectS=rectN+rect.height;
            } else rectN=rectW=rectE=rectS=-1;
            boolean rectChanged=false;
            if (q==hSB) vSB.setLock(false);
            final int protectAliR=!align.gap2col().isWide() ? getProtectFromColumn('N') : -1;
            final int x=mev.getX(), y=mev.getY();
            final int mCol=x/charW, mRow=y/lineH;
            if (id==MOUSE_CLICKED && component==rh_info_TA && x<12 && y<12) { setPrpty(StrapView.class,"xrefShownP",""); rh_info_W.setVisible(false); }
            final boolean isDouble=clckCount(e)>1, middle=(modi&BUTTON2_MASK)!=0, isPop=isPopupTrigger(e);
            final StrapProtein pRow=getProtein(mini(nRows-1,mRow));
            int newHeaderWidth=-1;
            if (id==MOUSE_CLICKED || id==MOUSE_RELEASED || id==MOUSE_MOVED || id==MOUSE_EXITED || id==MOUSE_ENTERED)  isMouseButtonPressed=false;
            if (id==MOUSE_PRESSED || id==MOUSE_DRAGGED)  isMouseButtonPressed=true;
            final int wheel=wheelRotation(e);
            if (id==MOUSE_PRESSED) { align.setMenuBar(null);align.setToolpane(null); }
            if (wheel!=0) {
                final int wheel2=alt ? wheel :4*wheel;
                if (ctrl) setFont("",wheel2,false);
                else {
                    if (q==pCenter) {
                        align.setAlignmentPanelWithFocus(this);
                        if (!shift && vSB.isVisible()) scrollRel('v',4*wheel2,false); else scrollRel('h',wheel2,true);
                    }
                    if (q==rowHeader) { final float f=rh_adjust+wheel*.1f; this.rh_adjust=f<0?0:f>1?1:f;  rowHeader.repaint(); }
                    if (q==hSB) scrollRel('h',wheel2,true);
                }
            }
            if (id==MOUSE_EXITED) {
                setTipTiming(-1,0);
                if (q==pCenter) { looksLikeDivider=false;  scrollPane.repaint(0,0,2*EM,9999);}
                if (q==header) ((JComponent)q).repaint();
            }
            if (id==MOUSE_MOVED) {
                if (q==TOOLBARS[0]) mouseInToolbarWhen=currentTimeMillis();
                doingDnd=rh_doingDnd=false;
            }
            if (q==pCenter && showsSequence()) {
                if (id==MOUSE_CLICKED||id==MOUSE_PRESSED||id==MOUSE_RELEASED||id==MOUSE_PRESSED) {
                    this.zoomZenter=null;
                    Arrays.fill(zoom_reduceHeight,-1);
                }
                final char isDrag=this.isDragged;
                final boolean edgeN, edgeS, edgeE, edgeW;
                if (rectW<rectE && rectN<rectS && isDrag==0) {
                    final int d=charW/2;
                    final int rN=rectN*lineH, rS=rectS*lineH, rE=rectE*charW, rW=rectW*charW;
                    final boolean in=rN-d<y && rS+d>y && rW-d<x && rE+d>x;
                    edgeN=in && Math.abs(rN-y)<d;
                    edgeS=in && Math.abs(rS-y)<d;
                    edgeE=in && Math.abs(rE-x)<d;
                    edgeW=in && Math.abs(rW-x)<d;
                } else edgeN=edgeS=edgeE=edgeW=false;
                if (mouseMovedCol!=mCol || mouseMovedRow!=mRow || resSelUnderM==null) {
                    mouseMovedCol=mCol;
                    mouseMovedRow=mRow;
                    resSelUnderM=getResidueSelectionsAtXY(x,y, VisibleIn123.SEQUENCE);
                }
                this.mouseOverRectangle=edgeN || edgeS || edgeE || edgeW;
                if (isPop) {
                    final ResidueSelection ss[]=ResidueSelectionUtils.filter(ResidueSelectionUtils.FILTER_NOT_DISABLED,this.resSelUnderM);
                    if (ss.length>0) {
                        setEventLocationOnScreen(e);
                        ResidueSelectionPopup.showContextMenu(ss, e);
                        setPrpty(StrapView.class,"popupShownS","");
                    } else {
                        setPrpty(StrapView.class,"popupShownR","");
                        getPopup(mouseOverRectangle).show(pCenter,x,y);
                    }
                    return null;
                } else if (id==MOUSE_CLICKED) {
                    final ResidueSelection ss[]=ResidueSelectionUtils.filter(ResidueSelectionUtils.FILTER_NOT_DISABLED,this.resSelUnderM);
                    if (isDouble) {
                        for(ResidueSelection s : ss) if (s instanceof ResidueAnnotation) align.editAnnotation((ResidueAnnotation)s);
                    } else if (ctrl && !shift && !alt && ss.length>0) {
                        for(ResidueSelection s : ss)  {
                            if (s!=resSel_cursor && vSEL.contains(s)) vSEL.remove(s);
                            else vSEL.add(s);
                        }
                    } else if (!middle && shift && !alt && !ctrl) {
                        final int countFeature=ResidueSelectionUtils.countType('F',ss);
                        final int countAnnotat=ResidueSelectionUtils.countType('A',ss);
                        if (sze(ss)>0 && selIntervallRow>=0) {
                            for(int row=mini(mRow,selIntervallRow); row<=maxi(mRow,selIntervallRow); row++) {
                                StrapProtein p=getProtein(row);
                                if (p!=null) {
                                    for(ResidueSelection s : p.getAllResidueSelections()) {
                                        final boolean take=
                                            s==resSel_cursor ? false :
                                            s instanceof ResidueAnnotation ? (ResidueAnnotation.isAnno(s)==(countAnnotat>0)) :
                                            countFeature==0 && countAnnotat==0;
                                        if (take) {
                                            final boolean bb[]=s.getSelectedAminoacids();
                                            final int offset=s.getSelectedAminoacidsOffset(), first=p.getResidueColumn(firstTrue(bb)+offset), last=p.getResidueColumn(lastTrue(bb)+offset);
                                            if (first<=maxi(mCol, selIntervallCol) && last>=mini(mCol, selIntervallCol)) vSEL.add(s);
                                        }
                                    }
                                }
                            }
                        } else {
                            if (pRow==mouseSelection().getProtein()) mouseSelection().set(pRow,colLastClicked,mCol);
                            else if (rectW <rectE) {
                                final int r0=mini(rowDraggedFrom, mRow);
                                final int r1=maxi(rowDraggedFrom, mRow+1);
                                final int c0=maxi(colDraggedFrom, mCol+1);
                                final int c1=mini(colDraggedFrom, mCol);
                                setRectRubberBand(c0,r0, c1-c0, r1-r0);
                            }
                        }
                    } else {
                        this.colLastClicked=mCol;
                        pCenter.setCursor(getCursr('D'));
                        align.setAlignmentPanelWithFocus(this);
                        if (isDrag==0 && (modi&BUTTON3_MASK)==0) {
                            if (!setCursorToColumn(getProtein(mRow),mCol)) {
                                StrapAlign.dispatchLater(StrapEvent.CURSOR_CLICKED,99);
                            }
                        }
                    }
                    if (ss.length==0 || !(ctrl||shift)) selIntervallCol=selIntervallRow=-1;
                    if (ss.length>0 && ctrl) { selIntervallRow=mRow; selIntervallCol=mCol; }
                }
                if (id==MOUSE_DRAGGED || id==MOUSE_PRESSED) {
                    final ResidueSelection[] resSelAtXY=getResidueSelectionsAtXY(x,y, VisibleIn123.SEQUENCE);
                    boolean would_exportDrg=false;
                    final boolean dragVertically=Math.abs(prevMouseX-x)<Math.abs(prevMouseY-y);
                    if (isDrag==0 && !doingDnd) {

                        boolean selectedResSelUnderMouse=false;
                        for(ResidueSelection sel : resSelAtXY) selectedResSelUnderMouse=selectedResSelUnderMouse||align.getSelectedObjectsV().contains(sel);
                        if (dragVertically || selectedResSelUnderMouse) {
                            final List v=getDndFilesV(pCenter);
                            v.clear();
                            for(ResidueSelection sel : resSelAtXY) {
                                adUniq(sel,v);
                                final String types="AFS";
                                for(int type=sze(types); --type>=0;) {
                                    final ResidueSelection aa[]=align.SELECTED_OBJECTS.residueSelections(types.charAt(type));
                                    if (cntains(sel,aa)) adAllUniq(aa,v);
                                }
                            }
                            if (sze(v)>0) {
                                if (id==MOUSE_DRAGGED) {
                                    exportDrg(pCenter, e, TransferHandler.COPY);
                                    doingDnd=true;
                                }
                                would_exportDrg=true;
                            }
                        }
                    }
                    if (!would_exportDrg) {
                        align.setAlignmentPanelWithFocus(this);
                        if (!alt && !middle) {
                            if (isDrag==0 && looksLikeDivider && !edgeW) isResizing=true;
                            if (isDrag==0 && !isResizing && (id!=MOUSE_PRESSED || resSelAtXY.length==0)) {
                                colDraggedFrom=mCol; //mousePressedRow
                                rowDraggedFrom=mRow;
                                int d=0;
                                if (edgeN) { rowDraggedFrom=rectS; d+='n'; }
                                if (edgeS) { rowDraggedFrom=rectN; d+='s'; }
                                if (edgeE) { colDraggedFrom=rectW; d+='e'; }
                                if (edgeW) { colDraggedFrom=rectE; d+='w'; }
                                if (d!=0) pCenter.setCursor(getCursr(d));
                                if (Math.abs(x-col2x(protectAliR))<charW/2) d='P';
                                this.isDragged= d!=0 ? (char)d  :  'A';
                            }
                        }
                    }
                }
                if (id==MOUSE_DRAGGED && !doingDnd && !alt && !middle) {
                    if (pRow!=null) {
                        final int col0=colDraggedFrom, row0=rowDraggedFrom;
                        final boolean all=isDrag=='A';
                        if (row0==mRow && all && y>=0 && y<pCenter.getHeight()) {
                            mouseSelection().set(pRow,col0,mCol);
                            setBnds(0,0,0,0, rectRBA);
                        } else if (isDrag=='P') {
                            setProtectFromColumn(maxi(1,get(mCol, align.gap2col().narrow2wideColumn())), false);
                        } else {
                            if (!_rectRubBand_isAdjusting) {
                                _rectRubBand_isAdjusting=true;
                                addMarchingAnt(ANTS_SEARCH, pCenter, (Rectangle)null,(Color)null, (Color)null );
                            }
                            mouseSelection().set(null,0,0);
                            int rn=rectN, rs=rectS, re=rectE, rw=rectW;
                            if (isDrag=='w' || isDrag=='s'+'w' || isDrag=='n'+'w' || all) { rw=mini(col0, mCol); isResizing=false; }
                            if (isDrag=='n' || isDrag=='n'+'e' || isDrag=='n'+'w' || all) { rn=mini(row0, mRow); }
                            if (isDrag=='s' || isDrag=='s'+'e' || isDrag=='s'+'w' || all) { rs=mini(nRows,maxi(row0, mRow)+1); }
                            if (isDrag=='e' || isDrag=='s'+'e' || isDrag=='n'+'e' || all) { re=maxi(col0, mCol); }
                            rectChanged=setBnds(rw, rn, re-rw, rs-rn,rectRBA);
                        }
                    }
                    if (isResizing) newHeaderWidth=x+pCenter.getX()+tLayout.columnWidth(0);
                    if (!looksLikeDivider) scrollToVisible(x,y);
                }
                if (id==MOUSE_MOVED) {
                    final ResidueAnnotation prevF=featureUnderM;
                    featureUnderM=null;
                    setTipTiming(500,0);
                    this.isDragged=0;
                    isResizing=false;
                    final boolean slider=x+pCenter.getX()<2*EM;
                    Cursor cursor=null;
                    if (looksLikeDivider!=slider) {
                        cursor= getCursr(slider ?'e':'D');
                        looksLikeDivider=slider;
                        scrollPane.repaint(0,0,2*EM,9999);
                    }
                    pCenter.setAutoscrolls(!looksLikeDivider);
                    if (cursor==null)  cursor=getCursr( x+pCenter.getX()<EM && !edgeW ? 'e' : ((edgeN?'n':0) + (edgeS?'s':0) + (edgeE?'e':0) + (edgeW?'w':0)) );
                    pCenter.setCursor(mCol==protectAliR ? getCursr('w') :  cursor!=null ? cursor : getCursr('D'));
                    for(int iS=sze(this.resSelUnderM); --iS>=0;) {
                        final ResidueSelection s=get(iS,this.resSelUnderM, ResidueSelection.class);
                        if (s instanceof ResidueAnnotation) {
                            final ResidueAnnotation a=(ResidueAnnotation)s;
                            final int style=a.getStyle();
                            if ((style&VisibleIn123.LINES)!=0) {
                                final int y0=row2y(mRow)+ResidueSelectionUtils.yT(style,cb.height);
                                if (y0<=y && y<=y0+2) { featureUnderM=a;  break; }
                            }
                            if (ResidueSelectionUtils.isShownIn(s,VisibleIn123.SEQUENCE)!=null) cursor=ChIcon.getCursor(IC_ANNO,pCenter);
                        }
                    }
                    if (prevF!=null && prevF!=featureUnderM) pCenter.repaint(0,row2y(findRow(prevF.getProtein()))+cb.height-3,99999,lineSkip);

                }
                if (id==MOUSE_PRESSED) {
                    button(BUT_SAVE_CURSOR).enabled(true);
                    align.setAlignmentPanelWithFocus(this);
                    mousePressedRow=mRow;
                    mousePressedCol=mCol;
                    pCenter.requestFocus();
                }
                if (rectChanged) {
                    pCenter.repaint();
                    hSB.repaint();
                    for(int i=sze(vLabelsRunningAnts); --i>=0;) {
                        final JComponent lab=get(i,vLabelsRunningAnts,JComponent.class);
                        if (lab!=null) {
                            lab.setForeground(rectW<rectE ? BLACK:C(DEFAULT_BACKGROUND));
                            setMarchingAnt(ANTS_SEARCH, lab, rectW<rectE ? new Rectangle(0,0,lab.getWidth()-1, lab.getHeight()-1) : null ,WHITE,RED);
                        }
                    }
                }
                if (id==MOUSE_RELEASED && _rectRubBand_isAdjusting) {
                    if (rect!=null) setRectRubberBand(rect.x,rect.y, rect.width, rect.height);
                    _rectRubBand_isAdjusting=false;
                }
                if ( rectW<rectE && ( id==MOUSE_RELEASED && isDrag!=0 || id==MOUSE_CLICKED && rectChanged)) {
                    if (!shift && !ctrl) vSEL.removeAll(align.SELECTED_OBJECTS.residueSelections(' '));
                    String msg=null;
                    for(ResidueSelection s:residueSelectionsInRectangle() ) {
                        if (s==resSel_cursor) continue;
                        if (ctrl) { vSEL.remove(s); msg="Ctrl-key: Unselected";}
                        else { if (shift) msg="Shift-key: Added"; vSEL.add(s); }
                    }
                    if (msg!=null) StrapAlign.drawMessage("@1"+ANSI_GREEN+msg+" all residues selections in rectangular region");
                }
            }
            if (id==MOUSE_DRAGGED && q==header) {
                final int w=tLayout.columnWidth(0);
                if (x>w-55 && x<w+55 || isResizing) { newHeaderWidth=x; isResizing=true;}  else isResizing=false;
                oneRowVisible=false;
                reduceHeightSave=0;
                setReduceHeight(reduceHeight+y-header.getHeight()/2);
                button(TOG_MAXIMIZE).s(false);
                if (align.TOG_MAXIMIZE_DIALOGS.s()) { align.TOG_MAXIMIZE_DIALOGS.s(false);  align.rightCollapse(false); }
            }
            if (newHeaderWidth>0) tLayout.setColWidth(false,0,  mini(getPanel().getWidth()*5/6,newHeaderWidth));
            if (component==rowHeader) {
                final StrapProtein ppVis[]=getVisibleProteins(), ppSel[]=align.getSelectedProteins();
                if (isPop && pRow!=null) {
                    setPrpty(StrapView.class,"popupShownP","");
                    StrapAlign.showContextMenu('P',oo(pRow),e); return null;
                }
                final Container parent2=rowHeader.getParent().getParent();
                final Point parentXY=SwingUtilities.convertPoint(rowHeader, x, y,parent2);
                final int px=parentXY.x;
                boolean dndStarted=false;
                final boolean setSelection=rh_draggedProtein2row==null;
                if (id==MOUSE_DRAGGED && !alt) {
                    StrapAlign.drawMessage("Info: Mouse-drag  up or down allows changing order of proteins. \n"+
                                           "Drag sidewards  to export protein to a target (See Wikipedia Drag-and-drop)");
                    if (rh_ppDrag==null) rh_ppDrag=idxOf(pRow,ppSel)>=0 ? a().getSelectedProteinsInVisibleOrder() : new StrapProtein[]{pRow};
                    final JFrame frame=align.getFrame();
                    final Point XY=SwingUtilities.convertPoint(rowHeader, x, y,frame);
                    final Component underMouse=SwingUtilities.getDeepestComponentAt(frame, XY.x, XY.y);
                    final boolean isDragTarget=underMouse instanceof Protein3d || gcp(KEY_DRAG_TARGET_DIRCTORY,underMouse)!=null;
                    final boolean isHorizontally=isMouseDrag && Math.abs(x-mouseDragFromX)>Math.abs(y-mouseDragFromY)+1;
                    if (null==gcp(KEY_LOOKS_LIKE_DIVIDER,rowHeader))
                    if (!rh_doingDnd && pRow!=null && (isHorizontally || px<4 || px>parent2.getWidth()-4) || rh_ppDrag!=null && isDragTarget) {
                        dndStarted=rh_doingDnd=true;
                        System.gc();
                        final List dndFile=getDndFilesV(rowHeader);
                        dndFile.clear();
                        for(StrapProtein p: rh_ppDrag) {
                            if (p!=null) adAllUniq(p.getDndFiles(), dndFile);
                            adUniq(p,dndFile);
                        }
                        final int dnd_y=mini(rowHeader.getHeight()-4,maxi(4,y));
                        final MouseEvent dndEv=new MouseEvent(rowHeader,MOUSE_DRAGGED,currentTimeMillis(), modi ,0,dnd_y,0,false);
                        exportDrg(rowHeader, isMacintosh() ? mev: dndEv, shift  ? TransferHandler.COPY_OR_MOVE :  TransferHandler.COPY);
                    }
                    if (!rh_doingDnd) {
                        rh_draggedDestRow=mRow;
                        if (rh_draggedSrcRow<0) {
                            rh_draggedSrcRow=mRow;
                            this.rh_countDraggedRows=1;
                            if (cntains(pRow,ppSel)) {
                                for(int r=mRow;--r>=0 && cntains(getProtein(r),ppSel);) rh_draggedSrcRow=r;
                                for(int r=mRow; r<nRows && cntains(getProtein(r),ppSel); r++) this.rh_countDraggedRows=r-this.rh_draggedSrcRow+1;
                            }
                            rh_lastDraggedDestRow=-1;
                        }
                        rowHeader.setCursor(getCursr('M'));
                        if (rh_lastDraggedDestRow!=rh_draggedDestRow ) {
                            rh_draggedProtein2row= moveIndices(ppVis.length,rh_draggedDestRow, rh_draggedSrcRow, this.rh_countDraggedRows);
                            if (setSelection && rh_draggedProtein2row!=null) {
                                rh_resSel_dragged.setSelectedRange(0,countColumns());
                                for(int i=this.rh_countDraggedRows; --i>=0;) {
                                    final StrapProtein ppRow[]=getProteinsInRow(i+rh_draggedSrcRow), p0=ppRow.length>0?ppRow[0]:null;
                                    if (p0!=null) { p0.addResidueSelection(rh_resSel_dragged); rh_resSel_dragged.setProtein(p0); }
                                }
                            }
                            rowHeader.repaint();
                            threadPR(PR_ONLY_CHANGES).run();
                            hSB.clearImage();
                            ChDelay.repaint(hSB,111);
                            rh_lastDraggedDestRow=rh_draggedDestRow;
                        }
                    }
                } else rh_doingDnd=false;
                if (id==MOUSE_MOVED  || id==MOUSE_RELEASED  || dndStarted ) {
                    rh_ppDrag=null;
                    if (rh_draggedProtein2row!=null) {
                        rh_draggedProtein2row=null;
                        if (!dndStarted) moveLines(rh_draggedSrcRow,mRow,this.rh_countDraggedRows);
                        threadPR(PR_ONLY_CHANGES).run();
                        rowHeader.repaint();
                    }
                }
                setTipTiming(id==MOUSE_EXITED || x>lineH ? -1 : 30,0);
                if (id==MOUSE_MOVED) {
                    if (pRow!=null) rowHeader.setCursor(x<lineH ? ChIcon.getCursor("Refs",rowHeader) : getCursr('H'));
                    rh_draggedSrcRow=-1;
                }
                if (id==MOUSE_CLICKED && pRow!=null) {
                    rowHeader.requestFocus();
                    if (x<lineH && !ctrl && !shift) {
                        rh_info_Show(pRow, xScreen(mev), yScreen(mev));
                        return null;
                    }
                    final Object p0=pRow;
                    if ( !middle && !isPop) {
                        if (ctrl) {
                            if (vSEL.contains(p0)) vSEL.remove(p0); else vSEL.add(p0);
                        } else if (shift) {
                            for(int i=mini(mRow,rh_clickedRow); i<=maxi(mRow,rh_clickedRow); i++)
                                if (i>=0 && i<ppVis.length) vSEL.add(ppVis[i]);
                        }
                        else { vSEL.clear(); vSEL.add(p0);}
                        final long durationRH[]=((ChPanel)rowHeader).getWhenPainted();
                        ChDelay.repaint(rowHeader,4*durationRH[DURATION],durationRH);
                    }
                    if (align.getTree().getParent().getWidth()>10) scrollToVis(p0,align.getTree());
                    rh_clickedRow=mRow;
                    highlightProteins((int[])null,(Color[])null);
                }
                if (!setSelection && rh_draggedProtein2row==null) for(StrapProtein p : ppVis)  p.removeResidueSelection(rh_resSel_dragged);
            }
            if (id==MOUSE_DRAGGED) {
                if (!isMouseDrag) {
                    isMouseDrag=true;
                    mouseDragFromX=x;
                    mouseDragFromY=y;
                }
            } else isMouseDrag=false;
            prevMouseX=x;
            prevMouseY=y;
        }
        if (vSEL_mc!=modiCount(vSEL)) new StrapEvent(StrapView.class,StrapEvent.OBJECTS_SELECTED).setParameter(Protein.class).run();
        if (id==MOUSE_RELEASED) {
            this.isDragged=0;
            isResizing=false;
        }
        return e;
    }

    public void actionPerformed(ActionEvent ev) {
        if (rowHeader==null) {
            StrapView v=a().getAlignmentPanelWithFocus();
            if (v!=null)actionPerformed(v,ev);
        } else actionPerformed(this,ev);
    }
    private static void actionPerformed(final StrapView view,  final ActionEvent ev) {
        final Object q=ev.getSource();
        final String cmd=ev.getActionCommand();
        final boolean isSelected=isSelctd(q);
        final HScrollBar hSB=view.hSB;
        final int bi=idxOf(q,_button);
        final int modi=ev.getModifiers();
        final int iAa=view.getIndexOfAminoAcidAtCursor();
        final Rectangle rect=view.getRectRubberBand();
        final int rectW=rect==null?-1:rect.x;
        final int rectE=rect==null?-1:rectW+rect.width;
        final boolean ctrl=isCtrl(ev);
        final StrapAlign align=a();
        final StrapProtein
            ppSel[]=align.getSelectedProteins(),
            ppAll[]=view.getVisibleProteins(),
            pCursor=view.getProteinAtCursor(),
            ppSelOrAll[]=ppSel.length>0 ? ppSel : ppAll,
            ppSelOrCursor[]= ppSel.length>0 ? ppSel :  pCursor!=null ? new StrapProtein[]{pCursor} : ppAll.length==1 ? ppAll :StrapProtein.NONE;
        if (bi==BUT_RESET_ALIGN) {
            button(BUT_RESET_ALIGN).t("","",dim(0,0)).revalidate();
            for(StrapProtein p : ppAll) {
                final int[] gaps=gcp(button(BUT_RESET_ALIGN),p,int[].class);
                if (gaps!=null) p.setResidueGap(gaps);
            }
            StrapAlign.dispatchLater(StrapEvent.ALIGNMENT_CHANGED,1);
        }
        if (q==butListHet) {
StrapProteinUtils.showChildObjects(HeteroCompound.class, rh_info_P);
rh_info_W.setVisible(false);
        }
        if (bi==BUT_PDB) showHelp(SimilarPdb.class);
        if (bi==BUT_INFER_3D) SimilarPdb.showTablePDB(ppSelOrAll,false);
        if (bi==BUT_INFER_3D_ALL) SimilarPdb.showTablePDB(ppSelOrAll,true);
        if (bi==BUT_SIMIL_SEARCH) SimilarPdb.startThreads(ppSelOrAll, true, null);
        if (cmd=="CLIPBOARD" && rectW<rectE) {
            final StrapProtein ppRegion[]=getProteinsInRectangle(view);
            final byte[][] gapped=StrapAlign.getGappedSequences(ppRegion,rectW,rectE,'-');
            final ByteArray sb=byteArr(SB_TIP);
            if (gapped!=null) for(byte[] gg:gapped) sb.append(gg,0,MAX_INT).append('\n');
            toClipbd(sb.newString());
        }
        if (cmd=="NEW_ANNO" || cmd=="NEW_ANNO_R") {
            final StrapProtein pp[];
            final int colFrom, colTo;
            if (cmd=="NEW_ANNO") {
                pp=new StrapProtein[]{view.getProtein(view.mousePressedRow)};
                if (pp[0]==null) return;
                colFrom=view.mousePressedCol;
                colTo=colFrom+1;
            } else {
                pp=getProteinsInRectangle(view);
                putln("pp=",pp);
                colFrom=rectW;
                colTo=rectE;
            }
            if (pp.length>0) {
                final List v=StrapAlign.getSelectedObjectsV();
                v.clear();
                for(StrapProtein p : pp) {
                    ResidueAnnotation a=new ResidueAnnotation(p);
                    final int idx0=p.getFirstResidueIndex();
                    a.setValue(ResidueAnnotation.GROUP,"Rectangle");
                    a.setValue(ResidueAnnotation.POS,(idx0+p.column2nextIndex(colFrom)+1)+"-"+(idx0+p.column2thisOrPreviousIndex(colTo-1)+1));
                    a.setColor(Color.MAGENTA);
                    a.setStyle(VisibleIn123.FULL_BACKGROUND);
                    a.setValue("Note","filled in rectangular region");
                    a.setValue(ResidueAnnotation.NAME,"rectangle");
                    p.addResidueSelection(a);
                    v.add(a);
                    if (cmd=="NEW_ANNO") align.editAnnotation(a);
                }
                new StrapEvent(view,StrapEvent.RESIDUE_SELECTION_CHANGED).run();
                new StrapEvent(view,StrapEvent.OBJECTS_SELECTED).run();
                align.SELECTED_OBJECTS.selectionMenu('A').COLOR_BUTTON.shw();
            }
        }
        if (cmd=="NARROW") {
            final StrapProtein pp[]=getProteinsInRectangle(view);
            for(StrapProtein p : pp) {
                pcp(cmd, new int[]{p.column2nextIndex(rectW), p.column2thisOrPreviousIndex(rectE-1)+1}, p);
            }
            DialogRenameProteins.instance(StrapView.class).rename(DialogRenameProteins.INFO_CUT_TERMINI,pp,cmd);
        }
        if (cmd=="U" || cmd=="L") {
            for(StrapProtein p:getProteinsInRectangle(view))
                for(int col=rectW; col<rectE; col++)
                    p.setResidueTypeToUpperOrLower(p.column2index(col),cmd=="U");
            new StrapEvent(view,StrapEvent.RESIDUE_TYPES_CHANGED).run();
        }
        if (cmd=="A") {
            final Object d=align.addDialog(DialogAlign.class,false);
            if (d instanceof ChRunnable) ((ChRunnable)d).run(DialogAlign.RADIO_REGION,null);
        }
        if ((cmd=="DG" || cmd=="IG") &&  rectW<rectE ) {
            for(StrapProtein p:getProteinsInRectangle(view)) {
                if (cmd=="DG") {
                    final int col0=p.column2thisOrPreviousIndex(rectW);
                    for(int iA=p.column2nextIndex(rectE-1); iA>col0; iA--) {
                        final int col=p.getResidueColumn(iA), col1=p.getResidueColumn(iA-1);
                        if (col>0) p.setResidueGap(iA, col1<0 || col1+2>rectW ?  col-rectE :  rectW-col1 +  maxi(0, col-rectE)-1);
                    }
                } else {
                    final int iA=p.column2nextIndex(rectW);
                    if (iA>=0) p.setResidueGap(iA, rectE-rectW+p.getResidueGap(iA));
                }
            }
            new StrapEvent(view,StrapEvent.ALIGNMENT_CHANGED).run();
        }
        if (cmd=="W") {
            final StrapView li=view;
            if (fWord==null) {
                fWord=new ChFrame("Export to Word-Processor");
                html_cbJalviewSS=toggl("Include PDB-secondary structure as feature");
                html_cbFileExt=toggl("Include file name suffices like \".pdb\"");
                html_cbPdbNumbering=toggl("Use the residue numbers in pdb-files if available");
                html_cbUL=toggl("Underline all annotations").s(true);
                html_cbProteinIcons=new ChButton[]{toggl("Show protein icons in Web-browser").s(true),toggl("Show protein icons in Word-Processor")};
                (html_iconSize=new ChTextField(" 20 ")).setEnabled(false);
                html_cb_imgAttributes=toggl("Set size for icon images ").s(true);
                html_cbFeatures=toggl("Include sequence features such as modified residues").s(true);
                tfWordCharsPerLine=new ChTextField("40").cols(5,true,true);
                html_butJalview=new ChButton("JALVIEW").t("<b>Export to<br>Jalview ...</b>").li(li);
                if (javaVsn()>14) {
                    requestImage("http://www.jalview.org/jalview.gif", html_butJalview);
                    pcp(KEY_REQUEST_IMAGE_MAX_WIDTH,intObjct(300),html_butJalview);
                    pcp(KEY_REQUEST_IMAGE_OPTIONS, longObjct(REQUEST_IMAGE_PACK|REQUEST_IMAGE_REMOVE_TEXT),html_butJalview);
                }
                html_tfFile=new ChTextField(" ");
                html_comboTarget=new ChCombo("Web-browser","Word-processor","clustalW-format","MSF-format");
                final JComponent
                    html_pan_iconTextProcessor=hBL(html_cb_imgAttributes.actionCollapse(false,html_iconSize).cb(), html_iconSize),
                    pAdv=vBpnl(html_butJalview, " ",align.b(CLASS_Texshade)),
                    advanced=vB(
                                hB("A Web-link for the alignment can be included into<ul>"+
                                   "<li> E-mails</li><li>Office documents</li><li> Web-pages</li></ul>"+
                                   "Advantage: A Web link is much shorter and easier to include into documents than the alignment itself."),
                                align.b(CLASS_DialogPublishAlignment),
                                "#JS",
                                hB("#TBem Options for Advanced Users", pAdv),vB(" "," "," "," "," "," ")
                                ),
                    bCustom=Customize.newButton(HtmlUtil.customize('A'),SequenceFeatures.customizeCSS_IE(),
                                                !Insecure.EXEC_ALLOWED?null: Customize.customize(Customize.webBrowser),
                                                !Insecure.EXEC_ALLOWED?null: Customize.customize(Customize.htmlEditors)
                                                ).t("").options(ChButton.ICON_SIZE),
                    msgAli=vBhB(
                                newLabelRectRegion(),
                                hB(pnl("<h2>Send  Alignment to </h2>",html_comboTarget),"#", bCustom),
                                html_cbFileExt.cb(),
                                html_cbPdbNumbering.cb(),
                                hBL("Fold lines after how many characters ?",pnl(tfWordCharsPerLine)),
                                " ",
                                vBhB("#TBem Options for Web-browser and Word-processor",
                                     html_cbUL.cb(),
                                     html_cbFeatures.cb(),
                                     html_cbProteinIcons[0].cb(),
                                     html_cbProteinIcons[1].actionCollapse(false,html_pan_iconTextProcessor).actionPack().cb(),
                                     html_pan_iconTextProcessor
                                     ),
                                " ",
                                pnl( enlargeFont(new ChButton("EXPORT_HTML").li(li).t(ChButton.GO).tt("Create HTML and send it to Browser.<br>Hold CTRL to see details"),1.6f)),
                                " ",
                                html_tfFile
                                ),
                    togCollapse=toggl("+",ChButton.NO_MARGIN).actionCollapse(false,advanced).actionCollapse(true,msgAli).actionPack(),
                    msg=vBhB(
                             msgAli,
                             hBL("Advanced options ",togCollapse),
                             advanced
                             );
                fWord.ad(msg);
            }
            fWord.shw(ChFrame.CENTER|ChFrame.PACK);
        }
        final boolean isJalview=cmd=="JALVIEW";
        if (isJalview || cmd=="EXPORT_HTML") {
            if (isJalview) {
                StrapProtein[] pp=align.getSelectedProteinsInVisibleOrder();
                if (pp.length==0) pp=view.getVisibleProteins();
                Object cb=null;
                for(StrapProtein p:pp) if (p.getResidueSecStrType()!=null) cb=html_cbJalviewSS.cb();
                final String s=getHlp(Export2Jalview.class);
                if (ChMsg.yesNo(vBpnl(s," ",cb,hBL("<br>Viewing the alignment in Jalview?")))) {
                    final int mode= html_cbJalviewSS.s() ? Export2Jalview.SECSTRU : 0;
                    new Export2Jalview().launchJalview(pp,mode);
                }
            } else {
                final StrapProtein[] pp=getProteinsInRectangleForAlignment(2);
                final int iTarget=html_comboTarget.i();
                final long iType;
                final String ext;
                final StringBuffer sb=new StringBuffer();
                int size=StrapProtein.getMaxColumn(pp)*pp.length;
                switch(iTarget) {
                case 2: iType=ExportAlignment.CLUSTALW;   ext=".clustalW"; break;
                case 3: iType=ExportAlignment.MSF;  ext=".msf"; break;
                default: iType=ExportAlignment.HTML; ext=".html"; sb.append("<!DOCTYPE HTML>\n<!-- saved from url=(0023)http://3d-alignment.eu/ -->\n"); size*=7;
                }
                sb.ensureCapacity(size);
                ExportAlignment.setIconSize( html_cbProteinIcons[1].s() && html_cb_imgAttributes.s()  ? atoi(html_iconSize) : 0);
                ExportAlignment.export(pp,
                                       iType | ExportAlignment.HTML_HEAD_BODY |
                                       (view.isShadingSecStru() ? ExportAlignment.SECONDARY_STRUCTURE:0) |
                                       ((html_cbFileExt.s())?ExportAlignment.FILE_SUFFIX:0) |
                                       ((html_cbPdbNumbering.s())?ExportAlignment.PDB_RESIDUE_NUMBER:0) |
                                       (html_cbUL.s() ? ExportAlignment.UNDERLINE_ANNOTATIONS : 0) |
                                       (iTarget<2 ? (html_cbProteinIcons[iTarget].s() ? ExportAlignment.PROTEIN_ICONS : 0) : 0) |
                                       (iTarget==0 ? ExportAlignment.CSS_BROWSER : 0) |
                                       (html_cbFeatures.s() ? ExportAlignment.SEQUENCE_FEATURES : 0),
                                       rectW<rectE ? new Range( rectW, rectE) : null,
                                       maxi(5,atoi(tfWordCharsPerLine)),
                                       null,sliderSimilarity().getValue()*.01f,sb);
                ExportAlignment.setIconSize(0);
                final File fOut=file(STRAPOUT+"/align"+ ++html_count+ext);
                wrte(fOut,sb);
                html_tfFile.t(" "+fOut).tools().underlineRefs(0);
                if (iTarget<2) {
                    if (!Insecure.EXEC_ALLOWED || true) {
                        if (iTarget==0) visitURL(url(fOut),modi);
                        else {
                            final JComponent msg=vBhB("Two options to open the alignment in the word processor (e.g. MS-Word):"+
                                                      "<ul><li>Drag the following file path with the mouse onto the word processor-icon. Then drop the file-path over the word-icon.</li>"+
                                                      "<li>Or open the file in the word processor</li></ul>", fOut);
                            ChMsg.msgDialog(msg);
                        }
                    } else {
                        htmlTidy(fOut);
                        final ChExec ex=new ChExec(0).setCommandsAndArgument(getCustomSettings(iTarget==1 ? Customize.htmlEditors : Customize.webBrowser),fOut);
                        if (ctrl) ex.log(Customize.customize(Customize.htmlEditors));
                        new Thread(ex).start();
                    }
                } else new AsciiView(fOut).showInFrame("",0);
            }
        }
        if (cmd=="WS") Strap.instance().run("WS",isSelected?"false":"true");
        if (bi==TOG_PLOT_ENTIRE_PANEL) StrapAlign.dispatchLater(StrapEvent.VALUE_OF_RESIDUE_CHANGED,111);
        if (bi==BUT_ALIGN) {
            if (gcp("C",button(BUT_ALIGN))==null) {
                for(Protein p:align.getProteins()) pcp(button(BUT_RESET_ALIGN),null,p);
                final StrapProtein[] pp=ppSel.length>1?ppSel:ppAll;
                if (sze(pp)>1) {
                    stopAllAlignments();
                    setButAlign(true);
                    final SequenceAligner aligner=StrapProteinUtils.alignerInstance();
                    lastAligner=aligner;
                    for(StrapProtein p:pp) pcp(button(BUT_RESET_ALIGN),p.getResidueGap().clone(),p);
                    final long mode=(modi&SHIFT_MASK)!=0 ? StrapProteinUtils.ALIGN_CHANGE_ORDER:0L;
                    newThrd(
                            StrapProteinUtils.threadAlignProteins(aligner,Arrays.asList(pp), CAN_BE_STOPPED, mode),
                            ChDelay.newThread(view,"AFTER_ALIGN", null)
                            ).start();
                }
            } else stopAllAlignments();
        }
        if (cmd=="FONT") {
            if (fFont==null) {
                final ChJList l=new ChJList(getAllFonts(),ChJTable.DEFAULT_RENDERER|ChJTable.ICON_ROW_HEIGHT).li(view);
                pcp(KEY_ACTION_COMMAND,ChRenderer.CMD_FOND_NAME,l);
                l.setSelO("Monospaced");
                Object s="change font size with ctrl-\"+\". You may need to use<br>number block of keyboard. Alternatively ctrl-mouse-wheel.";
                fFont=new ChFrame(cmd).ad(newCNSEW(new JScrollPane(l),"null",s,null,null));
            }
            fFont.shw(0);
        }
        if (cmd==ChRenderer.CMD_FOND_NAME) view.setFont(q.toString(),0,true);
        if (cmd=="CLOSE") { align.closeAlignmentPanel(view); }
        if (bi==BUT_FEATURE) {
            if (ctrl) StrapDAS.show(false);
            else SequenceFeatures.instance().showPanel();
        }
        if (q==choiceShading()) { // 5 2 3
            final int idx=choiceShading().i();
            view.shadeSecStru=view.showSolvAcc=false;
            view.iInfo=-1;
            ChDelay.runAfterMS(EDT,view.thread(T_ENABLED),111);
            final StrapProtein pp[]=view.getVisibleProteins();
            if (idx<5) {
                view.shadingColors=ShadingAA.getColors(idx);
                shadingColorsName=q.toString();
                view.isNucleotides=view.shadingColors==ShadingAA.ACTG;
            } else if (idx<5+2) {
                view.shadeSecStru=idx==5;
                view.showSolvAcc=idx==6;
            } else {
                view.iInfo=idx-5-2;
            }
            view.pCenter.repaint();
            view.hSB.clearImage().repaint();
            view.thread(T_CLEAR_SB).run();
            ChDelay.repaint(hSB,hSB.LAST[DURATION]*4);
            if (button(TOG_OUTLINE_HSB).s())  hSB.clearImage().repaint();
            ChDelay.revalidate(view.getPanel(),333, (long[])null);
            StrapAlign.dispatchLater(StrapEvent.AA_SHADING_CHANGED,111);
        }
        if (bi==TOG_MAXIMIZE) {
            align.rightCollapse(isSelected);
            view.getPanel().revalidate();
            if (isSelected) {
                view.reduceHeightSave=view.reduceHeight;
                view.setReduceHeight(0);
                align.TOG_MAXIMIZE_DIALOGS.s(false);
            } else view.setReduceHeight(view.reduceHeightSave);
            view.getPanel().revalidate();
        }
        if (bi==TOG_OUTLINE_HSB) {
            view.isShadingScrollbar=isSelected;
            hSB.clearImage().repaint();
        }
        if (bi==TOG_PROTECT_COL ) {
            final Rectangle r=view.getVisibleRowsAndCols();
            final int cc=r.x+r.width/2;
            setProtectFromColumn(get(cc, align.gap2col().narrow2wideColumn()), false);
        }
        if (bi==TOG_BRIGHTER || bi==TOG_SHADE_ONLY_SELECTED) { view.pCenter.repaint(); view.hSB.repaint(); }
        final boolean restoreC=bi==BUT_RESTORE_CURSOR|| equls("RESTORE",cmd,0);
        if (restoreC || bi==BUT_SAVE_CURSOR || equls("SAVE",cmd,0)) {
            final Object key=intObjct(modi+trailInt(cmd));
            if (restoreC) {
                final Object[] sc=view.mapSavedCursor.get(key);
                if (sc!=null) {
                    view.setCursorToResidue( (Protein)deref(sc[0]), atoi(sc[1]));
                    view.scrRectToVis((Rectangle)sc[2],false);
                }
                view.getAlignmentPane().requestFocus();
            } else {
                if (iAa>=0 && pCursor!=null) view.mapSavedCursor.put(key,new Object[]{weakRef(pCursor),intObjct(iAa),new Rectangle(view.getVisibleXYWH())});
                button(BUT_RESTORE_CURSOR).enabled(true);
            }
        }
        if (bi==BUT_WIRE) StrapAlign.newBackbone3D_View(ppSelOrCursor,0L);
        if (bi==BUT_SEQVISTA) {
            final UniqueList<File> v=new UniqueList(File.class);
            for(StrapProtein p : ppSelOrCursor)  if (p!=null && (p.countNucleotides()>0 || p.containsOnlyACTGNX())) v.add(p.getFile());
            Hyperrefs.toSeqvista(v.asArray());
        }
        if (bi==BUT_NUCL && sze(ppSelOrCursor)>0) {
            final UniqueList<StrapProtein> vGB=new UniqueList(StrapProtein.class) , vNoGB=new UniqueList(StrapProtein.class);
            EditDnaParentPane pane=null;
            for(StrapProtein p : ppSelOrCursor) {
                if (p!=null) (sze(p.getCDS())>0 && p.countNucleotides()==0 ? vGB : vNoGB).add(p);
            }
            if (sze(vGB)>0) {
                DialogGenbank gb=(DialogGenbank)align.addDialog(DialogGenbank.class,false);
                gb.setProteins(vGB.asArray());
                gb.run();
            }
            if (sze(vNoGB)>2) ChUtils.error("Please select only one nucleotide sequence");
            else if (sze(vNoGB)>0) {
                for(StrapProtein p : vNoGB.asArray()) {
                    if (pane==null)  pane=(EditDnaParentPane) align.addDialog(EditDnaParentPane.class,false);
                    if (p.countNucleotides()>0) pane.addEditor(p,true); else pane.setOrientation(p);
                }
            }
        }
        if (q==sliderSimilarity()) {
            StrapAlign.dispatchLater(StrapEvent.AA_SHADING_CHANGED,333);
            view.threadPR(PR_ONLY_CHANGES).run();
            button(TOG_OUTLINE_HSB).s(view.isShadingScrollbar=true);
            hSB.clearImage().repaint();
        }
    }
    /* <<< ActionEvent <<< */
    /* ---------------------------------------- */
    /* >>> Frequency >>> */
    private boolean freqValid[];
    private byte[][] freqGapped;
    private StrapProtein ppShading[]=StrapProtein.NONE;
    protected StrapProtein[] ppShading() { return ppShading;}
    protected short[][] computeFrequenciesAtColumn( int fromColumn0, int toColumn0) {
        final Object[] cached=cached();
        final int countColumns=countColumns();
        short[][] frequencyOfAA=(short[][])cached[MCSV_FREQ];
        final int mc=Protein.MODI_COUNTS_ALL[MCA_ALIGNMENT]+modiCount(button(TOG_SHADE_ONLY_SELECTED))+modiCount(StrapAlign.getSelectedObjectsV());
        if (countColumns>=sze(frequencyOfAA) || frequencies_modificationCount!=mc) {
            if (frequencyOfAA!=null && countColumns<=frequencyOfAA.length) Arrays.fill(frequencyOfAA,null);
            else cached[MCSV_FREQ]=frequencyOfAA=new short[countColumns+9][];
            frequencies_modificationCount=mc;
            if (freqValid!=null) Arrays.fill(freqValid,false);
        }
        if (countColumns>=sze(freqValid)) freqValid=new boolean[countColumns+99];
        StrapProtein[] vP=this.ppShading=button(TOG_SHADE_ONLY_SELECTED).s() ? a().getSelectedProteins() : getVisibleProteins();
        if (vP.length==0) vP=this.ppShading=getVisibleProteins();
        final int fromColumn=maxi(0,fromColumn0);
        final int toColumn=maxi(0,mini(countColumns,toColumn0,frequencyOfAA.length));
        boolean isValid=true;
        for(int iCol=toColumn; --iCol>=fromColumn;)  if (!freqValid[iCol]) { isValid=false; break;}
        if (!isValid) {
            if (sze(this.freqGapped)!=vP.length) this.freqGapped=new byte[vP.length][];
            for(int i=vP.length; --i>=0;) freqGapped[i]=vP[i]==null ? NO_BYTE : vP[i].getGappedSequences();
            final byte[][] ggg=this.freqGapped;
            for(int iCol=toColumn; --iCol>=fromColumn;) {
                if (freqValid[iCol]) continue;
                freqValid[iCol]=true;
                short ff[]=frequencyOfAA[iCol];
                if (ff==null) frequencyOfAA[iCol]=ff=new short['Z'-'A'+1]; else Arrays.fill(ff,(short)0);
                for(int iP=vP.length; --iP>=0;) {
                    final byte gs[]=ggg[iP];
                    final int b=iCol<gs.length ? (gs[iCol]|32) : 0;
                    if (b>='a' && b<='z') ff[b-'a']++;
                }
            }
        }
        return frequencyOfAA;
    }
    /* <<< Frequency <<< */
    /* ---------------------------------------- */
    /* >>> Row header >>> */
    private float rh_adjust;
    private int[] rh_draggedProtein2row;
    private static int[] rh_indices;
    private int rh_draggedSrcRow=-1, rh_draggedDestRow, rh_lastDraggedDestRow, rh_clickedRow=0, rh_countDraggedRows=1;
    private StrapProtein[] rh_ppDrag;
    private final AbstractResidueSelection rh_resSel_dragged=new AbstractResidueSelection();
    public JComponent getRowHeader() { return rowHeader;}
    {
        rh_resSel_dragged.setStyle(VisibleIn123.FULL_BACKGROUND);
        rh_resSel_dragged.setColor(C(0xFFffFF,99));
        rh_resSel_dragged.setVisibleWhere(VisibleIn123.SEQUENCE);
    }
    private static Window rh_info_W;
    private static ChTextArea rh_info_TA;
    private static StrapProtein rh_info_P;
    private static ChButton butListHet;
    private static void rh_info_Add(String id, StringBuffer sb, StringBuffer sbStru) {
        if (id==null || id.indexOf(':')<0 || sb.indexOf(id)>=0 || sbStru.indexOf(id)>=0) return;
        if (id.startsWith("PDB:")) {
            final String pdb=withoutLastComponent(id,'_');
            if (sbStru.indexOf(pdb)<0) sbStru.append(pdb).append(' ');
        } else sb.append(id).append(' ');
    }
    protected static void rh_info_Show(final StrapProtein p0, int xScreen, int yScreen) {
        final StrapView view=a().getAlignmentPanelWithFocus();
        final StrapProtein p=p0!=null ? p0 : rh_info_P;
        rh_info_P=p;
        if (p==null) return;
        final StringBuffer sb=new StringBuffer(333);
        if (rh_info_W==null) {
            butListHet=new ChButton().t("List child objects such","as DNA/RNA and ligands").li(view);
            pcp(KEY_PREF_SIZE_IF_DISABLED,dim(0,0),butListHet);
            final JComponent pSouth=pnl(butListHet, WHITE, WATCH_MOVIE+MOVIE_Export_Proteins+" "+MOVIE_Load_Proteins+" "+MOVIE_Context_Menu),
                pnl=newCNSEW(rh_info_TA=new ChTextArea(""),null,pSouth,null, "#RBB");
            (rh_info_W=new Window(a().getFrame())).add(pnl);
            setProcessEvt((ChProcessEvent)view, rh_info_TA);
            rh_info_TA.setCursor(getCursr('D'));
            setDragMovesWindow(rh_info_TA,rh_info_W);
            setDragMovesWindow(pSouth,rh_info_W);
            setDragMovesWindow(butListHet,rh_info_W);
        }
        final StringBuffer sbSeq=new StringBuffer(), sbRef=new StringBuffer(), sbStru=new StringBuffer();
        rh_info_Add(p.getPdbID(),sbSeq,sbStru);
        rh_info_Add(p.getAccessionID(),sbSeq,sbStru);
        for(String s : p.getSequenceRefs()) rh_info_Add(s,sbSeq,sbStru);
        for(String s : p.getDatabaseRefs()) rh_info_Add(s,sbSeq,sbStru);
        for(String ec : p.getEC()) if (ec!=null && ec.indexOf('-')<0) sbRef.append(' ').append(ec).append(' ');
        final String uniprot=p.getUniprotID();
        if (uniprot!=null) {
            final String up=uniprot.substring(uniprot.indexOf(':')+1);
            sbRef.append("DASTY:").append(up).append(' '); /* Rafael Jimenez  rajido@gmail.com */
            sbRef.append(" ARCHSCHEMAu:").append(up).append(' ');
        }
        for(String s:p.getSequenceRefs()) {
            final String pf=delPfx("PFAM:",s);
            if(pf!=s) sbRef.append(" ARCHSCHEMApf:").append(pf).append(' ');
        }
        int isFirst=0;
        sb.append("  ").append(filepathUnixStyle(p.getFile())).append('\n');
        for(StringBuffer s:new StringBuffer[]{sbStru,sbSeq,sbRef})  if (sze(s)>0) sb.append(isFirst++==0 ? "  " : "\n").append(s);
        butListHet.setEnabled(sze(p.run(PROVIDE_JTREE_CHILDS,null))>0);
        rh_info_TA.t(sb);
        rh_info_TA.tools().underlineRefs(0);
        ChRenderer.setSmallButtonXc(0, rh_info_TA);
        if (xScreen>=0 && yScreen>=0) rh_info_W.setLocation(xScreen-4,yScreen-4);
        rh_info_W.setVisible(true);
        butListHet.revalidate();
        rh_info_W.pack();
    }
    private int[] draggedProtein2row() { return rh_draggedProtein2row;}
    /* <<< Row header <<< */
    /* ---------------------------------------- */
    /* >>> Plotting >>> */
    private static UniqueList _vPl;
    protected static UniqueList<Object> getPlottersV() { return _vPl==null ? _vPl=new UniqueList(Object.class) : _vPl;}
    private final WeakHashMap<Object,double[][]> mapMiMa=new WeakHashMap();
    private void plot(Graphics g) { final Rectangle vis=getVisibleXYWH(); plot(g,pCenter,charB().width, vis.y,vis.y+vis.height); }
    protected void plot(Graphics g, Component component, double charWf, int yMin,int yMax) {
        if (sze(getPlottersV())==0) return;
        final Object[] cached=cached();
        final StrapAlign align=a();
        final int charW=(int)charWf, small=maxi(1,charW>>4), charH=charB().height, xMin, xMax;
        {
            final Shape c=g.getClip();
            final Rectangle  r= c!=null ? c.getBounds() : null;
            if (r!=null) { xMin=r.x; xMax=xMin+r.width;}
            else { xMin=0; xMax=MAX_INT;}
        }
        final short[][] ccc=component==pCenter ? (short[][])cached[MCSV_CHARACTERS] : null;
        for(Object hasValues : getPlottersV().asArray()) {
            final ValueOfResidue vor;
            final ValueOfAlignPosition vap;
            {
                final ValueOfAlignPosition vap_= hasValues instanceof ValueOfAlignPosition  ? (ValueOfAlignPosition) hasValues : null;
                final ValueOfResidue       vor_= hasValues instanceof ValueOfResidue        ? (ValueOfResidue) hasValues  : null;
                vap= vap_!=null && vap_.getValues()!=null  && vap_.getProteins()!=null ? vap_ : null;
                vor= vap==null && vor_!=null && vor_.getValues()!=null  && vor_.getProtein() !=null ? vor_ : null;
            }
            assert vap==null || vor==null;
            final double vv[]=vor!=null ? vor.getValues() : vap!=null ? vap.getValues() : null;
            if (vv==null) continue;
            int ymin=yMin,ymax=yMax;
            final StrapProtein p=vor!=null ? (StrapProtein)vor.getProtein() : null;
            if (component==pCenter) { /* plot in separate row */
                final int row=!button(TOG_PLOT_ENTIRE_PANEL).s() && hasValues instanceof HasProtein ? align.findRow(((HasProtein)hasValues).getProtein()) : -1;
                if (row>=0) { ymin=row*(lineSkip+charH);  ymax=ymin+lineSkip+charH; }
            }
            final Color color=hasValues instanceof Colored ? ((Colored)hasValues).getColor() : null;
            g.setColor(color!=null ? color : PINK);
            final int vvLength=mini(vor!=null ? p.countResidues(): countColumns(),vv.length);
            double[][]vv_mima=mapMiMa.get(hasValues);
            final double[] mima;
            final int modi=modiCount(vap!=null ? vap : vor);
            if (vv_mima==null || vv_mima[0]!=vv || modi!=(int)vv_mima[1][2]) {
                mapMiMa.put(hasValues,  vv_mima=new double[][]{vv,mima=new double[]{Double.MAX_VALUE,Double.MIN_VALUE,modi}});
                for(int i=vvLength; --i>=0;) {
                    if (vv[i]<mima[0]) mima[0]=vv[i];
                    if (vv[i]>mima[1]) mima[1]=vv[i];
                }
            } else mima=vv_mima[1];
            final double mi=mima[0], ma=mima[1],  h_ma_mi= ma==mi ? 1 : (ymax-ymin-4)/(ma-mi);
            final int toCol=(int)mini((int)((xMax)/charWf+1),countColumns()+1);
            for(int col=maxi(0,(int)(xMin/charWf-1)), lastY=-1; col<toCol; col++) {
                final int x=(int)(col*charWf);
                final int iA=vap!=null ? col : p.column2index(col);
                final double v=iA>=0 && iA<vvLength ? vv[iA] : Float.NaN;
                if (Double.isNaN(v)) { lastY=-1; continue;}
                final int y= ymax-(int) ((v-mi)*h_ma_mi)-2;
                if (y>0) {
                    if (lastY>0) g.drawLine(x+small, y, x-small,lastY);
                    g.fillRect(x+small,y-1,charW-small*2,3);
                }
                lastY=y;
                if (component==pCenter) markDirtyRect(true, ccc,(y-charH/2)/charH, col,1,2);
            }
        }
    }
    /* <<< Plotting <<< */
    /* ---------------------------------------- */
    /* >>>  Bar Chart >>> */
    private double[] vop_vv;
    private Object vopClass, vop_sharedInstance;
    private StrapProtein vopOtherProtein;
    private int vopFromColumn,vopToColumn;
    protected void setBarchart(Object c,Object sharedInstance,int fromCol, int toCol, StrapProtein p) {
        vopClass=c;
        vop_sharedInstance=sharedInstance;
        vopOtherProtein=p;
        vopFromColumn=fromCol;
        vopToColumn=toCol;
        rowHeader.repaint();
    }
    private void barChart(final Object clazz, Object sharedInstance, final Graphics gr, final StrapProtein pp[], final int rowFrom, final int rowTo) {
        if (clazz==null) return;
        final Rectangle cb=charB();
        final int w=rowHeader.getWidth();
        double vv[]=this.vop_vv;
        if (vv==null || vv.length!=pp.length) vv=this.vop_vv=new double[pp.length];
        double max=Double.MIN_VALUE,min=Double.MAX_VALUE;
        for(int i=rowTo; --i>=rowFrom;) {
            final double v=vv[i]=value(clazz,sharedInstance,pp[i],  vopOtherProtein );
            if (min>v)min=v;
            if (max<v)max=v;
        }
        if (max==Double.MIN_VALUE) return;
        gr.setColor(C(0xFFff00, 100));
        final Color COLOR_IS=C(0xffFFff,99);
        for(int i=rowTo; --i>=rowFrom;) {
            if (Double.isNaN(vv[i])) continue;
            final int width=(int)((vv[i]-min)/(max-min)*(w-9));
            final int y=i*(cb.height+lineSkip)+2;
            gr.fillRect( 2,y, width+3, cb.height-4 );
            final Color c=gr.getColor();
            gr.setColor(COLOR_IS);
            gr.drawString("(identical sequence)",strgWidth(gr, pp[i].getName())+2*ICON_HEIGHT,cb.y+y);
            gr.setColor(c);
        }
    }
    private double value(Object vopClass, Object sharedInstance, StrapProtein p,StrapProtein otherProt) {
        ValueOfProtein vop=p.getValueOfProtein();
        if (vop==null || vop.getClass()!=vopClass) p.setValueOfProtein(vop=HotswapUtil.instance(vopClass,ValueOfProtein.class,true));
        if (vop!=null) {
            vop.setProtein(p);
            setSharedParas(sharedInstance,vop);
            if (ProteinProteinValue.class.isAssignableFrom(vop.getClass())) {
                final ProteinProteinValue ppv=(ProteinProteinValue)vop;
                ppv.setOtherProtein(otherProt);
                ppv.setColumnRange(maxi(0,vopFromColumn),vopToColumn);
            }
            vop.compute();
            return vop.getValue();
        } else return Float.NaN;
    }
    /* <<< Bar Chart <<< */
    /* ---------------------------------------- */
    /* >>> Toolbar >>> */
    private static ChFrame fFont, fWord;
    private final static JComponent TOOLBARS[]={null,null};
    protected static JComponent[] getToolbars() {
        if (TOOLBARS[0]==null) {
            final Color bg=new javax.swing.table.JTableHeader().getBackground();
            button(BUT_EXPORT).setFont(new Font("Serif",Font.BOLD|Font.ITALIC,18));
            final boolean wm=UserProfile.isWebMode();
            final JComponent pButtons=pnl(
                                          bg,
                                          button(BUT_LAF).li(StrapAlign.li()),
                                          button(BUT_NUCL),
                                          button(BUT_SEQVISTA),
                                          button(BUT_PDB),
                                          button(BUT_WIRE),
                                          button(BUT_FEATURE),
                                          button(BUT_SAVE_CURSOR),
                                          button(BUT_RESTORE_CURSOR),
                                          button(BUT_ALIGN),
                                          button(BUT_RESET_ALIGN),
                                          button(BUT_EXPORT),
                                          button(TOG_MAXIMIZE),
                                          wm ? smallHelpBut(resourceAsURL(_CCS.replace('.','/')+"webMode.html")) : null
                                          );
            button(BUT_ALIGN).addMouseListener(instance().processEvent);
            final JComponent panel=newCNSEW(hB("#",bg,sliderSimilarity()), bg, null, pButtons, choiceShading());
            new CollapseComponent(panel,"N");
            panel.addMouseMotionListener(instance().processEvent);
            setButAlign(false);
            for(Component c1 : panel.getComponents()) {
                if (c1 instanceof JComponent) {
                    repaintOnMouseEnterAndExit(c1);
                    for(Component c2 : ((JComponent)c1).getComponents()) repaintOnMouseEnterAndExit(c2);
                }
            }
            TOOLBARS[0]=panel;
            TOOLBARS[1]=pButtons;
        }
        return TOOLBARS;
    }
    private static void setButAlign(boolean b) {
        button(BUT_ALIGN).t(b ? "Cancel" : "Align",b? "Computations" : "Proteins",dim(64,32));
        pcp("C", b ? "C":null,button(BUT_ALIGN));
        button(BUT_ALIGN).repaint();
    }

    /* <<<  <<< */
    /* ---------------------------------------- */
    /* >>>  >>> */

    public final int MCSV_RESSEL_RECT=1, MCSV_CHARACTERS=2, MCSV_FREQ=3;
    private long[] CACHE_MC=new long[10];
    private transient java.lang.ref.Reference<Object[]> rCached;
    protected Object[] cached() {
        Object[] r=rCached!=null ? rCached.get() : null;
        if (r==null) rCached=new java.lang.ref.SoftReference(r=new Object[10]);
        return r;
    }

    /* <<< Toolbar <<< */
    /* ---------------------------------------- */
    /* >>> ResidueSelection >>> */
    private MouseSelection mouseSelection;
    private static ResidueSelection resSelUnderM[];
    private static ResidueAnnotation featureUnderM;
    private ResidueSelection[] residueSelectionsInRectangle() {
        final List<ResidueSelection> v=new ArrayList();
        final Rectangle rect=getRectRubberBand();
        if (rect==null) return ResidueSelection.NONE;
        for(StrapProtein p : getProteinsInRectangle(this)) {
            final int iA0=p.column2nextIndex(maxi(0,rect.x)), iA1=p.column2thisOrPreviousIndex(rect.x+rect.width);
            if (iA0<0 || iA1<0) continue;
            for (ResidueSelection rs : p.getResidueSelectionsAt(VisibleIn123.ARROW_HEADS, iA0, iA1, VisibleIn123.ANYWHERE)) {
                if (rs!=null && !(rs instanceof VisibleIn123 && VisibleIn123.CURSOR==((VisibleIn123)rs).getStyle())) v.add(rs);
            }
        }
        return toArry(v,ResidueSelection.class);
    }
    protected ResidueSelection[] getResidueSelectionsAtXY(int x, int y,int where) {
        final Object[] cached=cached();
        final int row=y/(charB().height+lineSkip);
        final int col=x/charB().width;
        final StrapProtein protein=getProtein(row);
        if (protein==null) return ResidueSelection.NONE;
        final int MM[]=Protein.MODI_COUNTS_ALL;
        final long mc=MM[MCA_ALIGNMENT]+ MM[MC_RESIDUE_SELECTIONS_V]+MM[MC_SELECTED_AA]+((row)<<48)+(col<<32);
        ResidueSelection ss[]=(ResidueSelection[])cached[MCSV_RESSEL_RECT];
        if (ss==null || CACHE_MC[MCSV_RESSEL_RECT]!=mc) {
            CACHE_MC[MCSV_RESSEL_RECT]=mc;
            final int iA=protein.column2index(col);
            cached[MCSV_RESSEL_RECT]=ss=protein.getResidueSelectionsAt(VisibleIn123.ARROW_HEADS, iA,iA+1,where);
        }
        return  ss;
    }
    protected  MouseSelection mouseSelection() { if (mouseSelection==null) mouseSelection=new MouseSelection(); return mouseSelection;}
    /* <<< ResidueSelection <<< */
    /* ---------------------------------------- */
    /* >>> Menu >>> */
    JPopupMenu getPopup(boolean region){
        final StrapAlign align=StrapAlign.getInstance();
        if (region) {
            if (popupMenuRegion==null) {
                Object
                    bNarrow=new ChButton("NARROW").t("Hide sequence outside rectangle, keep sequence within rectangle ...").li(this).i(IC_SCISSOR),
                    oo[]={
                    "MENU FOR RECTANGULAR REGION",
                    null,null,
                    new Object[]{
                        "File",
                        new ChButton("CLIPBOARD").t("Copy to clipboard").li(this),
                        new ChButton("W").t("Export as alignment ...").li(this),
                        CLASS_DialogPhylogeneticTree,
                        "-",
                        bNarrow
                    },
                    new Object[]{
                        "Edit",
                        new ChButton("U").t("To upper case").li(this),
                        new ChButton("L").t("To lower case").li(this),
                        "-",
                        new ChButton("IG").li(this).t("Insert gaps"),
                        new ChButton("DG").li(this).t("Delete gaps"),
                        "-",
                        bNarrow
                    },
                    new Object[] {
                        "Align",
                        new ChButton("A").t("Align rectangular region ...").i(IC_ALIGN).li(this),
                        "Note: Sequences outside Rectangle will not be affected."
                    },
                    new Object[]{
                        "Apply",
                        CLASS_DialogPhylogeneticTree,
                        new ChButton("NEW_ANNO_R").t("Fill with new residue selections").i(IC_ANNO).li(this)
                    }
                };
                align.replaceClassByButton(oo,null);
                popupMenuRegion=jPopupMenu("rect-region-contextmenu", oo);
            }
            return popupMenuRegion;
        } else {
            if (popupMenu==null) {
                final ChButton butSR[]=new ChButton[6];
                for(int i=butSR.length; --i>=0;) {
                    butSR[i]=new ChButton((i<3?"SAVE":"RESTORE")+(i%3)).t((i<3?"Save":"Restore")+" cursor position "+(i%3+1)).li(this);
                }
                Object oo[]={
                    "MENU FOR THE ALIGNMENT WINDOW",
                    null,null,
                    new Object[]{
                        "New ",
                        StrapAlign.menuItem('n'),
                        StrapAlign.menuItem('N'),
                        "-",
                        new ChButton("NEW_ANNO").t("Create residue selection").i(IC_ANNO).li(this),
                    },
                    new Object[]{
                        "View",
                        "&Fc",CLASS_DialogHighlightPattern,
                        new ChButton("FONT").t("Font type").li(this),
                        align.CB_HIGHLIGHT_IDENTICAL_SEQUENCES,
                        button(TOG_BRIGHTER),
                        togAntialiasing(),
                        align.BUT_HIDE_GAPS
                    },
                    new Object[]{
                        "Sort",
                        CLASS_DialogSortProteins
                    },
                    new Object[]{
                        "Alignment",
                        CLASS_DialogAlign,
                        button(TOG_PROTECT_COL)
                    },
                    new Object[]{
                        "Cursor position",
                        butSR[0],butSR[1],butSR[2],"-",butSR[3],butSR[4],butSR[5]
                    },
                    "-",
                    new ChButton("CLOSE").t("Close alignment panel").i(IC_CLOSE).li(this)
                };
                align.replaceClassByButton(oo,null);
                popupMenu=jPopupMenu("Alignment-Contextmenu", oo);
            }
            return popupMenu;
        }
    }
    private JPopupMenu popupMenu,popupMenuRegion;
    /* <<< Menu <<< */
    /* ---------------------------------------- */
    /* >>> Cursor  >>> */
    private int cursorAA;
    private StrapProtein cursorP;
    public int getIndexOfAminoAcidAtCursor() { return cursorAA; }
    public int getColumnAtCursor() { StrapProtein p=getProteinAtCursor(); return p==null ? 0  : p.getResidueColumn(cursorAA);}
    public int getRowAtCursor() { return findRow(cursorP); }
    public StrapProtein getProteinAtCursor() { return cursorP;}
    private int lastRow=-1;
    private final AbstractResidueSelection resSel_cursor=new AbstractResidueSelection();
    {
        resSel_cursor.setColor(WHITE);
        resSel_cursor.setName(ResidueSelection.CURSOR);
        resSel_cursor.setImage(IC_CURSOR);
        resSel_cursor.setStyle(VisibleIn123.CURSOR);
        resSel_cursor.setVisibleWhere(VisibleIn123.SEQUENCE);
    }
    public boolean setCursorToColumn(Protein p,int iAa) { return setCursor(true, p, iAa); }
    public boolean setCursorToResidue(Protein p,int iAa) { return setCursor(false, p, iAa); }
    public Thread thread_setCursor(boolean asColumn, Protein p, int h0) {
        if (p==null) return null;
        return threadM("setCursor",this,new Object[]{boolObjct(asColumn), p,intObjct(h0)});
    }
    public boolean setCursor(boolean asColumn, Protein protein, final int h0) {
        final StrapProtein p=deref(protein,StrapProtein.class);
        if (p==null) return false;
        if (!isEDT()) {
            inEDT(thread_setCursor(asColumn, p,h0));
            return true;
        }
        final StrapProtein p_Last=this.cursorP;
        this.cursorP=p;
        final int lastCol=getColumnAtCursor();
        if (p_Last!=p) {
            if (p!=null) {
                Protein.incrementModificationCount(MC_CURSOR_PROTEIN,p);
                p.addResidueSelection(resSel_cursor);
            }
            if (p_Last!=null) {
                Protein.incrementModificationCount(MC_CURSOR_PROTEIN,p_Last);
                p_Last.removeResidueSelection(resSel_cursor);
            }
        }
        final int nAa=p!=null ? p.countResidues() : 0;
        if (nAa==0) return false;
        final int row=findRow(p), col;
        if (!asColumn) col=p.getResidueColumn(cursorAA= mini(nAa-1,maxi(0,h0)));
        else cursorAA=mini(nAa-1,maxi(0, p.column2nextIndex(col=maxi(0,h0))));
        resSel_cursor.setSelectedRange(cursorAA,cursorAA+1);
        if (rowHeader!=null) {
            final int H=charB().height+lineSkip;
            rowHeader.repaint(0,H*row,999,H);
            if (row!=lastRow) rowHeader.repaint(0,H*lastRow+H/4,H,H/2);
        }
        lastRow=row;
        final int evType=p!=p_Last ? StrapEvent.CURSOR_CHANGED_PROTEIN : col!=lastCol ? StrapEvent.CURSOR_MOVED_WITHIN_PROTEIN : -1;
        if (evType!=-1) new StrapEvent(this,evType).setParameters(new Object[]{p,intObjct(cursorAA),intObjct(row)}).run();
        return evType!=-1;
    }
    private StrapProtein pAtCursor_last;
    private int aminoAtCursor_last;
    private void maybeUpdateCursor() {
        final StrapProtein p=getProteinAtCursor();
        final int iA=getIndexOfAminoAcidAtCursor();
        if (pAtCursor_last!=p || aminoAtCursor_last!=iA) {
            pAtCursor_last=p;
            aminoAtCursor_last=iA;
            hSB.paintCursor();
            for(int i=0;i<2;i++) {
                if (vSB_yCursor>=0) vSB.repaint(0,vSB_yCursor-8, 16,16);
                if (i==0) vSB_yCursor=vSB_yCursor();
            }
            if (p!=null) {
                final String ch=p.getChainsAsString();
                ByteArray ba=p.cursorPosAsString(iA,true);
                if (sze(ch)>1) ba.append("  concatenated chains: ").append(ch);
                tLayout.setTitle(1,ba.newString());
                tLayout.setTitle(0,"Protein "+(1+getRowAtCursor()));
            }
        }
    }
    /* <<< Cursor <<< */
    /* ---------------------------------------- */
    /* >>> Proteins >>> */
    protected boolean addProteins( int row0, StrapProtein pp[], int options) {
        if (sze(pp)==0) return false;
        final int modi=getModificationCount();
        final int nRows=countRows();
        final List<StrapProtein[]> vRows=getList();
        final StrapProtein pp0[]=getProteins();
        for(int iP=0; iP<pp.length; iP++) {
            final StrapProtein p=pp[iP];
            if (p==null || !p.isInAlignment() || idxOf(p,pp,0,iP)>=0 ) continue;
            final int N=countRows();
            final int row=mini(row0+(N-nRows),N);
            if (cntains(p,pp0)) {
                if (row0>=0) {
                    final int inRow=findRow(p);
                    final StrapProtein[] v=getProteinsInRow(inRow);
                    if (v!=null) {
                        if (inRow!=row) {
                            vRows.remove(v);
                            vRows.add(row,v);
                        }
                        addProteinToRow(p,mini(sze(vRows)-1,row));
                    }
                }
            } else if (p.OBJECTS[StrapProtein.B_INTO_SAME_LINE]!=null && N>0) {
                addProteinToRow(p,N);
                p.OBJECTS[StrapProtein.B_INTO_SAME_LINE]=null;
                continue;
            } else {
                final StrapProtein[] p1=new StrapProtein[]{p};
                if (row0<0) vRows.add(p1); else  vRows.add(row,p1);
            }
        }
        return modi!=getModificationCount();
    }
    protected void oneLineManyProteins(StrapProtein pVisible,StrapProtein pp[]) {
        final StrapProtein proteins[]=pp.clone();
        final int iLine=findRow(pVisible);
        if (iLine>=0) {
            for( int i; (i=idxOf(pVisible,proteins))>=0;) proteins[i]=null;
            removeProteins(proteins);
            removeProteins(pVisible);
            final StrapProtein rowWithPVisible[]=getProteinsInRow(iLine);
            getList().remove(iLine);
            final StrapProtein[] ppRow=set1stElement(pVisible, pp,0,StrapProtein.class);
            getList().set(iLine,ppRow);
            for(int i=rowWithPVisible.length; --i>=0;) addProtein(iLine,rowWithPVisible[i]);
            removeEmptyRows();
            setCursorToResidue(getProteinAtCursor(), cursorAA);
            StrapAlign.dispatchLater(StrapEvent.PROTEINS_HIDDEN,111);
        }
    }
    private int update_isInAlignment_mc=-1;
    private void update_isInAlignment() {
        List<StrapProtein> vAdd=null, vRM=null;
        final StrapProtein pp[]=getProteins();
        for(StrapProtein p : a().getProteins()) {
            if (p.isInAlignment()) {
                if (!cntains(p,pp)) adUniq(p, vAdd==null ? vAdd=new ArrayList() : vAdd);
            } else {
                if (cntains(p,pp)) adUniq(p, vRM==null ? vRM=new ArrayList() : vRM);
            }
        }
        if (vRM!=null) { removeProteins(toArry(vRM, StrapProtein.class)); removeEmptyRows();}
        if (vAdd!=null) addProteins(-1,toArry(vAdd,StrapProtein.class), 0);
        final int mc=getModificationCount();
        if (update_isInAlignment_mc!=mc) {
            update_isInAlignment_mc=mc;
            final int nRows=countRows();
            if (nRows>10) setReduceHeight(reduceHeight-charB().height);
            pCenter.invalidate();
            ChDelay.runAfterMS(EDT,thread(T_ENABLED),111);
            for(int i=3; --i>=0;) ChDelay.revalidate(button(i==0 ? BUT_WIRE : i==1 ? BUT_NUCL : BUT_SEQVISTA),99,(long[])null);
            ChDelay.revalidate(pCenter,111,(long[])null );
            setCursorToResidue(this.cursorP,cursorAA);
        }
    }
    protected StrapProtein getProteinAtY(int y) { return get(y/maxi(1,getSpaceBetweenRows()+charB().height), getVisibleProteins(),StrapProtein.class);}
    protected StrapProtein[] getProteinsAtY(int y) {
        final StrapProtein p=getProteinAtY(y), pp[]=a().getSelectedProteins();
        if (p==null) return StrapProtein.NONE;
        final int i=idxOf(p,pp);
        if (i<0) return new StrapProtein[]{p};
        if (i==0) return pp;
        final StrapProtein[] ppNew=pp.clone();
        ppNew[0]=p;
        ppNew[i]=pp[0];
        return ppNew;
    }
    /* ---------------------------------------- */
    private StrapProtein[] ppR;
    protected StrapProtein[] getVisibleProteinsReordered() {
        final int[] drag=draggedProtein2row();
        final StrapProtein[] pp=getVisibleProteins();
        if (drag==null) return pp;
        if (ppR==null || ppR.length!=pp.length) ppR=new StrapProtein[pp.length];
        for(int i=drag.length; --i>=0;) {
            int row=drag[i];
            if (row>=0 && row<pp.length && i<pp.length) {
                ppR[row]=pp[i];
            }
        }
        return ppR;
    }
    /* <<< Proteins <<< */
    /* ---------------------------------------- */
    /* >>> DnD >>> */
    private boolean doingDnd, rh_doingDnd;
    /* <<< DnD <<< */
    /* ---------------------------------------- */
    /* >>> Rubber band >>> */
    private final Rectangle rectRubBand=new Rectangle();
    private int _rectRubBand_mc ,_rectRubBandP_mc;
    private Object[][] _rectRubBandP;
    private static boolean _rectRubBand_isAdjusting;
    private final static Rectangle rectRBA=new Rectangle();
    public void setRectRubberBand(int col, int row, int numCol, int numRow) {
        boolean success=false;
        if (numRow<=0||numCol<=0) {
            setBnds(0,0,0,0, rectRubBand);
            _rectRubBandP=null;
        } else {
            final Object rp[][]=new Object[numRow][];
            for(int iP=numRow; --iP>=0;) {
                final StrapProtein p=getProtein(iP+row);
                if (p==null) continue;
                final int i0=p.getFirstResidueIndex();
                final int col0=i0+p.column2nextIndex(col);
                final int col1=i0+p.column2thisOrPreviousIndex(col+numCol-1);
                if (col0<col1) {
                    rp[iP]=new Object[]{weakRef(p),new int[]{col0, col1}};
                    success=true;
                }
            }
            _rectRubBandP=rp;
        }
        for(int i=sze(vLabelsRunningAnts); --i>=0;) {
            final ChLabel lab=get(i,vLabelsRunningAnts,ChLabel.class);
            if (lab!=null) lab.fg(success ? BLACK:C(DEFAULT_BACKGROUND));
        }
        _rectRubBandP_mc++;
    }
    public Rectangle getRectRubberBand() {
        if (_rectRubBand_isAdjusting) return rectRBA;
        final Object[] cached=cached();
        final int mc=_rectRubBandP_mc+vROWS.getModificationCount()+Protein.MODI_COUNTS_ALL[MCA_ALIGNMENT];
        final Object[][] rp=_rectRubBandP;
        if (rp==null) { setBnds(0,0,0,0, rectRubBand); return null;}
        if (_rectRubBand_mc!=mc) {
            _rectRubBand_mc=mc;
            int rowMin=MAX_INT, rowMax=MIN_INT, colMax=MIN_INT, colMin=MAX_INT;
            for(Object[] row : rp) {
                final StrapProtein p=get(0,row, StrapProtein.class);
                if (p==null) continue;
                final int i0=p.getFirstResidueIndex();
                final int iRow=findRow(p);
                if (iRow<0) continue;
                final int[] ii=get(1,row,int[].class);
                final int col0=p.getResidueColumn(ii[0]-i0);
                final int col1=p.getResidueColumn(ii[1]-i0);
                if (col0>=0) {
                    if (col0<colMin) colMin=col0;
                    if (col0>colMax) colMax=col0;
                }
                if (col1>=0) {
                    if (col1<colMin) colMin=col1;
                    if (col1>colMax) colMax=col1;
                }
                rowMax=maxi(rowMax,iRow);
                rowMin=mini(rowMin,iRow);
            }
            final int w=maxi(0,colMax-colMin)+1;
            final int h=maxi(0,rowMax-rowMin)+1;
            final short[][] ccc=(short[][])cached[MCSV_CHARACTERS];
            for(int i=2; --i>=0;) {
                final Rectangle r=i==0?rectRBA : rectRubBand;
                markDirtyRect(false,ccc, r.x-1, r.y-1, r.width, r.height);
                markDirtyRect(false,ccc, r.x  , r.y  , r.width, r.height);
                setBnds(colMin, rowMin, w, h, r);
            }
            ChDelay.runAfterMS(EDT,threadPR(PR_ONLY_CHANGES),99);
        }
        return rectRubBand;
    }
    static Rectangle getRectangle(StrapView view0) {
        final StrapView v=view0!=null ? view0 : a().getAlignmentPanelWithFocus();
        return v==null ? null : v.getRectRubberBand();
    }

}
